#
# dialog_create --
#
#       Create a new tclSadie dialog window.
#           (based on "dialog_create" in Effective Tcl/Tk Programming)
#
# Arguments:
#       class -- The class of the toplevel window
#       win -- The name of the toplevel window
#
# Results:
#       Creates a window with a left pane for dialog content and 
#       a right pane for dialog controls.
#
proc dialog_create {class {win "auto"}} {
    if {$win == "auto"} {
        set count 0
        set win ".dialog[incr count]"
        while {[winfo exists $win]} {
            set win ".dialog[incr count]"
        }
    }

    # In case there is already a window by this name,
    # eliminate it!
    if {[winfo exists $win]} {
	destroy $win
    }
    
    toplevel $win -class $class

    frame $win.title
    frame $win.input
    frame $win.output
    frame $win.xsep -width 2 -borderwidth 1 -relief sunken
    frame $win.ysep -height 2 -borderwidth 1 -relief sunken
    frame $win.controls

    grid $win.title -row 0 -column 0 -columnspan 3 -padx 4 -pady 4 -sticky news
    grid $win.ysep -row 1 -column 0 -columnspan 3 -padx 4 -sticky ew
    grid $win.input -row 2 -column 0 -padx 4 -pady 4 -sticky news
    grid $win.output -row 3 -column 0 -padx 4 -pady 4 -sticky news
    grid $win.xsep -row 2 -column 1 -rowspan 2 -pady 4 -sticky ns
    grid $win.controls -row 2 -column 2 -rowspan 2 -padx 4 -pady 4 -sticky news

    grid columnconfigure $win 0 -weight 1
    grid rowconfigure $win 2 -weight 1

    wm title $win $class
    wm group $win .
    wm resizable $win 0 0

    return $win
}

#
# dialog_title --
#
#       Query the name of the title pane in a dialog window.
#
# Arguments:
#       win -- The name of the toplevel window
#
# Results:
#       Returns the name of the title pane.
#
proc dialog_title {win} {
    return "$win.title"
}

#
# dialog_input --
#
#       Query the name of the input pane in a dialog window.
#
# Arguments:
#       win -- The name of the toplevel window
#
# Results:
#       Returns the name of the input pane.
#
proc dialog_input {win} {
    return "$win.input"
}

#
# dialog_output --
#
#       Query the name of the output pane in a dialog window.
#
# Arguments:
#       win -- The name of the toplevel window
#
# Results:
#       Returns the name of the output pane.
#
proc dialog_output {win} {
    return "$win.output"
}

#
# dialog_controls --
#
#       Query the name of the control pane in a dialog window.
#
# Arguments:
#       win -- The name of the toplevel window
#
# Results:
#       Returns the name of the control pane.
#
proc dialog_controls {win} {
    return "$win.controls"
}

#
# dialog_wait --
#
#       Lock the program until user responds to the dialog.
#
# Arguments:
#       win -- The name of the toplevel window
#       varName -- Name of variable to monitor
#
# Results:
#       Grabs control of the window manager until $varName
#       is written to.
#
proc dialog_wait {win varName} {
    dialog_safeguard $win

    set x [expr [winfo rootx .]+50]
    set y [expr [winfo rooty .]+50]
    wm geometry $win "+$x+$y"

    wm deiconify $win
    
    set cmd {grab set $win}

    tkwait visibility $win

    if {[catch $cmd err] != 0} {
        global errorInfo
        set savedInfo $errorInfo
        
        destroy $win
        
        error $err $savedInfo
    }

    vwait $varName

    set cmd {grab release $win}
    if {[catch $cmd err] != 0} {
        global errorInfo
        set savedInfo $errorInfo
        
        destroy $win
        
        error $err $savedInfo
    }

    wm withdraw $win
}

bind modalDialog <ButtonPress> {
    wm deiconify %W
    raise %W
}

#
# dialog_safeguard --
#
#       Make sure that a grabbed window cannot get lost.
#
# Arguments:
#       win -- The name of the toplevel window
#
# Results:
#       Adds modalDialog binding to the window to ensure that
#       the window cannot get lost.
#
proc dialog_safeguard {win} {
    if {[lsearch [bindtags $win] modalDialog] < 0} {
        bindtags $win [linsert [bindtags $win] 0 modalDialog]
    }
}

#
# dialog_buttons --
#
#       Create a standard set of buttons for each dialog
#       in tclSadie.
#
# Arguments:
#       win -- The name of the toplevel window
#       okcmd -- The command to call if "OK" is pressed
#       cancelcmd -- The command to call if "Cancel" 
#           is pressed
#       helpcmd -- The command to call if "Help" is pressed
#
# Results:
#       Creates "OK", "Help", and "Cancel" buttons and
#       packs them vertically.
#
proc dialog_buttons {win okcmd cancelcmd {helpcmd "none"}} {
    button $win.ok -text "OK" -command $okcmd
    
    button $win.cancel -text "Cancel" -command $cancelcmd
    
    button $win.help -text "Help" -command $helpcmd
    if {$helpcmd == "none"} {
        $win.help configure -state disabled
    }
    
    grid $win.ok -pady 5 -padx 5 -sticky ew
    grid $win.cancel -pady 5 -padx 5 -sticky ew
    grid $win.help -pady 5 -padx 5 -sticky ew

    focus $win.ok

    wm protocol [winfo toplevel $win] WM_DELETE_WINDOW "$win.cancel invoke"
}

#
# dialog_set_confirm --
#
#       Set the confirmStatus variable.
#
# Arguments:
#       status -- Value to set variable to.
#
# Results:
#       Sets the confirmStatus variable, which should
#       triger a modal dialog to continue.
#
proc dialog_set_confirm {status} {
    global confirmStatus
    
    set confirmStatus $status
}

#
# dialog_create_stdrow --
#
#       Automatically create a row of standard components for use in a dialog
#       and grid them to the screen.
#
# Arguments:
#       win      -- The window to place the items in.
#       row      -- The row to place the items in.
#       typelist -- A list of items to place in the row.
#
# Results:
#       The specified items are created and gridded.
#
proc dialog_create_stdrow {win row typelist} {
    set column 0
    set sticky w
    set pady 2
    set padx 0

    foreach element $typelist {
	set span 1
	set usevar 0
	set usecmd 0

	set numargs [llength $element]
	set type [lindex $element 0]

	switch $type {
	    label {
		if {$numargs < 2} {
		    error "wrong number of args for type $type:  should be \"$type text \[variable\] \[span\]\""
		}

		set text [lindex $element 1]
		if {$numargs > 2} {
		    set variable [lindex $element 2]
		    if {$variable != "none"} {
			set usevar 1
		    }
		}
		if {$numargs > 3} {
		    set span [lindex $element 3]
		}

		set name ${win}.item${row}_$column
		if {$usevar} {
		    label $name -text $text -textvariable $variable
		} else {
		    label $name -text $text
		}
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }

	    imgchooser {
		set imgnum 0
		set index none
		
		if {$numargs < 2} {
		    error "wrong number of args for type $type:  should be \"$type variable [imgnum] \[span\]\""
		}

		set var [lindex $element 1]
		# If the variable is an array, parse into variable and index names
		set isarray [regexp {([^\(]*)\(([^\)]*)\)$} $var match var index]

		if {$numargs > 2} {
		    set imgnum [lindex $element 2]
		}
		if {$numargs > 3} {
		    set span [lindex $element 3]
		}

		set name ${win}.item${row}_$column
		create_image_selector $name $var $index $imgnum
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }

	    entry {
		set width 5
		
		if {$numargs < 2} {
		    error "wrong number of args for type $type:  should be \"$type variable \[width\] \[span\]\""
		}

		set var [lindex $element 1]
		if {$numargs > 2} {
		    set width [lindex $element 2]
		}
		if {$numargs > 3} {
		    set span [lindex $element 3]
		}

		set name ${win}.item${row}_$column
		entry $name -width $width -textvariable $var
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }

	    scale {
		set length 175
		set tickinterval 25
		
		if {$numargs < 3} {
		    error "wrong number of args for type $type:  should be \"$type variable orient \[length\]\""
		}

		set var [lindex $element 1]
		set orient [lindex $element 2]
		if {$numargs > 3} {
		    set length [lindex $element 3]
		}

		if {$numargs > 4} {
		    set span [lindex $element 4]
		}

		set name ${win}.item${row}_$column
		scale $name -variable $var -orient $orient -length $length -showvalue 0
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }
	    
	    radiobutton {
		if {$numargs < 4} {
		    error "wrong number of args for type $type:  should be \"$type text variable value \[command\] \[span\]\""
		}

		set text [lindex $element 1]
		set var [lindex $element 2]
		set value [lindex $element 3]
		if {$numargs > 4} {
		    set cmd [lindex $element 4]
		    if {$cmd != "NOP"} {
			set usecmd 1
		    }
		}
		if {$numargs > 5} {
		    set span [lindex $element 5]
		}

		set name ${win}.item${row}_$column
		if {$usecmd} {
		    radiobutton $name -variable $var -text $text -value $value -command $cmd
		} else {
		    radiobutton $name -variable $var -text $text -value $value
		}
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }

	    checkbutton {
		if {$numargs < 5} {
		    error "wrong number of args for type $type:  should be \"$type text variable onval offval \[command\] \[span\]\""
		}

		set text [lindex $element 1]
		set var [lindex $element 2]
		set onvalue [lindex $element 3]
		set offvalue [lindex $element 4]
		if {$numargs > 5} {
		    set cmd [lindex $element 5]
		    if {$cmd != "NOP"} {
			set usecmd 1
		    }
		}
		if {$numargs > 6} {
		    set span [lindex $element 6]
		}

		set name ${win}.item${row}_$column
		if {$usecmd} {
		    checkbutton $name -variable $var -text $text -onvalue $onvalue -offvalue $offvalue -command $cmd
		} else {
		    checkbutton $name -variable $var -text $text -onvalue $onvalue -offvalue $offvalue
		}
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }

	    chooser {
		if {$numargs < 3} {
		    error "wrong number of args for type $type:  should be \"$type variable valuelist \[span\]\""
		}

		set var [lindex $element 1]
		set valuelist [lindex $element 2]
		if {$numargs > 3} {
		    set span [lindex $element 3]
		}

		set name ${win}.item${row}_$column
		eval tk_optionMenu $name $var $valuelist
		grid $name -row $row -column $column -columnspan $span -padx $padx -pady $pady -sticky $sticky
	    }

	    interp_select {
		set span1 1
		set span2 1
		set span3 1
		
		if {$numargs < 3} {
		    error "wrong number of args for type $type:  should be \"$type typevar alphavar \[span1\] \[span2\] \[span3\]\""
		}

		set typevar [lindex $element 1]
		set alphavar [lindex $element 2]
		if {$numargs > 3} {
		    set span1 [lindex $element 3]
		}
		if {$numargs > 4} {
		    set span2 [lindex $element 4]
		}
		if {$numargs > 5} {
		    set span3 [lindex $element 5]
		}
		
		set name ${win}.item${row}_$column

		label ${name}_lbl -text "Interpolation: "
		label ${name}_alphalbl -text "  Alpha:"
		entry ${name}_alphaent -textvariable $alphavar -width 5
		radiobutton ${name}_r1 -text "Nearest Neighbor" -variable $typevar -value 1\
			-command [subst {
		    ${name}_alphaent configure -state disabled -fg [${name}_alphaent cget -bg]
		                         }]
		radiobutton ${name}_r2 -text "Bi-Linear" -variable $typevar -value 2\
			-command [subst {
		    ${name}_alphaent configure -state disabled -fg [${name}_alphaent cget -bg]
		                         }]
		radiobutton ${name}_r3 -text "Parametric Cubic" -variable $typevar -value 3\
			-command [subst {
		    ${name}_alphaent configure -state normal -fg [${name}_alphaent cget -fg]
		                         }]

		${name}_r1 invoke

		grid ${name}_lbl -row $row -column $column -columnspan $span1 -padx $padx -pady $pady -sticky $sticky
		grid ${name}_r1 -row $row -column [expr $column+$span1] -columnspan $span2 -padx $padx -pady $pady -sticky $sticky
		incr row
		grid ${name}_r2 -row $row -column [expr $column+$span1] -columnspan $span2 -padx $padx -pady $pady -sticky $sticky
		incr row
		grid ${name}_r3 -row $row -column [expr $column+$span1] -columnspan $span2 -padx $padx -pady $pady -sticky $sticky
		grid ${name}_alphalbl -row $row -column [expr $column+$span1+$span2] -columnspan $span3 -padx $padx -pady $pady -sticky $sticky
		grid ${name}_alphaent -row $row -column [expr $column+$span1+$span2+$span3] -columnspan 1 -padx $padx -pady $pady -sticky $sticky
	    }

	    skip {
		set span 1
	    }

	    default {
		error "Unknown element type ${type}.  Should be: label, imgchooser, entry, radiobutton, checkbutton, chooser, or skip."
	    }
	}
	incr column $span
    }

    return [expr $row + 1]
}




option add *Confirm*icon.bitmap questhead widgetDefault
option add *Confirm*mesg.wrapLength 4i widgetDefault

proc confirm_ask {mesg {ok "OK"} {cancel "Cancel"}} {
    global confirmStatus

    set top [dialog_create Confirm]

    set info [dialog_input $top]
    label $info.icon
    pack $info.icon -side left -padx 8 -pady 8
    label $info.mesg -text $mesg
    pack $info.mesg -side right -expand yes -fill both -padx 8 -pady 8


    set cntls [dialog_controls $top]
    button $cntls.ok -text $ok -command {set confirmStatus 1}
    button $cntls.cancel -text $cancel -command {set confirmStatus 0}

    grid $cntls.ok -pady 5 -padx 5 -sticky ew
    grid $cntls.cancel -pady 5 -padx 5 -sticky ew

    focus $cntls.ok
    wm protocol $top WM_DELETE_WINDOW "$cntls.cancel invoke"

    dialog_wait $top confirmStatus
    destroy $top
    return $confirmStatus
}

# Generic Dialog Box with Info and Controls Section
proc info_dialog_create { class {win "auto"} } {
  if {$win == "auto"} {
  	set count 0
	set win ".dialog[incr count]"
	
	while {[winfo exists $win]} {
		set win ".dialog[incr count]"
	}
  }

  toplevel $win -class $class -background LightGray
  
  frame $win.info -background LightGray
  pack $win.info -expand yes -fill both -padx 2 -pady 2
  
  frame $win.sep -height 2 -borderwidth 1 -relief sunken -background LightGray
  pack $win.sep -fill x -pady 4
  
  frame $win.controls -background LightGray
  pack $win.controls -fill x -padx 4 -pady 4
  
  wm title $win $class
  wm group $win .
  
  after idle [format {
    update idletasks
    wm minsize %s [winfo reqwidth %s] [winfo reqheight %s]
  } $win $win $win]
  
  return $win
}

proc info_dialog_info { win } {
	return "$win.info"
}

proc info_dialog_controls { win } {
	return "$win.controls"
}
