#
# tclSadie_filtermenu_init --
#
#       Initialize the parameters for the all of the filter menu dialogs.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters
#       for each dialog.
#
proc tclSadie_filtermenu_init {} {
    powerspect_init
    fftconv_init
    fftfilter_init
    fft_init
    fht_init
    conv_init
    gradientlib_init
    gradientuser_init
    varmap_init
    diversmap_init
    majorfilt_init
    medianfilt_init
    minfilt_init
    maxfilt_init
    condavgfilt_init
}



#
# powerspect_dialog --
#
#       Create a dialog box for computing a power spectrum for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc powerspect_dialog { } {
    global PowerspectVar
    global confirmStatus

    #
    # Determine the current image
    #
    set PowerspectVar(inimg,array) [get_current_image]
    if {$PowerspectVar(inimg,array) != "none"} {
	    upvar #0 $PowerspectVar(inimg,array) imgarray
	    set PowerspectVar(inimg,name) $imgarray(title)
	    set PowerspectVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .powerspectdlg]
    wm geometry $top +110+110
    wm title $top "Power Spectrum Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Power Spectrum of: "
    label $title.imglbl -fg blue -text $PowerspectVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {radiobutton "2-Dimensional" PowerspectVar(option) 0}
    dialog_create_stdrow $input 0 $row
    
    set row {}
    lappend row {radiobutton "Row Average" PowerspectVar(option) 1}
    dialog_create_stdrow $input 1 $row

    set row {}
    lappend row {radiobutton "Column Average" PowerspectVar(option) 2}
    dialog_create_stdrow $input 2 $row


    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable PowerspectVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        powerspect_run    
    }
}
#
# powerspect_run --
#
#       Define a new region based on the parameters from the
#       powerspect dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc powerspect_run {} {
    global PowerspectVar
    set out_sadie_addr [Sadie_Filter_PSpect PowerspectVar]
    create_display_image $PowerspectVar(outname) $out_sadie_addr
}
#
# powerspect_init --
#
#       Initialize the parameters for the powerspect dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc powerspect_init {} {
    global PowerspectVar
    
    set PowerspectVar(inimg,array) "none"
    set PowerspectVar(inimg,name) "none"
    set PowerspectVar(inimg,addr) "none"
    set PowerspectVar(option) 0
    set PowerspectVar(outname) "PowerSpect"
}



#
# fftconv_dialog --
#
#       Create a dialog box for performing an FFT Convolution on a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc fftconv_dialog { } {
    global FFTConvVar
    global confirmStatus

    #
    # Determine the current image
    #
    set FFTConvVar(inimg,array1) [get_current_image]
    if {$FFTConvVar(inimg,array1) != "none"} {
	    upvar #0 $FFTConvVar(inimg,array1) imgarray
	    set FFTConvVar(inimg,name1) $imgarray(title)
	    set FFTConvVar(inimg,addr1) $imgarray(sadie_addr)
    } else {
    	return
    }

    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .fftconvdlg]
    wm geometry $top +110+110
    wm title $top "FFT Convolution Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute FFT Convolution on: "
    label $title.imglbl -fg blue -text $FFTConvVar(inimg,name1)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Filter Image: "}
    lappend row {imgchooser FFTConvVar(inimg,addr2) 1}
    dialog_create_stdrow $input 0 $row

    set row {}
    lappend row {skip}
    lappend row {radiobutton "Spatial Kernel"  FFTConvVar(option) 0}
    dialog_create_stdrow $input 1 $row

    set row {}
    lappend row {skip}
    lappend row {radiobutton "Frequency Transfer Function"  FFTConvVar(option) 1}
    dialog_create_stdrow $input 2 $row


    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable FFTConvVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        fftconv_run    
    }
}
#
# fftconv_run --
#
#       Define a new region based on the parameters from the
#       FFTConv dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc fftconv_run {} {
    global FFTConvVar
    set out_sadie_addr [Sadie_Filter_FFTConv FFTConvVar]
    create_display_image $FFTConvVar(outname) $out_sadie_addr
}
#
# fftconv_init --
#
#       Initialize the parameters for the FFTConv dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc fftconv_init {} {
    global FFTConvVar
    
    set FFTConvVar(inimg,array1) "none"
    set FFTConvVar(inimg,name1) "none"
    set FFTConvVar(inimg,addr1) "none"
    set FFTConvVar(inimg,addr2) "none"
    set FFTConvVar(option) 0
    set FFTConvVar(outname) "FFTConv"
}



#
# fftfilter_dialog --
#
#       Create a dialog box for performing an FFT Filter on a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc fftfilter_dialog { } {
    global FFTFilterVar
    global confirmStatus

    #
    # Determine the current image
    #
    set FFTFilterVar(inimg,array) [get_current_image]
    if {$FFTFilterVar(inimg,array) != "none"} {
	    upvar #0 $FFTFilterVar(inimg,array) imgarray
	    set FFTFilterVar(inimg,name) $imgarray(title)
	    set FFTFilterVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }

    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .fftfilterdlg]
    wm geometry $top +110+110
    wm title $top "FFT Filter Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Perform FFT Filter on: "
    label $title.imglbl -fg blue -text $FFTFilterVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    label $input.funlbl -text "Filter Function: "
    tk_optionMenu $input.funmnu FFTFilterVar(function) "Cone" "Negative Exponential" "Gaussian" "Box" "Cylinder"
    grid $input.funlbl -row 0 -column 0 -columnspan 1 -sticky nws
    grid $input.funmnu -row 0 -column 1 -columnspan 4 -sticky nws

    set row {}
    lappend row {label "Frequency Radius: "}
    lappend row {entry FFTFilterVar(xradius)}
    lappend row {label "Cycles / Pixel"}
    dialog_create_stdrow $input 1 $row

    set row {}
    lappend row {skip}
    lappend row {entry FFTFilterVar(yradius)}
    lappend row {label "Cycles / Line"}
    dialog_create_stdrow $input 2 $row

    set row {}
    lappend row {label "Filter Type: "}
    lappend row {radiobutton "Low-Pass"  FFTFilterVar(type) 0 fftfilter_update 2}
    dialog_create_stdrow $input 3 $row

    set row {}
    lappend row {skip}
    lappend row {radiobutton "High Boost"  FFTFilterVar(type) 1 fftfilter_update 2}
    lappend row {label "Parameter:"}
    dialog_create_stdrow $input 4 $row

    entry $input.parament -width 5 -textvariable FFTFilterVar(parameter)
    set FFTFilterVar(entry,name) $input.parament
    set FFTFilterVar(entry,fg) [$input.parament cget -fg]
    set FFTFilterVar(entry,bg) [$input.parament cget -bg]
    grid $input.parament -row 4 -column 4 -sticky nws

    set FFTFilterVar(parameter) 1.0
    fftfilter_update
    
    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable FFTFilterVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        fftfilter_run    
    }
}
proc fftfilter_update {} {
    global FFTFilterVar

    if {$FFTFilterVar(type)} {
	$FFTFilterVar(entry,name) configure -state normal -fg $FFTFilterVar(entry,fg)
    } else {
	$FFTFilterVar(entry,name) configure -state disabled -fg $FFTFilterVar(entry,bg)
    }
}
#
# fftfilter_run --
#
#       Define a new region based on the parameters from the
#       FFTFilter dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc fftfilter_run {} {
    global FFTFilterVar

    set FFTFilterVar(parameter) [expr double($FFTFilterVar(type)) * $FFTFilterVar(parameter)]
    
    switch $FFTFilterVar(function) {
	"Cone" { set FFTFilterVar(option) 1}
	"Negative Exponential" { set FFTFilterVar(option) 2}
	"Gaussian" { set FFTFilterVar(option) 3}
	"Box" { set FFTFilterVar(option) 4}
	default { set FFTFilterVar(option) 5}
    }

    set out_sadie_addr [Sadie_Filter_FFTFilt FFTFilterVar]
    create_display_image $FFTFilterVar(outname) $out_sadie_addr
}
#
# fftfilter_init --
#
#       Initialize the parameters for the FFTFilter dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc fftfilter_init {} {
    global FFTFilterVar
    
    set FFTFilterVar(inimg,array) "none"
    set FFTFilterVar(inimg,name) "none"
    set FFTFilterVar(inimg,addr) "none"
    set FFTFilterVar(function) "Cone"
    set FFTFilterVar(type) 0
    set FFTFilterVar(option) 0
    set FFTFilterVar(xradius) 0.5
    set FFTFilterVar(yradius) 0.5
    set FFTFilterVar(parameter) 0.0
    set FFTFilterVar(outname) "FFTFilter"
}




#
# fft_dialog --
#
#       Create a dialog box for computing a fft for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc fft_dialog { } {
    global FFTVar
    global confirmStatus

    #
    # Determine the current image
    #
    set FFTVar(inimg,array) [get_current_image]
    if {$FFTVar(inimg,array) != "none"} {
	    upvar #0 $FFTVar(inimg,array) imgarray
	    set FFTVar(inimg,name) $imgarray(title)
	    set FFTVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .fftdlg]
    wm geometry $top +110+110
    wm title $top "FFT Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Fast Fourier Transform of: "
    label $title.imglbl -fg blue -text $FFTVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {radiobutton "Forward Transform" FFTVar(option) -1}
    dialog_create_stdrow $input 0 $row
    
    set row {}
    lappend row {radiobutton "Inverse Transform" FFTVar(option) 1}
    dialog_create_stdrow $input 1 $row
    

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable FFTVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        fft_run    
    }
}
#
# fft_run --
#
#       Define a new region based on the parameters from the
#       fft dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc fft_run {} {
    global FFTVar
    set out_sadie_addr [Sadie_Filter_FFT FFTVar]
    create_display_image $FFTVar(outname) $out_sadie_addr
    if {$FFTVar(option) == 1} {
	reload_sadie_image $FFTVar(inimg,array)
    }
}
#
# fft_init --
#
#       Initialize the parameters for the fft dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc fft_init {} {
    global FFTVar
    
    set FFTVar(inimg,array) "none"
    set FFTVar(inimg,name) "none"
    set FFTVar(inimg,addr) "none"
    set FFTVar(option) -1
    set FFTVar(outname) "FFT"
}



#
# fht_dialog --
#
#       Create a dialog box for computing a fht for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc fht_dialog { } {
    global FHTVar
    global confirmStatus

    #
    # Determine the current image
    #
    set FHTVar(inimg,array) [get_current_image]
    if {$FHTVar(inimg,array) != "none"} {
	    upvar #0 $FHTVar(inimg,array) imgarray
	    set FHTVar(inimg,name) $imgarray(title)
	    set FHTVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .fhtdlg]
    wm geometry $top +110+110
    wm title $top "FHT Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Fast Hartley Transform of: "
    label $title.imglbl -fg blue -text $FHTVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {radiobutton "Forward Transform" FHTVar(option) -1}
    dialog_create_stdrow $input 0 $row
    
    set row {}
    lappend row {radiobutton "Inverse Transform" FHTVar(option) 1}
    dialog_create_stdrow $input 1 $row
    

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable FHTVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        fht_run    
    }
}
#
# fht_run --
#
#       Define a new region based on the parameters from the
#       fht dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc fht_run {} {
    global FHTVar
    set out_sadie_addr [Sadie_Filter_FHT FHTVar]
    create_display_image $FHTVar(outname) $out_sadie_addr
}
#
# fht_init --
#
#       Initialize the parameters for the fht dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc fht_init {} {
    global FHTVar
    
    set FHTVar(inimg,array) "none"
    set FHTVar(inimg,name) "none"
    set FHTVar(inimg,addr) "none"
    set FHTVar(option) -1
    set FHTVar(outname) "FHT"
}



#
# conv_dialog --
#
#       Create a dialog box for computing a convolution for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc conv_dialog { } {
    global ConvVar
    global confirmStatus

    #
    # Determine the current image
    #
    set ConvVar(inimg,array) [get_current_image]
    if {$ConvVar(inimg,array) != "none"} {
	    upvar #0 $ConvVar(inimg,array) imgarray
	    set ConvVar(inimg,name) $imgarray(title)
	    set ConvVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .convdlg]
    wm geometry $top +110+110
    wm title $top "Convolution Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Convolution for: "
    label $title.imglbl -fg blue -text $ConvVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    radiobutton $input.ikernelrb -text "Image Kernel" -variable ConvVar(option) -value 1 -command conv_update
    radiobutton $input.mkernelrb -text "Mask Kernel" -variable ConvVar(option) -value 2 -command conv_update

    create_image_selector $input.ikernelimg ConvVar kimg,addr 1
    checkbutton $input.ikernelnorm -text "Normalize Kernel" -variable ConvVar(normalize) -onvalue 1 -offvalue 0

    radiobutton $input.linesrb1 -text 1 -variable ConvVar(lines) -value 1 -command conv_update
    radiobutton $input.linesrb3 -text 3 -variable ConvVar(lines) -value 3 -command conv_update
    radiobutton $input.linesrb5 -text 5 -variable ConvVar(lines) -value 5 -command conv_update
    radiobutton $input.linesrb7 -text 7 -variable ConvVar(lines) -value 7 -command conv_update
    label $input.lineslbl -text "Line(s)"

    radiobutton $input.pixrb1 -text 1 -variable ConvVar(pix) -value 1 -command conv_update
    radiobutton $input.pixrb3 -text 3 -variable ConvVar(pix) -value 3 -command conv_update
    radiobutton $input.pixrb5 -text 5 -variable ConvVar(pix) -value 5 -command conv_update
    radiobutton $input.pixrb7 -text 7 -variable ConvVar(pix) -value 7 -command conv_update
    label $input.pixlbl -text "Pixel(s)"

    label $input.factorlbl -text "Factor: 1/"
    entry $input.factorent -width 5 -textvariable ConvVar(factor)
    set ConvVar(entry,fg) [$input.factorent cget -fg]
    set ConvVar(entry,bg) [$input.factorent cget -bg]

    grid $input.ikernelrb -row 0 -column 0 -columnspan 2 -sticky nws -pady 2
    grid $input.ikernelimg -row 0 -column 2 -columnspan 5 -sticky nws -pady 2

    grid $input.ikernelnorm -row 1 -column 2 -columnspan 5 -sticky nws -pady 2

    grid $input.mkernelrb -row 3 -column 0 -columnspan 2 -sticky nws -pady 2
    grid $input.factorlbl -row 4 -column 0 -sticky nws -pady 2
    grid $input.factorent -row 4 -column 1 -sticky nws -pady 2

    for {set i 0} {$i < 4} {incr i} {
	set rb [expr 1+($i*2)]
	grid $input.linesrb$rb -row 3 -column [expr 2+$i] -sticky nws -pady 2
	grid $input.pixrb$rb -row 4 -column [expr 2+$i] -sticky nws -pady 2
    }
    grid $input.lineslbl -row 3 -column 6 -sticky nws -pady 2
    grid $input.pixlbl -row 4 -column 6 -sticky nws -pady 2

    for {set j 0} {$j < 7} {incr j} {
	for {set k 0} {$k < 7} {incr k} {
	    if {$k == 0} {
		set stick ne
	    } else {
		set stick nw
	    }
	    entry $input.${j}_${k} -width 5 -textvariable ConvVar(mask,${j},${k})
	    grid $input.${j}_${k} -row [expr $j+6] -column [expr $k] -sticky $stick
	}
    }

    grid rowconfigure $input 2 -minsize 10
    grid rowconfigure $input 5 -minsize 10

    conv_update
    
    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable ConvVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        conv_run    
    }
}
proc conv_update {} {
    global ConvVar

    set input [dialog_input .convdlg]
    
    if {$ConvVar(option) == 1} {
	$input.ikernelimg configure -state normal
	$input.ikernelnorm configure -state normal

	for {set i 1} {$i <= 7} {incr i 2} {
	    $input.linesrb$i configure -state disabled
	    $input.pixrb$i configure -state disabled
	}

	$input.factorent configure -state disabled -fg $ConvVar(entry,bg)

	for {set j 0} {$j < 7} {incr j} {
	    for {set k 0} {$k < 7} {incr k} {
		$input.${j}_${k} configure -state disabled -fg $ConvVar(entry,bg)
	    }
	}
    } else {
	$input.ikernelimg configure -state disabled
	$input.ikernelnorm configure -state disabled

	for {set i 1} {$i <= 7} {incr i 2} {
	    $input.linesrb$i configure -state normal
	    $input.pixrb$i configure -state normal
	}

	$input.factorent configure -state normal -fg $ConvVar(entry,fg)
	set ConvVar(factor) [format "%.1f" [expr double($ConvVar(lines)*$ConvVar(pix))]] 

	for {set j 0} {$j < 7} {incr j} {
	    for {set k 0} {$k < 7} {incr k} {
		if {$j < $ConvVar(lines)  &&  $k < $ConvVar(pix)} {
		    $input.${j}_${k} configure -state normal -fg $ConvVar(entry,fg)
		} else {
		    $input.${j}_${k} configure -state disabled -fg $ConvVar(entry,bg)
		}
	    }
	}
    }
}
#
# conv_run --
#
#       Define a new region based on the parameters from the
#       conv dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc conv_run {} {
    global ConvVar
    if {$ConvVar(option) == 1} {
	set out_sadie_addr [Sadie_Filter_ImageConv ConvVar]
    } else {
	set out_sadie_addr [Sadie_Filter_MaskConv ConvVar]
    }
    create_display_image $ConvVar(outname) $out_sadie_addr
}
#
# conv_init --
#
#       Initialize the parameters for the conv dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc conv_init {} {
    global ConvVar
    
    set ConvVar(inimg,array) "none"
    set ConvVar(inimg,name) "none"
    set ConvVar(inimg,addr) "none"
    set ConvVar(kimg,addr) "none"
    set ConvVar(option) 1
    set ConvVar(normalize) 1
    set ConvVar(lines) 3
    set ConvVar(pix) 3
    set ConvVar(factor) 9.0

    for {set j 0} {$j < 7} {incr j} {
	for {set k 0} {$k < 7} {incr k} {
	    set ConvVar(mask,${j},${k}) 1.0
	}
    }
    
    set ConvVar(outname) "Conv"
}



#
# gradientlib_dialog --
#
#       Create a dialog box for performing a library gradient operation on a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc gradientlib_dialog { } {
    global GradientlibVar
    global confirmStatus

    #
    # Determine the current image
    #
    set GradientlibVar(inimg,array) [get_current_image]
    if {$GradientlibVar(inimg,array) != "none"} {
	    upvar #0 $GradientlibVar(inimg,array) imgarray
	    set GradientlibVar(inimg,name) $imgarray(title)
	    set GradientlibVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }

    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .gradientlibdlg]
    wm geometry $top +110+110
    wm title $top "Library Gradient Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Library Gradient for: "
    label $title.imglbl -fg blue -text $GradientlibVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Gradient Operator: "}
    lappend row {chooser GradientlibVar(type) {"Sobel" "Sobel2" "Sobel3" "Roberts" "Roberts2" "Prewitt"}}
    dialog_create_stdrow $input 0 $row


    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Names: "
    entry $output.magnameent -width 20 -textvariable GradientlibVar(magname)
    entry $output.dirnameent -width 20 -textvariable GradientlibVar(dirname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.magnameent -row 0 -column 1 -sticky ew
    grid $output.dirnameent -row 1 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        gradientlib_run    
    }
}
#
# gradientlib_run --
#
#       Define a new region based on the parameters from the
#       Gradientlib dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc gradientlib_run {} {
    global GradientlibVar

    switch $GradientlibVar(type) {
	"Sobel" {set GradientlibVar(option) 0}
	"Sobel2" {set GradientlibVar(option) 1}
	"Sobel3" {set GradientlibVar(option) 2}
	"Roberts" {set GradientlibVar(option) 3}
	"Roberts2" {set GradientlibVar(option) 4}
	"Prewitt" {set GradientlibVar(option) 5}
    }
    
    set results [Sadie_Filter_LibGrad GradientlibVar]

    set mag_sadie_addr [lindex $results 0]
    create_display_image $GradientlibVar(magname) $mag_sadie_addr

    set dir_sadie_addr [lindex $results 1]
    create_display_image $GradientlibVar(dirname) $dir_sadie_addr
}
#
# gradientlib_init --
#
#       Initialize the parameters for the Gradientlib dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc gradientlib_init {} {
    global GradientlibVar
    
    set GradientlibVar(inimg,array) "none"
    set GradientlibVar(inimg,name) "none"
    set GradientlibVar(inimg,addr) "none"
    set GradientlibVar(option) 0
    set GradientlibVar(magname) "GradMag"
    set GradientlibVar(dirname) "GradDir"
}



#
# gradientuser_dialog --
#
#       Create a dialog box for computing a user gradient for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc gradientuser_dialog { } {
    global GradientuserVar
    global confirmStatus

    #
    # Determine the current image
    #
    set GradientuserVar(inimg,array) [get_current_image]
    if {$GradientuserVar(inimg,array) != "none"} {
	    upvar #0 $GradientuserVar(inimg,array) imgarray
	    set GradientuserVar(inimg,name) $imgarray(title)
	    set GradientuserVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .gradientuserdlg]
    wm geometry $top +110+110
    wm title $top "User Gradient Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute User Gradient for: "
    label $title.imglbl -fg blue -text $GradientuserVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    radiobutton $input.ikernelrb -text "Image Kernels" -variable GradientuserVar(option) -value 1 -command gradientuser_update
    radiobutton $input.mkernelrb -text "Mask Kernels" -variable GradientuserVar(option) -value 2 -command gradientuser_update

    label $input.himglbl -text "Horizontal: "
    create_image_selector $input.himg GradientuserVar himg,addr 1

    label $input.vimglbl -text "Vertical: "
    create_image_selector $input.vimg GradientuserVar vimg,addr 2

    grid $input.ikernelrb -row 0 -column 0 -columnspan 3 -sticky nws -pady 2
    grid $input.himglbl -row 0 -column 4 -columnspan 2 -sticky nes -pady 2
    grid $input.himg -row 0 -column 6 -columnspan 2 -sticky nws -pady 2

    grid $input.vimglbl -row 1 -column 4 -columnspan 2 -sticky nes -pady 2
    grid $input.vimg -row 1 -column 6 -columnspan 2 -sticky nws -pady 2

    grid $input.mkernelrb -row 2 -column 0 -columnspan 3 -sticky nws -pady 2

    label $input.hlbl -text "Horizontal: "
    label $input.vlbl -text "Vertical: "

    grid $input.hlbl -row 3 -column 0 -columnspan 3 -sticky ws -pady 2
    grid $input.vlbl -row 3 -column 4 -columnspan 3 -sticky ws -pady 2
    
    for {set j 0} {$j < 3} {incr j} {
	for {set k 0} {$k < 3} {incr k} {
	    entry $input.h_${j}_${k} -width 5 -textvariable GradientuserVar(hmask,${j},${k})
	    grid $input.h_${j}_${k} -row [expr $j+4] -column [expr $k] -sticky nw
	}
    }
    for {set j 0} {$j < 3} {incr j} {
	for {set k 0} {$k < 3} {incr k} {
	    entry $input.v_${j}_${k} -width 5 -textvariable GradientuserVar(vmask,${j},${k})
	    grid $input.v_${j}_${k} -row [expr $j+4] -column [expr $k+4] -sticky nw
	}
    }

    set GradientuserVar(entry,fg) [$input.h_0_0 cget -fg]
    set GradientuserVar(entry,bg) [$input.h_0_0 cget -bg]

    grid columnconfigure $input 3 -minsize 10

    gradientuser_update
    
    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Names: "
    entry $output.magnameent -width 20 -textvariable GradientuserVar(magname)
    entry $output.dirnameent -width 20 -textvariable GradientuserVar(dirname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.magnameent -row 0 -column 1 -sticky ew
    grid $output.dirnameent -row 1 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        gradientuser_run    
    }
}
proc gradientuser_update {} {
    global GradientuserVar

    set input [dialog_input .gradientuserdlg]
    
    if {$GradientuserVar(option) == 1} {
	$input.himg configure -state normal
	$input.vimg configure -state normal

	for {set j 0} {$j < 3} {incr j} {
	    for {set k 0} {$k < 3} {incr k} {
		$input.h_${j}_${k} configure -state disabled -fg $GradientuserVar(entry,bg)
		$input.v_${j}_${k} configure -state disabled -fg $GradientuserVar(entry,bg)
	    }
	}
    } else {
	$input.himg configure -state disabled
	$input.vimg configure -state disabled

	for {set j 0} {$j < 3} {incr j} {
	    for {set k 0} {$k < 3} {incr k} {
		$input.h_${j}_${k} configure -state normal -fg $GradientuserVar(entry,fg)
		$input.v_${j}_${k} configure -state normal -fg $GradientuserVar(entry,fg)
	    }
	}
    }
}
#
# gradientuser_run --
#
#       Define a new region based on the parameters from the
#       gradientuser dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc gradientuser_run {} {
    global GradientuserVar
    if {$GradientuserVar(option) == 1} {
	set results [Sadie_Filter_ImageGrad GradientuserVar]
    } else {
	set results [Sadie_Filter_MaskGrad GradientuserVar]
    }
    
    set mag_sadie_addr [lindex $results 0]
    create_display_image $GradientuserVar(magname) $mag_sadie_addr

    set dir_sadie_addr [lindex $results 1]
    create_display_image $GradientuserVar(dirname) $dir_sadie_addr
}
#
# gradientuser_init --
#
#       Initialize the parameters for the gradientuser dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc gradientuser_init {} {
    global GradientuserVar
    
    set GradientuserVar(inimg,array) "none"
    set GradientuserVar(inimg,name) "none"
    set GradientuserVar(inimg,addr) "none"
    set GradientuserVar(himg,addr) "none"
    set GradientuserVar(vimg,addr) "none"
    set GradientuserVar(option) 1

    for {set j 0} {$j < 3} {incr j} {
	for {set k 0} {$k < 3} {incr k} {
	    set GradientuserVar(hmask,${j},${k}) 0.0
	    set GradientuserVar(vmask,${j},${k}) 0.0
	}
    }
    
    set GradientuserVar(magname) "GradMag"
    set GradientuserVar(dirname) "GradDir"
}



#
# varmap_dialog --
#
#       Create a dialog box for computing a variance map for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc varmap_dialog { } {
    global VarmapVar
    global confirmStatus

    #
    # Determine the current image
    #
    set VarmapVar(inimg,array) [get_current_image]
    if {$VarmapVar(inimg,array) != "none"} {
	    upvar #0 $VarmapVar(inimg,array) imgarray
	    set VarmapVar(inimg,name) $imgarray(title)
	    set VarmapVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .varmapdlg]
    wm geometry $top +110+110
    wm title $top "Variance Map Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Variance Map on: "
    label $title.imglbl -fg blue -text $VarmapVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry VarmapVar(lines)}
    lappend row {label "Line(s)   "}
    lappend row {entry VarmapVar(pix)}
    lappend row {label "Pixel(s)   "}
    dialog_create_stdrow $input 0 $row

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable VarmapVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        varmap_run    
    }
}
#
# varmap_run --
#
#       Define a new region based on the parameters from the
#       varmap dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc varmap_run {} {
    global VarmapVar
    set out_sadie_addr [Sadie_Filter_VarMap VarmapVar]
    create_display_image $VarmapVar(outname) $out_sadie_addr
}
#
# varmap_init --
#
#       Initialize the parameters for the varmap dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc varmap_init {} {
    global VarmapVar
    
    set VarmapVar(inimg,array) "none"
    set VarmapVar(inimg,name) "none"
    set VarmapVar(inimg,addr) "none"
    set VarmapVar(lines) 3
    set VarmapVar(pix) 3
    set VarmapVar(outname) "VarMap"
}



#
# diversmap_dialog --
#
#       Create a dialog box for computing a diversity map for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc diversmap_dialog { } {
    global DiversmapVar
    global confirmStatus

    #
    # Determine the current image
    #
    set DiversmapVar(inimg,array) [get_current_image]
    if {$DiversmapVar(inimg,array) != "none"} {
	    upvar #0 $DiversmapVar(inimg,array) imgarray
	    set DiversmapVar(inimg,name) $imgarray(title)
	    set DiversmapVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .diversmapdlg]
    wm geometry $top +110+110
    wm title $top "Diversity Map Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Diversity Map on: "
    label $title.imglbl -fg blue -text $DiversmapVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry DiversmapVar(lines)}
    lappend row {label "Line(s)   "}
    lappend row {entry DiversmapVar(pix)}
    lappend row {label "Pixel(s)   "}
    dialog_create_stdrow $input 0 $row

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable DiversmapVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        diversmap_run    
    }
}
#
# diversmap_run --
#
#       Define a new region based on the parameters from the
#       diversmap dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc diversmap_run {} {
    global DiversmapVar
    set out_sadie_addr [Sadie_Filter_DivMap DiversmapVar]
    create_display_image $DiversmapVar(outname) $out_sadie_addr
}
#
# diversmap_init --
#
#       Initialize the parameters for the diversmap dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc diversmap_init {} {
    global DiversmapVar
    
    set DiversmapVar(inimg,array) "none"
    set DiversmapVar(inimg,name) "none"
    set DiversmapVar(inimg,addr) "none"
    set DiversmapVar(lines) 3
    set DiversmapVar(pix) 3
    set DiversmapVar(outname) "DivMap"
}



#
# majorfilt_dialog --
#
#       Create a dialog box for computing a majority filter for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc majorfilt_dialog { } {
    global MajorfiltVar
    global confirmStatus

    #
    # Determine the current image
    #
    set MajorfiltVar(inimg,array) [get_current_image]
    if {$MajorfiltVar(inimg,array) != "none"} {
	    upvar #0 $MajorfiltVar(inimg,array) imgarray
	    set MajorfiltVar(inimg,name) $imgarray(title)
	    set MajorfiltVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .majorfiltdlg]
    wm geometry $top +110+110
    wm title $top "Majority Filter Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Majority Filter on: "
    label $title.imglbl -fg blue -text $MajorfiltVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry MajorfiltVar(lines)}
    lappend row {label "Line(s)   " none 2}
    lappend row {entry MajorfiltVar(pix)}
    lappend row {label "Pixel(s)"}
    dialog_create_stdrow $input 0 $row

    set row {}
    lappend row {label "Center Pixel Weight: " none 2}
    lappend row {entry MajorfiltVar(weight)}
    lappend row {label "* Class Count" none 3}
    dialog_create_stdrow $input 1 $row
    
    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable MajorfiltVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        majorfilt_run    
    }
}
#
# majorfilt_run --
#
#       Define a new region based on the parameters from the
#       majorfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc majorfilt_run {} {
    global MajorfiltVar
    set out_sadie_addr [Sadie_Filter_MajFilt MajorfiltVar]
    create_display_image $MajorfiltVar(outname) $out_sadie_addr
}
#
# majorfilt_init --
#
#       Initialize the parameters for the majorfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc majorfilt_init {} {
    global MajorfiltVar
    
    set MajorfiltVar(inimg,array) "none"
    set MajorfiltVar(inimg,name) "none"
    set MajorfiltVar(inimg,addr) "none"
    set MajorfiltVar(lines) 3
    set MajorfiltVar(pix) 3
    set MajorfiltVar(weight) 1.0
    set MajorfiltVar(outname) "Majority Filter"
}



#
# medianfilt_dialog --
#
#       Create a dialog box for computing a median filter for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc medianfilt_dialog { } {
    global MedianfiltVar
    global confirmStatus

    #
    # Determine the current image
    #
    set MedianfiltVar(inimg,array) [get_current_image]
    if {$MedianfiltVar(inimg,array) != "none"} {
	    upvar #0 $MedianfiltVar(inimg,array) imgarray
	    set MedianfiltVar(inimg,name) $imgarray(title)
	    set MedianfiltVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .medianfiltdlg]
    wm geometry $top +110+110
    wm title $top "Median Filter Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Median Filter on: "
    label $title.imglbl -fg blue -text $MedianfiltVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry MedianfiltVar(lines)}
    lappend row {label "Line(s)   "}
    lappend row {entry MedianfiltVar(pix)}
    lappend row {label "Pixel(s)   "}
    dialog_create_stdrow $input 0 $row

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable MedianfiltVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        medianfilt_run    
    }
}
#
# medianfilt_run --
#
#       Define a new region based on the parameters from the
#       medianfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc medianfilt_run {} {
    global MedianfiltVar
    set out_sadie_addr [Sadie_Filter_MedFilt MedianfiltVar]
    create_display_image $MedianfiltVar(outname) $out_sadie_addr
}
#
# medianfilt_init --
#
#       Initialize the parameters for the medianfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc medianfilt_init {} {
    global MedianfiltVar
    
    set MedianfiltVar(inimg,array) "none"
    set MedianfiltVar(inimg,name) "none"
    set MedianfiltVar(inimg,addr) "none"
    set MedianfiltVar(lines) 3
    set MedianfiltVar(pix) 3
    set MedianfiltVar(outname) "Median Filter"
}



#
# minfilt_dialog --
#
#       Create a dialog box for computing a minimum filter for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc minfilt_dialog { } {
    global MinfiltVar
    global confirmStatus

    #
    # Determine the current image
    #
    set MinfiltVar(inimg,array) [get_current_image]
    if {$MinfiltVar(inimg,array) != "none"} {
	    upvar #0 $MinfiltVar(inimg,array) imgarray
	    set MinfiltVar(inimg,name) $imgarray(title)
	    set MinfiltVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .minfiltdlg]
    wm geometry $top +110+110
    wm title $top "Minimum Filter Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Minimum Filter on: "
    label $title.imglbl -fg blue -text $MinfiltVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry MinfiltVar(lines)}
    lappend row {label "Line(s)   "}
    lappend row {entry MinfiltVar(pix)}
    lappend row {label "Pixel(s)   "}
    dialog_create_stdrow $input 0 $row

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable MinfiltVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        minfilt_run    
    }
}
#
# minfilt_run --
#
#       Define a new region based on the parameters from the
#       minfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc minfilt_run {} {
    global MinfiltVar
    set out_sadie_addr [Sadie_Filter_MinFilt MinfiltVar]
    create_display_image $MinfiltVar(outname) $out_sadie_addr
}
#
# minfilt_init --
#
#       Initialize the parameters for the minfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc minfilt_init {} {
    global MinfiltVar
    
    set MinfiltVar(inimg,array) "none"
    set MinfiltVar(inimg,name) "none"
    set MinfiltVar(inimg,addr) "none"
    set MinfiltVar(lines) 3
    set MinfiltVar(pix) 3
    set MinfiltVar(outname) "MinFilt"
}



#
# maxfilt_dialog --
#
#       Create a dialog box for computing a maximum filter for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc maxfilt_dialog { } {
    global MaxfiltVar
    global confirmStatus

    #
    # Determaxe the current image
    #
    set MaxfiltVar(inimg,array) [get_current_image]
    if {$MaxfiltVar(inimg,array) != "none"} {
	    upvar #0 $MaxfiltVar(inimg,array) imgarray
	    set MaxfiltVar(inimg,name) $imgarray(title)
	    set MaxfiltVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .maxfiltdlg]
    wm geometry $top +110+110
    wm title $top "Maximum Filter Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Maximum Filter on: "
    label $title.imglbl -fg blue -text $MaxfiltVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry MaxfiltVar(lines)}
    lappend row {label "Line(s)   "}
    lappend row {entry MaxfiltVar(pix)}
    lappend row {label "Pixel(s)   "}
    dialog_create_stdrow $input 0 $row

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable MaxfiltVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        maxfilt_run    
    }
}
#
# maxfilt_run --
#
#       Define a new region based on the parameters from the
#       maxfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc maxfilt_run {} {
    global MaxfiltVar
    set out_sadie_addr [Sadie_Filter_MaxFilt MaxfiltVar]
    create_display_image $MaxfiltVar(outname) $out_sadie_addr
}
#
# maxfilt_init --
#
#       Initialize the parameters for the maxfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc maxfilt_init {} {
    global MaxfiltVar
    
    set MaxfiltVar(inimg,array) "none"
    set MaxfiltVar(inimg,name) "none"
    set MaxfiltVar(inimg,addr) "none"
    set MaxfiltVar(lines) 3
    set MaxfiltVar(pix) 3
    set MaxfiltVar(outname) "MaxFilt"
}



#
# condavgfilt_dialog --
#
#       Create a dialog box for computing a conditional average filter for a display
#       image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc condavgfilt_dialog { } {
    global CondavgfiltVar
    global confirmStatus

    #
    # Detercondavge the current image
    #
    set CondavgfiltVar(inimg,array) [get_current_image]
    if {$CondavgfiltVar(inimg,array) != "none"} {
	    upvar #0 $CondavgfiltVar(inimg,array) imgarray
	    set CondavgfiltVar(inimg,name) $imgarray(title)
	    set CondavgfiltVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .condavgfiltdlg]
    wm geometry $top +110+110
    wm title $top "Conditional Average Filter Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Compute Conditional Average Filter on: "
    label $title.imglbl -fg blue -text $CondavgfiltVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    set row {}
    lappend row {label "Window Size: "}
    lappend row {entry CondavgfiltVar(lines)}
    lappend row {label "Line(s)   "}
    lappend row {entry CondavgfiltVar(pix)}
    lappend row {label "Pixel(s)   "}
    dialog_create_stdrow $input 0 $row

    set row {}
    lappend row {label "Set Size Threshold: " none 2}
    lappend row {entry CondavgfiltVar(size)}
    dialog_create_stdrow $input 1 $row

    set row {}
    lappend row {label "Graylevel Difference Threshold: " none 3}
    lappend row {entry CondavgfiltVar(thresh)}
    dialog_create_stdrow $input 2 $row

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable CondavgfiltVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus

    destroy $top
    
    if {$confirmStatus == 1} {
        condavgfilt_run    
    }
}
#
# condavgfilt_run --
#
#       Define a new region based on the parameters from the
#       condavgfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       The new region is created.
#
proc condavgfilt_run {} {
    global CondavgfiltVar
    set out_sadie_addr [Sadie_Filter_CAvgFilt CondavgfiltVar]
    create_display_image $CondavgfiltVar(outname) $out_sadie_addr
}
#
# condavgfilt_init --
#
#       Initialize the parameters for the condavgfilt dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc condavgfilt_init {} {
    global CondavgfiltVar
    
    set CondavgfiltVar(inimg,array) "none"
    set CondavgfiltVar(inimg,name) "none"
    set CondavgfiltVar(inimg,addr) "none"
    set CondavgfiltVar(lines) 3
    set CondavgfiltVar(pix) 3
    set CondavgfiltVar(size) 0
    set CondavgfiltVar(thresh) 0.0
    set CondavgfiltVar(outname) "CondAvgFilt"
}
