#
# create_display_image --
#
#       Create a new display image for use in the tclSadie GUI.
#
# Arguments:
#       title -- The title of the display image
#       sadie_addr -- The SADIE address of the image to display
#
# Results:
#       A new display image is created with the supplied title
#       and sadie image.  An image array is constructed and
#       added to the global image list.
#
proc create_display_image {title sadie_addr} {
    global SadieVar

    # Only proceed if tk_photo is created successfully
    if {[catch {set photoname [image create photo -format sadie -data $sadie_addr]} err] == 0} {

	set out_title "#[incr SadieVar(counter)]: $title"
	Sadie_General_SetName $sadie_addr $out_title
        Sadie_General_IndexImage $sadie_addr
	set imgarrayname [create_image_array $out_title $sadie_addr $photoname]

	insert_new_image $imgarrayname
	create_image_window $imgarrayname

	# Make sure the image is scaled correctly
	change_scaling
    }
}

#
# insert_new_image --
#
#       Insert a new image information array into a list of
#       active images.
#
# Arguments:
#       imgName -- The name of the global image information array
#
# Results:
#       The result is that the image information array is inserted
#       at the top of the image list.  This image will be the
#       "current" image.
#
proc insert_new_image {imgName} {
    global SadieVar
      
    if {[lsearch -exact $SadieVar(imagelist) $imgName] < 0} {
        unhighlight_current_image
        set SadieVar(imagelist) [linsert $SadieVar(imagelist) 0 $imgName]
    } else {
        error "The image array \"$imgName\" is already included in the global image list"
    }
}


#
# create_image_array --
#
#       Create a new image information array.
#
# Arguments:
#       imgTitle -- The title of the image
#       sadieAddr -- The address of the SADIE image
#       photoName -- The name of the tk_photo for displaying
#           the SADIE image
#       imgName -- The name of the global image information array
#       imgWindow -- The name of the Tk window in which to
#           display the image
#
# Results:
#       A global image information array is created and
#       initialized.  The name of the array is returned.
#
proc create_image_array { imgTitle sadieAddr photoName {imgName "auto"} {imgWindow "auto"} } {
    global SadieVar
    
    if {$imgName == "auto"} {
        set count 0
        set imgName "imgarr[incr count]"
        global $imgName
        while {[info exists $imgName]} {
            set imgName "imgarr[incr count]"
            global $imgName
        }
    }

    if {$imgWindow == "auto"} {
        set count 0
        set imgWindow ".imgwin[incr count]"
        while {[winfo exists $imgWindow]} {
            set imgWindow ".imgwin[incr count]"
        }
    }

    upvar #0 $imgName imgarray

    # The SADIE address of the image
    set imgarray(sadie_addr) $sadieAddr

    # The name of the Tk window to display the image
    set imgarray(window) $imgWindow
    
    # The title of the image
    set imgarray(title) $imgTitle
    
    # The Tk photo name of the original display image
    set imgarray(tk_photo) $photoName

    # The Tk photo name of the zoomed display image
    set imgarray(tk_photo_zoom) [image create photo]

    # The zoom value for the displayed image
    set imgarray(zoom) 1
    
    # The scaling value for the displayed image
    # Always turn on scaling by default for a new image
    set SadieVar(getstretched) 1
    set imgarray(scaling) $SadieVar(getstretched)
    
    # The number of bands in the SADIE image
    set imgarray(bands) [Sadie_General_GetNumBands $imgarray(sadie_addr)]
    
    # The number of defined regions in the image
    set imgarray(region,count) 0

    # Indicator of whether to show the defined regions
    set imgarray(showregions) 0

    # The number of defined classes in the image
    set imgarray(class,count) 0
    
    # The number of defined control points in the image
    set imgarray(ctrlpt,count) 0

    # Indicator of whether to show the defined control points
    set imgarray(showctrlpts) 1

    return $imgName
}


#
# create_image_window --
#
#       Create a new image window for displaying an active SADIE
#       image.
#
# Arguments:
#       imgName -- The name of the global image information array
#
# Results:
#       A Tk window is created displaying the image.
#
proc create_image_window { imgName } {
    global SadieVar

    upvar #0 $imgName imgarray

    if {$SadieVar(fullcolor)} {
        #Attempt to create the image window with 24 bits/pixel
        toplevel $imgarray(window) -visual {best 24} -bd 5 -relief ridge
    } else {
        #Create the image window with the default depth of the window manager
        toplevel $imgarray(window) -bd 5 -relief ridge
    }

    set offset [expr 150 + (10 * ($SadieVar(counter) % 10))]
    wm geometry $imgarray(window) +${offset}+${offset}
  

    # 
    # Tell the window manager how to handle the window
    #
    wm title $imgarray(window) $imgarray(title)
    wm protocol $imgarray(window) \
        WM_DELETE_WINDOW "delete_image $imgName"
    wm sizefrom $imgarray(window) user
    
  
    $SadieVar(menu,winlist) add command \
        -label $imgarray(title) \
        -command "set_current_image $imgName"


    canvas $imgarray(window).imgcanvas \
        -borderwidth 0 -highlightthickness 0 \
        -xscrollcommand "update_xscroll $imgName" \
        -yscrollcommand "update_yscroll $imgName"
  
    scrollbar $imgarray(window).hscr \
        -command "$imgarray(window).imgcanvas xview" \
        -orient horizontal
    
    scrollbar $imgarray(window).vscr \
        -command "$imgarray(window).imgcanvas yview"

    frame $imgarray(window).empty -bd 0


    $imgarray(tk_photo_zoom) copy $imgarray(tk_photo)

    #
    # Display the tk_photo in the canvas
    #
    set h [image height $imgarray(tk_photo)]
    set w [image width $imgarray(tk_photo)]
    $imgarray(window).imgcanvas configure \
        -height $h \
        -width $w \
        -scrollregion "0 0 $w $h"
    $imgarray(window).imgcanvas create image 0 0 \
        -image $imgarray(tk_photo_zoom) \
        -anchor nw -tags disp_img

  

    grid $imgarray(window).imgcanvas -row 0 -column 0 -sticky nw
    grid $imgarray(window).vscr -row 0 -column 1 -sticky nws
    grid $imgarray(window).hscr -row 1 -column 0 -sticky ews
    grid $imgarray(window).empty -row 1 -column 1 -sticky news
    grid rowconfigure $imgarray(window) 0 -weight 1
    grid columnconfigure $imgarray(window) 0 -weight 1

    set maxwidth [expr [$imgarray(window).imgcanvas cget -width] \
        + [$imgarray(window).vscr cget -width] \
        + [expr 3 * [$imgarray(window).vscr cget -borderwidth]] \
	+ [expr 2 * [$imgarray(window) cget -borderwidth]]]
    set maxheight [expr [$imgarray(window).imgcanvas cget -height] \
        + [$imgarray(window).hscr cget -width] \
        + [expr 3 * [$imgarray(window).hscr cget -borderwidth]] \
	+ [expr 2 * [$imgarray(window) cget -borderwidth]]]

    wm maxsize $imgarray(window) $maxwidth $maxheight
    wm minsize $imgarray(window) $imgarray(zoom) $imgarray(zoom)

    update
    highlight_current_image
    update_gui

    #
    # If the window will not fit on the screen, reduce its size
    #
    set screenwidth [winfo screenwidth .]
    set screenheight [winfo screenheight .]
    set startwidth [expr int($screenwidth * 0.75)]
    set startheight [expr int($screenheight * 0.75)]

    if {$maxwidth < $startwidth || $maxheight < $startheight} {
	set startwidth $maxwidth
	set startheight $maxheight
    }

    wm geometry $imgarray(window) ${startwidth}x${startheight}+${offset}+${offset}
    bind $imgarray(window).imgcanvas <Motion> "image_mouse_motion $imgName %x %y"
    bind $imgarray(window).imgcanvas <Leave>  "image_mouse_leave $imgName"
    bind $imgarray(window).imgcanvas <Button-1> "image_mouse_buttondown1 $imgName %x %y"
    bind $imgarray(window).imgcanvas <ButtonRelease-1> "image_mouse_buttonup1 $imgName %x %y"
    bind $imgarray(window).imgcanvas <Configure> "constrain_size $imgName"

    #bind $imgarray(window) <ButtonRelease-1> "puts \<ButtonRelease\>"
    #bind $imgarray(window) <ButtonPress> "puts \<ButtonPress\>"
    #bind $imgarray(window) <Motion> "puts \<Motion\>"
    #bind $imgarray(window) <Enter> "puts \<Enter\>"
    #bind $imgarray(window) <Leave> "puts \<Leave\>"
    #bind $imgarray(window) <Configure> "puts \<Configure\>"
    
    update
}


#
# image_mouse_motion --
#
#       React to mouse motion over an image window.
#
# Arguments:
#       imgName -- The name of the global image information array
#       xmouse -- current x-position in the canvas
#       ymouse -- current y-position in the canvas
#
# Results:
#       Updates coordinate and pixel information on the main tclSadie
#       window and redraws selection drawing on image if necessary.
#
proc image_mouse_motion {imgName xmouse ymouse} {
    global SadieVar
    
    upvar #0 $imgName imgarray
    

    if {[lsearch -exact $SadieVar(imagelist) $imgName] == 0} {

	set coords [convert_mouse_coordinates $imgName $xmouse $ymouse]
	set xpos [lindex $coords 0]
	set ypos [lindex $coords 1]
	
	update_coordinates $imgName $xpos $ypos
	
        if { $SadieVar(tools,draw) == 1 } {

	    set SadieVar(tools,point2x) $xpos
	    set SadieVar(tools,point2y) $ypos

	    draw_shape $imgName
        }
    }

}

#
# image_mouse_leave --
#
#       React to mouse leaving image window.
#
# Arguments:
#       imgName -- The name of the global image information array
#
# Results:
#       Updates coordinate and pixel information on the main tclSadie
#       window to indicate that no image is under the mouse.
#
proc image_mouse_leave {imgName} {
    global SadieVar

    set SadieVar(label,xpos) " X: "
    set SadieVar(label,ypos) " Y: "
    set SadieVar(label,s_rval) " Z: "
    set SadieVar(label,s_gval) "    "
    set SadieVar(label,s_bval) "    "
    set SadieVar(label,d_rval) " Z: "
    set SadieVar(label,d_gval) "    "
    set SadieVar(label,d_bval) "    "
}

#
# image_mouse_buttondown1 --
#
#       React to mouse left button press over an image window.
#
# Arguments:
#       imgName -- The name of the global image information array
#       xmouse -- current x-position in the canvas
#       ymouse -- current y-position in the canvas
#
# Results:
#       Updates coordinate and pixel information on the main tclSadie
#       window and begins selection drawing on image if necessary.
#       Also, make this image the current image.
#
proc image_mouse_buttondown1 {imgName xmouse ymouse} {
    global SadieVar
      
    upvar #0 $imgName imgarray

    set_current_image $imgName

    set coords [convert_mouse_coordinates $imgName $xmouse $ymouse]
    set xpos [lindex $coords 0]
    set ypos [lindex $coords 1]
    
    update_coordinates $imgName $xpos $ypos
    
    set SadieVar(tools,point1x) $xpos
    set SadieVar(tools,point1y) $ypos
    set SadieVar(tools,point2x) $xpos
    set SadieVar(tools,point2y) $ypos
    set SadieVar(tools,draw) 1

    if {$SadieVar(tools,shape) == "point"} {
        definectrlpt_dialog
    }
}

#
# image_mouse_buttonup1 --
#
#       React to mouse left button release over an image window.
#
# Arguments:
#       imgName -- The name of the global image information array
#       xmouse -- current x-position in the canvas
#       ymouse -- current y-position in the canvas
#
# Results:
#       Completes selection drawing and calls appropriate procedure
#       to use defined selection.
#
proc image_mouse_buttonup1 {imgName xmouse ymouse} {
    global SadieVar
      
    upvar #0 $imgName imgarray

    set coords [convert_mouse_coordinates $imgName $xmouse $ymouse]
    set xpos [lindex $coords 0]
    set ypos [lindex $coords 1]
    
    update_coordinates $imgName $xpos $ypos
    
    if {![info exists SadieVar(tools,point1x)] || ![info exists SadieVar(tools,point1y)]} {
	return 0
    }

    set SadieVar(tools,point2x) $xpos
    set SadieVar(tools,point2y) $ypos
    set SadieVar(tools,draw) 0

    if {$SadieVar(tools,point2x) != $SadieVar(tools,point1x) \
            || $SadieVar(tools,point2y) != $SadieVar(tools,point1y)} {
        
        if { $SadieVar(tools,shape) == "line" } {
	    defineprofile_dialog
        } elseif { $SadieVar(tools,shape) == "rectangle" } {

	    if {$SadieVar(tools,point2x) != $SadieVar(tools,point1x) \
		    || $SadieVar(tools,point2y) != $SadieVar(tools,point1y)} {

		global WinsubsampVar

                if {$SadieVar(tools,point1x) < $SadieVar(tools,point2x)} {
                    set WinsubsampVar(from,pix) [expr $SadieVar(tools,point1x) + 1]
                    set WinsubsampVar(to,pix) [expr $SadieVar(tools,point2x) + 1]
                } else {
                    set WinsubsampVar(from,pix) [expr $SadieVar(tools,point2x) + 1]
                    set WinsubsampVar(to,pix) [expr $SadieVar(tools,point1x) + 1]
                }

                if {$SadieVar(tools,point1y) < $SadieVar(tools,point2y)} {
                    set WinsubsampVar(from,line) [expr $SadieVar(tools,point1y) + 1]
                    set WinsubsampVar(to,line) [expr $SadieVar(tools,point2y) + 1]
                } else {
                    set WinsubsampVar(from,line) [expr $SadieVar(tools,point2y) + 1]
                    set WinsubsampVar(to,line) [expr $SadieVar(tools,point1y) + 1]
                }

                set WinsubsampVar(from,band) 1
                set WinsubsampVar(to,band) $imgarray(bands)
                set WinsubsampVar(incr,band) 1
                set WinsubsampVar(incr,line) 1
                set WinsubsampVar(incr,pix) 1

                winsubsamp_dialog
            }

        } elseif { $SadieVar(tools,shape) == "poly" } {
	    defineregion_dialog
        }
    }
}


#
# draw_shape --
#
#       Draw a shape (rectangle, point, line, or polygon) on a display
#       image.
#
# Arguments:
#       imgName -- The name of the global image information array
#
# Results:
#       If a shape is already drawn on the display image, delete it.
#       Then draw the new shape.
#
proc draw_shape {imgName} {
    global SadieVar
    
    upvar #0 $imgName imgarray

    # make sure only one shape appears at a time
    if {[info exists imgarray(shape)]} {
	$imgarray(window).imgcanvas delete $imgarray(shape)
	unset imgarray(shape)
    }

    # Adjust the coordinates for zoom value
    set x1 [expr $SadieVar(tools,point1x) * $imgarray(zoom)]
    set y1 [expr $SadieVar(tools,point1y) * $imgarray(zoom)]
    set x2 [expr $SadieVar(tools,point2x) * $imgarray(zoom)]
    set y2 [expr $SadieVar(tools,point2y) * $imgarray(zoom)]
    
    if { $SadieVar(tools,shape) == "line" } {
	set imgarray(shape) [$imgarray(window).imgcanvas create line \
		$x1 $y1 $x2 $y2 \
		-width $imgarray(zoom) -fill yellow]
    } elseif { $SadieVar(tools,shape) == "rectangle" } {
	set imgarray(shape) [$imgarray(window).imgcanvas create rectangle \
		$x1 $y1 $x2 $y2 \
		-width $imgarray(zoom) -outline yellow]
    } elseif { $SadieVar(tools,shape) == "poly" } {
	set imgarray(shape) [$imgarray(window).imgcanvas create rectangle \
		$x1 $y1 $x2 $y2 \
		-width $imgarray(zoom) -outline yellow]
    }
}

#
# convert_mouse_coordinates --
#
#       Convert coordinates returned from mouse position in a display
#       image canvas to actual image coordinates.
#
# Arguments:
#       imgName -- The name of the global image information array
#       xmouse -- current mouse x-position in the canvas
#       ymouse -- current mouse y-position in the canvas
#
# Results:
#       Returns the cooresponding image x and y coordinates, after
#       adjusting for zoom value.
#
proc convert_mouse_coordinates {imgName xmouse ymouse} {
    upvar #0 $imgName imgarray

    set canv_name $imgarray(window).imgcanvas
    set canv_xpos [$canv_name canvasx $xmouse]
    set canv_ypos [$canv_name canvasy $ymouse]

    set xpos [expr int($canv_xpos / $imgarray(zoom))]
    set ypos [expr int($canv_ypos / $imgarray(zoom))]

    if {$xpos >= [image width $imgarray(tk_photo)]} {
	set xpos [expr [image width $imgarray(tk_photo)] - 1]
    } elseif {$xpos <=0} {
	set xpos 0
    }

    if {$ypos >= [image height $imgarray(tk_photo)]} {
	set ypos [expr [image height $imgarray(tk_photo)] - 1]
    } elseif {$ypos <=0} {
	set ypos 0
    }

    return "$xpos $ypos"
}

#
# update_coordinates --
#
#       Update coordinate and pixel readouts.
#
# Arguments:
#       imgName -- The name of the global image information array
#       xpos -- current x-position in the image
#       ypos -- current y-position in the image
#
# Results:
#       The main tclSadie window is updated wwith the current coordinates
#       and pixel readouts for the display image.
#
proc update_coordinates {imgName xpos ypos} {
    global SadieVar
    
    upvar #0 $imgName imgarray

    set sadie [Sadie_General_GetPixel $imgarray(sadie_addr) $xpos $ypos]
    set s_rval [lindex $sadie 0]
    set s_gval [lindex $sadie 1]
    set s_bval [lindex $sadie 2]

    set display [$imgarray(tk_photo) get $xpos $ypos]
    set d_rval [lindex $display 0]
    set d_gval [lindex $display 1]
    set d_bval [lindex $display 2]

    set SadieVar(label,xpos) " X: [expr $xpos + 1]"
    set SadieVar(label,ypos) " Y: [expr $ypos + 1]"

    if {$imgarray(bands) == 3} {
	set SadieVar(label,s_rval) " R: $s_rval"
	set SadieVar(label,s_gval) " G: $s_gval"
	set SadieVar(label,s_bval) " B: $s_bval"

	set SadieVar(label,d_rval) " R: $d_rval"
	set SadieVar(label,d_gval) " G: $d_gval"
	set SadieVar(label,d_bval) " B: $d_bval"
    } else {
	set SadieVar(label,s_rval) " Z: $s_rval"
	set SadieVar(label,s_gval) "   "
	set SadieVar(label,s_bval) "   "

	set SadieVar(label,d_rval) " Z: $d_rval"
	set SadieVar(label,d_gval) "   "
	set SadieVar(label,d_bval) "   "
    }
}

proc get_shape_coordinates {} {
    global SadieVar
      
    return "$SadieVar(tools,point1x) $SadieVar(tools,point1y) $SadieVar(tools,point2x) $SadieVar(tools,point2y)"
}

#
# set_current_image --
#
#       Make an image the current image.
#
# Arguments:
#       imgName -- The name of the global image information array
#
# Results:
#       Moves $imgName to the top of the global image list and
#       updates menus, windows, etc. to reflect this change.
#
proc set_current_image {imgName} {
    global SadieVar
      
    #
    # Make sure that $imgName is in the image list
    #
    set imgName_index [lsearch -exact $SadieVar(imagelist) $imgName]
    if {$imgName_index < 0} {
        error "The image array \"$imgName\" is not included in the image list \"$listName\""
    }
    

    upvar #0 $imgName imgarray


    # Move window to top of window manager in case it is buried
    raise $imgarray(window)
    
    #
    # Only proceed if $imgName is not at the top of the list
    #
    if {$imgName_index > 0} {
        #
        # First, handle the original current image
        #
        unhighlight_current_image

        # Delete $imgName from the list
        set SadieVar(imagelist) [lreplace $SadieVar(imagelist) $imgName_index $imgName_index]
        # Reinsert $imgName at the top of the list
        set SadieVar(imagelist) [linsert $SadieVar(imagelist) 0 $imgName]

	# turn on current image indicator
	highlight_current_image
    }

    # Update the GUI
    update_gui

}



#
# get_current_image --
#
#       Get the current image.
#
# Arguments:
#       None
#
# Results:
#       Return the array name of the current image, if the global image
#       list is not empty.  Otherwise, return "none".
#
proc get_current_image {} {
    global SadieVar
      
    if {[llength $SadieVar(imagelist)] > 0} {
        set curimg [lindex $SadieVar(imagelist) 0]
	return $curimg
    } else {
	return "none"
    }
}


#
# highlight_current_image --
#
#       Highlight the current image.
#
# Arguments:
#       None
#
# Results:
#       Highlights the current image, if the global image list is
#       not empty.
#
proc highlight_current_image {} {
    global SadieVar
      
    if {[llength $SadieVar(imagelist)] > 0} {
        set curimg [lindex $SadieVar(imagelist) 0]
        upvar #0 $curimg imgarray


        # Move window to top of window manager in case it is buried
        raise $imgarray(window)

        # turn on current image indicator
        $imgarray(window) configure -bg blue
    }
}

#
# unhighlight_current_image --
#
#       Make updates for an image that is no longer current.
#
# Arguments:
#       None
#
# Results:
#       Unhighlights the current image, if the image list is not
#       empty.
#
proc unhighlight_current_image {} {
    global SadieVar
	
    if {[llength $SadieVar(imagelist)] > 0} {
        set oldcurimg [lindex $SadieVar(imagelist) 0]
        upvar #0 $oldcurimg imgarray

        # delete any drawings from the image
        if {[info exists imgarray(shape)]} {
          $imgarray(window).imgcanvas delete $imgarray(shape)
          unset imgarray(shape)
        }
        
        # turn off current image indicator
        $imgarray(window) configure -bg [. cget -bg]
    }
}

#
# delete_image --
#
#       Delete an active image.
#
# Arguments:
#       imgName -- The name of the global image information array
#
# Results:
#       Deletes $imgName and associated variables from memory and
#       updates menus, windows, etc. to reflect this change.
#
proc delete_image {{imgName none}} {
    global SadieVar
    
    #
    # Only proceed if the global image list is non-empty
    #  
    if {[llength $SadieVar(imagelist)] > 0} {

        if {$imgName == "none"} {
            #
            # By default, delete current image
            #
            set imgName [lindex $SadieVar(imagelist) 0]
        }


        #
        # Make sure that $imgName is in the image list
        #
        set imgName_index [lsearch -exact $SadieVar(imagelist) $imgName]
        if {$imgName_index < 0} {
            error "The image array \"$imgName\" is not included in the image list \"$listName\""
        }

        upvar #0 $imgName imgarray

        # Delete the window and associated widgets
        destroy $imgarray(window)

        # Delete the tk_photo
        image delete $imgarray(tk_photo)

        # Delete the SADIE image
        Sadie_General_Close $imgarray(sadie_addr)

        # Delete image from displayed window list/menu  
        $SadieVar(menu,winlist) delete $imgarray(title)

        # Remove the image array name from the global image list
        set SadieVar(imagelist) [lreplace $SadieVar(imagelist) $imgName_index $imgName_index]

        # Delete the image array
        unset imgarray

        # Update the GUI
        update_gui

        highlight_current_image
    }
}


#
# delete_all_images --
#
#       Delete all active images.
#
# Arguments:
#       none
#
# Results:
#       Deletes all images and associated variables from memory and
#       updates menus, windows, etc. to reflect this change.
#
proc delete_all_images {} {
    global SadieVar

    foreach imgName $SadieVar(imagelist) {
        delete_image $imgName
    }

    # Update the GUI
    update_gui
}


#
# change_scaling --
#
#       Toggle the scaling property of the current image.
#
# Arguments:
#       None
#
# Results:
#       If the global image list is not empty, change the scaling
#       property of the image at the top of the list.
#
proc change_scaling {} {
    global SadieVar

    if {[llength $SadieVar(imagelist)] > 0} {
        set imgName [lindex $SadieVar(imagelist) 0]

        upvar #0 $imgName imgarray

        set imgarray(scaling) $SadieVar(getstretched)        
  
        image delete $imgarray(tk_photo)
        image create photo $imgarray(tk_photo) -format sadie -data $imgarray(sadie_addr)

	update_display_image $imgName
    }
}

#
# change_showregions --
#
#       Toggle the show regions property of the current image.
#
# Arguments:
#       None
#
# Results:
#       If the global image list is not empty, change the showregions
#       property of the image at the top of the list.
#
proc change_showregions {} {
    global SadieVar

    if {[llength $SadieVar(imagelist)] > 0} {
        set imgName [lindex $SadieVar(imagelist) 0]

        upvar #0 $imgName imgarray

        set imgarray(showregions) $SadieVar(showregions)        
  
	update_display_image $imgName
    }
}


#
# change_showctrlpts --
#
#       Toggle the show control points property of the current image.
#
# Arguments:
#       None
#
# Results:
#       If the global image list is not empty, change the showctrlpts
#       property of the image at the top of the list.
#
proc change_showctrlpts {} {
    global SadieVar

    if {[llength $SadieVar(imagelist)] > 0} {
        set imgName [lindex $SadieVar(imagelist) 0]

        upvar #0 $imgName imgarray

        set imgarray(showctrlpts) $SadieVar(showctrlpts)        
  
	update_display_image $imgName
    }
}


#
# zoom_in --
#
#       Zoom in on the current display image
#
# Arguments:
#       None
#
# Results:
#       If the global image list is not empty, double the zoom value
#       of the current display image, up to a maximum magnification
#       of 32X.
#
proc zoom_in {} {
    global SadieVar

    if {[llength $SadieVar(imagelist)] > 0} {
        set imgName [lindex $SadieVar(imagelist) 0]

        upvar #0 $imgName imgarray

	if {$imgarray(zoom) < 32} {

	    set imgarray(zoom) [expr 2 * $imgarray(zoom)]
	    
	    #
	    # Find the starting and ending points of the visible region
	    # of the canvas.
	    #
	    set beginxview [lindex [$imgarray(window).imgcanvas xview] 0]
	    set beginyview [lindex [$imgarray(window).imgcanvas yview] 0]
	    set endxview [lindex [$imgarray(window).imgcanvas xview] 1]
	    set endyview [lindex [$imgarray(window).imgcanvas yview] 1]

	    #
	    # Compute the ending points of the visible region of the 
	    # canvas after the zoom is performed.
	    # 
	    set newendxview [expr $beginxview + \
		    [expr double($endxview - $beginxview) / 4.0]]
	    set newendyview [expr $beginyview + \
		    [expr double($endyview - $beginyview) / 4.0]]

	    #
	    # Double the size of the canvas
	    #
	    set h [expr [image height $imgarray(tk_photo)] * $imgarray(zoom)]
	    set w [expr [image width $imgarray(tk_photo)] * $imgarray(zoom)]
	    $imgarray(window).imgcanvas configure \
		    -height $h \
		    -width $w \
		    -scrollregion "0 0 $w $h"
            
	    $imgarray(window).imgcanvas xview moveto $newendxview
	    $imgarray(window).imgcanvas yview moveto $newendyview
	    
	    #
	    # Set a new maximum window size that is double the previous value
	    #
	    set maxwidth [expr [$imgarray(window).imgcanvas cget -width] \
		    + [$imgarray(window).vscr cget -width] \
		    + [expr 3 * [$imgarray(window).vscr cget -borderwidth]] \
		    + [expr 2 * [$imgarray(window) cget -borderwidth]]]
	    set maxheight [expr [$imgarray(window).imgcanvas cget -height] \
		    + [$imgarray(window).hscr cget -width] \
		    + [expr 3 * [$imgarray(window).hscr cget -borderwidth]] \
		    + [expr 2 * [$imgarray(window) cget -borderwidth]]]
	    wm maxsize $imgarray(window) $maxwidth $maxheight
	    wm minsize $imgarray(window) $imgarray(zoom) $imgarray(zoom)
	    
	    #
	    # Constrain the scroll increment to the zoom value to ensure
	    # that partial pixels are not visible at the edges of the canvas.
	    #
	    $imgarray(window).imgcanvas configure \
		    -xscrollincrement $imgarray(zoom) \
		    -yscrollincrement $imgarray(zoom)

	    #
	    # Perform updates.
	    #
	    update_display_image $imgName
	    constrain_size $imgName        
	    update_gui

	}
    }
}


#
# zoom_out --
#
#       Zoom out on the current display image
#
# Arguments:
#       None
#
# Results:
#       If the global image list is not empty, half the zoom value
#       of the current display image, down to a minimum magnification
#       of 1X.
#
proc zoom_out {} {
    global SadieVar

    if {[llength $SadieVar(imagelist)] > 0} {
        set imgName [lindex $SadieVar(imagelist) 0]

        upvar #0 $imgName imgarray

	if {$imgarray(zoom) > 1} {

	    set imgarray(zoom) [expr $imgarray(zoom) / 2]
	    
	    #
	    # Find the starting and ending points of the visible region
	    # of the canvas.
	    #
	    set beginxview [lindex [$imgarray(window).imgcanvas xview] 0]
	    set beginyview [lindex [$imgarray(window).imgcanvas yview] 0]
	    set endxview [lindex [$imgarray(window).imgcanvas xview] 1]
	    set endyview [lindex [$imgarray(window).imgcanvas yview] 1]

	    #
	    # Compute the ending points of the visible region of the 
	    # canvas after the zoom is performed.
	    # 
	    set newendxview [expr $beginxview - \
		    [expr double($endxview - $beginxview) / 4.0]]
	    set newendyview [expr $beginyview - \
		    [expr double($endyview - $beginyview) / 4.0]]

	    #
	    # Half the size of the canvas
	    #
	    set h [expr [image height $imgarray(tk_photo)] * $imgarray(zoom)]
	    set w [expr [image width $imgarray(tk_photo)] * $imgarray(zoom)]
	    $imgarray(window).imgcanvas configure \
		    -height $h \
		    -width $w \
		    -scrollregion "0 0 $w $h"
            
	    $imgarray(window).imgcanvas xview moveto $newendxview
	    $imgarray(window).imgcanvas yview moveto $newendyview
	    
	    # 
	    # Set a new maximum window size that is double the previous value
	    #
	    set maxwidth [expr [$imgarray(window).imgcanvas cget -width] \
		    + [$imgarray(window).vscr cget -width] \
		    + [expr 3 * [$imgarray(window).vscr cget -borderwidth]] \
		    + [expr 2 * [$imgarray(window) cget -borderwidth]]]
	    set maxheight [expr [$imgarray(window).imgcanvas cget -height] \
		    + [$imgarray(window).hscr cget -width] \
		    + [expr 3 * [$imgarray(window).hscr cget -borderwidth]] \
		    + [expr 2 * [$imgarray(window) cget -borderwidth]]]

	    wm maxsize $imgarray(window) $maxwidth $maxheight
	    wm minsize $imgarray(window) $imgarray(zoom) $imgarray(zoom)
	    
	    #
	    # Constrain the scroll increment to the zoom value to ensure
	    # that partial pixels are not visible at the edges of the canvas.
	    #
	    $imgarray(window).imgcanvas configure \
		    -xscrollincrement $imgarray(zoom) \
		    -yscrollincrement $imgarray(zoom)

	    #
	    # Perform updates.
	    #
	    update_display_image $imgName
	    constrain_size $imgName
	    update_gui
	}
    }
}


#
# update_display_image --
#
#       Update the view of the display image to reflect a new
#       zoom value.
#
# Arguments:
#       imgName -- The name of the image array to update.
#
# Results:
#       Update the displayed tk photo image to reflect a new
#       zoom setting.  To conserve memory, make the tk photo
#       only as large as the visible size of the canvas and
#       update each time the window size changes or a scroll
#       occurs.
#
proc update_display_image {imgName} {
    global SadieVar

    upvar #0 $imgName imgarray

    #
    # Remove the old display image from the canvas
    #
    $imgarray(window).imgcanvas delete disp_img

    #
    # Determine the visible portion of the canvas
    #
    set beginxview [lindex [$imgarray(window).imgcanvas xview] 0]
    set beginyview [lindex [$imgarray(window).imgcanvas yview] 0]
    set endxview [lindex [$imgarray(window).imgcanvas xview] 1]
    set endyview [lindex [$imgarray(window).imgcanvas yview] 1]

    #
    # Determine the coordinates of the image that will be
    # displayed in the visible portion of the canvas.
    #
    set extract_x1 [expr round($beginxview * [image width $imgarray(tk_photo)])]
    set extract_y1 [expr round($beginyview * [image height $imgarray(tk_photo)])]
    set extract_x2 [expr round($endxview * [image width $imgarray(tk_photo)]) + 1]
    set extract_y2 [expr round($endyview * [image height $imgarray(tk_photo)]) + 1]
    if {[image width $imgarray(tk_photo)] < $extract_x2} {
        set extract_x2 [image width $imgarray(tk_photo)]
    }
    if {[image height $imgarray(tk_photo)] < $extract_y2} {
        set extract_y2 [image height $imgarray(tk_photo)]
    }

    #
    # Clear the old zoomed tk photo and recreate it with the
    # new zoomed view.
    #
    $imgarray(tk_photo_zoom) blank
    $imgarray(tk_photo_zoom) copy $imgarray(tk_photo) \
        -from $extract_x1 $extract_y1 $extract_x2 $extract_y2 \
        -zoom $imgarray(zoom) $imgarray(zoom)

    #
    # Insert the new display image into the canvas
    #
    set w [$imgarray(window).imgcanvas cget -width]
    set h [$imgarray(window).imgcanvas cget -height]
    $imgarray(window).imgcanvas create image \
        [expr $beginxview * $w] \
        [expr $beginyview * $h] \
        -image $imgarray(tk_photo_zoom) \
        -anchor nw -tags disp_img

    #
    # Redraw the defined regions, if they are visible
    #
    if {$imgarray(showregions)} {
	draw_regions $imgName
    } else {
	$imgarray(window).imgcanvas delete regions
    }

    #
    # Redraw the defined control points, if they are visible
    #
    if {$imgarray(showctrlpts)} {
	draw_ctrlpts $imgName
    } else {
	$imgarray(window).imgcanvas delete ctrlpts
    }

    #
    # Redraw the shape, if it is present
    #
    if {[info exists imgarray(shape)]} {
	draw_shape $imgName
    }
}


#
# update_xscroll --
#
#       Handle a scroll change in the x-direction for the
#       image canvas.
#
# Arguments:
#       imgName -- The name of the image array to update.
#       args -- The arguments returned by the
#                  canvas -xscrollcommand
#
# Results:
#       Update the scroll bars and canvas position, and then
#       update the displayed (and possibly zoomed) tk photo
#       image.
#
proc update_xscroll {imgName args} {
    upvar #0 $imgName imgarray
    
    eval $imgarray(window).hscr set $args
    update_display_image $imgName
}

#
# update_yscroll --
#
#       Handle a scroll change in the y-direction for the
#       image canvas.
#
# Arguments:
#       imgName -- The name of the image array to update.
#       args -- The arguments returned by the
#                  canvas -yscrollcommand
#
# Results:
#       Update the scroll bars and canvas position, and then
#       update the displayed (and possibly zoomed) tk photo
#       image.
#
proc update_yscroll {imgName args} {
    upvar #0 $imgName imgarray
    
    eval $imgarray(window).vscr set $args
    update_display_image $imgName
}

#
# constrain_size --
#
#       Constrain the size of the window that displays an image
#       to ensure that the canvas width and height are multiples
#       of the zoom factor.  This procedure is bound to resize
#       events (i.e. <Configure>) for image display windows.
#
# Arguments:
#       imgName -- The name of the image array to constrain.
#
# Results:
#       Records the time of the request for a size constraint on the
#       window, and sets up a call to "test_constrain" after 200ms.
#       This is necessary to avoid excessive size constraints when
#       a user is interactively resizing a window.  Also, if this
#       procedure is called due to a resize generated in the
#       "do_constrain" procedure, do not continue.
#
proc constrain_size {imgName} {
    global SadieVar

    if {![info exists SadieVar(constrain,recursive,$imgName)]} {
	set SadieVar(constrain,recursive,$imgName) 0
    }

    # If this procedure was called in a recursive fashion due to a computer
    # generated resize event in the "do_constrain" procedure, exit now
    # to avoid setting up an infinite loop.
    if {$SadieVar(constrain,recursive,$imgName) == 1} {
	return
    }

    if {![info exists SadieVar(constrain,waiting,$imgName)]} {
	set SadieVar(constrain,waiting,$imgName) 0
    }

    set time [clock clicks]

    # Record the time of the current request for a size constraint
    set SadieVar(constrain,last,$imgName) $time

    if {!$SadieVar(constrain,waiting,$imgName)} {
	# Generate a call to "test_constrain" after waiting
	# for 200ms.
	set SadieVar(constrain,waiting,$imgName) 1
	set id [after 200 test_constrain $imgName]
    }
}

#
# test_constrain --
#
#       Constrain the size of a window after ensuring that multiple
#       size constraint requests are not being generated multiple
#       times in rapid succession (i.e. during an interactive resize
#       of a window by a user).
#
# Arguments:
#       imgName -- The name of the image array to constrain.
#
# Results:
#       If a resize request has not been generated in the past
#       410ms, continue on with the size constraint by calling
#       the "do_constrain" procedure.  Otherwise, wait for
#       200ms and try again.
#
proc test_constrain {imgName} {
    global SadieVar

    set time [clock clicks]

    set delta [expr $time - $SadieVar(constrain,last,$imgName)]

    if {$delta < 410} {
	set id [after 200 test_constrain $imgName]
    } else {
	set SadieVar(constrain,waiting,$imgName) 0
	do_constrain $imgName
    }
}

#
# do_constrain --
#
#       Constrain the size of the window that displays an image
#       to ensure that the canvas width and height are multiples
#       of the zoom factor, after repetitive resize events have
#       been filtered out by the "test_constrain" procedure.
#
# Arguments:
#       imgName -- The name of the image array to constrain.
#
# Results:
#       If the canvas width/height is not a multiple of the zoom factor,
#       its value is changed to one of the following:
#          1) The zoom value
#          2) The next smaller or larger multiple of the zoom factor
#               (whichever is closer)
#          3) The width/height of the zoomed image
#
proc do_constrain {imgName} {
    global SadieVar
    
    upvar #0 $imgName imgarray

    update
    
    #
    # Determine the current size of the window
    #
    set winwidth [winfo width $imgarray(window)]
    set winheight [winfo height $imgarray(window)]

    #
    # Determine the current size of the canvas
    #
    set canvwidth [expr $winwidth \
	    - [$imgarray(window).vscr cget -width] \
	    - [expr 3 * [$imgarray(window).vscr cget -borderwidth]] \
	    - [expr 2 * [$imgarray(window) cget -borderwidth]]]
    set canvheight [expr $winheight \
	    - [$imgarray(window).hscr cget -width] \
	    - [expr 3 * [$imgarray(window).hscr cget -borderwidth]] \
	    - [expr 2 * [$imgarray(window) cget -borderwidth]]]

    if {$canvwidth <  $imgarray(zoom)} {
	# Ensure that at least one pixel is displayed in the x-dir
	set new_canvwidth $imgarray(zoom)
    } elseif {$canvwidth > [$imgarray(window).imgcanvas cget -width]} {
	# Ensure that the canvas is no wider than the display image
	set new_canvwidth [$imgarray(window).imgcanvas cget -width]
    } elseif {[expr $canvwidth % $imgarray(zoom)] < [expr $imgarray(zoom) / 2.0]} {
	# Drop the canvas width to the next smallest multiple of
	# the zoom factor
	set new_canvwidth [expr $canvwidth - ($canvwidth % $imgarray(zoom))]
    } else {
	# Increase the canvas width to the next larger multiple of
	# the zoom factor
	set new_canvwidth [expr $canvwidth + ($imgarray(zoom) - ($canvwidth % $imgarray(zoom)))]
    }
    
    if {$canvheight < $imgarray(zoom)} {
	# Ensure that at least one pixel is displayed in the y-dir
	set new_canvheight $imgarray(zoom)
    } elseif {$canvheight > [$imgarray(window).imgcanvas cget -height]} {
	# Ensure that the canvas is no taller than the display image
	set new_canvheight [$imgarray(window).imgcanvas cget -height]
    } elseif {[expr $canvheight % $imgarray(zoom)] < [expr $imgarray(zoom) / 2.0]} {
	# Drop the canvas height to the next smallest multiple of
	# the zoom factor
	set new_canvheight [expr $canvheight - ($canvheight % $imgarray(zoom))]
    } else {
	# Increase the canvas height to the next larger multiple of
	# the zoom factor
	set new_canvheight [expr $canvheight + ($imgarray(zoom) - ($canvheight % $imgarray(zoom)))]
    }

    #
    # Compute the new window size based on the new canvas size
    #
    set new_winwidth [expr $new_canvwidth \
	    + [$imgarray(window).vscr cget -width] \
	    + [expr 3 * [$imgarray(window).vscr cget -borderwidth]] \
	    + [expr 2 * [$imgarray(window) cget -borderwidth]]]
    set new_winheight [expr $new_canvheight \
	    + [$imgarray(window).hscr cget -width] \
	    + [expr 3 * [$imgarray(window).hscr cget -borderwidth]] \
	    + [expr 2 * [$imgarray(window) cget -borderwidth]]]

    #
    # Avoid setting up an infinite loop by only updating the geometry if a 
    # change is necessary
    #
    if {$new_winwidth != $winwidth || $new_winheight != $winheight} {
	# To avoid creating an infinite loop of resizes on the window,
	# indicate that the following resize event is generated by the computer
	# and may cause a recursive call to the constrain_size procedures.
	set SadieVar(constrain,recursive,$imgName) 1

	# Tell the window manager to treat the resize as if a user had caused it
	wm sizefrom $imgarray(window) user
	wm geometry $imgarray(window) ${new_winwidth}x${new_winheight}
	update

	# Reset the indicator since any recursive calls should have completed after
	# the update
	set SadieVar(constrain,recursive,$imgName) 0
    }
}

#
# reload_sadie_image --
#
#       Reload a tclSadie display image from the corresponding
#       SADIE image.
#
# Arguments:
#       imgName -- The name of the image array
#
# Results:
#       The current display image is destroyed, and then reloaded
#       from its sadie address.
#
proc reload_sadie_image {imgName} {
    upvar #0 $imgName imgarray

    image delete $imgarray(tk_photo)
    image create photo $imgarray(tk_photo) -format sadie -data $imgarray(sadie_addr)
    update_display_image $imgName
    tk_messageBox -type ok -message "Warning: The image:\n   $imgarray(title)\nmay have changed!"
}


#
# create_image_selector --
#
#       Create an image selection widget.
#
# Arguments:
#       button_name -- The name of the widget.
#       variable_name -- The name of the global variable in
#           which to store image sadie address.
#       index -- If $variable_name is an array, this is its
#           index.
#       imgnum -- The number of the initial image to display
#           in the widget.
#
# Results:
#       A menubutton and associated menu are created and initialized
#       with the images from the global image list.  The title of images
#       are displayed in the widget, while the sadie address of the
#       active image is stored in the supplied global variable.
#
proc create_image_selector {button_name variable_name {index "none"} {imgnum 0}} {
    global SadieVar    
    
    #
    # Create the widget
    #
    set m "${button_name}.menu"
    menubutton $button_name -menu $m -relief raised -borderwidth 3 \
        -justify left -anchor w
    menu $m -tearoff 0
    
    foreach imagename $SadieVar(imagelist) {
        upvar #0 $imagename imgarray
        
        $m add command -label $imgarray(title) \
            -command "update_image_selector $button_name $variable_name \
                $index \"$imgarray(title)\" $imgarray(sadie_addr)"
        
    }
    
    #
    # Initialize the widget
    #
    if {[llength $SadieVar(imagelist)] < [expr $imgnum + 1]} {
        set imgnum 0
    }
    
    upvar #0 [lindex $SadieVar(imagelist) $imgnum] imgarray
    $button_name configure -text $imgarray(title)
    upvar #0 $variable_name var
    if {$index != "none"} {
        set var($index) $imgarray(sadie_addr)
    } else {
        set var $imgarray(sadie_addr)
    }
}

#
# update_image_selector --
#
#       Update an image selection widget.
#
# Arguments:
#       button_name -- The name of the widget.
#       variable_name -- The name of the global variable in
#           which to store image sadie address.
#       index -- If $variable_name is an array, this is its
#           index.
#       title -- The title of the image that is to be made
#           active.
#       address -- The sadie address of the image that is
#           to be made active.
#
# Results:
#       The image selection widget is updated by setting the 
#       label of the button to $title and the value of 
#       $variablename($index) to $address.
#
proc update_image_selector {button_name variable_name index title address} {
    $button_name configure -text $title
    
    upvar #0 $variable_name var
    if {$index != "none"} {
        set var($index) $address
    } else {
        set var $address
    }
}


#
# create_ctrlpt_selector --
#
#       Create a control point selection widget.
#
# Arguments:
#       button_name -- The name of the widget.
#       variable_name -- The name of the global variable in
#           which to store control point coordinates.
#       x_index -- The index of the x-coordinate.
#       y_index -- The index of the y-coordinate.
#       control_Img -- Image name to pick control points from.
#
# Results:
#       A menubutton and associated menu are created and initialized
#       with the control points for images from the global image list.
#
proc create_ctrlpt_selector {button_name variable_name x_index y_index control_Img} {
    global SadieVar
    global ControlptVar
    
    #
    # Create the widget
    #
    set m "${button_name}.menu"
    menubutton $button_name -menu $m -relief raised -borderwidth 1 \
        -text "Set"
    menu $m -tearoff 0

    set j 0
    foreach imagename $SadieVar(imagelist) {
        upvar #0 $imagename imgarray
	incr j
		
	if { $imgarray(title) == $control_Img } {
	    if { $imgarray(ctrlpt,count) > 0 } {
		# We don't need the following lines anymore since we decide
		# on the reference and transformation images before hand.
		# So we only need to show control points in the drop-down list.
		
		# set s $m.submenu$j
		# $m add cascade -label $imgarray(title) -menu $s
		# menu $s -tearoff 0
		
		for {set i 1} {$i <= $imgarray(ctrlpt,count)} {incr i} {
		    $m add command -label $imgarray(ctrlpt,$i,name) \
			    -command [subst {
			upvar #0 $variable_name var
			set var($x_index) [expr [lindex $imgarray(ctrlpt,$i,coords) 0] + 1]
			set var($y_index) [expr [lindex $imgarray(ctrlpt,$i,coords) 1] + 1 ]
		    }]
		}
	    } else {
		bind $button_name <Button-1> {
		    tk_messageBox -type ok -title "Warning" -icon info \
			    -message "No control points defined in this image."
		}
	    }
	}
    }
}


#
# create_new_region --
#
#       Define a new region in a display image.
#
# Arguments:
#       imgName -- The name of the image array to update.
#       x1,y1,x2,y2 -- The boundaries of the new region.
#       regionName -- The name of the new region.
#
# Results:
#       The new region is defined and added to the image
#       information array.
#
proc create_new_region {imgName x1 y1 x2 y2 regionName} {
    global SadieVar

    upvar #0 $imgName imgarray
# puts stderr "just goes into create_new_region\n"
    incr imgarray(region,count)
    set n $imgarray(region,count)
    
    set imgarray(region,$n,name) $regionName
    set imgarray(region,$n,coords) "$x1 $y1 $x2 $y2"

    # Delete the temporary selection region
    if {[info exists imgarray(shape)]} {
	$imgarray(window).imgcanvas delete $imgarray(shape)
	unset imgarray(shape)
    }
#puts stderr "before update_display_image\n"
    update_display_image $imgName
    update_gui
#puts stderr "create_new_region finished\n"
}


#
# draw_regions --
#
#       Draw defined regions to the screen so that they appear in
#       their parent display image.
#
# Arguments:
#       imgName -- The name of the image array to update.
#
# Results:
#       All of the defined regions for the image are drawn to the
#       screen at the appropriate zoom setting.
#
proc draw_regions {imgName} {
    global SadieVar

    upvar #0 $imgName imgarray

    $imgarray(window).imgcanvas delete regions

    for {set n 1} {$n <= $imgarray(region,count)} {incr n} {
	set x1 [expr [lindex $imgarray(region,$n,coords) 0] * $imgarray(zoom)]
	set y1 [expr [lindex $imgarray(region,$n,coords) 1] * $imgarray(zoom)]
	set x2 [expr [lindex $imgarray(region,$n,coords) 2] * $imgarray(zoom)]
	set y2 [expr [lindex $imgarray(region,$n,coords) 3] * $imgarray(zoom)]
	
	set imgarray(region,$n,object) [$imgarray(window).imgcanvas create rectangle \
		$x1 $y1 $x2 $y2 -tags regions -width $imgarray(zoom) \
		-outline $SadieVar(color,regions)]

	register_canvas_info $imgarray(window).imgcanvas $imgarray(region,$n,object) \
		"Region: \\\"$imgarray(region,$n,name)\\\""
    }

}





#
# create_new_ctrlpt --
#
#       Define a new control point in a display image.
#
# Arguments:
#       imgName -- The name of the image array to update.
#       x1,y1 -- The coordinates of the new ctrlpt.
#       ctrlptName -- The name of the new ctrlpt.
#
# Results:
#       The new ctrlpt is defined and added to the image
#       information array.
#
proc create_new_ctrlpt {imgName x1 y1 ctrlptName} {
    global SadieVar

    upvar #0 $imgName imgarray

    incr imgarray(ctrlpt,count)
    set n $imgarray(ctrlpt,count)
    
    set imgarray(ctrlpt,$n,name) $ctrlptName
    set imgarray(ctrlpt,$n,coords) "$x1 $y1"

    update_display_image $imgName
    update_gui
}


#
# draw_ctrlpts --
#
#       Draw defined ctrlpts to the screen so that they appear in
#       their parent display image.
#
# Arguments:
#       imgName -- The name of the image array to update.
#
# Results:
#       All of the defined ctrlpts for the image are drawn to the
#       screen at the appropriate zoom setting.
#
proc draw_ctrlpts {imgName} {
    global SadieVar

    upvar #0 $imgName imgarray

    $imgarray(window).imgcanvas delete ctrlpts

    for {set n 1} {$n <= $imgarray(ctrlpt,count)} {incr n} {
	set x1 [expr [lindex $imgarray(ctrlpt,$n,coords) 0] * $imgarray(zoom)]
	set y1 [expr [lindex $imgarray(ctrlpt,$n,coords) 1] * $imgarray(zoom)]
	
	set imgarray(ctrlpt,$n,object) [$imgarray(window).imgcanvas create rectangle \
		$x1 $y1 $x1 $y1 -tags ctrlpts -width $imgarray(zoom) \
		-outline $SadieVar(color,ctrlpts)]

	register_canvas_info $imgarray(window).imgcanvas $imgarray(ctrlpt,$n,object) \
		"Control Point: \\\"$imgarray(ctrlpt,$n,name)\\\""
    }

}


proc set_image_title {imgName title} {
    global SadieVar
    upvar #0 $imgName imgarray

    # Delete the old menu entry
    $SadieVar(menu,winlist) delete $imgarray(title)

    Sadie_General_SetName $imgarray(sadie_addr) $title
    set imgarray(title) $title
    wm title $imgarray(window) $imgarray(title)

    $SadieVar(menu,winlist) add command \
        -label $imgarray(title) \
        -command "set_current_image $imgName"

    update_gui
}

proc rename_image_dialog {} {
    global RenameVar
    global confirmStatus

    #
    # Determine the current image
    #
    set RenameVar(inimg,array) [get_current_image]
    if {$RenameVar(inimg,array) != "none"} {
	    upvar #0 $RenameVar(inimg,array) imgarray
	    set RenameVar(inimg,name) $imgarray(title)
	    set RenameVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }

    set index [string first ":" $RenameVar(inimg,name)]
    set RenameVar(number) [string range $RenameVar(inimg,name) 1 [expr $index - 1]]
    set RenameVar(outname) [string range $RenameVar(inimg,name) [expr $index + 2] end]
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .renamedlg]
    wm geometry $top +110+110
    wm title $top "Rename Image Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Rename the image: "
    label $title.imglbl -fg blue -text $RenameVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    #
    # Create the dialog output section
    #
    set output [dialog_output $top]

    label $output.outnamelbl -text "Image Name: "
    entry $output.outnameent -width 20 -textvariable RenameVar(outname)

    grid columnconfigure $output 1 -weight 1
    grid $output.outnamelbl -row 0 -column 0 -sticky w
    grid $output.outnameent -row 0 -column 1 -sticky ew
    
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus
    if {$confirmStatus == 1} {
        set_image_title $RenameVar(inimg,array) "#$RenameVar(number): $RenameVar(outname)"
    }


    destroy $top

}
















