#
# tclSadie_printing_init --
#
#       Initialize the parameters for the all of the printing dialogs.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters
#       for each dialog.
#
proc tclSadie_printing_init {} {
    print_sessionlog_init
    print_image_init
}



#
# print_sessionlog_dialog --
#
#       Create a dialog box for printing the session log.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc print_sessionlog_dialog { } {
    global PrintlogVar
    global confirmStatus

    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .printlogdlg]
    wm geometry $top +110+110
    wm title $top "Print Sessionlog Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Print the SessionLog"
    grid $title.titlelbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    radiobutton $input.printer \
	    -text "Send to printer with command: " \
	    -variable PrintlogVar(type) -value "printer"
    entry $input.printerCmd -textvariable PrintlogVar(command)
    radiobutton $input.file \
	    -text "Save output in file: " \
	    -variable PrintlogVar(type) -value "file"
    entry $input.fileName -textvariable PrintlogVar(file)

    grid $input.printer -row 0 -columnspan 3 -sticky w
    grid $input.printerCmd -row 1 -columnspan 3 -sticky ew
    grid rowconfigure $input 2 -minsize 6
    grid $input.file -row 3 -columnspan 3 -sticky w
    grid $input.fileName -row 4 -columnspan 3 -sticky ew

    bind $input.printerCmd <FocusIn> "$input.printer invoke"
    bind $input.fileName <FocusIn> "$input.file invoke"


    set row {}
    lappend row {label "Page X: "}
    lappend row {entry PrintlogVar(pagex)}
    lappend row {label "Page Y: "}
    lappend row {entry PrintlogVar(pagey)}
    dialog_create_stdrow $input 5 $row

    set row {}
    lappend row {label "Page Anchor: "}
    lappend row {entry PrintlogVar(anchor)}
    dialog_create_stdrow $input 6 $row
    
    #
    # Create the dialog output section
    #
    set output [dialog_output $top]
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus
    if {$confirmStatus == 1} {
        print_sessionlog_run    
    }


    destroy $top
}
#
# print_sessionlog_run --
#
#       Obtain postscript output from the sessionlog and either print it or save to file.
#
# Arguments:
#       None
#
# Results:
#       The postscript output is returned.
#
proc print_sessionlog_run {} {
    global PrintlogVar

    set x0 0
    set y0 0
    set x1 [winfo width .sessionlog.canvas]
    set y1 [winfo height .sessionlog.canvas]

    foreach {x0 y0 x1 y1} [.sessionlog.canvas bbox all] {}
    set w [expr $x1-$x0]
    set h [expr $y1-$y0]

    # Find the number of output pages
    set pagecount 0
    for {set j 0} {$j <= $h} {incr j 800} {
	for {set k 0} {$k <= $w} {incr k 580} {
	    if {[.sessionlog.canvas find enclosed $k $j [expr $k+579] [expr $j+799]] != ""} {
		incr pagecount
	    }
	}
    }


    
    set fid [open $PrintlogVar(file) w]

    # Create the postscript header information
    puts $fid "%!PS-Adobe-2.1"
    puts $fid "%%Title: tclSadie SessionLog output"
    puts $fid "%%Creator: tclSadie"
    puts $fid "%%Pages: $pagecount"
    puts $fid "%%EndComments"
    puts $fid "%%EndProlog"
    puts $fid "\n\n\n"

    set pagecount 0
    for {set j 0} {$j <= $h} {incr j 800} {
	for {set k 0} {$k <= $w} {incr k 580} {
	    if {[.sessionlog.canvas find overlapping $k $j [expr $k+579] [expr $j+799]] != ""} {
		incr pagecount

		set args ""
		if {$PrintlogVar(pagex) != ""} {
		    lappend args -pagex
		    lappend args $PrintlogVar(pagex)
		}    
		if {$PrintlogVar(pagey) != ""} {
		    lappend args -pagey
		    lappend args $PrintlogVar(pagey)
		}    
		if {$PrintlogVar(anchor) != ""} {
		    lappend args -pageanchor
		    lappend args $PrintlogVar(anchor)
		}    

		set postscript [eval .sessionlog.canvas postscript -colormode gray \
			-x $k -y $j -width 580 -height 800 $args
		]

		puts $fid "%%Page: $pagecount $pagecount\n"
		
		# Find the EOF and remove it
		set end [expr [string first "%%EOF" $postscript] - 1]
		puts $fid [string range $postscript 0 $end]
		puts $fid "\n\n\n"

	    }
	}
    }


    # Put the trailer
    puts $fid "%%Trailer"

    close $fid
    
    if {$PrintlogVar(type) == "printer"} {
	set output [eval exec $PrintlogVar(command) $PrintlogVar(file)]
	tk_messageBox -type ok -message "Postscript output sent to printer with:\n$PrintlogVar(command) $PrintlogVar(file)"
	file delete $PrintlogVar(file)
    } else {
	tk_messageBox -type ok -message "Postscript output saved to:\n$PrintlogVar(file)"
    }
    
}
#
# print_sessionlog_init --
#
#       Initialize the parameters for the print sessionlog dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc print_sessionlog_init {} {
    global PrintlogVar
    
    set PrintlogVar(type) "printer"
    set PrintlogVar(command) "lpr"
    set PrintlogVar(file) "/tmp/printfile.ps"

    set PrintlogVar(pagex) "1i"
    set PrintlogVar(pagey) "10i"
    set PrintlogVar(anchor) "nw"
}





#
# print_image_dialog --
#
#       Create a dialog box for printing a SADIE image.
#
# Arguments:
#       None
#
# Results:
#       The dialog box is created and grabs input until either
#       OK or CANCEL button is pressed by user.
#
proc print_image_dialog { } {
    global PrintimgVar
    global confirmStatus

    #
    # Determine the current image
    #
    set PrintimgVar(inimg,array) [get_current_image]
    if {$PrintimgVar(inimg,array) != "none"} {
	    upvar #0 $PrintimgVar(inimg,array) imgarray
	    set PrintimgVar(inimg,name) $imgarray(title)
	    set PrintimgVar(inimg,addr) $imgarray(sadie_addr)
    } else {
    	return
    }
    
    #
    # Create the dialog 
    #
    set top [dialog_create Dialog .printimgdlg]
    wm geometry $top +110+110
    wm title $top "Print Image Dialog"
    wm protocol $top WM_DELETE_WINDOW {dialog_set_confirm 0}
    bind $top <Return> {dialog_set_confirm 1}
    
    #
    # Create the dialog title section
    #
    set title [dialog_title $top]
    label $title.titlelbl -text "Print the Image:"
    label $title.imglbl -fg blue -text $PrintimgVar(inimg,name)
    grid $title.titlelbl $title.imglbl -sticky news

    #
    # Create the dialog input section
    #
    set input [dialog_input $top]

    radiobutton $input.printer \
	    -text "Send to printer with command: " \
	    -variable PrintimgVar(type) -value "printer"
    entry $input.printerCmd -textvariable PrintimgVar(command)
    radiobutton $input.file \
	    -text "Save output in file: " \
	    -variable PrintimgVar(type) -value "file"
    entry $input.fileName -textvariable PrintimgVar(file)

    grid $input.printer -row 0 -columnspan 3 -sticky w
    grid $input.printerCmd -row 1 -columnspan 3 -sticky ew
    grid rowconfigure $input 2 -minsize 6
    grid $input.file -row 3 -columnspan 3 -sticky w
    grid $input.fileName -row 4 -columnspan 3 -sticky ew

    bind $input.printerCmd <FocusIn> "$input.printer invoke"
    bind $input.fileName <FocusIn> "$input.file invoke"


    #
    # Create the dialog output section
    #
    set output [dialog_output $top]
    
    #
    # Create the dialog controls
    #
    set cntls [dialog_controls $top]
    dialog_buttons $cntls {dialog_set_confirm 1} {dialog_set_confirm 0}
    
    
    #
    # Wait for a response
    #
    dialog_wait $top confirmStatus
    if {$confirmStatus == 1} {
        print_image_run    
    }


    destroy $top
}
#
# print_image_run --
#
#       Obtain postscript output from the image and either print it or save to file.
#
# Arguments:
#       None
#
# Results:
#       The postscript output is returned.
#
proc print_image_run {} {
    global PrintimgVar

    Sadie_FileIO_SaveEPS $PrintimgVar(inimg,addr) $PrintimgVar(file) 8

    # Read the generated EPS data
    set fid [open $PrintimgVar(file) r]
    set eps [read $fid]
    close $fid

    # Now rewrite the file with the full postscript info
    set fid [open $PrintimgVar(file) w]
    
    # Create the postscript header information
    puts $fid "%!PS-Adobe-2.1"
    puts $fid "%%Title: tclSadie image \"$PrintimgVar(inimg,name)\""
    puts $fid "%%Creator: tclSadie"
    puts $fid "%%Pages: 1"
    puts $fid "%%EndComments"
    puts $fid "%%EndProlog"
    puts $fid "\n\n\n"

    puts $fid "%%Page: 1 1\n"
    puts $fid "/EPSsave save def"
    puts $fid "60 60 translate"
    puts $fid $eps
    puts $fid "EPSsave restore"
    puts $fid "showpage"
    
    # Put the trailer
    puts $fid "%%Trailer"

    close $fid
    
    
    if {$PrintimgVar(type) == "printer"} {
	set output [eval exec $PrintimgVar(command) $PrintimgVar(file)]
	tk_messageBox -type ok -message "Postscript output sent to printer with:\n$PrintimgVar(command) $PrintimgVar(file)"
	file delete $PrintimgVar(file)
    } else {
	tk_messageBox -type ok -message "Postscript output saved to:\n$PrintimgVar(file)"
    }
    
}
#
# print_image_init --
#
#       Initialize the parameters for the print image dialog.
#
# Arguments:
#       None
#
# Results:
#       A global variable array is created to store the parameters.
#
proc print_image_init {} {
    global PrintimgVar
    
    set PrintimgVar(type) "printer"
    set PrintimgVar(command) "lpr"
    set PrintimgVar(file) "/tmp/printfile.ps"
}









##
## Buggy printing procedures for windows 95
##



proc page_args { array } {
  upvar #0 $array ary

  # First we check whether we have a valid hDC
  # (perhaps we can later make this also an optional argument, defaulting to 
  #  the default printer)
  set attr [ printer attr ]
  foreach attrpair $attr {
    set key [lindex $attrpair 0]
    set val [lindex $attrpair 1]
    switch -exact $key {
      "hDC"       { set ary(hDC) $val }
      "copies"    { if { $val >= 0 } { set ary(copies) $val } }
      "page dimensions" {
                          set wid [lindex $val 0]
                          set hgt [lindex $val 1]
                          if { $wid > 0 } { set ary(pw) $wid }
                          if { $hgt > 0 } { set ary(pl) $hgt }
                        }
      "page margins"    {
                          if { [scan [lindex $val 0] %d tmp] > 0 } {
			    set ary(lm) [ lindex $val 0 ]
			    set ary(tm) [ lindex $val 1 ]
			    set ary(rm) [ lindex $val 2 ]
			    set ary(bm) [ lindex $val 3 ]
			  }
                        }
      "resolution"      {
                          if { [scan [lindex $val 0] %d tmp] > 0 } {
			    set ary(resx) [ lindex $val 0 ]
			    set ary(resy) [ lindex $val 1 ]
			  } else {
			    set ary(resx) 200	;# Set some defaults for this...
			    set ary(resy) 200
			  }
                        }
    }
  }
}



################################################################
## Start a printer job
################################################################
proc start_print_job { } {
  global hdc
  set hdc [printer dialog select]
  printer job start -name gdi
  printer page start
}

################################################################
## Finish a printer job and unload the printer driver
################################################################
proc end_print_job { } {
  global hdc
  printer page end
  printer job end
  printer close
  set hdc 0x0
}

################################################################
## Main routine to produce printed pages.
################################################################
proc gdi_print { reporttitle {window screen} {reporttail "End of print" } } {
  global hdc
  global attr
  global status

  # This is not a fully general test, but should get the idea across...
  start_print_job

  page_args attr
  if { [info exist attr(lm)] == 0 } { set attr(lm) 1000 }
  if { [info exist attr(rm)] == 0 } { set attr(rm) 1000 }
  if { [info exist attr(tm)] == 0 } { set attr(tm) 1000 }
  if { [info exist attr(bm)] == 0 } { set attr(bm) 1000 }
  if { [info exist attr(pw)] == 0 } { set attr(pw) 8500 }
  if { [info exist attr(pl)] == 0 } { set attr(pl) 11000 }
  if { [info exist attr(resx)] == 0 } { set attr(resx) 300 }
  if { [info exist attr(resy)] == 0 } { set attr(resy) 300 }

  set tm [expr $attr(tm) / $attr(resy) * 72 ]
  set bm [expr $attr(bm) / $attr(resy) * 72 ]
  set rm [expr $attr(rm) / $attr(resx) * 72 ]
  set lm [expr $attr(lm) / $attr(resx) * 72 ]
  set pw [expr $attr(pw) / $attr(resx) * 72 ]
  set pl [expr $attr(pl) / $attr(resy) * 72 ]

  # First, we'll raise the window of interest, 
  # and update everything so the screen will look OK.
  if { $window != "screen" } {
    raise [ winfo toplevel $window ]
  }
  update

  # Print the title
  set wid [expr $pw ]
  set hgt [gdi text $hdc $lm $tm -text $reporttitle -anchor nw -width $wid -font { Arial 18 } ]

  # Ensure the scaling is correct
  set maxx [ expr $pw - $rm ]
  set maxy [ expr $pl - $bm ]
  # set dest "$lm [expr $tm + $hgt + 10] $maxx $maxy"
  set dest "$lm [expr $tm + $hgt + 10] $maxx"
  if { $window == "screen" } {
      set retval [gdi copybits $hdc -screen -destination $dest]
  } else {
      set retval [gdi copybits $hdc -window $window -destination $dest]
  }

  # Now, if we have retval, let's put out some text below that!
  # A second hgt is not needed below because the top corner of the text is the
  # anchor (nw). The second 10 is used to provide a wee bit of spacing
  set endtexttop [ expr [lindex $retval 3] + $tm + $hgt + 10 + 10 ]
  gdi text $hdc $lm $endtexttop -text $reporttail -anchor nw -width $wid -font { Arial 18 }

  end_print_job
}


proc print_image {} {
    # Get the current image
    set imgName [get_current_image]
    if {$imgName != "none"} {
	upvar #0 $imgName imgarray
	set canv $imgarray(window).imgcanvas
    } else {
    	return
    }

    gdi_print "Image: $imgarray(title)" $canv
}
