#
# BindYview --
#
#       Bind multiple widgets to a single vertical scrollbar.
#
# Arguments:
#       lists -- The list of widgets to bind
#       args -- The arguements from the scrollbar
#
# Results:
#       Allows all of the widgets listed in $lists to be scrolled
#       in the y-direction by the same scrollbar.
#
proc BindYview { lists args } {
    foreach l $lists {
	eval {$l yview} $args
    }
}

#
# BindXview --
#
#       Bind multiple widgets to a single horizontal scrollbar.
#
# Arguments:
#       lists -- The list of widgets to bind
#       args -- The arguements from the scrollbar
#
# Results:
#       Allows all of the widgets listed in $lists to be scrolled
#       in the x-direction by the same scrollbar.
#
proc BindXview { lists args } {
    foreach l $lists {
	eval {$l xview} $args
    }
}

#
# create_class_widget --
#
#       Create a widget for modifying and displaying the defined classes
#       for an image.
#
# Arguments:
#       win -- The name of the class widget
#       arrayName -- The name of the image array
#
# Results:
#       Creates the widget and sets up bindings to allow the classes
#       to be modified.
#
proc create_class_widget {win arrayName} {
    upvar #0 $arrayName imgarray

    global bugwarn

    # Give the user a bug warning about the bug that shows up
    # in update_class_widget procedure under CDE
    if {![info exists bugwarn($win)]} {
	set bugwarn($win) 1

	tk_messageBox -type ok -icon warning -title "Bug Warning!" \
		-message "Warning:  A possible bug exists under the CDE window manager \
	                  that may cause this dialog to stop responding to button presses. \
	                  If this happens, try moving the window around with the mouse and \
	                  then press the button again."
    }

    set checksize 20

    frame $win

    frame $win.size
    label $win.size.lbl -text "Number of Classes: "
    entry $win.size.ent -width 5 -textvariable "${arrayName}(tempclasscount)"
    label $win.size.units -text " ( 1 - 32 ) "
    grid $win.size.lbl $win.size.ent $win.size.units -sticky w

    bind $win.size.ent <KeyRelease> "change_class_size $win $arrayName"

    label $win.classlbl -text "Class Definitions: "
    
    #
    # Create scrollable canvases to hold the class definition frames
    #
    canvas $win.regions -width 10 -height 10 -yscrollcommand "$win.yscr set"
    canvas $win.classes -width 10 -height 10 -xscrollcommand "$win.xscr set"
    canvas $win.checks -yscrollcommand "$win.yscr set"

    frame $win.regions.f -bd 0
    frame $win.classes.f -bd 0
    frame $win.checks.f -bd 0

    #
    # Embed the class definition frames in canvases
    #
    $win.regions create window 0 0 -anchor nw -window $win.regions.f
    $win.classes create window 0 0 -anchor nw -window $win.classes.f
    $win.checks create window 0 0 -anchor nw -window $win.checks.f

    # Scroll both $win.checks and $win.classes in the x-direction
    scrollbar $win.xscr -orient horizontal -command [list BindXview [list $win.checks $win.classes]]

    # Scroll both $win.regions and $win.checks in the y-direction
    scrollbar $win.yscr -orient vertical -command [list BindYview [list $win.regions $win.checks]]

    #
    # Initialize the class definitions
    #
    class_size_init $arrayName

    for {set n 1} {$n <= $imgarray(class,count)} {incr n} {
	label $win.classes.f.$n -text $n
	
	grid $win.classes.f.$n -row 0 -column [expr $n - 1] -sticky news
	grid columnconfigure $win.classes.f [expr $n - 1] -minsize [expr round($checksize * 1.5)] -weight 1
	grid columnconfigure $win.checks.f [expr $n - 1] -minsize [expr round($checksize * 1.5)] -weight 1
	
	for {set m 1} {$m <= $imgarray(region,count)} {incr m} {
	    if {$n == 1} {
		label $win.regions.f.$m -text $imgarray(region,$m,name)

		grid $win.regions.f.$m -row [expr $m - 1] -column 0 -sticky w
		grid rowconfigure $win.regions.f [expr $m - 1] -minsize $checksize -weight 1
		grid rowconfigure $win.checks.f [expr $m - 1] -minsize $checksize -weight 1
	    }
	    
	    if {$imgarray(class,${n},${m}) == 1} {
		# If the class contains this region, put a checkmark
		label $win.checks.f.${n}_${m} -image checkmark -relief ridge -bd 1
	    } else {
		# Otherwise, leave box blank
		label $win.checks.f.${n}_${m} -image blank -relief ridge -bd 1
	    }
	    grid $win.checks.f.${n}_${m} -row [expr $m - 1] -column [expr $n - 1] -sticky ns
	    bind $win.checks.f.${n}_${m} <ButtonRelease-1> [subst {toggle_region $win.checks.f $arrayName $n $m}]
	}
    }

    grid $win.size -row 0 -column 0 -columnspan 4 -pady 8 -sticky w
    grid $win.classlbl -row 1 -column 1 -columnspan 3 -sticky w
    grid $win.regions -row 3 -column 0 -padx 2 -sticky ne
    grid $win.classes -row 2 -column 1 -sticky ws
    grid $win.checks -row 3 -column 1 -sticky nw
    grid $win.xscr -row 4 -column 1 -sticky new
    grid $win.yscr -row 3 -column 2 -sticky nes
    grid columnconfigure $win 3 -weight 1
}

#
# change_class_size --
#
#       Change the number of defined classes for an image, and
#       update the class widget to reflect the change.
#
# Arguments:
#       win -- The name of the class widget
#       arrayName -- The name of the image array
#
# Results:
#       Changes the number of classes, initializes any new class
#       definitions, and updates the GUI.
#
proc change_class_size {win arrayName} {
    upvar #0 $arrayName imgarray

    set checksize 20

    set newsize $imgarray(tempclasscount)
    
    #
    # Make sure the new size is reasonable
    #
    if {[regexp {^[0-9]+$} $newsize] == 0} {
	set newsize $imgarray(region,count)
    } elseif {$newsize < 1  ||  $newsize > 32} {
	set newsize $imgarray(region,count)
    }

    #
    # Only continue if the class count has changed
    #
    if {$newsize != $imgarray(class,count)} {
	if {$newsize < $imgarray(class,count)} {
	    for {set n [expr $newsize + 1]} {$n <= $imgarray(class,count)} {incr n} {
		destroy $win.classes.f.$n
		for {set m 1} {$m <= $imgarray(region,count)} {incr m} {
		    destroy $win.checks.f.${n}_${m}
		}
	    }
	} else {
	    for {set n [expr $imgarray(class,count) + 1]} {$n <= $newsize} {incr n} {
		label $win.classes.f.$n -text $n
		grid $win.classes.f.$n -row 0 -column [expr $n - 1] -sticky news
		grid columnconfigure $win.classes.f [expr $n - 1] -minsize [expr round($checksize * 1.5)] -weight 1
		grid columnconfigure $win.checks.f [expr $n - 1] -minsize [expr round($checksize * 1.5)] -weight 1

		for {set m 1} {$m <= $imgarray(region,count)} {incr m} {
		    if {![info exists imgarray(class,${n},${m})]} {
			if {$n == $m} {
			    set imgarray(class,${n},${m}) 1
			} else {
			    set imgarray(class,${n},${m}) 0
			}
		    }

		    if {$imgarray(class,${n},${m}) == 1} {
			label $win.checks.f.${n}_${m} -image checkmark -relief ridge -bd 1
		    } else {
			label $win.checks.f.${n}_${m} -image blank -relief ridge -bd 1
		    }
		    grid $win.checks.f.${n}_${m} -row [expr $m - 1] -column [expr $n - 1] -sticky ns
		    bind $win.checks.f.${n}_${m} <ButtonRelease-1> [subst {toggle_region $win.checks.f $arrayName $n $m}]
		}
	    }
	}

	set imgarray(class,count) $newsize

	#
	# Always re-create the last checkbox so that the window size will
	# be updated correctly.
	#
	set n $imgarray(class,count)
	set m $imgarray(region,count)
	destroy $win.checks.f.${n}_${m}
	if {$imgarray(class,${n},${m}) == 1} {
	    label $win.checks.f.${n}_${m} -image checkmark -relief ridge -bd 1
	} else {
	    label $win.checks.f.${n}_${m} -image blank -relief ridge -bd 1
	}
	grid $win.checks.f.${n}_${m} -row [expr $m - 1] -column [expr $n - 1] -sticky ns
	bind $win.checks.f.${n}_${m} <ButtonRelease-1> [subst {toggle_region $win.checks.f $arrayName $n $m}]

	update_class_widget $win $arrayName
    }
    
}


#
# class_size_init --
#
#       Initialize class size and definitions.
#
# Arguments:
#       arrayName -- The name of the image array
#
# Results:
#       Sets the number of classes to the number of defined
#       regions for the image, and initializes each class to
#       exclusively contain its corresponding region.
#
proc class_size_init {arrayName} {
    upvar #0 $arrayName imgarray

    set imgarray(class,count) $imgarray(region,count)
    set imgarray(tempclasscount) $imgarray(class,count)

    for {set n 1} {$n <= $imgarray(class,count)} {incr n} {
	for {set m 1} {$m <= $imgarray(region,count)} {incr m} {
	    if {![info exists imgarray(class,${n},${m})]} {
		if {$n == $m} {
		    set imgarray(class,${n},${m}) 1
		} else {
		    set imgarray(class,${n},${m}) 0
		}
	    }
	}
    }
}


#
# update_class_widget --
#
#       Update the size and layout of a class widget.
#
# Arguments:
#       win -- The name of the class widget
#       arrayName -- The name of the image array
#
# Results:
#       Waits for the class widget elements to become visible on the
#       screen, and then makes sure that the layout and sizing
#       are correct.
#
proc update_class_widget {win arrayName} {
    upvar #0 $arrayName imgarray

    set checksize 20

    #
    # Wait for the last checkbox to be displayed to the screen
    #
    # Bug Notification: This causes hangs under CDE -- If the ok or cancel
    #    buttons do not respond on the dialog, move the dialog with the mouse
    #    and try again!
    tkwait visibility $win.checks.f.$imgarray(class,count)_$imgarray(region,count)

    # Find the size of each checkbox
    #set bbox [grid bbox $win.checks.f 0 0]
    #set incr [lindex $bbox 3]
    set xincr [expr round($checksize * 1.5)]
    set yincr $checksize
    
    # Find the size of the checkbox frame
    set checkwidth [expr $xincr * $imgarray(class,count)]
    set checkheight [winfo reqheight $win.checks.f]

    # Update the scroll width and height of the checkbox canvas/frame
    $win.checks configure -scrollregion "0 0 $checkwidth $checkheight" \
	    -yscrollincrement $yincr \
	    -xscrollincrement $xincr

    # Determine and set an appropriate height and width for the
    # checkbox canvas/frame
    set max_x $imgarray(class,count)
    if {$max_x > 6} {
	set max_x 6
    }
    set max_y $imgarray(region,count)
    if {$max_y > 6} {
	set max_y 6
    }
    set width [expr $xincr * $max_x]
    set height [expr $yincr * $max_y]
    $win.checks configure -width $width -height $height


    # Find the size of the region list frame
    set regionwidth [winfo reqwidth $win.regions.f]
    set regionheight $checkheight
    # Update the scroll width and height of the region list canvas/frame
    $win.regions configure -scrollregion "0 0 $regionwidth $regionheight" \
	    -yscrollincrement $yincr
    # Set an appropriate height and width for the region list canvas/frame
    $win.regions configure -width $regionwidth -height $height


    # Find the size of the class list frame
    set classwidth $checkwidth
    set classheight [winfo reqheight $win.classes.f]
    # Update the scroll width and height of the class list canvas/frame
    $win.classes configure -scrollregion "0 0 $classwidth $classheight" \
	    -xscrollincrement $xincr
    # Set an appropriate height and width for the class list canvas/frame
    $win.classes configure -width $width -height $classheight
}

#
# toggle_region --
#
#       Add or remove a region from a class.
#
# Arguments:
#       win -- The name of the class widget
#       arrayName -- The name of the image array
#       class -- The class number
#       region -- The region number
#
# Results:
#       Adds or deletes the region to the class, and updates
#       the gui to show the change.
#
proc toggle_region {win arrayName class region} {
    upvar #0 $arrayName imgarray

    if {$imgarray(class,${class},${region}) > 0} {
	set imgarray(class,${class},${region}) 0
	$win.${class}_${region} configure -image blank
    } else {
	set imgarray(class,${class},${region}) 1
	$win.${class}_${region} configure -image checkmark
    }
}
