/** @file ProgMeter.c
 * Callbacks to allow libsadie to display a pop-up progress meter window.
 * If compiled with support for callback functions, the SADIE library can
 * display the progress of certain operations with a progress meter implemented
 * by the application using it; here is tclSadie's implementation.  The
 * @c Tcl_EvalObjv Tcl library function does most of the work of passing
 * the necessary commands to the interpreter for each operation.  We
 * pass it temporary arrays of Tcl objects that are treated as command
 * templates: sequences of slots with pre-defined roles that we fill with
 * pointers to the objects. 
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: ProgMeter.c,v 2.4 2005/03/22 21:11:15 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include "tclsadie.h"

/** Define command template slots for the meter creation command. */
enum create_template_index
{
  ADD_GUICMD, ADD_TITLE, max_add_index
};

/** Define command template slots for the progress meter update command. */
enum update_template_index
{
  UPD_GUICMD, UPD_GNUM, UPD_VALUE, max_upd_index
};

/** Define command template slots for the meter destruction command. */
enum destroy_template_index
{
  DEL_GUICMD, DEL_GNUM, max_del_index
};

/**
 * Recover the Tcl interpreter pointer from an opaque pointer.
 * The SADIE library knows nothing about the data structures peculiar to
 * particular applications, but will pass the callbacks an opaque pointer,
 * which the application should have initialized to something it can
 * manipulate.  Here we interpret it as a pointer to the Tclsadie_app_data
 * structure that should have the interpreter pointer as one of its members.
 * @param[in] data  the opaque pointer.
 * @return A pointer to the Tcl interpreter, or NULL on error.
 */
static Tcl_Interp *
recover_interp (void * data)
{
  tclsadie_app_datap_t private_data;

  if (data == NULL)
    return NULL;
  private_data = (tclsadie_app_datap_t) data;
  return private_data->main_interp;
}

/**
 * Create a pop-up progress meter widget.
 * There should already be a Tcl/Tk command that will create the widget
 * within the current GUI in its initial state (no progress yet).  Here we
 * invoke it, passing it a title string for the widget.
 * @param[in] title  the title string.
 * @param[in] data   an opaque pointer to the saved tclSadie state data.
 * @return An integer identifying the progress meter widget, zero on failure.
 */
int
ProgMeter_Create (const char * title, void * data)
{
  Tcl_Interp *interp;
  Tcl_Obj *tclcmd[max_add_index];
  int gaugenum = 0;
  int err = 1;

  interp = recover_interp (data);
  if (interp == NULL)
    return 0;
  tclcmd[ADD_GUICMD] = Tcl_NewStringObj ("gauge_gui_add", -1);
  if (tclcmd[ADD_GUICMD] != NULL)
    {
      Tcl_IncrRefCount (tclcmd[ADD_GUICMD]);
      tclcmd[ADD_TITLE] = Tcl_NewStringObj (title, -1);
      if (tclcmd[ADD_TITLE] != NULL)
        {
          Tcl_IncrRefCount (tclcmd[ADD_TITLE]);
          if ((Tcl_EvalObjv (interp, max_add_index, tclcmd, TCL_EVAL_GLOBAL)
               != TCL_OK)
              || (Tcl_GetIntFromObj (interp,
                                     Tcl_GetObjResult (interp),
                                     &gaugenum)
                  != TCL_OK))
            gaugenum = 0;
          else
            err = 0;
          Tcl_DecrRefCount (tclcmd[ADD_TITLE]);
        }
      Tcl_DecrRefCount (tclcmd[ADD_GUICMD]);
    }
  if (err)
    Tcl_BackgroundError (interp);
  return gaugenum;
}

/**
 * Modify the progress displayed by an existing progress meter.
 * The progress meter should already exist, and there should be a Tcl command
 * that will update it, given an ID number and an indication of the current
 * percentage progress.  This invokes the command, and returns a value
 * showing if the user clicked on a cancel button within the widget since
 * the previous update.
 * @param[in] gaugenum the code identifying the particular progress meter.
 * @param[in] value    the percentage progress the meter should display.
 * @param[in] data     an opaque pointer to the saved tclSadie state data.
 * @return A non-zero value for cancellation (by user selection or error).
 */
int
ProgMeter_Update (int gaugenum, double value, void * data)
{
  Tcl_Interp * interp;
  Tcl_Obj * tclcmd[max_upd_index];
  int cancelval = -1;
  int err = 1;

  interp = recover_interp (data);
  if (interp == NULL)
    return 0;
  tclcmd[UPD_GUICMD] = Tcl_NewStringObj ("gauge_gui_update", -1);
  if (tclcmd[UPD_GUICMD] != NULL)
    {
      Tcl_IncrRefCount (tclcmd[UPD_GUICMD]);
      tclcmd[UPD_GNUM] = Tcl_NewIntObj (gaugenum);
      if (tclcmd[UPD_GNUM] != NULL)
        {
          Tcl_IncrRefCount (tclcmd[UPD_GNUM]);
          tclcmd[UPD_VALUE] = Tcl_NewDoubleObj (value);
          if (tclcmd[UPD_VALUE] != NULL)
            {
              Tcl_IncrRefCount (tclcmd[UPD_VALUE]);
              if ((Tcl_EvalObjv (interp, max_upd_index, tclcmd,
                                 TCL_EVAL_GLOBAL)
                   != TCL_OK)
                  || (Tcl_GetIntFromObj (interp,
                                         Tcl_GetObjResult (interp),
                                         &cancelval)
                      != TCL_OK))
                cancelval = -1;
              else
                err = 0;
              Tcl_DecrRefCount (tclcmd[UPD_VALUE]);
            }
          Tcl_DecrRefCount (tclcmd[UPD_GNUM]);
        }
      Tcl_DecrRefCount (tclcmd[UPD_GUICMD]);
    }
  if (err)
    Tcl_BackgroundError (interp);
  return cancelval;
}


/**
 * Destroy a pop-up progress meter widget.
 * There should already be a Tcl/Tk command that will destroy the widget
 * within the current GUI, given its ID number, so we invoke it, returning
 * any error status.
 * @param[in] data     an opaque pointer to the saved tclSadie state data.
 * @param[in] gaugenum the code identifying the particular progress meter.
 * @return An error status code (zero on failure).
 */
int
ProgMeter_Destroy (int gaugenum, void * data)
{
  Tcl_Interp * interp;
  Tcl_Obj * tclcmd[max_del_index];
  int status = 0;

  interp = recover_interp (data);
  if (interp == NULL)
    return 0;
  tclcmd[DEL_GUICMD] = Tcl_NewStringObj ("gauge_gui_delete", -1);
  if (tclcmd[DEL_GUICMD] != NULL)
    {
      Tcl_IncrRefCount (tclcmd[DEL_GUICMD]);
      tclcmd[DEL_GNUM] = Tcl_NewIntObj (gaugenum);
      if (tclcmd[DEL_GNUM] != NULL)
        {
          Tcl_IncrRefCount (tclcmd[DEL_GNUM]);
          status = (Tcl_EvalObjv (interp, max_del_index, tclcmd,
                                  TCL_EVAL_GLOBAL)
                    == TCL_OK);
          Tcl_DecrRefCount (tclcmd[DEL_GNUM]);
        }
      Tcl_DecrRefCount (tclcmd[DEL_GUICMD]);
    }
  if (!status)
    Tcl_BackgroundError (interp);
  return status;
}
