/** @file Sadie_Classify.c
 * A Tcl/Tk interface to the SADIE image classification functions.
 * The SADIE library implements some of the classification operations
 * described in Chapter 9 of R. A. Schowengerdt @e Remote @e Sensing
 * (2nd. ed.) Academic Press: 1997.  In addition to the images themselves,
 * they require definitions of regions of interest, which are recorded
 * neither within the SADIE library, nor within the Tcl code, but within
 * an independent hash table (see @ref Sadie_Index.c for the implementation
 * details). 
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Classify.c,v 2.5 2005/03/22 22:07:11 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <stdio.h>
#include <string.h>
#include <sadie.h>
#include "tclsadie.h"
#include "Sadie_Index.h"
#include "tclSadie_ROI.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** The global hash table structures, used for ROI storage. */
extern sad_doclistp_t global_doclist;

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixaddr[] = DEFAULT_IN_IMAGE_ADDR_INDEX;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
static const char roiprop[] = ROI_PROPERTY_NAME;
/*@}*/

/**
 * Set mean and standard deviation vectors from bands within image ROIs.
 * We need a summary of the training areas for image classification,
 * represented by a single image and a list of regions of interest (ROIs).
 * A Tcl array records the user's assignments of particular areas to
 * particular classes.  The summary vectors are supposed to record the
 * means and standard deviations for each band (so the vector length is
 * always the same as the number of bands, and the index of each element
 * matches the index of the corresponding band).  Here we compute the
 * mean and standard deviation vectors for a single class, iterating down
 * the list of ROIs and accumulating an avearge of the means and standard
 * deviations of those regions assigned to the class.  The average must
 * include weightings for the sizes of the different regions, and a term
 * to represent the between-region as well as within-region variance.
 * @param[in] interp       the Tcl interpreter.
 * @param[in] tclimgarray  classification array name (a Tcl object).
 * @param[in] ixclass      index number of the class to summarize.
 * @param[in] img          the image being classified.
 * @param[in] regionlisthead  a list of training areas.
 * @param[out] mean_vec    band means for the class.
 * @param[out] sd_vec      band standard deviations.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
set_class_summary_vec (Tcl_Interp *interp, Tcl_Obj *tclimgarray,
                       int ixclass, IMAGE *img, ROIPtr regionlisthead,
                       double *mean_vec, double *sd_vec)
{
  uint32_t nreg, iband, jband, kband;
  uint64_t m;
  int included;
  ROIPtr region;
  char index[SLEN];

  if ((!interp) || (!tclimgarray) || (!img) || (!regionlisthead)
      || (!mean_vec) || (!sd_vec))
    return TCL_ERROR;
  m = 0;
  for (iband = 0; iband < img->nbnd; iband++)
    {
      mean_vec[iband] = 0;
      sd_vec[iband] = 0;
    }
  for (nreg = 1, region = regionlisthead;
       region != NULL;
       region = region->next, nreg++)
    {
      /* Read the input boolan imgarray(class,iclass,nreg) */
      snprintf (index, SLEN - 1, "class,%d,%u", ixclass, nreg);
      if (GetSadieBooleanFromObj2 (interp, tclimgarray, (const char *) index,
                                   &included) != TCL_OK)
        return TCL_ERROR;
      if (included)
        {
          for (jband = 0; jband < img->nbnd; jband++)
            {
              mean_vec[jband] += region->nval * region->mean[jband];
              sd_vec[jband] += region->nval
                * ((region->mean[jband] * region->mean[jband])
                   + (region->dev[jband] * region->dev[jband]));
            }
          m += region->nval;
        }
    }
  if (m == 0)
    return TCL_ERROR;
  for (kband = 0; kband < img->nbnd; kband++)
    {
      mean_vec[kband] /= m;
      sd_vec[kband] = sqrt ((sd_vec[kband] / m)
                            - (mean_vec[kband] * mean_vec[kband]));
    }
  return TCL_OK;
}

/**
 * Level-slice classifier (box classifier).
 * The LVLSLICE function in the SADIE library performs a simple level-slice
 * classification, but needs additional help computing the level ranges for
 * each band and class from the mean vectors of a set of training areas;
 * Note that "the level-slice classifier does not fit remote-sensing data
 * particularly well." (R.A. Schowengerdt @e Remote @e Sensing (2nd. ed.)
 * Academic Press: 1997: p. 412).
 * The Tcl data should conform to the patterns:
 * - @c main_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c main_array_name(numclasses) the overall number of classes.
 * - @c main_array_name(thresh) box limits are +/- thresh sigma units.
 * - @c main_array_name(inimg,array) classification array name.
 * - @c main_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * Please refer to set_class_summary_vec for details of the Tcl
 * classification array.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the main array name).
 * @return @c TCL_ERROR on serious errors only, otherwise @c TCL_OK.
 */
int
Sadie_Classify_LvlSliceCmd (ClientData client_data, Tcl_Interp *interp,
                            int objc, Tcl_Obj * const objv[])
{
  Tcl_Obj *tclindexname = NULL;
  Tcl_Obj *tclimgarray = NULL;
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int numclasses, iclass;
  uint32_t iband;
  PIXEL thresh;
  PIXEL *grng = NULL;
  PIXEL *base = NULL;
  ROIPtr regionlist = NULL;
  double *mean = NULL;
  double *sd = NULL;
  char msg[SLEN];
  const char *dkey;
  int err = TCL_OK;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDkeyFromObj2 (interp, objv[1], ixaddr, &dkey)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "numclasses", &numclasses)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "thresh", &thresh)
          != TCL_OK))
    return TCL_ERROR;
  if (numclasses == 0)
    {
      MESSAGE ('E', " No classes defined.");
      return TCL_OK;
    }
  /* Read the input array name */
  tclindexname = Tcl_NewStringObj ("inimg,array", -1);
  Tcl_IncrRefCount (tclindexname);
  tclimgarray = Tcl_ObjGetVar2 (interp, objv[1], tclindexname,
                                TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  Tcl_DecrRefCount (tclindexname);
  if (tclimgarray == NULL)
    return TCL_ERROR;
  else
    Tcl_IncrRefCount (tclimgarray);
  /* Allocate space for the class boundary array */
  grng = calloc (((size_t)numclasses * inimg->nbnd * 2), sizeof (PIXEL));
  if (grng != NULL)
    {
      mean = calloc (inimg->nbnd, sizeof (double));
      if (mean == NULL)
        free (grng);
      else
        {
          sd = calloc (inimg->nbnd, sizeof (double));
          if (sd == NULL)
            {
              free (mean);
              free (grng);
            }
        }
    }
  if (sd == NULL)
    {
      Tcl_DecrRefCount (tclimgarray);
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Could not allocate memory!", -1);
      return TCL_ERROR;
    }
  regionlist = sad_get_doc_property (global_doclist, dkey, roiprop);
  if (regionlist == NULL)
    {
      free (sd);
      free (mean);
      free (grng);
      Tcl_DecrRefCount (tclimgarray);
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Could not find training areas.", -1);
      return TCL_ERROR;
    }
  for (iclass = 0; (err == TCL_OK) && (iclass < numclasses); iclass++)
    {
      err = set_class_summary_vec (interp, tclimgarray, iclass + 1,
                                   inimg, regionlist, mean, sd);
      if (err != TCL_OK)
        {
          snprintf (msg, SLEN - 1,
                    " No training area defined for class %d.", iclass + 1);
          MESSAGE ('E', msg);
        }
      else
        {
          for (iband = 0; iband < inimg->nbnd; iband++)
            {
              base
                = (grng
                   + (((size_t) iclass * inimg->nbnd * 2) + (iband * 2)));
              *base
                = (PIXEL) (mean[iband] - (thresh * sd[iband]));
              *(base + 1)
                = (PIXEL) (mean[iband] + (thresh * sd[iband]));
            }
        }
    }
  free (mean);
  free (sd);
  if (err == TCL_OK)
    LVLSLICE (inimg, grng, numclasses, &outimg);
  Tcl_DecrRefCount (tclimgarray);
  free (grng);
  return (err == TCL_OK)
    ? SetSadieImageObj2 (interp, objv[1], ixname, outimg)
    : TCL_OK;
}

/**
 * Minimum distance image classification, with several distance metric options.
 * Given a set of regions within an image, a number of classes (not
 * greater than the number of regions), an assignment table that shows
 * which regions belong to which class, and a complete unclassified
 * image, this uses the SADIE library function MINDIST to generate a
 * classified image, where the class assigned to each pixel is the one
 * whose mean vector is closest to the corresponding point in the
 * unclassified image.  The "closest" criterion is the minimum
 * distance in feature space, according to a particular distance
 * metric, and there are several options for this metric (including
 * Euclidian, city block, and Mahalanobis).
 * The Tcl data should conform to the patterns:
 * - @c main_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c main_array_name(numclasses) the overall number of classes.
 * - @c main_array_name(option) the distance metric code (see @c sadie.h ).
 * - @c main_array_name(inimg,array) classification array name.
 * - @c main_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * The Tcl classification array follows the pattern
 * @c classification_array_name(class,class_number,region_number) and
 * has what are essentially Boolean values: for a particular
 * @c class_number and @c region_number the array element is non-zero
 * if the region is in the class.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the main array name).
 * @return @c TCL_ERROR on serious errors only, otherwise @c TCL_OK.
 */
int
Sadie_Classify_MinDistCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  int iclass, nreg;
  uint32_t jband, kband;
  uint64_t m, n;
  size_t a, b;
  Tcl_Obj *tclindexname = NULL;
  Tcl_Obj *tclimgarray = NULL;
  char msg[SLEN];
  const char *dkey;
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int numclasses, option, included;
  char index[SLEN];
  ROIPtr region = NULL;
  ROIPtr regionlist = NULL;
  double *mean = NULL;
  double *cov = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDkeyFromObj2 (interp, objv[1], ixaddr, &dkey)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "numclasses", &numclasses)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  if (numclasses == 0)
    {
      MESSAGE ('E', " No classes defined.");
      return TCL_OK;
    }
  /* Read the input array name */
  tclindexname = Tcl_NewStringObj ("inimg,array", -1);
  Tcl_IncrRefCount (tclindexname);
  tclimgarray = Tcl_ObjGetVar2 (interp, objv[1], tclindexname,
                                TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  Tcl_DecrRefCount (tclindexname);
  if (tclimgarray == NULL)
    return TCL_ERROR;
  else
    Tcl_IncrRefCount (tclimgarray);
  mean = calloc (((size_t)inimg->nbnd * numclasses), sizeof (double));
  if (mean != NULL)
    {
      cov = calloc (((size_t)inimg->nbnd * inimg->nbnd), sizeof (double));
      if (cov == NULL)
        free (mean);
    }
  if (cov == NULL)
    {
      Tcl_DecrRefCount (tclimgarray);
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Memory request failed!", -1);
      return TCL_ERROR;
    }                                                                         
  regionlist = sad_get_doc_property (global_doclist, dkey, roiprop);
  if (regionlist == NULL)
    {
      free (cov);
      free (mean);
      Tcl_DecrRefCount (tclimgarray);
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Could not find training areas.", -1);
      return TCL_ERROR;
    }                                                                         
  m = 0;
  for (iclass = 0; iclass < numclasses; iclass++)
    {
      n = 0;
      for (nreg = 1, region = regionlist;
           region != NULL;
           region = region->next, nreg++)
        {
          /* Read the input Boolean imgarray(class,iclass,nreg) */
          snprintf (index, SLEN - 1, "class,%d,%d", iclass + 1, nreg);
          if (GetSadieBooleanFromObj2 (interp, tclimgarray,
                                       (const char *) index, &included)
              != TCL_OK)
            {
              free (cov);
              free (mean);
              Tcl_DecrRefCount (tclimgarray);
              return TCL_ERROR;
            }
          if (included)
            {
              for (jband = 0; jband < inimg->nbnd; jband++)
                mean[jband + iclass * inimg->nbnd]
                  += region->nval * region->mean[jband];
              for (a = 0; a < (inimg->nbnd * inimg->nbnd); a++)
                cov[a] += region->nval * region->cov[a];
              n += region->nval;
            }
        }
      if (n > 0)
        {
          for (kband = 0; kband < inimg->nbnd; kband++)
            mean[kband + iclass * inimg->nbnd] /= n;
          m += n;
        }
      else
        {
          free (cov);
          free (mean);
          Tcl_DecrRefCount (tclimgarray);
          snprintf (msg, SLEN - 1, " No training area defined for class %d.",
                    iclass + 1);
          MESSAGE ('E', msg);
          return TCL_OK;
        }
    }
  Tcl_DecrRefCount (tclimgarray);
  for (b = 0; b < (inimg->nbnd * inimg->nbnd); b++)
    cov[b] /= m;
  MINDIST (inimg, option, numclasses, mean, cov, &outimg);
  free (mean);
  free (cov);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}


/**
 * An interface to the MAXLIKE function in the SADIE library (not working).
 * The MAXLIKE function should provide a Bayesian probabilistic image
 * classification, but at the moment is just a placeholder.
 */
int
Sadie_Classify_MaxLikeCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  int iclass, nreg;
  uint64_t n;
  size_t a, b;
  Tcl_Obj *tclindexname = NULL;
  Tcl_Obj *tclimgarray = NULL;
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int numclasses, included;
  double theprob;
  double *prob = NULL;
  double *cov = NULL;
  ROIPtr region = NULL;
  ROIPtr regionlist = NULL;
  char msg[SLEN];
  char index[SLEN];
  const char *dkey;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDkeyFromObj2 (interp, objv[1], ixaddr, &dkey)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "numclasses", &numclasses)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "prob", &theprob)
          != TCL_OK))
    return TCL_ERROR;
  if (numclasses == 0)
    {
      MESSAGE ('E', " No classes defined.");
      return TCL_OK;
    }
  /* Read the input array name */
  tclindexname = Tcl_NewStringObj ("inimg,array", -1);
  Tcl_IncrRefCount (tclindexname);
  tclimgarray = Tcl_ObjGetVar2 (interp, objv[1], tclindexname,
                                TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  Tcl_DecrRefCount (tclindexname);
  if (tclimgarray == NULL)
    return TCL_ERROR;
  else
    Tcl_IncrRefCount (tclimgarray);
  prob = calloc ((size_t)numclasses, sizeof (double));
  if (prob != NULL)
    {
      cov = calloc(((size_t)numclasses * inimg->nbnd * inimg->nbnd),
                   sizeof (double));
      if (cov == NULL)
        free (prob);
    }
  if (cov == NULL)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Memory request failed!", -1);
      Tcl_DecrRefCount (tclimgarray);
      return TCL_ERROR;
    }
  regionlist = sad_get_doc_property (global_doclist, dkey, roiprop);
  if (regionlist == NULL)
    {
      free (cov);
      free (prob);
      Tcl_DecrRefCount (tclimgarray);
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Could not find training areas.", -1);
      return TCL_ERROR;
    }                                                                         
  for (iclass = 0; iclass < numclasses; iclass++)
    {
      prob[iclass] = theprob;
      n = 0;
      for (nreg = 1, region = regionlist;
           region != NULL;
           region = region->next, nreg++)
        {
          /* Read the input Boolean imgarray(class,iclass,nreg) */
          snprintf (index, SLEN - 1, "class,%d,%d", iclass + 1, nreg);
          if (GetSadieBooleanFromObj2 (interp, tclimgarray,
                                       (const char *) index, &included)
              != TCL_OK)
            {
              free (prob);
              free (cov);
              Tcl_DecrRefCount (tclimgarray);
              return TCL_ERROR;
            }
          if (included)
            {
              n += region->nval;
              for (a = 0; a < (inimg->nbnd * inimg->nbnd); a++)
                cov[a + iclass * inimg->nbnd * inimg->nbnd]
                  += region->nval * region->cov[a];
            }
        }
      if (n > 0)
        {
          for (b = 0; b < (inimg->nbnd * inimg->nbnd); b++)
            cov[b + iclass * inimg->nbnd * inimg->nbnd] /= n;
        }
      else
        {
          free (prob);
          free (cov);
          Tcl_DecrRefCount (tclimgarray);
          snprintf (msg, SLEN - 1, " No training area defined for class %d.",
                    iclass + 1);
          MESSAGE ('E', msg);
          return TCL_OK;
        }
    }
  Tcl_DecrRefCount (tclimgarray);
  MAXLIKE (inimg, numclasses, cov, prob, &outimg);
  free (prob);
  free (cov);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Image classification by an iterative K-means clustering algorithm.
 * This is simply an interface to the SADIE library function CLUSTER.
 * Starting with a certain number of arbitrarily defined clusters, we classify
 * each point by assigning it to the cluster whose mean vector is closest
 * in feature space.  We then re-evaluate the cluster mean vectors and
 * iterate the process, re-assigning points between the clusters.  Adjustable
 * parameters allow clusters whose means become close to merge, and clusters
 * that end up holding too few points to vanish.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c tcl_array_name(iter) the maximum number of iterations.
 * - @c tcl_array_name(classes) the number of initial clusters.
 * - @c tcl_array_name(min) minimum number of points in a cluster.
 * - @c tcl_array_name(incr) spacing of the points (pixel units).
 * - @c tcl_array_name(merging) threshold for merging adjacent clusters.
 * - @c tcl_array_name(outlier) threshold for excluding distant points.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Classify_ClusterCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int iter, classes, min, incr;
  PIXEL merging, outlier;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "iter", &iter)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "classes", &classes)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "min", &min)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "incr", &incr)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "merging", &merging)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "outlier", &outlier)
          != TCL_OK))
    return TCL_ERROR;
  CLUSTER (inimg, iter, classes, min, merging, 1, incr, incr, outlier,
           &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Segment an image into regions with similar greylevels.
 * As well as assigning labels to groupings of similar pixels, this will
 * optionally downsample the image before a coarse segmentation, then refine
 * the labelmap to the resolution of the original image: see the SADIE
 * library SEGMENT and REFINE functions for details.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) unclassified image ID.
 * - @c tcl_array_name(steps) number of downsampling and upsampling steps.
 * - @c tcl_array_name(sigma) the greylevel variance threshold.
 * - @c tcl_array_name(thresh) the greylevel difference threshold.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output label map name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Classify_SegmentCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
    IMAGE *inimg = NULL;
    IMAGE *outimg = NULL;
    int steps;
    double sigma;
    PIXEL thresh;

    if (objc != 2)
      {
        Tcl_WrongNumArgs (interp, 1, objv, atitle);
        return TCL_ERROR;
      }
    if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
         != TCL_OK)
        || (GetSadieIntFromObj2 (interp, objv[1], "steps", &steps)
            != TCL_OK)
        || (GetSadieDoubleFromObj2 (interp, objv[1], "sigma", &sigma)
            != TCL_OK)
        || (GetSadiePixelFromObj2 (interp, objv[1], "thresh", &thresh)
            != TCL_OK))
      return TCL_ERROR;
    SEGMENT (inimg, steps, sigma, thresh, &outimg);
    return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Assign realistic mean values to an arbitrary label map.
 * Given an image and a label map defining regions within it, generate a new
 * image, assigning values to all the pixels within a particular region
 * the mean value from the corresponding region in the original image,
 * using the SADIE library SIGMAP function.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lblmap,addr) arbitrary label map image ID. 
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) the original image ID.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output map name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Classify_SigMapCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *lblmap = NULL;
  IMAGE *origimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "lblmap,addr", &lblmap)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "origimg,addr", &origimg)
          != TCL_OK))
    return TCL_ERROR;
  SIGMAP (lblmap, origimg, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Classify_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Classify_LvlSlice",
                        Sadie_Classify_LvlSliceCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Classify_MinDist",
                        Sadie_Classify_MinDistCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Classify_MaxLike",
                        Sadie_Classify_MaxLikeCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Classify_Cluster",
                        Sadie_Classify_ClusterCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Classify_Segment",
                        Sadie_Classify_SegmentCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Classify_SigMap",
                        Sadie_Classify_SigMapCmd, NULL, NULL);
  return TCL_OK;
}

