/** @file Sadie_Contrast.c
 * A Tcl/Tk interface to the SADIE contrast manipulation functions.
 * The SADIE library includes several operations to re-scale pixel
 * values, notably contrast stretching operations.  Unfortunately
 * several of these use shared global data structures within the
 * library, and are not thread-safe or re-entrant even in versions
 * that protect other shared data with locks: see the list of
 * extern declarations within this file for details.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Contrast.c,v 2.6 2005/03/25 07:23:44 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "tclsadie.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** @name SADIE library globals
 * Global variables within the SADIE library that augment function arguments.
 */
/*@{*/
extern uint32_t nlev; /**< Used by STRETCH, TFQUANT & WTFMAJ. */
extern PIXEL gain; /**< Used by STRETCH & TFLINEAR. */
extern PIXEL bias; /**< Used by STRETCH & TFLINEAR. */
extern PIXEL gmin; /**< Used by STRETCH, TFQUANT, TFSAT, TFSCALE & WTFMAJ. */
extern PIXEL gmax; /**< Used by STRETCH, TFINVERT, TFQUANT, TFSAT & TFSCALE. */
extern PIXEL thresh; /**< Used by STRETCH, TFTHRESH, WINDOW & WTFCAVG. */
extern PIXEL gbrk[2][4]; /**< Used by STRETCH & TFPLT. */
extern PIXEL *table; /**< Used by STRETCH & TFTABLE. */
/*@}*/

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixaddr[] = DEFAULT_IN_IMAGE_ADDR_INDEX;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
/*@}*/

/**
 * Normalize the contrast within an image.
 * Taking an arbitrary input image, this generates an output image whose
 * pixels have the specified mean and standard deviation (optionally
 * down-sampling in the process), using the SADIE library function CONTNORM.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c tcl_array_name(incr) downsampling spacing (pixel units).
 * - @c tcl_array_name(sigma) desired standard deviation.
 * - @c tcl_array_name(mean) desired mean.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) re-scaled image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_NormStretchCmd (ClientData client_data, Tcl_Interp *interp,
                               int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t incr;
  double sigma;
  PIXEL mean;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr", &incr)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "sigma", &sigma)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "mean", &mean)
          != TCL_OK))
    return TCL_ERROR;
  CONTNORM (inimg, mean, sigma, incr, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Apply a histogram equalization to an image.
 * Given an image and the number of histogram bins, this uses the SADIE
 * library function HISTEQ to re-map the image values using the
 * empirically determined value distribution as a look-up table, optionally
 * downsampling the image in the process.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(res) the number of histogram bins.
 * - @c tcl_array_name(incr) downsampling spacing (pixel units).
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) equalized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_HisteqStretchCmd (ClientData client_data, Tcl_Interp *interp,
                                 int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t res, incr;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "res", &res)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr", &incr)
          != TCL_OK))
    return TCL_ERROR;
  HISTEQ (inimg, res, incr, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Apply a linear contrast stretch to an image.
 * We select either the SADIE library function TFSCALE, which uses linear
 * interpolation between the specified maximum and minimum levels, or
 * TFLINEAR, which uses @a new_value = @a bias + @a gain * @a old_value
 * for each pixel.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(minorbias) minimum for TFSCALE, or TFLINEAR bias.
 * - @c tcl_array_name(maxorgain) maximum for TFSCALE, or TFLINEAR gain.
 * - @c tcl_array_name(type) zero for TFSCALE, non-zero for TFLINEAR.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_LinearStretchCmd (ClientData client_data, Tcl_Interp *interp,
                                 int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  double minorbias, maxorgain;
  int type;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "minorbias", &minorbias)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "maxorgain", &maxorgain)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "type", &type)
          != TCL_OK))
    return TCL_ERROR;
  if (type == 0)
    {
      gmin = (PIXEL) minorbias;
      gmax = (PIXEL) maxorgain;
      STRETCH (inimg, TFSCALE, &outimg);
    }
  else
    {
      bias = (PIXEL) minorbias;
      gain = (PIXEL) maxorgain;
      STRETCH (inimg, TFLINEAR, &outimg);
    }
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Apply a log transform to an image.
 * This is simply an interface to the SADIE library function TFLOG,
 * which applies the log10 function to the pixel values, while retaining
 * values less than or equal to zero.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_LogStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
    return TCL_ERROR;
  STRETCH (inimg, TFLOG, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Apply a three-segment piecewise linear transformation to an image.
 * We define the linear segments by the values (breakpoints) of their
 * ends in both the original image and the transformed image, so for
 * three segments we need two sets of four breakpoint values.  The
 * SADIE library function TFPLT handles the actual transformation.
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(in1) breakpoint 1 in the original image.
 * - @c tcl_array_name(in2) breakpoint 2 in the original image.
 * - @c tcl_array_name(in3) breakpoint 3 in the original image.
 * - @c tcl_array_name(in4) breakpoint 4 in the original image.
 * - @c tcl_array_name(out1) breakpoint 1 in the transformed image.
 * - @c tcl_array_name(out2) breakpoint 2 in the transformed image.
 * - @c tcl_array_name(out3) breakpoint 3 in the transformed image.
 * - @c tcl_array_name(out4) breakpoint 4 in the transformed image.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_PWLStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "in1", gbrk[0])
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "in2", gbrk[0] + 1)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "in3", gbrk[0] + 2)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "in4", gbrk[0] + 3)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "out1", gbrk[1])
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "out2", gbrk[1] + 1)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "out3", gbrk[1] + 2)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "out4", gbrk[1] + 3)
          != TCL_OK))
    return TCL_ERROR;
  STRETCH (inimg, TFPLT, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}


/**
 * Apply a square root transformation to an image.
 * This is simply an interface to the SADIE library function TFROOT,
 * which takes the square root of the pixel values, not generating
 * complex values if they are negative.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_RootStretchCmd (ClientData client_data, Tcl_Interp *interp,
                               int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
    return TCL_ERROR;
  STRETCH (inimg, TFROOT, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Square all the image pixel values.
 * This is an interface to the SADIE library function TFSQUARE,
 * which computes the squares of all the pixel values.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) squared image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_SqrStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
    return TCL_ERROR;
  STRETCH (inimg, TFSQUARE, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Re-map image greylevels using an arbitrary look-up table.
 * We require not only an image to transform, but a Tcl array (not a list)
 * whose indices include the numeric value of each greylevel in the range
 * found within the image, assigned a the new value to replace this in the
 * transformed image.  The SADIE library function TFTABLE handles the
 * actual trasformation.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(index,(numeric_value)) individual look-up table values.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_TblStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  uint32_t i;
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  char index[SLEN];

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
    return TCL_ERROR;
  nlev = (uint32_t)floor(inimg->gmax) - (uint32_t)floor(inimg->gmin) + 1;
  table = calloc (nlev, sizeof(PIXEL));
  if (table == NULL)
    return TCL_ERROR;
  for (i = 0; i < nlev; i++)
    {
      /* Read the input double array(index,i) */
      snprintf(index, SLEN - 1, "index,%u", i);
      if (GetSadiePixelFromObj2 (interp, objv[1], (const char *) index,
                                 table + i) != TCL_OK)
        {
          free (table);
          return TCL_ERROR;
        }
    }
  STRETCH (inimg, TFTABLE, &outimg);
  free (table);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Quantize image pixel values to a specified number of greylevels.
 * Given a image whose greylevels are continuous (down to the numerical
 * precision of the @c PIXEL type), this generates a new image with at most
 * the specified number of levels between a maximum and minimum value,
 * using the SADIE library function TFQUANT.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(steps) the number of levels.
 * - @c tcl_array_name(gmin) the maximum level.
 * - @c tcl_array_name(gmax) the minimum level.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) quantized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_QntStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "steps", &nlev)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "gmax", &gmax)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "gmin", &gmin)
          != TCL_OK))
    return TCL_ERROR;
  STRETCH (inimg, TFQUANT, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Apply a saturation stretch to an image.
 * Re-assign image pixel values above a specified threshold to the maximum
 * value in the image, and those below a specified threshold to the minimum
 * value in the image, using the SADIE library function TFSAT.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(gmin) set to the minimum below this level.
 * - @c tcl_array_name(gmax) set to the maximum above this level.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) thresholded image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_SatStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "gmax", &gmax)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "gmin", &gmin)
          != TCL_OK))
    return TCL_ERROR;
  STRETCH (inimg, TFSAT, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Convert a continuous image to a binary (two-valued) one.
 * This maps all the pixel values below a theshold to the current minimum
 * value, and all the pixel values greater than or equal to the threshold
 * to the current maximum value, using the SADIE library TFTHRESH function.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(threshold) values set to minimum below, maximum above.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) two-valued image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_ThreshStretchCmd (ClientData client_data, Tcl_Interp *interp,
                                 int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "thresh", &thresh)
          != TCL_OK))
    return TCL_ERROR;
  STRETCH (inimg, TFTHRESH, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Transform an image with the Local Range Modification algorithm.
 * The SADIE library function VSTRETCH does all the real work, evaluating
 * a maximal linear stretch for fixed-size square regions that completely
 * tile the image, interpolating between these to give a per-pixel
 * transformation, and applying this to the image.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) original image ID.
 * - @c tcl_array_name(blocksize) the size of the square regions (in pixels).
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_AdaptStretchCmd (ClientData client_data, Tcl_Interp *interp,
                                int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t blocksize;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "blocksize", &blocksize)
          != TCL_OK))
    return TCL_ERROR;
  VSTRETCH (inimg, blocksize, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Transform the greylevels in an image to match those in a reference image.
 * Given an arbitrary image to be transformed and a reference image
 * (of identical size), the SADIE library function MATCH can apply a
 * reference stretch according two one of two selectable algorithms:
 * either by inverting the cumulative distribution function (CDF) of
 * the reference image, or by using the observed joint distribution of
 * greylevels in the two images to derive an averaged look-up table
 * (LUT).  The SADIE library refers to the LUT option as the "least
 * squares transformation", but it does not do any kind of regression.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) the original image ID.
 * - @c tcl_array_name(inimg,addr1) the reference image ID.
 * - @c tcl_array_name(nlevels) the number of LUT entries or CDF bins.
 * - @c tcl_array_name(option) one of CDF or LSQ (as defined in @c sadie.h).
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Contrast_RefStretchCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  uint32_t nlevels;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "nlevels", &nlevels)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  MATCH (inimg1, inimg2, nlevels, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_Contrast_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_NormStretch",
                        Sadie_Contrast_NormStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_HisteqStretch",
                        Sadie_Contrast_HisteqStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_LinearStretch",
                        Sadie_Contrast_LinearStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_LogStretch",
                        Sadie_Contrast_LogStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_PWLStretch",
                        Sadie_Contrast_PWLStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_RootStretch",
                        Sadie_Contrast_RootStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_SqrStretch",
                        Sadie_Contrast_SqrStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_TblStretch",
                        Sadie_Contrast_TblStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_QntStretch",
                        Sadie_Contrast_QntStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_SatStretch",
                        Sadie_Contrast_SatStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_ThreshStretch",
                        Sadie_Contrast_ThreshStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_AdaptStretch",
                        Sadie_Contrast_AdaptStretchCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Contrast_RefStretch",
                        Sadie_Contrast_RefStretchCmd, NULL, NULL);
  return TCL_OK;
}
