/** @file Sadie_FileIO.c
 * A Tcl/Tk interface to the SADIE file input/output operations.
 * Please note that tclSadie also uses an add-on package for Tk called Img.
 * Some of the operations that are disabled here may nevertheless work
 * in tclSadie if it uses Img instead; for more information see
 * http://sourceforge.net/projects/tkimg/ for downloads and the
 * latest versions, but these days it should really be a standard
 * additional option in whatever Tcl/Tk distribution you are using.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_FileIO.c,v 2.10 2005/03/22 20:53:41 mmunro Exp $
*/

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sadie.h>
#include "tclsadie.h"

/**
 * Read a SADIE-formatted file into a newly allocated image.
 * The SADIE library function DISK2IMG should read files in either
 * the SADIE version 0 or version 1 format, and will allocate space for
 * the image.  We set the Tcl interpreter result to an ID string
 * representing the new image.  The Tcl object array always contains
 * the following elements:
 * - @c objv[0] the command name.
 * - @c objv[1] filesystem path to the file (as a string, not a list).
 * - @c objv[2] the file size.
 * - @c objv[3] a title for the newly read image.
 * 
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always four).
 * @param[in]  objv the Tcl object array.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_OpenCmd (ClientData client_data, Tcl_Interp* interp,
                      int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;
  uint64_t filesize;

  if (objc != 4)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "filename filesize title");
      return TCL_ERROR;
    }
  /* Read the size of the file */
  if (GetSadieUint64FromObj (interp, objv[2], &filesize) != TCL_OK)
    return TCL_ERROR;
  DISK2IMG (Tcl_GetString (objv[1]), 0, filesize, &theimg);
  return SetSadieNamedImageResult (interp, theimg, objv[3]);
}

/**
 * An interface to the TIFF2IMG function in the SADIE library (disabled).
 * This should read TIFF files into SADIE images, but is currently disabled.
 */
int
Sadie_FileIO_OpenTIFFCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;
  
  if (objc != 3)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "filename title");
      return TCL_ERROR;
    }
  /* TIFF2IMG (Tcl_GetString (objv[1]), 0, &theimg); */
  return SetSadieNamedImageResult (interp, theimg, objv[2]);
}

/**
 * An interface to the PNM2IMG function in the SADIE library (disabled).
 * This should read PNM files into SADIE images, but is currently disabled.
 */
int
Sadie_FileIO_OpenPNMCmd (ClientData client_data, Tcl_Interp* interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 3)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "filename title");
      return TCL_ERROR;
    }
  /* PNM2IMG (argv[1],&theimg); */
  return SetSadieNamedImageResult (interp, theimg, objv[2]);
}

/**
 * Import data from an arbitrarily defined image file into a SADIE image.
 * This is a generic image file reader that should handle most formats
 * using uncompressed regular arrays of pixels.  The SADIE library
 * function IMPORT allocates a new image to hold the imported data as
 * well as reading from the file.  We set the Tcl interpreter result
 * to an ID string representing the new image.  The lengthy array of 
 * Tcl object arguments have the following meanings:
 * - @c objv[0]  the command name.
 * - @c objv[1]  filesystem path to the file (as a string, not a list).
 * - @c objv[2]  the number of initial bytes (header bytes) to skip.
 * - @c objv[3]  the number of bands in the image.
 * - @c objv[4]  the number of lines in each image band.
 * - @c objv[5]  the number of pixels in each image line.
 * - @c objv[6]  the number of bits per pixel.
 * - @c objv[7]  format code: one of BIL, BSQ, or BIP from @c sadie.h.
 * - @c objv[8]  if subsampling is required (a Boolean code).
 * - @c objv[9]  first band in the subsample (counting from 1).
 * - @c objv[10] last band in the subsample (counting from 1).
 * - @c objv[11] increment between sampled bands (1 if contiguous).
 * - @c objv[12] the first line in the subsample (counting from 1).
 * - @c objv[13] the last line in the subsample (counting from 1).
 * - @c objv[14] increment between sampled lines (1 if contiguous).
 * - @c objv[15] the first pixel in the subsample (counting from 1).
 * - @c objv[16] the last pixel in the subsample (counting from 1).
 * - @c objv[17] increment between sampled pixels (1 if contiguous).
 * - @c objv[18] the file size.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp   the main Tcl interpreter.
 * @param[in]  objc     number of Tcl object arguments (always 19).
 * @param[in]  objv     the array of Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_ImportCmd (ClientData client_data, Tcl_Interp* interp,
                        int objc, Tcl_Obj * const objv[])
{
  int option;
  uint64_t filesize;
  uint64_t hdrsize;
  uint32_t nbnd, nlin, npix, nbit;
  int subsamp;
  uint32_t fromband, toband, incrband;
  uint32_t fromline, toline, incrline;
  uint32_t frompixel, topixel, incrpixel;
  IMAGE *theimg = NULL;

  if (objc < 19)
    {
      Tcl_WrongNumArgs (interp, 1, objv, 
                        "name hdrsize nbnd nlin npix nbit option subsamp fromband toband incrband fromline toline incrline frompixel topixel incrpixel filesize");
      return TCL_ERROR;
    }
  if ((GetSadieUint64FromObj (interp, objv[2], &hdrsize) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[3], &nbnd) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[4], &nlin) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[5], &npix) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[6], &nbit) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[7], &option) != TCL_OK)
      || (Tcl_GetBooleanFromObj (interp, objv[8], &subsamp) != TCL_OK))
    return TCL_ERROR;
  if (subsamp)
    {
      if ((GetSadieUintFromObj (interp, objv[9], &fromband) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[10], &toband) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[11], &incrband) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[12], &fromline) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[13], &toline) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[14], &incrline) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[15], &frompixel) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[16], &topixel) != TCL_OK)
          || (GetSadieUintFromObj (interp, objv[17], &incrpixel) != TCL_OK))
        return TCL_ERROR;
      if (fromband > 0)
        fromband--;
      if (toband > 0)
        toband--;
      if (fromline > 0)
        fromline--;
      if (toline > 0)
        toline--;
      if (frompixel > 0)
        frompixel--;
      if (topixel > 0)
        topixel--;
    }
  else
    {
      fromband = 0;
      fromline = 0;
      frompixel = 0;
      toband = nbnd-1;
      toline = nlin-1;
      topixel = npix-1;
      incrband = 1;
      incrline = 1;
      incrpixel = 1;
    }
  if (GetSadieUint64FromObj (interp, objv[18], &filesize) != TCL_OK)
    return TCL_ERROR;
  IMPORT (Tcl_GetString (objv[1]), 0, (short)option, filesize, hdrsize,
          nbnd, nlin, npix, nbit, fromband, fromline, frompixel,
          incrband, incrline, incrpixel, toband, toline, topixel, &theimg);
  return SetSadieNamedImageResult (interp, theimg, objv[1]);
}

/**
 * Write a SADIE (version 1) formatted image file.
 * Given a image ID string and the specified number of bits per pixel,
 * we use the SADIE library function IMG2DISK to write a SADIE v. 1
 * formatted image to the specified file (version 1 is the post-2003
 * SADIE file format; the original format is version 0, and isn't
 * handled here).  The Tcl object arguments are in an array with
 * the following layout:
 * - @c objv[0] the command name.
 * - @c objv[1] the image ID string.
 * - @c objv[2] filesystem path to the file (as a string, not a list).
 * - @c objv[3] the desired number of bits per pixel.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always four).
 * @param[in]  objv  the array containing the Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_SaveCmd(ClientData client_data, Tcl_Interp* interp,
                     int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;
  uint32_t bitsperpix;

  if (objc != 4)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "image_ID filename bitsperpixel");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[3], &bitsperpix) != TCL_OK))
    return TCL_ERROR;
  IMG2DISK (theimg, Tcl_GetString (objv[2]), bitsperpix, 0, 0);
  return TCL_OK;
}

/**
 * Write an Encapsulated PostScript formatted image file.
 * Given a image ID string and the specified number of bits per pixel,
 * we use the SADIE library function IMG2EPSF to write an EPS
 * formatted image to the specified file.  The Tcl object arguments
 * are in an array with the following layout:
 * - @c objv[0] the command name.
 * - @c objv[1] the image ID string.
 * - @c objv[2] filesystem path to the file (as a string, not a list).
 * - @c objv[3] the desired number of bits per pixel.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always four).
 * @param[in]  objv  the array holding the Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_SaveEPSCmd (ClientData client_data, Tcl_Interp* interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;
  uint32_t bitsperpix;

  if (objc != 4)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID filename bitsperpixel");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[3], &bitsperpix) != TCL_OK))
    return TCL_ERROR;
  IMG2EPSF (theimg, Tcl_GetString (objv[2]), bitsperpix, 0, 0);
  return TCL_OK;
}

/**
 * Write a TIFF formatted image file.
 * Given a image ID string and the specified number of bits per pixel,
 * we use the SADIE library function IMG2TIFF to write a TIFF
 * formatted image to the specified file.  This is probably better
 * handled by the Img add-on package for Tk.  The Tcl object argument
 * array is arranged as follows:
 * - @c objv[0] the command name.
 * - @c objv[1] the image ID string.
 * - @c objv[2] filesystem path to the file (as a string, not a list).
 * - @c objv[3] the desired number of bits per pixel.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always four).
 * @param[in]  objv  an array holding the Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_SaveTIFFCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;
  uint32_t bitsperpix;

  if (objc != 4)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID filename bitsperpixel");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[3], &bitsperpix) != TCL_OK))
    return TCL_ERROR;
  IMG2TIFF (theimg, Tcl_GetString (objv[2]), bitsperpix, 0, 0);
  return TCL_OK;
}

/**
 * An interface to the IMG2PNM function in the SADIE library (disabled).
 * This should write PNM files from SADIE images, but is currently disabled.
 */
int
Sadie_FileIO_SavePNMCmd (ClientData client_data, Tcl_Interp* interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 3)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID filename");
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
    return TCL_ERROR;
  /* IMG2PNM (theimg, Tcl_GetString (objv[2])); */
  return TCL_OK;
}

/**
 * Dump pixel values to a text file.
 * Given a image ID string, we use the SADIE library function IMG2TEXT
 * to dump the formatted numeric values of the image pixels to the
 * specified file.  The Tcl objects passed through the array argument
 * have the following meanings:
 * - @c objv[0] the command name.
 * - @c objv[1] the image ID string.
 * - @c objv[2] filesystem path to the file (as a string, not a list).
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always three).
 * @param[in]  objv  the Tcl object array.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_SaveTEXTCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 3)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID filename");
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
    return TCL_ERROR;
  IMG2TEXT (theimg, Tcl_GetString (objv[2]), 0, 0);
  return TCL_OK;
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_FileIO_Init (Tcl_Interp *interp) {
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_Open",
                        Sadie_FileIO_OpenCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_OpenTIFF",
                        Sadie_FileIO_OpenTIFFCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_OpenPNM",
                        Sadie_FileIO_OpenPNMCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_Import",
                        Sadie_FileIO_ImportCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_Save",
                        Sadie_FileIO_SaveCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_SaveEPS",
                        Sadie_FileIO_SaveEPSCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_SaveTIFF",
                        Sadie_FileIO_SaveTIFFCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_SavePNM",
                        Sadie_FileIO_SavePNMCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_FileIO_SaveTEXT",
                        Sadie_FileIO_SaveTEXTCmd, NULL, NULL);
  return TCL_OK;
}
