/** @file Sadie_Filter.c
 * Local filtering operations (convolutions), Fourier and related transforms.
 * Here are Tcl/Tk interfaces to most of the local convolution
 * kernel-based filtering operations provided by the SADIE library, as
 * well as the Fourier and Hartley transforms and the operations
 * derived from them (such as frequency-domain convolutions and power
 * spectrum determinations).  Please note that some of the operations
 * are not thread-safe or even re-entrant, using shared global
 * variables in the SADIE library (see the list of external variable
 * declarations for details).
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Filter.c,v 2.6 2005/03/22 21:41:38 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "tclsadie.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** @name SADIE library globals
 * Global variables within the SADIE library that augment function arguments.
 */
/*@{*/
extern uint32_t nlev; /**< Used by STRETCH, TFQUANT & WTFMAJ. */
extern uint32_t csize; /**< Used by WINDOW & WTFCAVG. */
extern double weight; /**< Used by WINDOW & WTFMAJ. */
extern double *count; /**< Used by WTFMAJ. */
extern PIXEL gmin; /**< Used by STRETCH, TFQUANT, TFSAT, TFSCALE & WTFMAJ. */
extern PIXEL gmax; /**< Used by STRETCH, TFINVERT, TFQUANT, TFSAT & TFSCALE. */
extern PIXEL thresh; /**< Used by STRETCH, TFTHRESH, WINDOW & WTFCAVG. */
/*@}*/

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixaddr[] = DEFAULT_IN_IMAGE_ADDR_INDEX;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
/*@}*/

/**
 * Compute an image power spectrum, either 2D, or 1D (row or column averages).
 * We use either the SADIE library function PSPECT (for the
 * two-dimensional spectrum) or PSPECT1D (for one-dimensional row or
 * column averages), which in turn use a Fast Fourier
 * Transform-derived power spectrum.  The option codes to select 2-D,
 * 1-D rows, or 1-D columns are @c NOR_PSPECT, @c ROW_PSPECT, or
 * @c COL_PSPECT respectively, defined in @c sadie.h.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to analyse.
 * - @c tcl_array_name(option) select 2-D, 1-D rows, or 1-D columns.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) the spectrum image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_PSpectCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE* inimg = NULL;
  IMAGE* outimg = NULL;
  int option;
  
  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  if (option == NOR_PSPECT) 
    PSPECT (inimg, &outimg);
  else
    PSPECT1D (inimg, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Convolve a image with a function via a Fast Fourier Transform.
 * We use the SADIE library function FCONVL to convolve one (arbitrary) image
 * with another whose values represent either a set of spatial weights
 * (such as a sampled point spread function), or a sampled frequency-domain
 * transfer function.  The codes @c PSF and @c OTF defined in @c sadie.h
 * specify which option to use.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) input image ID string.
 * - @c tcl_array_name(inimg,addr2) ID string for sampled function image.
 * - @c tcl_array_name(option) spatial or frequency-domain function.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) convolution result name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_FFTConvCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  FCONVL (inimg1, inimg2, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Convolve an image with a pre-defined function.
 * The SADIE library includes a function FILTER that uses a Fast Fourier
 * Transform to convolve an image with members of a family of filter functions
 * that specify an amplitude for a particular combination of two parameters
 * giving @a x and @a y spatial frequencies, defined in the interval [0, 1/2],
 * and two scaling parameters giving the radii of the filter in the frequency
 * domain, in cycles per pixel.  It also pre-defines some such functions,
 * so here we just select one from several of these pre-defined functions,
 * and convolve the image with it.  Note that the filter selection codes
 * differ from those in @c sadie.h:
 * -# @c FLTCONE cone filter.
 * -# @c FLTEXP negative exponential filter.
 * -# @c FLTGAUSS Gaussian filter.
 * -# @c FLTBOX box filter.
 * - @c FLTCYL cylinder filter is the default for all other values.
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(option) the filter selection code.
 * - @c tcl_array_name(xradius) the filter @a x radius.
 * - @c tcl_array_name(yradius) the filter @a y radius.
 * - @c tcl_array_name(parameter) subtract the filter from this, if non-zero.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) convolution result name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_FFTFiltCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int option;
  double xradius, yradius, parameter;

  double (*flt) (double, double, double, double); 

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "xradius", &xradius)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "yradius", &yradius)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "parameter", &parameter)
          != TCL_OK))
    return TCL_ERROR;
  switch (option)
    {
    case 1:
      flt = FLTCONE;
      break;
    case 2:
      flt = FLTEXP;
      break;
    case 3:
      flt = FLTGAUSS;
      break;
    case 4:
      flt = FLTBOX;
      break;
    default:
      flt = FLTCYL;
      break;
    }
  FILTER (inimg, flt, xradius, yradius, parameter, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Compute a forward or inverse two-dimensional Fourier transform of an image.
 * The SADIE library function FFT2D implements a two-dimensional Fast Fourier
 * Transform algorithm, using the codes @c FORWARD or @c INVERSE (defined in
 * @c sadie.h ) to show the direction of the transform.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(option) the direction selection code.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transform result name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_FFTCmd (ClientData client_data, Tcl_Interp *interp,
                     int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  FFT2D (inimg, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Compute a forward or inverse two-dimensional Hartley transform of an image.
 * The SADIE library function FHT2D implements a two-dimensional Fast Hartley
 * Transform algorithm, using the codes @c FORWARD or @c INVERSE (defined in
 * @c sadie.h ) to show the direction of the transform.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(option) the direction selection code.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transform result name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_FHTCmd (ClientData client_data, Tcl_Interp *interp,
                     int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  FHT2D (inimg, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Convolve an image with a kernel defined in a Tcl array.
 * The SADIE library function SCONVL convolves an image with a two-
 * dimensional spatial kernel.  Here we feed it the kernel weights
 * from Tcl array elements, first dividing them by a constant we
 * specify, and apply it to the specified image.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(lines) the diameter of the kernel in lines.
 * - @c tcl_array_name(pix) the diameter of the kernel in pixels.
 * - @c tcl_array_name(factor) divide the kernel weights by this factor.
 * - @c tcl_array_name(mask,(line_number),(pixel_number)) individual weights.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) convolved result name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_MaskConvCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
    uint32_t j, k;
    uint32_t lines, pix;
    size_t a;
    IMAGE *inimg = NULL;
    IMAGE *outimg = NULL;
    double factor, wt;
    PIXEL *psf = NULL;
    char index[SLEN];

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "factor", &factor)
          != TCL_OK))
    return TCL_ERROR;
  psf = calloc (((size_t) lines * pix), sizeof(PIXEL));
  if (psf == NULL)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Cannot allocate memory!", -1);
      return TCL_ERROR;
    }                                                                         
  for (a=0, j=0; j<lines; j++)
    {
      for (k=0; k<pix; k++)
        {
          /* Read the input double array(mask,j,k) */
          snprintf (index, SLEN - 1, "mask,%d,%d", j, k);
          if (GetSadieDoubleFromObj2 (interp, objv[1], (const char *) index,
                                      &wt) == TCL_OK)
            psf[a++] = (PIXEL) (wt / factor);
          else
            {
              free (psf);
              return TCL_ERROR;
            }
        }
    }
  SCONVL (inimg, psf, lines, pix, &outimg);
  free(psf);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Convolve an image with a spatial kernel defined in another image.
 * The SADIE library function SCONVL convolves an image with a two-
 * dimensional spatial kernel.  Here we feed it the kernel weights
 * from another SADIE image, first dividing them by a constant we
 * specify, and apply it to the specified image.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(kimg,addr) the ID of the kernel image.
 * - @c tcl_array_name(normalize) whether to normalize weights to sum to 1.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) convolved result name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_ImageConvCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  uint32_t j, k;
  size_t n, a, b, c;
  int normalize;
  uint32_t lines, pix;
  double factor;
  PIXEL pfactor;
  PIXEL *psf = NULL;
  IMAGE *inimg = NULL;
  IMAGE *kimg = NULL;
  IMAGE *outimg = NULL;
    
  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "kimg,addr", &kimg)
          != TCL_OK)
      || (GetSadieBooleanFromObj2 (interp, objv[1], "normalize", &normalize)
          != TCL_OK))
    return TCL_ERROR;
  lines = kimg->nlin;
  pix = kimg->npix;
  n = (size_t)lines * pix;
  psf = calloc (n, sizeof (PIXEL));		
  if (psf == NULL)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Cannot allocate memory!", -1);
      return TCL_ERROR;
    }                                                                         
  for (a=0, j=0; j<lines; j++)
    {
      for (k=0; k<pix; k++)
        psf[a++] = kimg->data[0][j][k];
    }
  if (normalize)
    {
      factor = 0.0;
      for (b = n; b != 0; factor += psf[--b]);
      pfactor = (PIXEL) factor;
      for (c = n; c != 0; psf[--c] /= pfactor);
    }
  SCONVL (inimg, psf, lines, pix, &outimg);
  free(psf);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/** Fixed string indexes into whatever Tcl array holds the component names. */
static const char *comp_label_index[2] = { "magname", "dirname" };

/**
 * Convolve the image with one of a set of pre-defined gradient operators.
 * The SADIE library includes a generic gradient function, GRADIENT, and a
 * small set of 3 X 3 gradient kernels that the EDGE function can supply
 * to it to provide some well-known gradient operators.  Here we apply
 * EDGE to an image, generating both a magnitude and a direction image,
 * and specify the particular gradient operator by a code that should be
 * one of the constants @c SOBEL, @c SOBEL2, @c SOBEL3, @c ROBERTS,
 * @c ROBERTS2, or @c PREWITT, defined in the @c sadie.h header.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(option) the operator selection code.
 * - @c tcl_array_name(magname) name for resulting gradient magnitude image.
 * - @c tcl_array_name(dirname) name for resulting gradient direction image.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_LibGradCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *comp_img[2] = { NULL, NULL };
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  EDGE (inimg, (short)option, comp_img, comp_img + 1);
  return SetSadieComponentObj2 (interp, objv[1], comp_label_index, comp_img);
}

/** The width of the gradient operator kernel, in pixels. */
#define GRADIENT_KERNEL_SIZE 3
/** Use this constant in most places that need the gradient kernel width. */
static const short gradient_kernel_size = GRADIENT_KERNEL_SIZE;
/** The storage required for the gradient operator kernel, in pixels. */
#define GRADIENT_KERNEL_AREA 9

/**
 * Apply a gradient operator defined in Tcl array elements to an image.
 * The SADIE library function GRADIENT generates paired gradient direction
 * and magnitude images from an input image and two square convolution
 * kernels, defining the components of the gradient in the @a x and @a y
 * (horizontal and vertical) directions.  Here we apply it to an image,
 * taking the kernel weights from Tcl array elements.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(hmask,(row_index),(column_index)) horizontal weights.
 * - @c tcl_array_name(vmask,(row_index),(column_index)) vertical weights.
 * - @c tcl_array_name(magname) name for resulting gradient magnitude image.
 * - @c tcl_array_name(dirname) name for resulting gradient direction image.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int Sadie_Filter_MaskGradCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  int j, k, n;
  IMAGE *inimg = NULL;
  IMAGE *comp_img[2] = { NULL, NULL };
  PIXEL hpsf[GRADIENT_KERNEL_AREA], vpsf[GRADIENT_KERNEL_AREA];
  char index[SLEN];
  double hweight, vweight;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if (GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
    return TCL_ERROR;
  for (n = 0, j = 0; j < gradient_kernel_size; j++)
    {
      for (k = 0; k < gradient_kernel_size; k++)
        {
          /* Read the input PIXEL array(hmask,j,k) */
          snprintf (index, SLEN - 1, "hmask,%d,%d", j, k);
          if (GetSadieDoubleFromObj2 (interp, objv[1],
                                      (const char *) index, &hweight)
              != TCL_OK)
            return TCL_ERROR;
          hpsf[n] = (PIXEL) hweight;
          /* Read the input PIXEL array(vmask,j,k) */
          snprintf(index, SLEN - 1, "vmask,%d,%d", j, k);
          if (GetSadieDoubleFromObj2 (interp, objv[1],
                                      (const char *) index, &vweight)
              != TCL_OK)
            return TCL_ERROR;
          vpsf[n++] = (PIXEL) vweight;
	}
    }
  GRADIENT (inimg, vpsf, hpsf, gradient_kernel_size, comp_img, comp_img + 1);
  return SetSadieComponentObj2 (interp, objv[1], comp_label_index, comp_img);
}

/**
 * Apply a gradient operator defined by a pair of images to a third image.
 * We interpret the pixel values in a pair of SADIE images as discrete
 * versions of the functions that define the gradient in the orthogonal
 * @a x (horizontal) and @a y (vertical) directions: the paired square
 * convolution kernels the SADIE function GRADIENT can use to derive paired
 * gradient magnitude and direction images from the input image.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of image to filter.
 * - @c tcl_array_name(himg,addr) ID of the image holding horizontal weights.
 * - @c tcl_array_name(vimg,addr) ID of the image holding vertical weights.
 * - @c tcl_array_name(magname) name for resulting gradient magnitude image.
 * - @c tcl_array_name(dirname) name for resulting gradient direction image.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_ImageGradCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  uint32_t lines, pix, j, k;
  size_t n, a;
  short ssize;
  PIXEL *hpsf = NULL;
  PIXEL *vpsf = NULL;
  IMAGE *inimg = NULL;
  IMAGE *himg = NULL;
  IMAGE *vimg = NULL;
  IMAGE *comp_img[2] = { NULL, NULL };

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "himg,addr", &himg)
          != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "vimg,addr", &vimg)
          != TCL_OK))
    return TCL_ERROR;
  if ((himg->nlin != himg->npix) || (himg->nlin != vimg->nlin)
      || (himg->nlin != vimg->npix))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Image kernels must be square and equal in size.", -1);
      return TCL_ERROR;
    }
  lines = himg->nlin;
  pix = himg->npix;
  n = (size_t)lines * pix;
  ssize = (short) n;
  if ((size_t) ssize != n)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Image kernels too large for the SADIE GRADIENT function.", -1);
      return TCL_ERROR;
    }
  hpsf = calloc (n, sizeof (PIXEL));		
  if (hpsf == NULL)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Cannot allocate memory!", -1);
      return TCL_ERROR;
    }
  vpsf = calloc (n, sizeof (PIXEL));		
  if (vpsf == NULL)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Cannot allocate memory!", -1);
      free (hpsf);
      return TCL_ERROR;
    }
  for (a = 0, j = 0; j < lines; j++)
    {
      for (k = 0; k < pix; k++)
        {
          hpsf[a] = himg->data[0][j][k];
          vpsf[a++] = vimg->data[0][j][k];
        }
    }
  GRADIENT(inimg, vpsf, hpsf, ssize, comp_img, comp_img + 1);
  free (vpsf);
  free (hpsf);
  return SetSadieComponentObj2 (interp, objv[1], comp_label_index, comp_img);
}

/**
 * Apply one of the SADIE library's local windowing operations to an image.
 * The SADIE library defines several functions for local operations and a
 * single function, WINDOW, to apply one of them to an image.  Here we
 * invoke WINDOW with an image and parameters taken from Tcl variables,
 * and with a pointer to the desired operation that's already set to
 * the appropriate SADIE function (in most cases we'll be using this from
 * from a function that implements a Tcl command, and this will have picked
 * the appropriate function).
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of the input image.
 * - @c tcl_array_name(lines) the number of lines in the window operator.
 * - @c tcl_array_name(pix) the number of pixels per line in the window.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @param[in]  wtf   the particular local operation to apply.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static int
apply_sadie_window (Tcl_Interp *interp, int objc, Tcl_Obj * const objv[],
                    PIXEL (*wtf)(PIXEL *, uint32_t, uint32_t))
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t lines, pix;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK))
    return TCL_ERROR;
  WINDOW (inimg, wtf, lines, pix, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Compute a local variance map from a sliding sub-window within an image.
 * The SADIE library function WTFVAR computes the variance within a
 * window: use the WINDOW function to apply it, via the
 * apply_sadie_window() function.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_VarMapCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  return apply_sadie_window (interp, objc, objv, WTFVAR);
}

/**
 * Compute minima for a sliding sub-window within an image.
 * The SADIE library function WTFMIN finds the minimum value within a
 * fixed-size window: use the WINDOW function to apply it, via the
 * apply_sadie_window() function, to every sliding window position in
 * an image.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_MinFiltCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  return apply_sadie_window (interp, objc, objv, WTFMIN);
}

/**
 * Compute maxima for a sliding sub-window within an image.
 * The SADIE library function WTFMAX finds the maximum value within a
 * fixed-size window: use the WINDOW function to apply it, via the
 * apply_sadie_window() function, to every sliding window position in
 * an image.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_MaxFiltCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  return apply_sadie_window (interp, objc, objv, WTFMAX);
}

/**
 * Compute greylevel diversity for a sliding sub-window within an image.
 * The SADIE library function WTFDIVER computes greylevel diversity
 * within a window: use the WINDOW function to apply it, via the
 * apply_sadie_window() function.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_DivMapCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  return apply_sadie_window (interp, objc, objv, WTFDIVER);
}

/**
 * Apply a local majority filter to an image.
 * The SADIE library function WTFMAJ implements a local majority
 * filter (a little more elaborate than a modal filter, using the
 * specified weighting parameter to handle the treatment of the mid-point
 * of the distribution of values within a fixed-size window).  It requires
 * some messy initialization of global variables within the SADIE library,
 * including allocating an array in which to summarize the value
 * distribution.  We set the parameters from Tcl variables, assign all the
 * necessary global variables, and let the standard SADIE library function
 * WINDOW apply the filter to the image via a sliding window.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of the input image.
 * - @c tcl_array_name(lines) the number of lines in the sliding window.
 * - @c tcl_array_name(pix) the number of pixels per line in the window.
 * - @c tcl_array_name(weight) for the mid-range distribution bin total.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_MajFiltCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t lines, pix;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
      != TCL_OK)
  || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
      != TCL_OK)
  || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
      != TCL_OK)
  || (GetSadieDoubleFromObj2 (interp, objv[1], "weight", &weight)
      != TCL_OK))
    return TCL_ERROR;
  RANGE (inimg);
  nlev = (uint32_t)inimg->gmax - (uint32_t)inimg->gmin + 1;
  gmin = inimg->gmin;
  gmax = inimg->gmax;
  count = NULL;
  count = calloc ((size_t) nlev , sizeof (double));
  if (count == NULL)
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "Cannot allocate memory!", -1);
      return TCL_ERROR;
    }                                                                         
  WINDOW (inimg, WTFMAJ, lines, pix, &outimg);
  free(count);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Apply a median filter to an image.
 * The SADIE library function MEDIAN finds the median of the values for each
 * position of a sliding window within an image.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of the input image.
 * - @c tcl_array_name(lines) the number of lines in the sliding window.
 * - @c tcl_array_name(pix) the number of pixels per line in the window.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_MedFiltCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t lines, pix;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK))
    return TCL_ERROR;
  MEDIAN (inimg, lines, pix, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Determine the conditional average of an image.
 * The SADIE library function WTFCAVG implements a conditional average
 * operation within a fixed-sized window.  If none of the conditions
 * are met, this is identical to the value of the pixel at the center
 * of the window.  If there is no limit on the number of divergent
 * pixels, it is the simple mean of all those pixels whose values are
 * less than the specified threshold from the value of the center
 * pixel.  If there are pixels whose values lie beyond this threshold
 * from the center pixel (divegent pixels), the total count of
 * divergent pixels is greater than the limit, and the range of their
 * values is less than the threshold, it is the simple mean of the
 * divergent pixels. We set the parameters from Tcl variables, assign
 * all the necessary global variables (and should maintain the
 * convention that the @c csize variable is zero if there is no limit
 * on the number of divergent pixels), and let the standard SADIE
 * library function WINDOW apply the filter to the image via a sliding
 * window.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of the input image.
 * - @c tcl_array_name(lines) the number of lines in the sliding window.
 * - @c tcl_array_name(pix) the number of pixels per line in the window.
 * - @c tcl_array_name(size) zero, or the divergent pixel count limit.
 * - @c tcl_array_name(thresh) threshold for comparing pixel values.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_CAvgFiltCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t lines, pix;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "size", &csize)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "thresh", &thresh)
          != TCL_OK))
    return TCL_ERROR;
  WINDOW (inimg, WTFCAVG, lines, pix, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Filter_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Filter_PSpect",
                        Sadie_Filter_PSpectCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_FFTConv",
                        Sadie_Filter_FFTConvCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_FFTFilt",
                        Sadie_Filter_FFTFiltCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_FFT",
                        Sadie_Filter_FFTCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_FHT",
                        Sadie_Filter_FHTCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_MaskConv",
                        Sadie_Filter_MaskConvCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_ImageConv",
                        Sadie_Filter_ImageConvCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_LibGrad",
                        Sadie_Filter_LibGradCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_ImageGrad",
                        Sadie_Filter_ImageGradCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_MaskGrad",
                        Sadie_Filter_MaskGradCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_VarMap",
                        Sadie_Filter_VarMapCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_DivMap",
                        Sadie_Filter_DivMapCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_MajFilt",
                        Sadie_Filter_MajFiltCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_MedFilt",
                        Sadie_Filter_MedFiltCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_MinFilt",
                        Sadie_Filter_MinFiltCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_MaxFilt",
                        Sadie_Filter_MaxFiltCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Filter_CAvgFilt",
                        Sadie_Filter_CAvgFiltCmd, NULL, NULL);
  return TCL_OK;
}

