/** @file Sadie_General.c
 * Application and image data utilities.
 * A miscellaneous collection of Tcl/Tk interfaces to SADIE image data
 * (recovering and setting image attributes, disposing of images), and
 * to the housekeeping operations needed by the SADIE library in the
 * context of the whole application (initialization and finalization).
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_General.c,v 2.3 2005/03/22 21:16:14 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include <string.h>
#include "tclsadie.h"
#include "Sadie_Index.h"

/** The global hash table structures, associating images and their metadata. */
extern sad_doclistp_t global_doclist;

/**
 * Dispose of an image and all its associated metadata such as ROIs.
 * We delete the image's entry from the global hash table, leaving this
 * operation to invoke the usual SADIE library image cleanup function,
 * RELIMG.
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_CloseCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  sad_dispose_doc (global_doclist, Tcl_GetString (objv[1]));
  return TCL_OK;
}

/**
 * SADIE library initialization.
 * Invoke the operation to initialize the SADIE library from Tcl code.
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always one).
 * @param[in]  objv  Tcl object argument list (no arguments, just the name).
 * @return @c TCL_OK unconditionally.
 */ 
int
Sadie_General_InitializeProgCmd (ClientData client_data, Tcl_Interp *interp,
                                 int objc, Tcl_Obj * const objv[])
{
  INIT ();
  return TCL_OK;
}

/**
 * SADIE library termination.
 * Invoke the operation to terminate the SADIE library from Tcl code.
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always one).
 * @param[in]  objv  Tcl object argument list (no arguments, just the name).
 * @return @c TCL_OK unconditionally.
 */ 
int
Sadie_General_FinishProgCmd (ClientData client_data, Tcl_Interp *interp,
                             int objc, Tcl_Obj * const objv[])
{
  FINI ();
  return TCL_OK;
}

/**
 * Pass the recorded width of a SADIE image to Tcl.
 * We recover the number of pixels per line from the SADIE image,
 * and assign this to the Tcl interpreter result object (as a long value).
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_WidthCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  Tcl_SetLongObj (Tcl_GetObjResult (interp), theimg->npix);
  return TCL_OK;
}

/**
 * Pass the recorded height of a SADIE image to Tcl.
 * We recover the number of lines per band from the SADIE image,
 * and assign this to the Tcl interpreter result object (as a long value).
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_HeightCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  Tcl_SetLongObj (Tcl_GetObjResult (interp), theimg->nlin);
  return TCL_OK;
}

/**
 * Pass the newly determined minimum SADIE image pixel value to Tcl.
 * We determine the image-wide minimum pixel value in the SADIE image,
 * and assign this to the Tcl interpreter result object (as a double value).
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_GminCmd (ClientData client_data, Tcl_Interp *interp,
                       int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  RANGE (theimg);
  Tcl_SetDoubleObj (Tcl_GetObjResult (interp), theimg->gmin);
  return TCL_OK;
}

/**
 * Pass the newly determined minimum SADIE image pixel value to Tcl.
 * We determine the image-wide minimum pixel value in the SADIE image,
 * and assign this to the Tcl interpreter result object (as a double value).
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_GmaxCmd (ClientData client_data, Tcl_Interp *interp,
                       int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  RANGE (theimg);
  Tcl_SetDoubleObj (Tcl_GetObjResult (interp), theimg->gmax);
  return TCL_OK;
}

/**
 * Pass the recorded number of bands in a SADIE image to Tcl.
 * We recover the number of bands per image from the SADIE image,
 * and assign this to the Tcl interpreter result object (as a long value).
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_GetNumBandsCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  Tcl_SetLongObj (Tcl_GetObjResult (interp), theimg->nbnd);
  return TCL_OK;
}

/** The number of bands to include in pixel value displays. */
#define MAX_DISPLAY_CHANNEL 3
/** Use this constant for most references to the limit on displayed bands. */
static const int max_display_channel = MAX_DISPLAY_CHANNEL;

/**
 * Make a Tcl list from the values of a sub-set of bands at a single point.
 * Given a particular position (pixel position within a line and line number),
 * find the values for a sub-set of bands, of size @c max_display_channel;
 * if the image has fewer than this number of bands, duplicate the value
 * of the first band to make up the missing values.  Set the Tcl interpreter
 * result to the sub-set of values, as a list of doubles.  The Tcl object
 * arguments required are passed in an array with the following arrangement:
 * - @c objv[0] the command name.
 * - @c objv[1] the image ID string.
 * - @c objv[2] the pixel position within a line.
 * - @c objv[3] the line number.
 * 
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always four).
 * @param[in]  objv  the array of Tcl object arguments.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_GetPixelCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  uint32_t xval, yval, band;
  IMAGE *theimg = NULL;
  Tcl_Obj * channel_subset[MAX_DISPLAY_CHANNEL];
  int err = TCL_OK;

  if (objc != 4)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID xcoord ycoord");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)) || (theimg->nbnd == 0))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj (interp, objv[2], &xval) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[3], &yval) != TCL_OK))
    return TCL_ERROR;
  band = 0;
  while ((err == TCL_OK)
         && (band < theimg->nbnd) && (band < max_display_channel))
    {
      channel_subset[band]
        = Tcl_NewDoubleObj (theimg->data[band][yval][xval]);
      if (channel_subset[band] == NULL)
        err = TCL_ERROR;
      else
        band++;
    }
  if (err != TCL_OK)
    {
      while (band != 0)
        Tcl_DecrRefCount (channel_subset[--band]);
    }
  else
    {
      while (band < max_display_channel)
        channel_subset[band++] = channel_subset[0];
      Tcl_SetListObj (Tcl_GetObjResult (interp),
                      max_display_channel, channel_subset);
    }
  return err;
}

/**
 * Set the SADIE image text from a Tcl string value.
 * Replace the text in a SADIE image structure with new text taken
 * from a Tcl string value, truncating it if it is too long.  The Tcl
 * object arguments required are passed in an array with the following
 * arrangement:
 * - @c objv[0] the command name.
 * - @c objv[1] the image ID string.
 * - @c objv[2] the new text string.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always three).
 * @param[in]  objv  the Tcl object argument array.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_SetNameCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *theimg;

  if (objc != 3)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID name");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  strncpy (theimg->text, Tcl_GetString(objv[2]), TLEN - 1);
  return TCL_OK;
}

/**
 * Register a SADIE image within the global hash table.
 * We insert a new entry in the global hash table, representing the image
 * and a table of additional properties (initially blank).
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the image ID string).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_IndexImageCmd (ClientData client_data, Tcl_Interp *interp,
                             int objc, Tcl_Obj * const objv[])
{
  int err = 0;
  IMAGE *theimg;
  char msg[SLEN];
  const char *dkey;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "image_ID");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &theimg) != TCL_OK)
      || (GetSadieDkeyFromObj (objv[1], &dkey) != TCL_OK)
      || (!CHECKIMG (theimg)))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  err = sad_put_doc (global_doclist, dkey, 8, theimg);
  if (err)
    {
      snprintf (msg, SLEN - 1,
                "Sadie_General_IndexImage failed to index the image: ID = %s, error = %d",
                Tcl_GetString (objv[1]), err);
      Tcl_SetStringObj (Tcl_GetObjResult (interp), msg, -1);
      return TCL_ERROR;
    }
  return TCL_OK;
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_General_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_General_Close",
                        Sadie_General_CloseCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_InitializeProg",
                        Sadie_General_InitializeProgCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "Sadie_General_FinishProg",
                       Sadie_General_FinishProgCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_Width",
                        Sadie_General_WidthCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_Height",
                        Sadie_General_HeightCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_Gmin",
                        Sadie_General_GminCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_Gmax",
                        Sadie_General_GmaxCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_GetNumBands",
                        Sadie_General_GetNumBandsCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_GetPixel",
                        Sadie_General_GetPixelCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_SetName",
                        Sadie_General_SetNameCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_General_IndexImage",
                        Sadie_General_IndexImageCmd, NULL, NULL);
  return TCL_OK;
}

