/** @file Sadie_Geometry.c
 * A Tcl/Tk interface to the SADIE image warping and resampling functions.
 * The SADIE library provides some geometric transformations (from
 * simple re-scalings and rotations to fitting polynomial distortion
 * models with variable numbers of terms), and the corresponding
 * re-sampling operations to generate the transformed images.  For
 * more details see Chapter 7, Section 4 "Distortion Correction" in
 * R. A. Schowengerdt @e Remote @e Sensing (2nd. ed.) Academic Press:
 * 1997.  Here we define some interfaces to Tcl code, without adding
 * any additional operations.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Geometry.c,v 2.6 2005/03/22 22:08:40 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <stdio.h>
#include <string.h>
#include <sadie.h>
#include "tclsadie.h"

/** @name Global polynomial definitions
 * The one copy of the polynomial terms shared by all the functions.
/*@{*/
/** The maximum number of polynomial terms we can handle. */
#define MAX_POLY_TERM 6
static double xpolyterms[MAX_POLY_TERM] = { 0.0, 1.0, 0.0, 0.0, 0.0, 0.0 };
static double ypolyterms[MAX_POLY_TERM] = { 0.0, 0.0, 1.0, 0.0, 0.0, 0.0 };
static int numpolyterms = 3; /**< Number of valid terms, <= MAX_POLY_TERM .*/
/*@}*/

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixaddr[] = DEFAULT_IN_IMAGE_ADDR_INDEX;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
/*@}*/

/**
 * Resample an image using the averages within a sliding rectangular window.
 * The SADIE library function RESAMPL resamples an image at the specified
 * horizontal and vertical intervals via averages within a sliding rectangular
 * window, whose size can also be specified.  We set these parameters from
 * Tcl variables and invoke it.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) input image ID string.
 * - @c tcl_array_name(lines) the number of lines in the sliding window.
 * - @c tcl_array_name(pix) the number of pixels per line in the window.
 * - @c tcl_array_name(incrlines) the line resampling interval.
 * - @c tcl_array_name(incrpix) the pixel resampling interval.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_AvgSubCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t lines, pix, incrlines, incrpix;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incrlines", &incrlines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incrpix", &incrpix)
          != TCL_OK))
    return TCL_ERROR;
  RESAMPL (inimg, lines, pix, incrlines, incrpix, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Extract a sub-image from an image, with optional scaling (by decimation).
 * The SADIE library function SUBSAMPL extracts a sub-image from a SADIE
 * image, optionally shrinking the extracted image (selecting
 * pixels at the specified horizontal and vertical spacings, without
 * any averaging or interpolation).
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c tcl_array_name(from,band) the first band to include in the subimage.
 * - @c tcl_array_name(to,band) the last band to include in the subimage.
 * - @c tcl_array_name(incr,band) band spacing interval (1 for each band).
 * - @c tcl_array_name(from,line) the first line to include in the subimage.
 * - @c tcl_array_name(to,line) the last line to include in the subimage.
 * - @c tcl_array_name(incr,line) line spacing interval (1 for no shrinking).
 * - @c tcl_array_name(from,pixel) the first pixel to include in the subimage.
 * - @c tcl_array_name(to,pixel) the last pixel to include in the subimage.
 * - @c tcl_array_name(incr,pixel) pixel spacing interval (1 for no shrinking).
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_WinSubCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t fromband, frompix, fromline;
  uint32_t toband, topix, toline;
  uint32_t incrband, incrpix, incrline;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "from,band", &fromband)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "from,pix", &frompix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "from,line", &fromline)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "to,band", &toband)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "to,pix", &topix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "to,line", &toline)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr,band", &incrband)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr,pix", &incrpix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr,line", &incrline)
          != TCL_OK))
    return TCL_ERROR;
  SUBSAMPL (inimg, fromband - 1, fromline - 1, frompix - 1,
            incrband, incrline, incrpix,
            ((toband - fromband) + 1),
            ((toline - fromline) + 1),
            ((topix - frompix) + 1), &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Re-scale an image using cubic, bilinear, or nearest-neighbor interpolation.
 * The SADIE library function GEOMWARP can apply a wide range of polynomial
 * transformations to an image, and can use three different re-sampling
 * methods in generating the transformed image.  Here we use the EXPAND
 * function, which invokes GEOMWARP to handle the simple specific case
 * of uniform but independent scalings in the horizontal and vertical
 * directions, unless the scale factors are trivial integer multipliers,
 * in which case it uses the DUPL SADIE function instead.  Despite the Tcl
 * names, the @c lines and @c pix indices are to scale factors, not sizes.
 * The @a option variable determines the interpolation technique: negative
 * values for cubic interpolation (with the value being interpreted as the
 * free parameter @a alpha for this method), zero values imply bilinear
 * interpolation, and positive values nearest-neighbor interpolation.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c tcl_array_name(lines) the vertical scale factor.
 * - @c tcl_array_name(pix) the horizontal scale factor.
 * - @c tcl_array_name(option) the interpolation option.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) resampled image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_ScaleCmd (ClientData client_data, Tcl_Interp* interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  double linescale, pixscale, option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "lines", &linescale)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "pix", &pixscale)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  EXPAND (inimg, linescale, pixscale, option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Rotate an image using cubic, bilinear, or nearest-neighbor interpolation.
 * The SADIE library function GEOMWARP can apply a wide range of
 * polynomial transformations to an image, and can use three different
 * re-sampling methods in generating the transformed image.  Here we
 * use the ROTATE function, which uses the GEOMCOEF function to derive
 * the polynomial coefficients equivalent to the specified rotation,
 * before invoking GEOMWARP itself, or recognizes the special cases
 * that TRNSFORM can handle via simple flips.  The @a angle parameter is in
 * degrees; the @a option variable determines the interpolation
 * technique: negative values for cubic interpolation (with the value
 * being interpreted as the free parameter @a alpha for this method),
 * zero values imply bilinear interpolation, and positive values
 * nearest-neighbor interpolation.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) raw image ID string.
 * - @c tcl_array_name(angle) the angle by which to rotate the image.
 * - @c tcl_array_name(option) the interpolation option.
 * - @c tcl_array_name(graylevel) the greyvalue with which to fill blanks.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) rotated image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_RotateCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  double angle, option;
  PIXEL greylevel;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "angle", &angle)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "graylevel", &greylevel)
          != TCL_OK))
    return TCL_ERROR;
  ROTATE (inimg, angle, greylevel, option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Rotate or reflect an image by pre-defined amounts.
 * The SADIE library function TRNSFORM flips an image according to a small
 * number of simple transformations, selected by one of a set of codes
 * defined in @c sadie.h :
 * - @c FLIP0  reflect about a vertical axis.
 * - @c FLIPM45 reflect about a -45 degree diagonal.
 * - @c FLIPP45 reflect about a +45 degree diagonal.
 * - @c FLIP90 reflect about a horizontal axis.
 * - @c ROTPM0 copy without transformation.
 * - @c ROTP90 rotate +90 degrees.
 * - @c ROTM90 rotate -90 degrees.
 * - @c ROT180 rotate 180 degrees.
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) input image ID string.
 * - @c tcl_array_name(option) the transformation code number.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_MirrorCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  TRNSFORM (inimg, option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Derive a best-fit polynomial from pairs of image and control points.
 * Given a set of pairs of points (one in image coordinates, the other
 * in the desired reference frame, such as ground control points in map
 * coordinates), the SADIE library function GEOMCOEF determines a polynomial
 * that defines a good mapping between the two coordinate frames, with
 * "good" implying some minimum of the squared errors (defferences between
 * actual point positions and those from the transformation defined by the
 * polynomial).  The number of point pairs used by GEOMCOEF can vary,
 * as can the number of terms in the polynomial it generates (provided they
 * are no ambiguities by having this greater than the number of point pairs),
 * but when providing the coordiates via Tcl array elements we always
 * give the maximum possible number and ignore those that are unnecesary.
 * At the moment the polynomial terms end up in global arrays, horribly
 * usafe from the point of re-entrancy or thread safety.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(pointcount) the number of point pairs.
 * - @c tcl_array_name(termcount) the number of polynomial terms.
 * - @c tcl_array_name(trans,x,(point_number)) distorted frame x coordinates.
 * - @c tcl_array_name(trans,y,(point_number)) distorted frame y coordinates.
 * - @c tcl_array_name(ref,x,(point_number)) reference frame x coordinates.
 * - @c tcl_array_name(ref,y,(point_number)) reference frame y coordinates.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_ContPtsCmd (ClientData client_data, Tcl_Interp* interp,
                           int objc, Tcl_Obj * const objv[])
{
  int npts, nterms;
  double transx[MAX_POLY_TERM], refx[MAX_POLY_TERM];
  double transy[MAX_POLY_TERM], refy[MAX_POLY_TERM];
  char transxindex[SLEN], refxindex[SLEN];
  char transyindex[SLEN], refyindex[SLEN];
  int i;
    
  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieIntFromObj2 (interp, objv[1], "pointcount", &npts)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "termcount", &nterms)
          != TCL_OK))
    return TCL_ERROR;
  for (i = 0; i < MAX_POLY_TERM; i++)
    {
      snprintf (transxindex, SLEN - 1, "trans,x,%d", i + 1);
      snprintf (transyindex, SLEN - 1, "trans,y,%d", i + 1);
      snprintf (refxindex, SLEN - 1, "ref,x,%d", i + 1);
      snprintf (refyindex, SLEN - 1, "ref,y,%d", i + 1);
      if ((GetSadieDoubleFromObj2 (interp, objv[1], transxindex, transx + i)
           != TCL_OK)
          || (GetSadieDoubleFromObj2 (interp, objv[1], transyindex, transy + i)
              != TCL_OK)
          || (GetSadieDoubleFromObj2 (interp, objv[1], refxindex, refx + i)
              != TCL_OK)
          || (GetSadieDoubleFromObj2 (interp, objv[1], refyindex, refy + i)
              != TCL_OK))
        return TCL_ERROR;
    }
  GEOMCOEF (transx, transy, refx, refy, (short)npts, (short)nterms,
            xpolyterms, ypolyterms);
  numpolyterms = nterms;
  return TCL_OK;
}  

/**
 * Use a previously determined polynomial distortion model to warp an image.
 * Given a distorted image and a series of polynomial terms that
 * define the distortion relative to some reference frame, the SADIE
 * library function GEOMWARP generates a new image representing the
 * distorted image in the reference frame with the distortion removed,
 * using interpolation guided by the distortion model.  The @a option
 * variable determines the interpolation technique: negative values
 * for cubic interpolation (with the value being interpreted as the
 * free parameter @a alpha for this method), zero values imply
 * bilinear interpolation, and positive values nearest-neighbor
 * interpolation.  Note that the polynomial terms come from an
 * uprotected global array, so there are issues with re-entrancy and
 * thread safety.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name((DEFAULT_IN_IMAGE_ADDR_INDEX) distorted image ID.
 * - @c tcl_array_name(lines) the number of lines in the interpolated image.
 * - @c tcl_array_name(pix) pixels per line in the interpolated image.
 * - @c tcl_array_name(option) the interpolation option.
 * - @c tcl_array_name(graylevel) the greyvalue with which to fill blanks.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) the warped image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_GeomWarpCmd (ClientData client_data, Tcl_Interp* interp,
                            int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  double option;
  PIXEL greylevel;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "graylevel", &greylevel)
          != TCL_OK))
    return TCL_ERROR;
  GEOMWARP (inimg, inimg->nbnd, lines, pix, xpolyterms, ypolyterms,
            (short) numpolyterms, option, greylevel, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Tile two images edge-to-edge horizontally, with an optional vertical offset.
 * The SADIE library function HMOSAIC creates a new image from two
 * input images arranged next to each other horizontally, optionally
 * staggering them in the vertical direction by the specified number
 * of lines, and filling any spaces between the copied component
 * images and their enclosing frame with pixels of the specified
 * greyvalue.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) left input image ID string.
 * - @c tcl_array_name(inimg,addr2) right input image ID string.
 * - @c tcl_array_name(offset) vertical offset for the right image.
 * - @c tcl_array_name(fill) the value for otherwise blank pixels.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) tiled image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_HMosaicCmd (ClientData client_data, Tcl_Interp* interp,
                           int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  int32_t offset;
  PIXEL fill;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "offset", &offset)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "fill", &fill)
          != TCL_OK))
    return TCL_ERROR;
  HMOSAIC (inimg1, inimg2, offset, fill, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Tile two images edge-to-edge vertically, with an optional horizontal offset.
 * The SADIE library function VMOSAIC creates a new image from two
 * input images arranged next to each other vertically, optionally
 * staggering them in the horizontal direction by the specified number
 * of pixel columns, and filling any spaces between the copied
 * component images and their enclosing frame with pixels of the
 * specified greyvalue.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) top input image ID string.
 * - @c tcl_array_name(inimg,addr2) bottom input image ID string.
 * - @c tcl_array_name(offset) horizontal offset for the bottom image.
 * - @c tcl_array_name(fill) the value for otherwise blank pixels.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) combined image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_VMosaicCmd (ClientData client_data, Tcl_Interp* interp,
                           int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  int32_t offset;
  PIXEL fill;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "offset", &offset)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "fill", &fill)
          != TCL_OK))
    return TCL_ERROR;
  VMOSAIC (inimg1, inimg2, offset, fill, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/** Number of images to combine. */
#define N_MOSAIC_IMAGES 2

/**
 * Combine multiple images into a single mosaic.
 * The SADIE library function MOSAIC can potentially combine various
 * numbers of individual images into a mosaic, but this Tcl interface
 * restricts these to one small constant (at the moment, only two
 * images).  It places the individual images at offsets specified in
 * pixel units relative to the top left corner of the mosaic image,
 * filling any otherwise blank areas with pixels of the specified
 * greyvalue.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) first input image ID string.
 * - @c tcl_array_name(inimg,addr2) second input image ID string.
 * - @c tcl_array_name(lines) number of lines in the mosaic.
 * - @c tcl_array_name(pix) number of pixel columns in the mosaic.
 * - @c tcl_array_name(offset1,lines) vertical offset for the first image.
 * - @c tcl_array_name(offset1,pix) horizontal offset for the first image.
 * - @c tcl_array_name(offset2,lines) vertical offset for the second image.
 * - @c tcl_array_name(offset2,pix) horizontal offset for the second image.
 * - @c tcl_array_name(fill) the value for otherwise blank pixels.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) combined image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_GenMosaicCmd (ClientData client_data, Tcl_Interp* interp,
                             int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg[N_MOSAIC_IMAGES];
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  uint32_t linebgn[N_MOSAIC_IMAGES], pixbgn[N_MOSAIC_IMAGES];
  PIXEL fill;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", inimg)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", inimg + 1)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "offset1,lines", linebgn)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "offset2,lines", linebgn + 1)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "offset1,pix", pixbgn)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "offset2,pix", pixbgn + 1)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "fill", &fill)
          != TCL_OK))
    return TCL_ERROR;
  MOSAIC (inimg, N_MOSAIC_IMAGES, lines, pix, linebgn, pixbgn, fill, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Insert paste one image on top of another.
 * Given two images, the SADIE library function INSERT treats one as
 * the background and the other as the (opaque) foreground, pasting
 * the foreground on top of the background at the specified pixel
 * unit offsets.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) background image ID string.
 * - @c tcl_array_name(inimg,addr2) foreground image ID string.
 * - @c tcl_array_name(offset,lines) vertical foreground image offset.
 * - @c tcl_array_name(offset,pix) horizontal foreground image offset.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) combined image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Geometry_InsertCmd (ClientData client_data, Tcl_Interp* interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  uint32_t offsetlines, offsetpix;
  PIXEL fill;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "offset,lines", &offsetlines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "offset,pix", &offsetpix)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "fill", &fill)
          != TCL_OK))
    return TCL_ERROR;
  INSERT (inimg1, inimg2, offsetlines, offsetpix, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_Geometry_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_AvgSub",
                        Sadie_Geometry_AvgSubCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_WinSub",
                        Sadie_Geometry_WinSubCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_Scale",
                        Sadie_Geometry_ScaleCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_Rotate",
                        Sadie_Geometry_RotateCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_Mirror",
                        Sadie_Geometry_MirrorCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_ContPts",
                        Sadie_Geometry_ContPtsCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_GeomWarp",
                        Sadie_Geometry_GeomWarpCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_HMosaic",
                        Sadie_Geometry_HMosaicCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_VMosaic",
                        Sadie_Geometry_VMosaicCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_GenMosaic",
                        Sadie_Geometry_GenMosaicCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Geometry_Insert",
                        Sadie_Geometry_InsertCmd, NULL, NULL);
  return TCL_OK;
}
