/** @file Sadie_Image.c
 * Tk support for SADIE images.
 * The image format support within Tk is extensible without requiring
 * any changes to the installed Tk or Tcl libraries.  Applications can
 * use certain data structures pre-defined within the Tk API to
 * register their own image handling callbacks with Tk, so here we
 * define callbacks to handle SADIE images.  It isn't necessary to
 * define all possible callbacks, and for the string interface we can
 * get away with using the image ID strings rather than strings that
 * directly encode the image content.  The only hidden complications
 * arise from changes to the Tk API after version 8.3: newer versions
 * of the image handling interface break older code, and whether the
 * image format is upper or lower case starts to have significance.
 * Please refer to any documentation for the @c Tk_CreatePhotoImageFormat
 * Tk function for more information.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Image.c,v 2.4 2005/03/22 21:37:39 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <string.h>
#include <sadie.h>
#include "tclsadie.h"

/* Forward declaration of the SADIE string match callback. */
static int
MatchSADIE (Tcl_Obj * data, Tcl_Obj * format,
            int * widthPtr, int * heightPtr, Tcl_Interp * interp);

/* Forward declaration of the SADIE string input callback. */
static int
ReadSADIE  (Tcl_Interp * interp, Tcl_Obj * data, Tcl_Obj * format,
            Tk_PhotoHandle imageHandle,
            int destX, int destY, int width, int height,
            int srcX, int srcY);

/* Forward declaration of the SADIE file output callback. */
static int
WriteSADIE (Tcl_Interp * interp, const char * fileName,
            Tcl_Obj * format, Tk_PhotoImageBlock * blockPtr);

/** The structure Tk uses to hold application-provided callback pointers. */
Tk_PhotoImageFormat tkImgFmtSADIE =
  {
    "sadie", /**< Name of the image format. */
    NULL, /**< File match procedure (unused). */
    MatchSADIE, /**< String match procedure. */
    NULL, /**< File read procedure (unused). */
    ReadSADIE, /**< String read procedure. */
    WriteSADIE, /**< File write procedure. */
    NULL  /**< String write procedure (unused). */
  };

/**
 * Callback procedure to inform Tk if it can decode SADIE image IDs.
 * The procedure for Tk to call when it is searching for an image file
 * format handler suitable for reading data from a given string; this
 * conforms to the @c Tk_ImageStringMatchProc prototype.
 * @param[in]  data  the object encoding the SADIE image.
 * @param[in]  format  a user-specified format (unused).
 * @param[out] widthPtr  where to store the image width.
 * @param[out] heightPtr where to store the image height.
 * @param[in]  interp  the Tcl interpreter.
 * @return 1 if the Tcl object @a data encodes a SADIE image, otherwise 0.
 */
static int
MatchSADIE (Tcl_Obj *data, Tcl_Obj *format, int *widthPtr,
            int *heightPtr, Tcl_Interp *interp)
{
  IMAGE *img;

  if ((GetSadieImageFromObj (data, &img) != TCL_OK) || (!CHECKIMG (img)))
    return 0;
  else
    {
      *widthPtr = img->npix;
      *heightPtr = img->nlin;
      return 1;
    }
}

/** @name Constants for display images
 * Constants used in translating SADIE images into Tk photo images.
 */
/*@{*/
static const double stretchMin = 0;
static const double stretchMax = 255;
static const double stretchTol = 1e-5;
/*@}*/

/** @name Support functions for display images
 * Utility functions used in translating SADIE images into Tk photo images.
 */
/*@{*/
/**
 * Scale and replicate a single SADIE pixel value within the Tk photo image.
 * The pixel value may be remapped to the limits @c SMIN and @c SMAX defined
 * within @c sadie.h.
 * @param[in] pixelPtr  points to the start of the destiation pixel range.
 * @param[in] corner  the value to replicate.
 * @param[in] n  the number of copies.
 */
static void
fill_bytes (void *pixelPtr, PIXEL corner, size_t n)
{
  double scaled;

  scaled = ((corner <= SMIN)
            ? 0
            : (((corner >= SMAX)
                ? SMAX
                : corner)
               - SMIN));
  memset (pixelPtr,
          (int)(floor ((scaled * (stretchMax - stretchMin))
                       / (SMAX - SMIN))),
          n);
}

/**
 * Scale and transcribe pixels from a three-band SADIE image to a Tk photo.
 * Interpret the first three bands of a SADIE image as color channels to be
 * scaled and copied into a color Tk photo image.
 * @param[in] pixelPtr  points to the start of the Tk destiation pixel range.
 * @param[in] img  points to the SADIE source image.
 */
static void
stretch_rgb (unsigned char *pixelPtr, IMAGE *img)
{
  size_t j, k;
  double factor, offset;

  factor = (stretchMax - stretchMin) / (double)(img->gmax - img->gmin);
  offset = stretchMin - (img->gmin * factor);
  for (j = 0; j < img->nlin; j++)
    {
      for (k = 0; k < img->npix; k++)
        {
          *pixelPtr++ = (unsigned char)(factor * img->data[0][j][k] + offset);
          *pixelPtr++ = (unsigned char)(factor * img->data[1][j][k] + offset);
          *pixelPtr++ = (unsigned char)(factor * img->data[2][j][k] + offset);
        }
    }
}

/**
 * Transcribe pixels from a three-band SADIE image to a Tk photo image.
 * Interpret the first three bands of a SADIE image as color channels to be
 * copied unchanged into a color Tk photo image.
 * @param[in] pixelPtr  points to the start of the Tk destiation pixel range.
 * @param[in] img  points to the SADIE source image.
 */
static void
transcribe_rgb (unsigned char *pixelPtr, IMAGE *img)
{
  size_t j, k;

  for (j = 0; j < img->nlin; j++)
    {
      for (k = 0; k < img->npix; k++)
        {
          *pixelPtr++ = (unsigned char) img->data[0][j][k];
          *pixelPtr++ = (unsigned char) img->data[1][j][k];
          *pixelPtr++ = (unsigned char) img->data[2][j][k];
        }
    }
}

/**
 * Scale and transcribe pixels from a single-band SADIE image to a Tk photo.
 * Interpret the first band of a SADIE image as pixel values to be
 * scaled and copied into a monochrome Tk photo image.
 * @param[in] pixelPtr  points to the start of the Tk destiation pixel range.
 * @param[in] img  points to the SADIE source image.
 */
static void
stretch_greyscale (unsigned char *pixelPtr, IMAGE *img)
{
  size_t j, k;
  double factor, offset;

  factor = (stretchMax - stretchMin) / (double)(img->gmax - img->gmin);
  offset = stretchMin - (img->gmin * factor);
  for (j = 0; j < img->nlin; j++)
    {
      for (k = 0; k < img->npix; k++)
        *pixelPtr++ = (unsigned char)(factor * img->data[0][j][k] + offset);
    }
}

/**
 * Transcribe pixels from a single-band SADIE image to a Tk photo.
 * Interpret the first band of a SADIE image as pixel values to be
 * copied unchanged into a monochrome Tk photo image.
 * @param[in] pixelPtr  points to the start of the Tk destiation pixel range.
 * @param[in] img  points to the SADIE source image.
 */
static void
transcribe_greyscale (unsigned char *pixelPtr, IMAGE *img)
{
  size_t j, k;

  for (j = 0; j < img->nlin; j++)
    {
      for (k = 0; k < img->npix; k++)
        *pixelPtr++ = (unsigned char) img->data[0][j][k];
    }
}
/*@}*/

/**
 * Callback procedure to read a Tk photo image from a SADIE image.
 * Tk calls this procedure to read SADIE image data from a string into
 * a photo image; this conforms to the @c Tk_ImageStringReadProc
 * prototype.
 * @param[in]  interp  the Tcl interpreter.
 * @param[in]  data  the object encoding the SADIE image.
 * @param[in]  format  a user-specified format (unused).
 * @param[out] imageHandle  the photo image to write into.
 * @param[out] heightPtr  where to store the image height.
 * @param[in]  destX  left edge of destination region in photo image.
 * @param[in]  destY  top edge of destination region in photo image.
 * @param[in]  width  horizontal size of the destination region.
 * @param[in]  height  vertical size of the destination region.
 * @param[in]  srcX  left edge of source region in SADIE image.
 * @param[in]  srcY  top edge of source region in SADIE image.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static int
ReadSADIE (Tcl_Interp *interp, Tcl_Obj *data, Tcl_Obj *format,
           Tk_PhotoHandle imageHandle, int destX, int destY,
           int width, int height, int srcX, int srcY)
{
  int option, need_stretch;
  size_t nbytes;
  Tk_PhotoImageBlock block;
  IMAGE *img;
  int err = TCL_OK;
  Tcl_Obj *arrayobjPtr = NULL;
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;

  arrayobjPtr = Tcl_NewStringObj (SADIEVAR_ARRAYNAME, -1);
  Tcl_IncrRefCount (arrayobjPtr);
  indexobjPtr = Tcl_NewStringObj ("getstretched", -1);
  Tcl_IncrRefCount (indexobjPtr);
  valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr, 0);
  if (valueobjPtr == NULL)
    option = 1;
  else
    {
      Tcl_IncrRefCount (valueobjPtr);
      Tcl_GetIntFromObj (interp, valueobjPtr, &option);
      Tcl_DecrRefCount (valueobjPtr);
    }
  Tcl_DecrRefCount (indexobjPtr);
  Tcl_DecrRefCount (arrayobjPtr);
  if (GetSadieImageFromObj (data, &img) != TCL_OK)
    return TCL_ERROR;
  else
    Tcl_ResetResult (interp);
  if (!CHECKIMG (img))
    {
      Tcl_AppendStringsToObj (Tcl_GetObjResult (interp),
                              "Couldn't identify SADIE image from input string.",
                              NULL);
      err = TCL_ERROR;
    }
  if ((int) img->nbnd <= 0)
    {
      Tcl_AppendStringsToObj (Tcl_GetObjResult (interp),
                              "Image bands not within display range.", NULL);
      err = TCL_ERROR;
    }
  if ((int) img->nlin <= 0)
    {
      Tcl_AppendStringsToObj (Tcl_GetObjResult (interp),
                              "Image lines not within display range.", NULL);
      err = TCL_ERROR;
    }
  if ((int) img->npix <= 0)
    {
      Tcl_AppendStringsToObj (Tcl_GetObjResult (interp),
                              "Image pixels not within display range.", NULL);
      err = TCL_ERROR;
    }
  if (err != TCL_OK)
    return TCL_ERROR;
  if (width > img->npix)
    width = img->npix;
  if (height > img->nlin)
    height = img->nlin;
  if ((width <= 0) || (height <= 0))
    return TCL_OK;
  block.width = width;
  block.height = height;
  if (img->nbnd != 3)
    {
      block.pixelSize = 1;
      block.offset[0] = 0;
      block.offset[1] = 0;
      block.offset[2] = 0;
      block.offset[3] = 0;
    }
  else
    {
      block.pixelSize = 3;
      block.offset[0] = 0;
      block.offset[1] = 1;
      block.offset[2] = 2;
      block.offset[3] = 0;
    }
  block.pitch = block.pixelSize * img->npix;
  nbytes = height * block.pitch;
  block.pixelPtr = (unsigned char *) ckalloc (nbytes);
  Tk_PhotoExpand(imageHandle, destX + width, destY + height);
  RANGE (img);
  need_stretch = ((option == 1)
                  && ((fabs (stretchMin - img->gmin) > stretchTol)
                      || (fabs (stretchMax - img->gmax) > stretchTol)));
  if (need_stretch && (img->gmin == img->gmax))
    fill_bytes (block.pixelPtr, img->data[0][0][0], nbytes);
  else
    {
      if (img->nbnd == 3)
        {
          if (need_stretch)
            stretch_rgb (block.pixelPtr, img);
          else
            transcribe_rgb (block.pixelPtr, img);
        }
      else
        {
          if (need_stretch)
            stretch_greyscale (block.pixelPtr, img);
          else
            transcribe_greyscale (block.pixelPtr, img);
        }
    }
#if (TK_MAJOR_VERSION > 8) || ((TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION > 3))
  Tk_PhotoPutBlock (imageHandle, &block, destX, destY, width, height,
                    TK_PHOTO_COMPOSITE_SET);
#else /* (TK_MAJOR_VERSION < 8) || ((TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION <= 3)) */
  Tk_PhotoPutBlock (imageHandle, &block, destX, destY, width, height);
#endif
  return TCL_OK;
}

/** @name Support functions for SADIE images
 * Utility functions used in translating Tk photo images into SADIE images.
 */
/*@{*/

/**
 * Transcribe pixels from a Tk photo image to a three-band SADIE image.
 * Assuming a Tk photo image has three color channels and an unused alpha
 * channel, copy the values into a three-band SADIE image.
 * @param[in] img  points to the SADIE source image.
 * @param[in] blockPtr  points to the Tk image data structure.
 */
static void
write_rgb_SADIE (IMAGE *img, Tk_PhotoImageBlock *blockPtr)
{
  size_t h, w;
  unsigned char *bytePtr;
  
  bytePtr = blockPtr->pixelPtr + blockPtr->offset[0];
  for (h = 0; h < blockPtr->height; h++)
    {
      for (w = 0; w < blockPtr->width; w++)
        {
          img->data[0][h][w] = (PIXEL) *(bytePtr++);
          img->data[1][h][w] = (PIXEL) *(bytePtr++);
          img->data[2][h][w] = (PIXEL) *(bytePtr++);
          bytePtr++; /* Skip any alpha channel data. */
        }
    }	
}

/**
 * Transcribe pixels from a Tk photo image to a single-band SADIE image.
 * Assuming there is a monochrome Tk photo image, copy the values into
 * one band of a SADIE image.
 * @param[in] img  points to the SADIE source image.
 * @param[in] blockPtr  points to the Tk image data structure.
 */
static void
write_greyscale_SADIE (IMAGE *img, Tk_PhotoImageBlock *blockPtr)
{
  size_t h, w;
  unsigned char *bytePtr;
  
  bytePtr = blockPtr->pixelPtr + blockPtr->offset[0];
  for (h = 0; h < blockPtr->height; h++)
    {
      for (w = 0; w < blockPtr->width; w++)
          img->data[0][h][w] = (PIXEL) *(bytePtr++);
    }	
}
/*@}*/

/**
 * Callback procedure to copy a Tk photo image into a new SADIE image.
 * The @c Tk_ImageFileWriteProc prototype to which this conforms is
 * intended for callback procedures that write disk files in
 * particular image formats, but here we substitute a function that
 * creates a new SADIE image from the Tk photo image and returns the
 * SADIE image ID string in the Tcl interpreter result.
 * @param[in]  interp  the Tcl interpreter.
 * @param[in]  fileName  file name for the image.
 * @param[in]  format  a user-specified format (unused).
 * @param[in]  blockPtr  points to the Tk image data structure.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static int
WriteSADIE (Tcl_Interp *interp, const char *fileName,
            Tcl_Obj *format, Tk_PhotoImageBlock *blockPtr)
{
  char msg[SLEN];
  int greenOffset, blueOffset;
  Tcl_Obj *objPtr = NULL;
  IMAGE *img = NULL;
    
  greenOffset = blockPtr->offset[1] - blockPtr->offset[0];
  blueOffset = blockPtr->offset[2] - blockPtr->offset[0];
  if ((greenOffset == 1) && (blueOffset == 2) && (blockPtr->pixelSize == 4)
      && (blockPtr->pitch == (blockPtr->width * 4)))
    {
      GETMEM (3, blockPtr->height, blockPtr->width, &img);
      if (img == NULL)
        return TCL_ERROR;
      write_rgb_SADIE (img, blockPtr);
    }
  else if (blockPtr->pixelSize == 1)
    {
      GETMEM (1, blockPtr->height, blockPtr->width, &img);
      if (img == NULL)
        return TCL_ERROR;
      write_greyscale_SADIE (img, blockPtr);
    }
  else
    {
      return TCL_ERROR;
    }
  if (NAMES)
    {
      MESSAGE('I', "");
      MESSAGE('I', "TKIMGSADIE");
      MESSAGE('I', "");
      snprintf(msg, SLEN - 1,
               " Input file:                                 %s", fileName);
      MESSAGE('I', msg); 
      snprintf(msg, SLEN - 1,
               " Number of bands, lines, pixels:             %lu, %lu, %lu",
               (unsigned long) img->nbnd,
               (unsigned long) img->nlin,
               (unsigned long) img->npix);
      MESSAGE('I', msg);
      snprintf(msg, SLEN - 1,
               " Number of bits/pixel/band:                  %d", 8);
      MESSAGE('I', msg);
      snprintf(msg, SLEN - 1,
               " Graylevel minimum, maximum:                 %10.4e, %10.4e",
               img->gmin, img->gmax);
      MESSAGE('I', msg);
      MESSAGE('I', " ...............");
    }
  strncpy (img->text, fileName, TLEN - 1);	
  objPtr = NewSadieImageObj (img);
  if (objPtr != NULL)
    Tcl_SetObjResult (interp, objPtr);
  return TCL_OK;
}

/**
 * Define a new Tk image format that handles SADIE images.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_Image_Init (Tcl_Interp *interp)
{
  Tk_CreatePhotoImageFormat (&tkImgFmtSADIE);
  return TCL_OK;
}
