/** @file Sadie_Index.c
 * Indexes to application-wide data structures such as lists of images.
 * SADIE represents images as structures that contain the pixel data
 * and a limited range of other information.  A simple in-memory
 * database can however associate arbitrary metadata with the images,
 * and track additional information not associated with a particular
 * image.  It is structured as a two-level hash table indexed by strings
 * from which the Fowler/Noll/Vo (FNV) algorithm produces short hashes.
 * See Landon Curt Noll's web site for extensive background and sample
 * source code: 
 *   http://www.isthe.com/chongo/tech/comp/fnv/
 * Each item at the upper level of the table will generally represent
 * a single image that SADIE is displaying, thought of as a SADIE
 * document.  Nested within this at the lower level of the hierarchy
 * will be the properties of the image.  However the implementation is
 * sufficiently general that you can stuff more or less anything that
 * can be referenced by an opaque pointer and indexed by a sting into
 * the table.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: Sadie_Index.c,v 1.1 2005/03/22 20:17:14 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "logerr.h"
#if HAVE_PTHREADS
#include <pthread.h>
/** Turn on local flag for POSIX Threads support. */
#define POSIX_THREADS 1
#else
/** Turn off local flag for POSIX Threads support. */
#define POSIX_THREADS 0
#endif /* HAVE_PTHREADS */

#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <sadie.h>
#include "Sadie_Index.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

static const int maxsize = 14; /**< Maximum table size is pow (2, maxsize). */
static const int minsize = 4; /**< Minimum table size is pow (2, minsize). */
static const uint32_t fnv32hash_magic_init = 0x811c9dc5;  /**< For FNV hash. */

typedef struct Sad_property * sad_propertyp_t; /**< property pointers. */
typedef struct Sad_docitem * sad_docitemp_t; /**< top-level item pointers. */
typedef struct Sad_doc * sad_docp_t;  /**< top-level list pointers. */
typedef struct Sad_propertyitem * sad_propertyitemp_t; /**< property lists. */

/** A hash table implements the list of documents. */
struct Sad_doclist
{
#ifdef HAVE_PTHREADS
  pthread_mutex_t lock;  /**< A single lock for all table accesses. */
#else
  int lock[6];  /**< Arbitrary padding if we can't use real locking. */
#endif /* HAVE_PTHREADS */
  uint32_t log2nbuckets; /**< Log to base 2 of number of hash buckets. */
  sad_docitemp_t *buckets; /**< Array of list heads, indexed by hash values. */
};

/** A node of the singly-linked list of documents hanging off a hash bucket. */
struct Sad_docitem
{
  size_t keylen; /**< Key string length in bytes (including terminal null). */
  const char *key;  /**< Unique key string (UTF-8 or ASCII). */
  sad_docp_t doc;   /**< Points at image and its properties. */
  sad_docitemp_t next;  /**< Link to next list item. */
};

/** A pairing of the useful data with a pointer to a function that frees it. */
struct Sad_property
{
  void *data; /**< Opaque property data. */
  sad_property_data_destructor_t datadestruct; /**< Destructor function. */
};

/** One node of the singly-linked property list hanging off a hash bucket. */
struct Sad_propertyitem
{
  size_t keylen; /**< Key string length in bytes (including terminal null). */
  const char *key;  /**< Unique key string. */
  sad_propertyp_t prop;  /**< Points at the property (metadata object). */ 
  sad_propertyitemp_t next;  /**< Link to next list item. */
};

/** Each document contains a SADIE image with its property hash table. */
struct Sad_doc
{
  uint32_t log2nbuckets; /**< Log to base 2 of number of hash buckets. */
  sad_propertyitemp_t *buckets; /**< Hash table of document properties. */
  const IMAGE *baseimg; /**< Points to the base SADIE image. */
};

#ifndef HAVE_PTHREADS
/** @name Locking function placeholders
 * If we do not require POSIX thread support, use dummy locking functions.
 */
/*@{*/
static int pthread_mutex_lock (void * dummy) { return 0; }
static int pthread_mutex_unlock (void * dummy) { return 0; }
static int pthread_mutex_destroy (void * dummy) { return 0; }
static int pthread_mutex_init (void * dummy1, void * dummy2) { return 0; }
/*@}*/
#endif

/**
 * Hash a string to a number using the Fowler/Noll/Vo (FNV) algorithm.
 * The hash uses the explicitly 32-bit variation of version 1a of the
 * algorithm, chopped to fit the actual number of potential hash buckets.
 * As a side-effect we evaluate the string length.
 * @param[in] str  points to the null-terminated string to be hashed.
 * @param[in] log2len  the maximum number of buckets is 2^log2len.
 * @param[out] slen  where to deposit the computed string length.
 * @return  The hash value.
 */
static size_t
hash_string (const char *str, const uint32_t log2len, size_t *slen) 
{
  uint32_t hash;
  uint32_t mask;
  unsigned char* usp;
 
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, ((str != NULL) && (slen != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= log2len) && (log2len <= maxsize)));
  hash = fnv32hash_magic_init;
  usp = (unsigned char *) str;
  *slen = 1; /* At least 1 (terminating null). */
  while (*usp)
    {
      (*slen)++;
      hash ^= (uint32_t) *usp++;
      hash += ((hash << 1) + (hash << 4) + (hash << 7)
               + (hash << 8) + (hash << 24));
    }
  mask = (1 << log2len) - 1;
  return (size_t) (hash & mask);
}

/**
 * Remove a property and its key (generally at the lower level of the table).
 * This just frees storage: other functions tidy up the property lists.
 * @param[in,out] thepropertyitem  to be destroyed.
 */
static void
dispose_propertyitem (sad_propertyitemp_t thepropertyitem)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (thepropertyitem != NULL));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((thepropertyitem->key != NULL) == (thepropertyitem->keylen > 0)));
  free ((char *) thepropertyitem->key);
  if (thepropertyitem->prop != NULL)
    {
      logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                     (thepropertyitem->prop->datadestruct != NULL));
      (*(thepropertyitem->prop->datadestruct)) (thepropertyitem->prop->data);
      free (thepropertyitem->prop);
    }
  free (thepropertyitem);
}

/**
 * Reclaim storage from within an item at the upper level of the table.
 * We not only dispose of the top-level item, but all the  property lists
 * that may be hanging below it, and the SADIE image that may be associated
 * with it.
 * @param[in,out] thedoc  the data and properties to be destroyed.
 */
static void
dispose_doc (sad_docp_t thedoc)
{
  size_t i, nbuckets;
  sad_propertyitemp_t *buck;
  sad_propertyitemp_t d, v;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (thedoc != NULL));
  RELIMG ((IMAGE**) &thedoc->baseimg);
  nbuckets = (size_t) (1 << thedoc->log2nbuckets);
  for (i = nbuckets, buck = thedoc->buckets; i; i--, buck++)
    {
      d = *buck;
      while (d)
        {
          v = d;
          d = d->next;
          dispose_propertyitem (v);
        }
    }
  free (thedoc->buckets);
  free (thedoc);
}

/**
 * Remove an item and its key (generally at the upper level of the table).
 * @param[in,out] thedocitem  the top-level item to be destroyed.
 */
static void
dispose_docitem (sad_docitemp_t thedocitem)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (thedocitem != NULL));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((thedocitem->key != NULL) == (thedocitem->keylen > 0)));
  free ((char *) thedocitem->key);
  if (thedocitem->doc != NULL)
    dispose_doc (thedocitem->doc);
  free (thedocitem);
}

/**
 * Tear down a complete hash table.
 * We iterate through the entire table, reclaiming storage at all levels,
 * but can't reset the pointer to it to NULL.
 * @param[in,out] thedoclist  points to the hash table to be destroyed.
 */
void
dispose_doclist (sad_doclistp_t thedoclist)
{
  int err = 0;
  size_t i, nbuckets;
  sad_docitemp_t *buck;
  sad_docitemp_t d, v;

  if (thedoclist)
    {
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                     ((thedoclist->buckets != NULL)
                      && (thedoclist->log2nbuckets >= minsize)
                      && (thedoclist->log2nbuckets <= maxsize)));
      if (POSIX_THREADS)
        {
          err = pthread_mutex_lock (&thedoclist->lock);
          if (err)
            {
              logerr_err (LOGERR_FUNC,
                          "could not acquire lock to dispose of doclist: error = %d",
                          err);
              return;
            }
        }
      nbuckets = (size_t) (1 << thedoclist->log2nbuckets);
      for (i = nbuckets, buck = thedoclist->buckets; i; i--, buck++)
        {
          d = *buck;
          while (d)
            {
              v = d;
              d = d->next;
              dispose_docitem (v);
            }
        }
      free (thedoclist->buckets);
      if (POSIX_THREADS)
        {
          err = pthread_mutex_unlock (&thedoclist->lock);
          if (err)
            {
              logerr_err (LOGERR_FUNC,
                          "could not release lock to dispose of doclist: error = %d",
                          err);
              return;
            }
          err = pthread_mutex_destroy (&thedoclist->lock);
          if (err)
            {
              logerr_err (LOGERR_FUNC,
                          "failed to destroy the doclist lock during disposal: error = %d",
                          err);
              return;
            }
        }
      free (thedoclist);
    }
}      

/**
 * Construct a new upper-level hash table entry.
 * The new entry can contain an optional SADIE image, and will have an empty
 * table for property lists, of a size that is a power of two.
 * @param[in] logpsize  property list hash table will have 2^logpsize buckets.
 * @param[in] img  points to the image to associate with the entry.
 * @return  A pointer to the new entry.
 */
static sad_docp_t
make_doc (int logpsize, const IMAGE *img)
{
  size_t nbuckets;
  sad_docp_t newdoc = NULL;

  newdoc = malloc (sizeof (struct Sad_doc));
  if (newdoc)
    {
      if (logpsize < minsize)
        logpsize = minsize;
      if (logpsize > maxsize)
        logpsize = maxsize;
      newdoc->log2nbuckets = logpsize;
      nbuckets = (size_t) (1 << logpsize);
      newdoc->buckets = calloc (nbuckets, sizeof (sad_propertyitemp_t));
      if (newdoc->buckets != NULL)
        newdoc->baseimg = img;
      else
        {
          free (newdoc);
          newdoc = NULL;
        }
    }
  return newdoc;
}

/**
 * Construct a new lower-level property.
 * We don't interpret this in any way, since all we have are opaque
 * pointers to some kind of data structure & destructor about which
 * we have no other information.
 * @param[in] yourdestruct  points to the destructor function.
 * @param[in] yourdata  points to some kind of data structure.
 * @return  A pointer to the new property.
 */
sad_propertyp_t
make_property (sad_property_data_destructor_t yourdestruct, void *yourdata)
{
  sad_propertyp_t newproperty;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((yourdestruct != NULL) && (yourdata != NULL)));
  newproperty = malloc (sizeof (struct Sad_property));
  if (newproperty)
    {
      newproperty->data = yourdata;
      newproperty->datadestruct = yourdestruct;
    }
  return newproperty;
}
 
/**
 * Create a new empty hash table.
 * If we're in a multi-threaded environment we not only allocate
 * storage for the table, but also initialize the lock used to serialize
 * all accesses to it.
 * @param[in] logsize  allocate 2^logsize hash buckets.
 * @return  A pointer to the successfully created table, or NULL on failure.
 */
sad_doclistp_t
make_doclist (int logsize)
{
  int err = 0;
  size_t nbuckets;
  sad_doclistp_t thedoclist;

  thedoclist = malloc (sizeof (struct Sad_doclist));
  if (thedoclist)
    {
      if (POSIX_THREADS)
        {
          err = pthread_mutex_init (&thedoclist->lock, NULL);
          if (err)
            {
              logerr_err (LOGERR_FUNC,
                          "could not initialize the doclist lock: error = %d",
                          err);
              free (thedoclist);
              return NULL;
            }
        }
      if (logsize < minsize)
        logsize = minsize;
      if (logsize > maxsize)
        logsize = maxsize;
      thedoclist->log2nbuckets = logsize;
      nbuckets = (size_t) (1 << logsize);
      thedoclist->buckets = calloc (nbuckets, sizeof (sad_docitemp_t));
      if (thedoclist->buckets == NULL)
        {
          if (POSIX_THREADS)
            {
              err = pthread_mutex_destroy (&thedoclist->lock);
              if (err)
                logerr_err (LOGERR_FUNC,
                            "failed to destroy the doclist lock during bucket allocation failure: error = %d",
                            err);
            }
          free (thedoclist);
          thedoclist = NULL;
        }
    }
  return thedoclist;
}

/**
 * Make a new top-level table entry and its key.
 * We not only create the entry and its key string (being careful to copy
 * the string rather than merely pass a pointer to a location that may vanish),
 * and insert the new data storage, but also append the list of any
 * subsequent items in the same hash bucket.
 * @param[in] dkeylen  the length of the key string;
 * @param[in] dkey  copy the key string from here.
 * @param[in] doc the data storage and properties for this entry.
 * @param[in] dinext  head of the list of subsequent items in the same bucket.
 * @return  the newly created top-level entry.
 */
static sad_docitemp_t
make_docitem (const size_t dkeylen, const char *dkey, const sad_docp_t docp,
              const sad_docitemp_t dinext)
{
  sad_docitemp_t newitem;
  char *newkey;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((dkeylen > 0) && (dkey != NULL) && (docp != NULL)));
  newitem = malloc (sizeof (struct Sad_docitem));
  if (newitem)
    {
      newkey = calloc (dkeylen, 1);
      if ((newkey == NULL) || (strncpy (newkey, dkey, dkeylen) == NULL))
        {
          free (newitem);
          newitem = NULL;
        }
      else
        {
          newitem->keylen = dkeylen;
          newitem->key = (const char *) newkey;
          newitem->doc = docp;
          newitem->next = dinext;
        }
    }
  return newitem;
}

/**
 * Make a new lower-level table entry and its key.
 * We not only create the entry and its key string (being careful to copy
 * the string rather than merely pass a pointer to a location that may vanish),
 * and insert the new data storage, but also append the list of any
 * subsequent items in the same hash bucket.
 * @param[in] pkeylen  the length of the key string;
 * @param[in] pkey  copy the key string from here.
 * @param[in] property  the data storage for this entry.
 * @param[in] pinext  the head of the list of subsequent items in the same bucket.
 * @return  the newly created lower-level entry.
 */
static sad_propertyitemp_t
make_propertyitem (const size_t pkeylen, const char *pkey,
                   const sad_propertyp_t property,
                   const sad_propertyitemp_t pinext)
{
  sad_propertyitemp_t newitem;
  char *newkey;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((pkeylen > 0) && (pkey != NULL) && (property != NULL)));
  newitem = malloc (sizeof (struct Sad_propertyitem));
  if (newitem)
    {
      newkey = calloc (pkeylen, 1);
      if ((newkey == NULL) || (strncpy (newkey, pkey, pkeylen) == NULL))
        {
          free (newitem);
          newitem = NULL;
        }
      else
        {
          newitem->keylen = pkeylen;
          newitem->key = (const char *) newkey;
          newitem->prop = property;
          newitem->next = pinext;
        }
    }
  return newitem;
}

/**
 * Insert a new top-level entry into the table.
 * Given a table and a key string, we search for an existing entry with the
 * same key.  If it exists, we delete it (and all its lower-level entries),
 * replacing it with a new empty entry.  If it does not exist, we insert
 * a new empty entry at the appropriate place.  We use an FNV hash code
 * to pick the initial hash bucket within which to search, then use ordinary
 * string comparisons to match keys within the ordered list of entries
 * within the bucket.  If locking is supported, we try to lock the entire
 * hash table during the insertion.
 * @param[out] doclist  points to the hash table.
 * @param[in]  dkey  points to the key for the new item.
 * @param[in]  logpsize allocate a sub-table of 2^logpsize buckets in the item.
 * @param[in]  img the SADIE image associated with the top-level entry.
 * @return  Zero on success, an error code on failure.
 */
int
sad_put_doc (sad_doclistp_t doclist, const char *dkey,
             int logpsize, const IMAGE *img)
{
  int err = 0;
  int failure = 0;
  sad_docp_t newdoc;
  size_t dkeylen;
  size_t hashval;
  sad_docitemp_t *dihandle;
  sad_docitemp_t newitem;
  int uninserted;
  int order;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doclist != NULL) && (dkey != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= doclist->log2nbuckets)
                  && (doclist->log2nbuckets <= maxsize)));
  newdoc = make_doc (logpsize, img);
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newdoc != NULL));
  if (POSIX_THREADS)
    {
      err = pthread_mutex_lock (&doclist->lock);
      if (err)
        {
          logerr_err (LOGERR_FUNC,
                      "could not acquire lock to insert doc %s: error = %d",
                      dkey, err);
          return err;
        }
    }
  uninserted = 1;
  hashval = hash_string (dkey, doclist->log2nbuckets, &dkeylen);
  dihandle = &doclist->buckets[hashval];
  while ((*dihandle != NULL) && uninserted)
    {
      order = strcmp (dkey, (*dihandle)->key);
      if (order > 0)
        dihandle = &(*dihandle)->next;
      else
        {
          if (order < 0)
            newitem = make_docitem (dkeylen, dkey, newdoc, *dihandle);
          else
            {            
              newitem = make_docitem (dkeylen, dkey, newdoc,
                                      (*dihandle)->next);
              dispose_docitem (*dihandle);
            }
          failure = (newitem == NULL);
          uninserted = 0;
          *dihandle = newitem;
          dihandle = &newitem->next;
        }
    }
  if (uninserted)
    {
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*dihandle == NULL));      
      *dihandle = make_docitem (dkeylen, dkey, newdoc, NULL);
    }
  if (POSIX_THREADS)
    {
      err = pthread_mutex_unlock (&doclist->lock);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "could not release lock to insert doc %s: error = %d",
                    dkey, err);
    }
  return (failure) ? failure : err;
}

/**
 * Explicitly delete a top-level entry from the table.
 * Given a hash table and the string key to one of the top-level
 * entries, delete the entry, its data, and any of the lower level
 * entries (properties).  If locking is supported, we try to lock the
 * entire hash table during the deletion.
 * @param[in,out] doclist  points to the hash table.
 * @param[in] dkey  points to the key for the entry to be deleted.
 * @return  Zero on success, an error code on failure.
 */
int
sad_dispose_doc (sad_doclistp_t doclist, const char *dkey)
{
  int err = 0;
  size_t dkeylen;
  size_t hashval;
  sad_docitemp_t *dihandle;
  sad_docitemp_t tail;
  int order;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doclist != NULL) && (dkey != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= doclist->log2nbuckets)
                  && (doclist->log2nbuckets <= maxsize)));
  if (POSIX_THREADS)
    {
      err = pthread_mutex_lock (&doclist->lock);
      if (err)
        {
          logerrno (LOGERR_ERR, err, LOGERR_FUNC,
                    "could not acquire the lock to delete doc %s",
                    dkey);
          return err;
        }
    }
  order = 1;
  hashval = hash_string (dkey, doclist->log2nbuckets, &dkeylen);
  dihandle = &doclist->buckets[hashval];
  while ((*dihandle != NULL) && (order > 0))
    {
      order = strcmp (dkey, (*dihandle)->key);
      if (order > 0)
        dihandle = &(*dihandle)->next;
      else if (order == 0)
        {
          tail = (*dihandle)->next;
          dispose_docitem (*dihandle);
          *dihandle = tail;
        }
    }
  if (order > 0)
    logerr_warning (LOGERR_FUNC,
                    "could not find the item to be deleted: doc=%s",
                    dkey);
  if (POSIX_THREADS)
    {
      err = pthread_mutex_unlock (&doclist->lock);
      if (err)
        logerrno (LOGERR_ERR, err, LOGERR_FUNC,
                  "could not release the lock after deleting doc %s",
                  dkey);
    }
  return (order > 0) ? -1 : err;
}

/**
 * Insert an entry into the lower-level table (property list).
 * We use a FNV hash to find the correct hash bucket, then a simple
 * string comparison to search the entries within it, which form
 * a linked list always sorted in ascending key order.
 * We can skip keys that are less than the new entry key, delete
 * and replace any matching key, or insert a new entry just before the
 * first observed to have a greater key. 
 * @param[in,out] docitem the upper-level entry in whose hash table we search.
 * @param[in] pkey  the string key to the lower-level entry (property name).
 * @param[in] prop  encapsulates the lower-level entry's data (property value).
 * @return  Zero on success, an error code on failure.
 */
static int
put_docitem_property (sad_docitemp_t docitem,
                      const char *pkey, const sad_propertyp_t prop)
{
  int err = 0;
  size_t pkeylen;
  size_t hashval;
  sad_propertyitemp_t *pihandle;
  sad_propertyitemp_t newitem;
  int order;
  int uninserted = 1;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((docitem != NULL) && (pkey != NULL) && (prop != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= docitem->doc->log2nbuckets)
                  && (docitem->doc->log2nbuckets <= maxsize)));
  hashval = hash_string (pkey, docitem->doc->log2nbuckets, &pkeylen);
  pihandle = &docitem->doc->buckets[hashval];
  while ((*pihandle != NULL) && uninserted)
    {
      order = strcmp (pkey, (*pihandle)->key);
      if (order > 0)
        pihandle = &(*pihandle)->next;
      else
        {
          if (order < 0)
            newitem = make_propertyitem (pkeylen, pkey, prop, *pihandle);
          else
            {            
              newitem = make_propertyitem (pkeylen, pkey, prop, (*pihandle)->next);
              if (newitem)
                dispose_propertyitem (*pihandle);
            }
          if (newitem == NULL)
            return -1;
          uninserted = 0;
          *pihandle = newitem;
          pihandle = &newitem->next;
        }
    }
  if (uninserted)
    {
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*pihandle == NULL));      
      *pihandle = make_propertyitem (pkeylen, pkey, prop, NULL);
      err = (*pihandle == NULL);
    }
  return err;  
}

/**
 * Recover the data (property value) from an entry in the lower-level table.
 * Use a hash to find the right bucket in the table, then a linear search
 * through the list of entries within it.
 * @param[in] docitem  the upper-level entry in whose hash sub-table we search.
 * @param[in] pkey  the string key to the lower-level entry (property name).
 * @return  A pointer to the entry data (property value), NULL on failure.
 */
static sad_propertyp_t
get_docitem_property (sad_docitemp_t docitem, const char *pkey)
{
  size_t pkeylen;
  size_t hashval;
  sad_propertyitemp_t pip;
  sad_propertyp_t theprop = NULL;
  int compare;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, ((docitem != NULL) && (pkey != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= docitem->doc->log2nbuckets)
                  && (docitem->doc->log2nbuckets <= maxsize)));
  hashval = hash_string (pkey, docitem->doc->log2nbuckets, &pkeylen);
  pip = docitem->doc->buckets[hashval];
  compare = -1;
  while ((pip != NULL) && (compare = strcmp (pkey, pip->key)) != 0)
    pip = pip->next;
  if (pip != NULL)
    theprop = pip->prop;
  return theprop;
}

/**
 * Add a new lower-level entry (property) to a existing top-level table entry.
 * Given a table and a key string, we search for an existing entry
 * with the same key.  When found, we search beneath it for a second
 * sub-key (property name).  If this exists, we replace it with a new
 * property, otherwise we insert the new entry at the appropriate
 * place, including the key, data (property value), and data
 * destructor function.  Both searches use an FNV hash code to pick
 * the initial hash bucket, then use ordinary string comparisons to
 * match keys within the ordered list of entries within the buckets.
 * If locking is supported, we try to lock the entire hash table
 * during the insertion.
 * @param[in,out] doclist  points to the hash table.
 * @param[in] dkey  the key for the existing top-level item.
 * @param[in] pkey  the key string for the new sub-entry (property name).
 * @param[in] destruct  a destructor function for the data we're adding.
 * @param[in] data  is an opaque pointer to the sub-item data (property value).
 * @return  Zero on success, an error code on failure.
 */
int
sad_put_doc_property (sad_doclistp_t doclist, const char *dkey,
                      const char *pkey,
                      sad_property_data_destructor_t destruct,
                      void *data)
{
  int err = 0;
  int failure = 0;
  size_t hashval;
  size_t dkeylen;
  sad_docitemp_t dip;
  sad_propertyp_t prop;
  int compare;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doclist != NULL) && (dkey != NULL)
                  && (pkey != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= doclist->log2nbuckets)
                  && (doclist->log2nbuckets <= maxsize)));
  prop = make_property (destruct, data);
  if (prop == NULL)
    {
      logerr_err (LOGERR_FUNC,
                  "could not create a property: doc = %s, prop = %s",
                  dkey, pkey);
      return -1;
    }
  if (POSIX_THREADS)
    {
      err = pthread_mutex_lock (&doclist->lock);
      if (err)
        {
          logerr_err (LOGERR_FUNC,
                      "could not acquire lock to set a property: doc = %s, prop = %s, error = %d",
                      dkey, pkey, err);
          return err;
        }
    }
  hashval = hash_string (dkey, doclist->log2nbuckets, &dkeylen);
  dip = doclist->buckets[hashval];
  compare = -1;
  while ((dip != NULL) && (compare = strcmp (dkey, dip->key)) != 0)
    dip = dip->next;
  if (dip != NULL)
    failure = put_docitem_property (dip, pkey, prop);
  else
    {
      logerr_err (LOGERR_FUNC,
                  "could not find the document to set a property: doc = %s, prop = %s, hv = %d",
                  dkey, pkey, hashval);
      failure = -1;
    }
  if (POSIX_THREADS)
    {
      err = pthread_mutex_unlock (&doclist->lock);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "could not release lock to set a property: doc = %s, prop = %s, error = %d",
                    dkey, pkey, err);
    }
  return (failure) ? failure : err;
}

/**
 * Given a hash table, a top-level entry key, and a sub-key, recover some data.
 * The hash table is organized in two levels.  We use the first key to
 * find a top-level entry (generally representing the document or
 * image), then use a second key to search the sub-table within it
 * (generally its properties) and return the associated data (property
 * value).  If locking is supported, we try to lock the entire hash
 * table during the retrieval.
 * @param[in] doclist  points to the hash table.
 * @param[in] dkey  the key for the existing top-level item.
 * @param[in] pkey  the key string for the new sub-entry (property name).
 * @return  An opaque pointer to the data on success, NULL on failure.
 */
void *
sad_get_doc_property (const sad_doclistp_t doclist, const char *dkey,
                      const char *pkey)
{
  int err = 0;
  sad_propertyp_t theprop = NULL;
  size_t hashval;
  size_t dkeylen;
  sad_docitemp_t dip;
  int compare;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((doclist != NULL) && (dkey != NULL) && (pkey != NULL)));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((minsize <= doclist->log2nbuckets)
                  && (doclist->log2nbuckets <= maxsize)));
  if (POSIX_THREADS)
    {
      err = pthread_mutex_lock (&doclist->lock);
      if (err)
        {
          logerr_err (LOGERR_FUNC,
                      "could not acquire lock to get a property: doc = %s, prop = %s, error = %d",
                      dkey, pkey, err);
          return NULL;
        }
    }
  hashval = hash_string (dkey, doclist->log2nbuckets, &dkeylen);
  dip = doclist->buckets[hashval];
  compare = -1;
  while ((dip != NULL) && ((compare = strcmp (dkey, dip->key)) != 0))
    dip = dip->next;
  if (dip != NULL)
    theprop = get_docitem_property (dip, pkey);
  else
    logerr_err (LOGERR_FUNC,
                "could not find the document to get a property: doc = %s, prop = %s, hv = %d",
                dkey, pkey, hashval);
  if (POSIX_THREADS)
    {
      err = pthread_mutex_unlock (&doclist->lock);
      if (err)
        logerr_err (LOGERR_FUNC,
                    "could not release lock to get a property: doc = %s, prop = %s, error = %d",
                    dkey, pkey, err);
    }
  return (err || (theprop == NULL)) ? NULL : theprop->data;
}

      
