/** @file Sadie_Multi.c
 * A Tcl/Tk interface to SADIE operations on multiple images and band formats.
 * The SADIE library supports operations that can merge multiple images into
 * single images, or split a single multi-band image into multiple images; it
 * can also handle principal-components based manipulations of image bands,
 * and changes to the formats used to store multi-band images.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Multi.c,v 2.5 2005/03/22 22:03:33 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <stdio.h>
#include <string.h>
#include <sadie.h>
#include "tclsadie.h"

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixaddr[] = DEFAULT_IN_IMAGE_ADDR_INDEX;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
/*@}*/

/**
 * Arithmetic operations on image pairs.
 * Given a pair of existing images, an arithmetic operator code, and a
 * pair of scalar weights, generate a new image by combining
 * corresponding pixels in the two images with the operator after
 * scaling them with the weights, using the SADIE library function
 * COMBINE.  The self-explanatory operator codes are defined in
 * @c sadie.h :
 * - @c ADD
 * - @c SUBTRACT
 * - @c MULTIPLY
 * - @c DIVIDE
 * - @c MINIMUM
 * - @c MAXIMUM
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) the first image ID.
 * - @c tcl_array_name(inimg,addr1) the second image ID.
 * - @c tcl_array_name(option)  the code for a binary arithmetic operator.
 * - @c tcl_array_name(weight1)  the weight to apply to the first image.
 * - @c tcl_array_name(weight2)  the weight to apply to the second image.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) newly computed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_CombineCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  int option;
  double weight1, weight2;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "weight1", &weight1)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "weight2", &weight2)
          != TCL_OK))
    return TCL_ERROR;
  COMBINE (inimg1, inimg2, (short)option, weight1, weight2, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Logical operations on image pairs.
 * Given a pair of existing images, and a logical operator code,
 * generate a binary mask image by combining corresponding pixels in
 * the two images with the operator, using the SADIE library function
 * COMPARE.  The operator codes are defined in @c sadie.h :
 * - @c LT  Less than.
 * - @c LE  Less than or equal to.
 * - @c GT  Greater than.
 * - @c GE  Greater than or equal to.
 * - @c EQ  Equal to.
 * - @c NE  Not equal to.
 * - @c AND Logical and.
 * - @c OR  Logical or.
 * - @c XOR Logical exclusive or.
 * - @c NOT Logical not (of the first input image).
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) the first image ID.
 * - @c tcl_array_name(inimg,addr2) the second image ID.
 * - @c tcl_array_name(option)  the code for a logical operator.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) new mask image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_CompareCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  COMPARE (inimg1, inimg2, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Generate a principal component transformation of a multi-band image.
 * Given a multi-band image, this generates a new image of identical
 * dimensions where the bands represent loadings on the eigenvectors
 * of the band covariance matrix of the original image (the Karhunen-Love
 * decomposition, yielding EOFs: empirical orthogonal functions).
 * The EOFs are ordered by the corresponding eigenvalues, and the bands
 * in the output image contain the original image data projected on each
 * EOF in turn, so the first output band is the first principal component
 * image and so on; the SADIE library function PCT handles all the actual
 * computation.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) a multi-band image ID.
 * - @c tcl_array_name(incr) spatial subsampling increment.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) transformed image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_PCTCmd (ClientData client_data, Tcl_Interp *interp,
                    int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t incr;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr", &incr)
          != TCL_OK))
    return TCL_ERROR;
  PCT (inimg, incr, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Transform a multi-band image by a decorrelation stretch.
 * From the principal components transform of a multi-band image (identical
 * to the one provided by Sadie_Multi_PCTCmd), normalize the varience
 * of all the components, then do an inverse transformation to project
 * it back into an image with bands having the same significance as in
 * the original image; the SADIE library function DECORSTR handles everything.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) a multi-band image ID.
 * - @c tcl_array_name(incr) spatial subsampling increment.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) stretched image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_PCSCmd (ClientData client_data, Tcl_Interp *interp,
                    int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  uint32_t incr;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr", &incr)
          != TCL_OK))
    return TCL_ERROR;
  DECORSTR (inimg, incr, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Generate an image that uses hue-saturation-value color coding.
 * Given an image where the first three bands represent red, green and blue
 * channels respectively, use the SADIE library function RGBTOHSV to generate
 * a three-band image, where the bands encode the same color information
 * as hue, saturation and value.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of RGB encoded image.
 * - @c tcl_array_name(incr) spatial subsampling increment.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) the HSV image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_RGB2HSVCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK))
    return TCL_ERROR;
  RGBTOHSV (inimg, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Decode an image that uses hue-saturation-value color coding.
 * Given an image where the first three bands represent hue, saturation and
 * value codes, use the SADIE library function HSVTORGB to generate
 * a three-band image, where the bands encode the same color information
 * as red, green and blue channels.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of HSV encoded image.
 * - @c tcl_array_name(incr) spatial subsampling increment.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) the RGB image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_HSV2RGBCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK))
    return TCL_ERROR;
  HSVTORGB (inimg, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Generate a set of images, one for each band in a multi-band image.
 * We use the SADIE library function BILTOBSQ to split a multi-band
 * image into a set of single-band images, giving these names taken
 * from Tcl array elements; if there are @a N bands, the names would
 * be in @a N Tcl array elements with indices all beginning with
 * "outname" and ending with the band number; the input image ID would
 * be in the same array:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) multi-band image ID.
 * - @c tcl_array_name(outname(i)) @c outname1, @c outname2, ... @c outnameN.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_BIL2BSQCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE **outband = NULL;
  Tcl_Obj **bandobj = NULL;
  size_t i, a;
  char imagename[SLEN];
  int err = TCL_OK;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK))
    return TCL_ERROR;
  outband = calloc (inimg->nbnd, sizeof (IMAGE *));
  if (outband == NULL)
    return TCL_ERROR;
  bandobj = calloc (inimg->nbnd, sizeof (Tcl_Obj *));
  if (bandobj == NULL)
    {
      free (outband);
      return TCL_ERROR;
    }
  BILTOBSQ (inimg, outband);
  for (i = 0; ((err == TCL_OK) && (i < inimg->nbnd)); i++)
    {
      snprintf (imagename, SLEN - 1, "outname%lu",
                (long unsigned int) (i + 1));
      bandobj[i] = NewSadieImageObj2 (interp, objv[1], (const char *)imagename,
                                      outband[i]);
      if (bandobj[i] != NULL)
        Tcl_IncrRefCount (bandobj[i]);
      else
        err = TCL_ERROR;
    }
  if (err == TCL_OK)
    Tcl_SetObjResult (interp, Tcl_NewListObj (inimg->nbnd, bandobj));
  for (a = 0; a < inimg->nbnd; a++)
    Tcl_DecrRefCount (bandobj[a]);
  free (bandobj);
  free (outband);
  return err;
}

/**
 * Create a three-band image from three individual single-band images.
 * We use the SADIE library function BSQTOBIL to merge the three specified
 * single-band images into one multi-band image, but don't manage a full
 * inverse of Sadie_Multi_BIL2BSQCmd.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) ID of the image to become band 1.
 * - @c tcl_array_name(inimg,addr2) ID of the image to become band 2.
 * - @c tcl_array_name(inimg,addr3) ID of the image to become band 3.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) multi-band image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_BSQ2BILCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inband[3] = { NULL, NULL, NULL };
  IMAGE *outimg = NULL;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", inband)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", inband + 1)
          != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr3", inband + 2)
          != TCL_OK))
    return TCL_ERROR;
  BSQTOBIL (inband, 3, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/** Fixed string indexes into whatever Tcl array holds the component names. */
static const char *comp_label_index[2] = { "outname1", "outname2" };

/**
 * Split a complex image into two components.
 * Given an image that encodes complex pixel values in its first two
 * bands, use the SADIE library function COM2REAL to generate a pair
 * of output images according to an option code taking one of the
 * values defined in @c sadie.h : @c REIMAG makes the first image the
 * real component and the second the imaginary, @c MAGPHA makes the
 * first the magnitude and the second the phase.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of the complex image.
 * - @c tcl_array_name(option) real/imaginary, magnitude/phase option code.
 * - @c tcl_array_name((comp_label_index[1])) the first component image name.
 * - @c tcl_array_name((comp_label_index[2])) the second component image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_Comp2RealCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *comp_img[2] = { NULL, NULL };
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  COM2REAL (inimg, (short)option, comp_img, comp_img + 1);
  return SetSadieComponentObj2 (interp, objv[1], comp_label_index, comp_img);
}

/**
 * Generate a complex image from two real images.
 * Given a pair of real images, use the SADIE library function
 * REAL2COM to generate a complex image, interpreting the image pair
 * as the components specified by an option code taking one of the
 * values defined in @c sadie.h : @c REIMAG makes the first image the
 * real component and the second the imaginary, @c MAGPHA makes the
 * first the magnitude and the second the phase.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) ID of the first component image.
 * - @c tcl_array_name(inimg,addr2) ID of the second component image.
 * - @c tcl_array_name(option) real/imaginary, magnitude/phase option code.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) the complex image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_Real2CompCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  int option;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK))
    return TCL_ERROR;
  REAL2COM (inimg1, inimg2, (short)option, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Change the storage order of bands, lines or pixels in an image.
 * From an existing image the SADIE library function ROTATECUBE generates a
 * new image with the interpretation of the image dimensions changed, for
 * example mapping the bands of one scanline of a hyperspectral image
 * onto the pixel dimension of a two-dimensional display, showing how the spectral
 * characteristics vary along the line.  Two sets of flags control the
 * image transformations: @a sign flags take the value 0 to keep the
 * original order of values along an axis, or 1 to invert it; @a dimension
 * flags specify which axis of the input image to map onto an axis
 * of the output image: 0 specifies the band axis, 1 the line axis and 2 the
 * pixel axis.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) ID of the input image.
 * - @c tcl_array_name(x,sign) the sign flag for the @a x axis.
 * - @c tcl_array_name(y,sign) the sign flag for the @a y axis.
 * - @c tcl_array_name(z,sign) the sign flag for the @a z axis.
 * - @c tcl_array_name(x,cubedim) the dimension flag for the @a x axis.
 * - @c tcl_array_name(y,cubedim) the dimension flag for the @a y axis.
 * - @c tcl_array_name(z,cubedim) the dimension flag for the @a z axis.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) the output image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Multi_RotateCubeCmd (ClientData client_data, Tcl_Interp *interp,
                           int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int xdir, xaxis;
  int ydir, yaxis;
  int zdir, zaxis;

  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "x,sign", &xdir)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "y,sign", &ydir)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "z,sign", &zdir)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "x,cubedim", &xaxis)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "y,cubedim", &yaxis)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "z,cubedim", &zaxis)
          != TCL_OK))
    return TCL_ERROR;
  ROTATECUBE (inimg, (short)xdir, (short)ydir, (short)zdir,
              (short)xaxis, (short)yaxis, (short)zaxis, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_Multi_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Multi_Combine",
                        Sadie_Multi_CombineCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_Compare",
                        Sadie_Multi_CompareCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_PCT",
                        Sadie_Multi_PCTCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_PCS",
                        Sadie_Multi_PCSCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_RGB2HSV",
                        Sadie_Multi_RGB2HSVCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_HSV2RGB",
                        Sadie_Multi_HSV2RGBCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_BIL2BSQ",
                        Sadie_Multi_BIL2BSQCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_BSQ2BIL",
                        Sadie_Multi_BSQ2BILCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_Comp2Real",
                        Sadie_Multi_Comp2RealCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_Real2Comp",
                        Sadie_Multi_Real2CompCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Multi_RotateCube",
                        Sadie_Multi_RotateCubeCmd, NULL, NULL);
  return TCL_OK;
}

