/** @file Sadie_NewFunctions.c
 * A Tcl/Tk interface to the SADIE operations that synthesize images.
 * It is often useful to synthesize an image by a simple algorithm
 * rather than by importing some kind of sensor data, not only for
 * having test patterns to check the results of applying sequences of
 * image processing operations, or simulated noise for modeling
 * processes of image degradation, but also to act as masks or windows
 * for operations such as Fourier transforms.  The SADIE library
 * includes sever operations for synthesizing all these types of
 * image, and we define the Tcl/Tk interface to them here.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_NewFunctions.c,v 2.6 2005/03/22 21:42:33 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "tclsadie.h"

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
/*@}*/

/**
 * Synthesize an image with a horizontal sine wave pattern.
 * The SADIE library function SINEWAVE scales a sine function so that
 * its maxima are at the value @c SMAX and its minima are at @c SMIN
 * (as defined in @c sadie.h), with a period of the specified number
 * of pixels, and uses it to synthesize the values of a line of pixels,
 * which it repeats for the specified number of lines vertically to
 * produce a pattern of vertical light and dark bands.  The phase is
 * the displacement of the sine function's origin from the start of a
 * line, in pixel units.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(period) period of the sine function (pixel units).
 * - @c tcl_array_name(phase) horizontal offset of the function origin.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_SineWaveCmd (ClientData client_data, Tcl_Interp *interp,
                                int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  double period, phase;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "period", &period)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "phase", &phase)
          != TCL_OK))
    return TCL_ERROR;
  SINEWAVE (lines, pix, period, phase, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Synthesize a two-dimensional chirp test pattern.
 * The SADIE library function CHIRP synthesizes a two-dimensional sinusoidal
 * pattern with the frequency of the sinusoid increasing with distance from
 * the center of the pattern.  The pattern itself is periodic, with the
 * specified period in pixel units being the interval between pattern
 * centers.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(period) repetition interval for the pattern.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_ChirpCmd (ClientData client_data, Tcl_Interp *interp,
                             int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  double period;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "period", &period)
          != TCL_OK))
    return TCL_ERROR;
  CHIRP (lines, pix, period, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Synthesize a radially-symmetrical sine test pattern.
 * The SADIE library function SINESTAR synthesizes a two-dimensional
 * sine test pattern that uses polar coordinates with an origin in at
 * the center of the image.  For each concentric circle, it sets
 * the image intensity to the sine of the angle scaled by a frequency
 * factor in hertz (cycles round the circle).
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(period) the radial repetition frequency of the pattern.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_SineStarCmd (ClientData client_data, Tcl_Interp *interp,
                                int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  double period;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "period", &period)
          != TCL_OK))
    return TCL_ERROR;
  SINESTAR (lines, pix, period, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Synthesize a test pattern of alternating black and white rectangles.
 * The SADIE library function CHECKER synthesizes an image tiled with
 * regularly alternating black and white rectangles of the specified
 * size, starting from the top left corner.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(checkpix) horizontal rectangle size (pixel units).
 * - @c tcl_array_name(checklines) vertical rectangle size (pixel units).
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_CheckerboardCmd (ClientData client_data, Tcl_Interp *interp,
                                    int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  uint32_t checklines, checkpix;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "checklines", &checklines)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "checkpix", &checkpix)
          != TCL_OK))
    return TCL_ERROR;
  CHECKER (lines, pix, checklines, checkpix, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Synthesize a uniform horizontal gradient image.
 * The SADIE library function GRAYSCAL synthesizes an image of the specified
 * size filled with a uniform horizontal gradient from a starting greylevel
 * to an ending greylevel (imposing a limitation that the start must be
 * darker than the end).
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(number) number of greylevels (broken at present?).
 * - @c tcl_array_name(min) starting value for the gradient.
 * - @c tcl_array_name(max) ending value for the gradient.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_GrayScaleCmd (ClientData client_data, Tcl_Interp *interp,
                                 int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix, number;
  PIXEL min, max;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "number", &number)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "min", &min)
          != TCL_OK)
      || (GetSadiePixelFromObj2 (interp, objv[1], "max", &max)
          != TCL_OK))
    return TCL_ERROR;
  GRAYSCAL (lines, pix, number, min, max, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Synthesize a windowing function image.
 * FUNCTION in the SADIE library sythesizes images with values
 * determined by one of a family of functions intended to be useful
 * for windowing and similar operations.  They generally have radial
 * symmetry about the image center (except for the similarly centered
 * box function), modified by different scale factors for the
 * horizontal and vertical axes, and values scaled to range from @c
 * SMIN to @c SMAX (defined in @c sadie.h).  An type code, also
 * defined in @c sadie.h, selects the particular function to use :
 * - @c CONE uniform radial gradient (2D triangle),
 * - @c NEGEXP 2D negative exponential,
 * - @c GAUSS 2D Gaussian,
 * - @c BOX centered square or rectangle,
 * - @c CYLINDER centered circle or elipse,
 * - @c HANNING 2D Hanning window (a modified truncated cosine),
 * - @c HAMMING 2D Hamming window (another modified truncated cosine),
 * - @c KAISER Kaiser window (with @ alpha specifying sidelobe attenuation),
 * - @c SINC 2D version of the sinc function.
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(type) function type code.
 * - @c tcl_array_name(radiuslines) defines the vertical scaling.
 * - @c tcl_array_name(radiuspix) defines the horizontal scaling.
 * - @c tcl_array_name(alpha) sidelobe attenuation for the Kaiser window.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_FunctionCmd (ClientData client_data, Tcl_Interp *interp,
                                int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  int type;
  double radiuslines, radiuspix, alpha;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "type", &type)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "radiuslines", &radiuslines)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "radiuspix", &radiuspix)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "alpha", &alpha)
          != TCL_OK))
    return TCL_ERROR;
  FUNCTION (lines, pix, radiuslines, radiuspix, alpha, (short)type, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Synthesize an image filled with random noise.
 * The SADIE library function RANDOM synthesizes an image of the specified
 * size filled with noise generated as random numbers from either a uniform
 * or a Gaussian distribution, depending on whether a type parameter
 * equals @c UNIFORM or @c GAUSSIAN respectively (constants defined in
 * @c sadie.h).
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(lines) the number of lines in the synthesized image.
 * - @c tcl_array_name(pix) pixels per line in the synthesized image.
 * - @c tcl_array_name(type) the uniform or Gaussian type code.
 * - @c tcl_array_name(deviation) Gaussian standard deviation, uniform scale.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) synthesized image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_NewFunctions_RandomCmd (ClientData client_data, Tcl_Interp *interp,
                              int objc, Tcl_Obj * const objv[])
{
  IMAGE *outimg = NULL;
  uint32_t lines, pix;
  int type;
  double deviation;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieUintFromObj2 (interp, objv[1], "lines", &lines)
       != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "pix", &pix)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "type", &type)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "deviation", &deviation)
          != TCL_OK))
    return TCL_ERROR;
  RANDOM (lines, pix, (short)type, deviation, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_NewFunctions_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_SineWave",
                        Sadie_NewFunctions_SineWaveCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_Chirp",
                        Sadie_NewFunctions_ChirpCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_SineStar",
                        Sadie_NewFunctions_SineStarCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_Checkerboard",
                        Sadie_NewFunctions_CheckerboardCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_GrayScale",
                        Sadie_NewFunctions_GrayScaleCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_Function",
                        Sadie_NewFunctions_FunctionCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_NewFunctions_Random",
                        Sadie_NewFunctions_RandomCmd, NULL, NULL);
  return TCL_OK;
}






