/** @file Sadie_Plot.c
 * Tcl/Tk interfaces to plotting and region of interest operations.
 * The operations to display histograms and other pixel value statistics
 * (and analyse regions of interest within images) are largely defined
 * elsewhere (including callbacks that must be pre-registered with the
 * SADIE library and utilities defined within tclSadie), but we define
 * the high-level Tcl/Tk interfaces to these here.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Plot.c,v 2.4 2005/03/22 21:01:45 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <stdio.h>
#include <string.h>
#include <sadie.h>
#include "tclsadie.h"
#include "bresenham.h"
#include "tclSadie_ROI.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/**
 * Plot the profiles of image values between two specified image points.
 * A callback that handles most of the Tcl/Tk drawing interface should
 * already be registered with the SADIE library (see the
 * @c tclSadie_plot.c file for details), and the PLOT function invokes
 * this.  However it doesn't extract the values to be plotted from
 * the image itself, so here we must define a temporary array to
 * hold them (extracting each band independently), and track positions
 * along the profile line with the Bresenham algorithm (there is no
 * interpolation, so the number of profile points will not equal the
 * Euclidean distance between the points in pixel units).  An option
 * code determines the scaling on the plot (one of the constants @c UNIT,
 * @c NORM or @c LOG defined in @c sadie.h).  The Tcl object arguments
 * that specify all the plot details are in the following arrangement:
 * - @c objv[0] the command name.
 * - @c objv[1] locate the profile within the image with this ID.
 * - @c objv[2] the horizontal coordinate of the profile start.
 * - @c objv[3] the vertical coordinate of the profile start.
 * - @c objv[4] the horizontal coordinate of the profile end.
 * - @c objv[5] the vertical coordinate of the profile end.
 * - @c objv[6] a scaling option code.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always four).
 * @param[in]  objv  the array of Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_Plot_ProfileCmd (ClientData client_data, Tcl_Interp *interp,
                       int objc, Tcl_Obj * const objv[])
{
  int startx, endx;
  int starty, endy;
  int option, off_the_end;
  size_t delta_x, delta_y, point, i, j;
  IMAGE *inimg = NULL;
  PIXEL *arrays = NULL;
  char minlbl[SLEN], maxlbl[SLEN];
  bresenham_linep_t profile = NULL;
  uint32_t profile_col, profile_row;

  if (objc != 7)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "image_id start_x start_y end_x end_y option_code");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &inimg) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[2], &startx) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[3], &starty) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[4], &endx) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[5], &endy) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[6], &option) != TCL_OK))
    return TCL_ERROR;
  /* check input */
  if ((option != UNIT) && (option != NORM) && (option != LOG))
    {
      MESSAGE ('E', " Option must be one of UNIT, NORM, or LOG.");
      return TCL_ERROR;
    }
  /* Count points*/
  delta_x = (startx < endx)
    ? ((size_t) endx - startx) : ((size_t) startx - endx);
  delta_y = (starty < endy)
    ? ((size_t) endy - starty) : ((size_t) starty - endy);
  point = 1 + ((delta_x < delta_y) ? delta_y : delta_x);
  /* Allocate memory for plotting */
  arrays = calloc (inimg->nbnd * point, sizeof (PIXEL));
  if (arrays == NULL)
    {
      MESSAGE ('E', " Error allocating memory for plotting!");
      return TCL_ERROR;
    }
  /* Fill the plot array */
  for (i = 0; i < inimg->nbnd; i++)
    {
      off_the_end = 0;
      profile = bresenham_from_s_coord (startx, starty, endx, endy);
      for (j = 0; (!off_the_end && (j < point)); j++)
        {
          off_the_end = bresenham_next (profile, &profile_col, &profile_row);
          arrays[i*point+j] = inimg->data[i][profile_row][profile_col];
        }
      demolish_bresenham (&profile);
      if (!off_the_end || j != point)
        {
          MESSAGE ('E',
                   " unexpected error when retrieving data for the profile.");
          free (arrays);
          return TCL_ERROR;
        }
    }
  snprintf (minlbl, SLEN - 1, "(%d,%d)", startx, starty);
  snprintf (maxlbl, SLEN - 1, "(%d,%d)", endx, endy);
  PLOT (arrays, inimg->nbnd, (uint32_t)point, minlbl, maxlbl, (short)option);
  if (NAMES)
    {
      MESSAGE ('I', " ");
      MESSAGE ('I', " ...............");
    }
  free(arrays);
  return TCL_OK;
}

/**
 * Compute summary statistics and value distribution historams for an image.
 * The SADIE library function STATS computes a variety of statistics
 * for the pixel values of an image, including the histograms of the
 * value distributions in the first few bands (using the same
 * Tcl/Tk-aware callback used indirectly by Sadie_Plot_ProfileCmd).
 * However it requires provision of some temporary workspace for
 * accumulating the statistics for each band, and we augment it here
 * with the RANGE function for finding maxima and minima, and the
 * COVAR function for the optional band covariance matrix.  The range
 * option code specifies how to handle the specified range of pixel
 * values, as one of the constants @c IGNORE, @c INCLUDE or @c EXCLUDE
 * defined in @c sadie.h.  The array holding the Tcl object arguments
 * is arranged as follows:
 * - @c objv[0] the command name.
 * - @c objv[1] compute statistics on the image with this ID.
 * - @c objv[2] the sub-sampling increment (pixel units).
 * - @c objv[3] the number of histogram bins for each band.
 * - @c objv[4] set 1 if a covarience matrix is needed, 0 otherwise.
 * - @c objv[5] the range option code.
 * - @c objv[6] the start of the range of pixel values.
 * - @c objv[7] the end of the range of pixel values.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always four).
 * @param[in]  objv  the array of Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_Plot_StatisticsCmd (ClientData client_data, Tcl_Interp *interp,
                          int objc, Tcl_Obj * const objv[])
{
  IMAGE *in = NULL;
  int incr, nbins, matrix;
  double *mean = NULL;
  double *dev = NULL;
  double *cov = NULL;
  double *cor = NULL;
  PIXEL *gmin = NULL;
  PIXEL *gmax = NULL;
  PIXEL *grng = NULL;
  uint64_t *zero = NULL;
  int rangeoption;
  PIXEL rangemin, rangemax;
  double tmp_rangemin, tmp_rangemax;

  if (objc != 8)
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "image_id increment n_bins matrix range_option range_min range_max");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &in) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[2], &incr) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[3], &nbins) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[4], &matrix) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[5], &rangeoption) != TCL_OK)
      || (Tcl_GetDoubleFromObj (interp, objv[6], &tmp_rangemin) != TCL_OK)
      || (Tcl_GetDoubleFromObj (interp, objv[7], &tmp_rangemax) != TCL_OK))
    return TCL_ERROR;
  rangemin = (PIXEL) tmp_rangemin;
  rangemax = (PIXEL) tmp_rangemax;
  if (!CHECKIMG(in))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  if ((rangeoption != IGNORE) && (rangeoption != INCLUDE)
      && (rangeoption != EXCLUDE))
    {
      MESSAGE ('E', " Option must be one of IGNORE, INCLUDE, or EXCLUDE.");
      return TCL_ERROR;
    }
  RANGE (in);
  /* allocate space */
  if ((mean = calloc (in->nbnd, sizeof (double)))
      && (dev = calloc (in->nbnd, sizeof (double)))
      && (gmin = calloc (in->nbnd, sizeof (PIXEL)))
      && (gmax = calloc (in->nbnd, sizeof (PIXEL)))
      && (grng = calloc (in->nbnd, sizeof (PIXEL)))
      && (zero = calloc (in->nbnd, sizeof (uint64_t)))
      && (cov = calloc ((size_t) in->nbnd * in->nbnd, sizeof (double)))
      && (cor = calloc ((size_t) in->nbnd * in->nbnd, sizeof (double))))
    {
      STATS (in, incr, nbins, (short) rangeoption, rangemin, rangemax,
             mean, dev, gmin, gmax, grng, zero);
      if (matrix)
        COVAR (in, incr, cov, cor);
    }
  else
    MESSAGE ('E', " Memory request failed.");
  if (cov)
    free(cov);
  if (cor)
    free(cor);
  if (mean)
    free(mean);
  if (dev)
    free(dev);
  if (gmin) 
    free(gmin);
  if (gmax)
    free(gmax);
  if (grng)
    free(grng);
  if (zero)
    free(zero);
  return TCL_OK;
}

/**
 * Define a rectangular region of interest within an image.
 * Although the SADIE library does not support Regions of Interest (ROIs),
 * an extension within tclSadie allows definition of multiple rectangular
 * ROIs associated with particular images.  Here we add a new ROI to an
 * image using this extension (see the file tclSadie_ROI.c for more details).
 * The array of Tcl object arguments is arranged in the following pattern:
 * - @c objv[0] the command name.
 * - @c objv[1] compute statistics on the image with this ID.
 * - @c objv[2] the sub-sampling increment (pixel units).
 * - @c objv[3] the number of histogram bins for each band.
 * - @c objv[4] set 1 if a covarience matrix is needed, 0 otherwise.
 * - @c objv[5] the horizontal coordinate of one ROI corner.
 * - @c objv[6] the vertical coordinate of one ROI corner.
 * - @c objv[7] the horizontal coordinate of the opposite ROI corner.
 * - @c objv[8] the vertical coordinate of the opposite ROI corner.
 * - @c objv[9] the ROI name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the main Tcl interpreter.
 * @param[in]  objc  the number of Tcl object arguments (always four).
 * @param[in]  objv  the array of Tcl object arguments.
 * @return @c TCL_OK, or @c TCL_ERROR on errors.
 */
int
Sadie_Plot_CreateROICmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *in = NULL;
  const char *name;
  uint32_t incr, nlev;
  int matrix;
  uint32_t point1x, point2x, point1y, point2y;
  uint32_t xbgn, ybgn, xend, yend;
  const char *dkey;

  if (objc != 10 )
    {
      Tcl_WrongNumArgs (interp, 1, objv,
                        "image_id increment n_levels matrix point1_x point1_y point2_x point2_y name");
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj (objv[1], &in) != TCL_OK)
      || (GetSadieDkeyFromObj (objv[1], &dkey) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[2], &incr) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[3], &nlev) != TCL_OK)
      || (Tcl_GetIntFromObj (interp, objv[4], &matrix) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[5], &point1x) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[6], &point1y) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[7], &point2x) != TCL_OK)
      || (GetSadieUintFromObj (interp, objv[8], &point2y) != TCL_OK))
    return TCL_ERROR;
  name = Tcl_GetString (objv[9]);
  if (!CHECKIMG (in))
    {
      Tcl_SetStringObj (Tcl_GetObjResult (interp),
                        "couldn't identify SADIE image from input string.",
                        -1);
      return TCL_ERROR;
    }
  RANGE (in);
  xbgn = min(point1x, point2x);
  xend = max(point1x, point2x);
  ybgn = min(point1y, point2y);
  yend = max(point1y, point2y);
  CREATE_ROI (in, name, dkey, xbgn, ybgn, xend, yend, matrix, incr, nlev);
  return TCL_OK;
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_Plot_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Plot_Profile",
                        Sadie_Plot_ProfileCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "Sadie_Plot_Statistics",
                       Sadie_Plot_StatisticsCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "Sadie_Plot_CreateROI",
                       Sadie_Plot_CreateROICmd, NULL, NULL);
  return TCL_OK;
}

