/** @file Sadie_Tools.c
 * Destriping, contour lines, pseudo-relief, RMS differences and scattergrams.
 * Tcl/Tk interfaces to a collection of various utility functions in the
 * SADIE library.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: Sadie_Tools.c,v 2.4 2005/03/22 22:04:50 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sadie.h>
#include "tclsadie.h"

/** @name Fixed Tcl array indices
 * C versions of fixed constant array index strings used in the Tcl code.
 */
/*@{*/
static const char atitle[] = DEFAULT_TCL_ARRAY_TITLE;
static const char ixaddr[] = DEFAULT_IN_IMAGE_ADDR_INDEX;
static const char ixname[] = DEFAULT_OUT_IMAGE_NAME_INDEX;
/*@}*/

/**
 * Destriping to remove systematic artefacts caused by detector variations.
 * The SADIE library function DESTRIPE provides several different options
 * to remove the horizontal or vertical stripe artefacts that can appear
 * on images acquired with a sensor that is composed of multiple physical
 * detectors with slightly different physical characteristics.  The file
 * @c sadie.h holds the constants that specify which option to use:
 * the option for the type of line matching can take the values :
 * - @c DESMNE to make only bias adjustments to the detector means,
 * - @c DESSTD to in addition standardize the variances, and
 * - @c DESCDF to match the cumulative distribution functions.
 *
 * The reference option can take the values
 * - @c DESSNG to use a single detector as the reference for adjustments, or
 * - @c DESAVG to use the mean of all the detectors.
 *
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) the degraded image ID.
 * - @c tcl_array_name(option)  the line matching option.
 * - @c tcl_array_name(reference) the reference option.
 * - @c tcl_array_name(period) the stripe period (number of detectors).
 * - @c tcl_array_name(detector) the reference detector number.
 * - @c tcl_array_name(number) cumulative distribution function levels.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) destriped image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Tools_DestripeCmd (ClientData client_data, Tcl_Interp *interp,
                         int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int option, reference, period, detector, number;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "option", &option)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "reference", &reference)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "period", &period)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "detector", &detector)
          != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "number", &number)
          != TCL_OK))
    return TCL_ERROR;
  DESTRIPE (inimg, (short)option, (short)reference, (short)period,
            (short)detector, (short)number, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Generate a binary image showing contour lines from an input image.
 * Given an arbitrary input image, this draws a crude contour map of
 * the pixel values as black and white areas on a new binary image,
 * using the SADIE library function CONTOUR.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) the input image ID.
 * - @c tcl_array_name(number) the number of contour lines.
 * - @c tcl_array_name(width) the (fractional) contour line width.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) contour image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Tools_ContMapCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  int number;
  double width;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieIntFromObj2 (interp, objv[1], "number", &number)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "width", &width)
          != TCL_OK))
    return TCL_ERROR;
  CONTOUR (inimg, (short)number, width, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Generate a pseudo-relief image from a DEM or arbitrary image.
 * This treats the input image (regardless of actual source) as a digital
 * elevation model, and given some additional parameters for illumination
 * geometry, generates a shaded relief image from it, using the SADIE
 * library function RELIEF (for realistic relief, the input pixel
 * values and the @a distance parameter should be in the same physical
 * units).
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(DEFAULT_IN_IMAGE_ADDR_INDEX) the DEM image ID.
 * - @c tcl_array_name(phi) horizontal illumination angle (radians from E).
 * - @c tcl_array_name(theta) illumination altitude (radians upwards).
 * - @c tcl_array_name(dist) physical distance between input pixels.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) relief image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Tools_ReliefCmd (ClientData client_data, Tcl_Interp *interp,
                       int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg = NULL;
  IMAGE *outimg = NULL;
  double phi, theta, dist;
  
  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], ixaddr, &inimg)
       != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "phi", &phi)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "theta", &theta)
          != TCL_OK)
      || (GetSadieDoubleFromObj2 (interp, objv[1], "dist", &dist)
          != TCL_OK))
    return TCL_ERROR;
  phi *= (PI / 180);
  theta *= (PI / 180);
  RELIEF (inimg, phi, theta, dist, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * A scattergram from the joint distribution of pixel values in two images.
 * Given two bands from a multi-band image, or two images of equal size,
 * we use the SADIE library function SCATTER to generate a scattergram
 * from the values of pixels at corresponding locations.
 * The Tcl data should conform to the patterns:
 * - @c tcl_array_name(inimg,addr1) the first image ID.
 * - @c tcl_array_name(inimg,addr2) a second image ID (may equal the first).
 * - @c tcl_array_name(band1) first image band number (counting from 1).
 * - @c tcl_array_name(band2) second image band number (counting from 1).
 * - @c tcl_array_name(incr) spatial sampling interval (pixels).
 * - @c tcl_array_name(res) the number of 2D histogram bins.
 * - @c tcl_array_name(DEFAULT_OUT_IMAGE_NAME_INDEX) relief image name.
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Tools_ScatterCmd (ClientData client_data, Tcl_Interp *interp,
                        int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  IMAGE *outimg = NULL;
  uint32_t band1, band2, incr, res;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "band1", &band1)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "band2", &band2)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "incr", &incr)
          != TCL_OK)
      || (GetSadieUintFromObj2 (interp, objv[1], "res", &res)
          != TCL_OK))
    return TCL_ERROR;
  SCATTER (inimg1, inimg2, --band1, --band2, incr, res, &outimg);
  return SetSadieImageObj2 (interp, objv[1], ixname, outimg);
}

/**
 * Compute the root-mean-square difference between two images.
 * Given a pair of images, we use the SADIE library function RMSDIFF to
 * find the RMS difference between them (a single scalar value), and
 * return this in the Tcl interpreter result.
 * - @c tcl_array_name(inimg,addr1) the first image ID.
 * - @c tcl_array_name(inimg,addr2) a second image ID (may equal the first).
 *
 * @param[in]  client_data  (unused).
 * @param[in]  interp       the Tcl interpreter.
 * @param[in]  objc  number of Tcl object arguments (always two).
 * @param[in]  objv  Tcl object argument list (the Tcl array name).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Sadie_Tools_RMSCmd (ClientData client_data, Tcl_Interp *interp,
                    int objc, Tcl_Obj * const objv[])
{
  IMAGE *inimg1 = NULL;
  IMAGE *inimg2 = NULL;
  double diff;

  if (objc != 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, atitle);
      return TCL_ERROR;
    }
  if ((GetSadieImageFromObj2 (interp, objv[1], "inimg,addr1", &inimg1)
       != TCL_OK)
      || (GetSadieImageFromObj2 (interp, objv[1], "inimg,addr2", &inimg2)
          != TCL_OK))
    return TCL_ERROR;
  RMSDIFF (inimg1, inimg2, &diff);
  Tcl_SetObjResult (interp, Tcl_NewDoubleObj (diff));
  return TCL_OK;
}

/**
 * Make the commands defined in this file accessible to the Tcl interpreter.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success.
 */
int
Sadie_Tools_Init (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "Sadie_Tools_Destripe",
                        Sadie_Tools_DestripeCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Tools_ContMap",
                        Sadie_Tools_ContMapCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Tools_Relief",
                        Sadie_Tools_ReliefCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Tools_Scatter",
                        Sadie_Tools_ScatterCmd, NULL, NULL);
  Tcl_CreateObjCommand (interp, "Sadie_Tools_RMS",
                        Sadie_Tools_RMSCmd, NULL, NULL);
  return TCL_OK;
}

