/** @file bresenham.c
 * An implementation of Bresenham's line drawing algorithm.
 * Many different web resources reference a web page by Colin Flanagan, 
 * currently hosted at
 * http://www.cs.helsinki.fi/group/goa/mallinnus/lines/bresenh.html
 * which is a very lucid description of the derivation of the algorithm.
 * Remember that here we're using row & column notation rather than
 * left-handed Cartesian x,y coordinates, but in any case we define
 * arbitrary internal coordinates, that get mapped to rows and columns
 * as needed.  The original reference is: Bresenham, J.E. (1965)
 * "Algorithm for computer control of a digital plotter".  IBM Systems
 * Journal 4(1) 1965, p. 25--30.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2004
 * @version $Id: bresenham.c,v 1.1 2005/03/22 20:23:43 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <sadie.h>
#include <math.h>
#include "logerr.h"
#include "bresenham.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** An integer of at least 33 bits: holds any signed OR unsigned 32 bit int. */
typedef int64_t bcoord_t;

/** A coordinate accessor function. */
typedef bcoord_t (* coord_accessorp_t) (bresenham_linep_t bline);

/** A point iterator function. */
typedef void (* point_iteratorp_t) (bresenham_linep_t bline);

/** Coordinates for anywhere in a 32-bit signed or unsigned Cartesian frame. */
struct Bpoint
{
  bcoord_t x;  /**< First internal coordinate. */
  bcoord_t y;  /**< Second internal coordinate. */
};

/** A context capturing the state of the Bresenham line drawing algorithm. */
struct Bresenham_line
{
  struct Bpoint first;    /**< The first point on the line. */
  struct Bpoint current;  /**< The current point on the line. */
  struct Bpoint last;     /**< The last point on the line. */
  int past_the_end;       /**< Boolean true iff there are no more points. */
  bcoord_t epsilon;       /**< The algorithm's error term. */
  bcoord_t delta_x;       /**< Absolute first to last x distance. */
  bcoord_t delta_y;       /**< Absolute first to last y distance. */
  coord_accessorp_t get_col; /**< Points to the column coordinate function. */
  coord_accessorp_t get_row; /**< Points to the row coordinate function. */
  point_iteratorp_t next_pt; /**< Points to the state iterating function. */
};

/**
 * Advance to the next state for lines of positive slope.
 * The classic Bresenham updating step, only valid in the one octant
 * of the Cartesian plane centered on the first point of the line.
 * @param[in,out] bline  points to the structure recording the state to update.
 */
static void
next_positive (bresenham_linep_t bline)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((bline != NULL) && (bline->current.x < bline->last.x)));
  bline->epsilon += bline->delta_y;
  if ((2 * bline->epsilon) >= bline->delta_x)
    {
      bline->epsilon -= bline->delta_x;
      (bline->current.y)++;
    }
  (bline->current.x)++;
}

/**
 * Advance to the next state for lines of negative slope.
 * The classic Bresenham updating step, only valid in the one octant
 * of the Cartesian plane centered on the first point of the line.
 * @param[in,out] bline  points to the structure recording the state to update.
 */
static void
next_negative (bresenham_linep_t bline)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 ((bline != NULL) && (bline->current.x < bline->last.x)));
  bline->epsilon += bline->delta_y;
  if ((2 * bline->epsilon) <= (-bline->delta_x))
    {
      bline->epsilon += bline->delta_x;
      (bline->current.y)--;
    }
  (bline->current.x)++;
}

/**
 * Access what the Bresenham algorithm is treating as the x-coordinate.
 * @param[in] bline  points to the structure recording the state to examine.
 */
static bcoord_t
fetch_x (bresenham_linep_t bline)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (bline != NULL));
  return bline->current.x;
}

/**
 * Access what the Bresenham algorithm is treating as the y-coordinate.
 * @param[in] bline  points to the structure recording the state to examine.
 */
static bcoord_t
fetch_y (bresenham_linep_t bline)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (bline != NULL));
  return bline->current.y;
}

/**
 * Reflect the Bresenham algorithm's x-coordinate about the line's first point.
 * @param[in] bline  points to the structure recording the state to examine.
 */
static bcoord_t
fetch_neg_x (bresenham_linep_t bline)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (bline != NULL));
  return (2 * bline->first.x - bline->current.x);
}

/**
 * Construct a Bresenham line context from row and column coordinates.
 * The simple versions of the algorithm only work for some orderings
 * of the end points of the line and some gradients, so
 * implementations commonly reflect or transpose the coordinates to
 * reduce them to the regions where the simple versions operate.  Here
 * we want to respect the direction of the original line in the order
 * in which we generate the coordinate pairs, so we can't use
 * solutions that include simple transpositions of the first and last
 * points.  Instead, we use pointers to tiny accessor functions to map
 * the data manipulated by the algorithm onto the row and column
 * coordinates understood by the calling application.  Pointers to
 * simple little functions can also encode the choice between the two
 * basic variations of the algorithm (lines of positive and negative
 * slope).
 * @param[in] first_col  the column-coordinate of the first point.
 * @param[in] first_row  the row-coordinate of the first point.
 * @param[in] last_col  the column-coordinate of the last point.
 * @param[in] last_row  the row-coordinate of the last point.
 * @return  A pointer to the state of the Bresenham context at the first point.
 */
static bresenham_linep_t
bresenham_from_bcoord (bcoord_t first_col, bcoord_t first_row,
                       bcoord_t last_col, bcoord_t last_row)
{
  bcoord_t delta_row, delta_col, abs_delta_row, abs_delta_col;
  bresenham_linep_t bline;
  int transposition_needed, reflection_needed;

  bline = malloc(sizeof(struct Bresenham_line));
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (bline != NULL));
  delta_row = last_row - first_row;
  delta_col = last_col - first_col;
  abs_delta_row = (delta_row >= 0) ? delta_row : (-delta_row);
  abs_delta_col = (delta_col >= 0) ? delta_col : (-delta_col);
  transposition_needed = (abs_delta_row < abs_delta_col);
  reflection_needed = (delta_col < (-delta_row));
  if (transposition_needed)
    {
      bline->first.x = first_col;
      bline->first.y = first_row;
    }
  else
    {
      bline->first.x = first_row;
      bline->first.y = first_col;
    }
  if (reflection_needed)
    {
      if (transposition_needed)
        {
          bline->last.x = first_col - delta_col;
          bline->last.y = last_row;
          bline->get_row = fetch_y;
          bline->get_col = fetch_neg_x;
        }
      else
        {
          bline->last.x = first_row - delta_row;
          bline->last.y = last_col;
          bline->get_row = fetch_neg_x;
          bline->get_col = fetch_y;
        }
      if ((delta_row > 0) || (delta_col > 0))
        bline->next_pt = next_positive;
      else
        bline->next_pt = next_negative;
    }
  else
    {
      if (transposition_needed)
        {
          bline->last.x = last_col;
          bline->last.y = last_row;
          bline->get_row = fetch_y;
          bline->get_col = fetch_x;
        }
      else
        {
          bline->last.x = last_row;
          bline->last.y = last_col;
          bline->get_row = fetch_x;
          bline->get_col = fetch_y;
        }
      if ((delta_row < 0) || (delta_col < 0))
        bline->next_pt = next_negative;
      else
        bline->next_pt = next_positive;
    }
  bline->delta_x = bline->last.x - bline->first.x;
  bline->delta_y = bline->last.y - bline->first.y;
  bline->past_the_end = 0;
  bline->epsilon = 0;
  bline->current = bline->first;
  return bline;
}

/**
 * Wrapper to construct a Bresenham line context from signed coordinates.
 * The context records the state of the algorithm that will generate pairs
 * of discrete coordinates for a series of steps approximating the line linking
 * the first and last points.
 * @see bresenham_from_bcoord
 * @param[in] first_col  the column-coordinate of the first point.
 * @param[in] first_row  the row-coordinate of the first point.
 * @param[in] last_col  the column-coordinate of the last point.
 * @param[in] last_row  the row-coordinate of the last point.
 * @return  A pointer to the state of the Bresenham context at the first point.
 */
bresenham_linep_t
bresenham_from_s_coord (int32_t first_col, int32_t first_row,
                        int32_t last_col, int32_t last_row)
{
  return bresenham_from_bcoord (first_col, first_row, last_col, last_row);
}

/**
 * Wrapper to construct a Bresenham line context from unsigned coordinates.
 * The context records the state of the algorithm that will generate pairs
 * of discrete coordinates for a series of steps approximating the line linking
 * the first and last points.
 * @see bresenham_from_bcoord
 * @param[in] first_col  the column-coordinate of the first point.
 * @param[in] first_row  the row-coordinate of the first point.
 * @param[in] last_col  the column-coordinate of the last point.
 * @param[in] last_row  the row-coordinate of the last point.
 * @return  A pointer to the state of the Bresenham context at the first point.
 */
bresenham_linep_t
bresenham_from_coord (uint32_t first_col, uint32_t first_row,
                      uint32_t last_col, uint32_t last_row)
{
  return bresenham_from_bcoord (first_col, first_row, last_col, last_row);
}

/**
 * Iterate the Bresenham line context, generating the next coordinate pair.
 * As constructed, the line context will always represent at least one
 * point (even if the first point coincides with the last point), and
 * so we can iterate it at least once; however once we've obtained the
 * coordinates of the last point further iterations are incorrect
 * (think of the context as being just off the end of the line).  An
 * iteration produces the current coordinate pair in the sequence of
 * discrete points that approximate the straight line joining the
 * first to the last point, and modifies the internals of the context
 * to prepare the next point.  The return value is a Boolean flag that
 * is false iff further iterations are possible.
 * @param[in,out] bline points to the structure recording the state to iterate.
 * @param[out] col  where to deposit the current column coordinate.
 * @param[out] row  where to deposit the current row coordinate.
 * @return  True if off the end of the line, false if we can iterate again.
 */
int
bresenham_next (bresenham_linep_t bline, uint32_t * col, uint32_t * row)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (bline && col && row
                  && bline->get_row && bline->get_col && bline->next_pt));
  if (bline->past_the_end)
    logerr_debug (LOGERR_FUNC, "iterating off the end of a line");
  else
    {
      *col = (uint32_t) ((*(bline->get_col)) (bline));
      *row = (uint32_t) ((*(bline->get_row)) (bline));
      bline->past_the_end = ((bline->current.x == bline->last.x)
                             && (bline->current.y == bline->last.y));
      if (!bline->past_the_end)
        (*(bline->next_pt)) (bline);
    }
  return bline->past_the_end;
}

/**
 * Iterate the Bresenham line context, generating the next coordinate pair.
 * Just a variation for signed coordinates.
 * @see bresenham_next
 * @param[in,out] bline points to the structure recording the state to iterate.
 * @param[out] col  where to deposit the current column coordinate.
 * @param[out] row  where to deposit the current row coordinate.
 * @return  True if off the end of the line, false if we can iterate again.
 */
int
bresenham_next_s (bresenham_linep_t bline, int32_t * col, int32_t * row)
{
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC,
                 (bline && col && row
                  && bline->get_row && bline->get_col && bline->next_pt));
  if (bline->past_the_end)
    logerr_debug (LOGERR_FUNC, "iterating off the end of a line");
  else
    {
      *col = (int32_t) ((*(bline->get_col)) (bline));
      *row = (int32_t) ((*(bline->get_row)) (bline));
      bline->past_the_end = ((bline->current.x == bline->last.x)
                             && (bline->current.y == bline->last.y));
      if (!bline->past_the_end)
        (*(bline->next_pt)) (bline);
    }
  return bline->past_the_end;
}

/**
 * Destroy a Bresenham line context.
 * A simple free() would suffice to reclaim the storage allocated by the
 * current Bresenham line context constructors, but we don't want to
 * guarantee this for all future versions of this code.
 * @param[out] blinehandle  points to the pointer to be freed (and set NULL).
 */
void
demolish_bresenham (bresenham_linep_t * blinehandle)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 ((blinehandle != NULL) && (*blinehandle != NULL)));
  free (*blinehandle);
  *blinehandle = NULL;
}

