/** @file logerr.c
 * Support for error reporting macros.
 * In the logerr.h we define a collection of error reporting macros
 * that are supposed to be flexible, so we can in principle define
 * our own error handling functions and to do things like pop up
 * dialog boxes within a GUI, or add to the system error logs.  The
 * application as a whole shares two private pointers to error
 * reporting functions that can either be set to defaults (which
 * simply print to standard error), or overridden by arbitrary
 * callbacks.  Errors beyond a redefinable severity threshold abort
 * the entire application (the codes are arranged in decreasing order
 * of severity, so these will be low numbers than the codes for which
 * we report an error without aborting).  The functions that register
 * callbacks and change the severity threshold use a single mutex
 * for access control in multithreaded applications, but this is
 * probably overkill since the updating operations are pretty much
 * atomic in any case.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: logerr.c,v 1.1 2005/03/22 20:27:45 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "logerr.h"
#if HAVE_PTHREADS
#include <pthread.h>
#endif /* HAVE_PTHREADS */

/** Points to an application-wide function for reporting most errors. */
static logerr_message_poster_t logerr_messenger = NULL;

/** Points to an application-wide function for reporting system errors. */
static logerr_perror_poster_t logerr_perror_messenger = NULL;

/** Errors of this severity or worse cause the application to abort. */
static int logerr_abort_threshold = LOGERR_ERR;

#if HAVE_PTHREADS
/** Used to lock the callback pointers and threshold variable. */
static pthread_mutex_t logerr_modlock = PTHREAD_MUTEX_INITIALIZER;

/** Use genuine a genuine lock operation in a multi-threaded environment. */
#define logerr_lock pthread_mutex_lock (&logerr_modlock)

/** Use genuine a genuine unlock operation in a multi-threaded environment. */
#define logerr_unlock pthread_mutex_unlock (&logerr_modlock)

#else /* !HAVE_PTHREADS */
/** Use genuine a dummy lock operation in a single-threaded environment. */
#define logerr_lock

/** Use genuine a dummy unlock operation in a single-threaded environment. */
#define logerr_unlock
#endif /* !HAVE_PTHREADS */

/**
 * The default general error reporting function.
 * We simply print to standard error.
 * @param[in] priority  the error severity (not used here).
 * @param[in] title  where in the application the error originated.
 * @param[in] message  the error description.
 */ 
static void
default_messenger (int priority, const char * title, const char * message)
{
  if (title != NULL)
    fprintf(stderr, "%s: ", title);
  else
    fprintf(stderr, "From unknown: ");
  if (message != NULL)
    fprintf(stderr, "%s\n", message);
  else
    fprintf(stderr, "(unspecified message)\n");    
}

/**
 * The default general system error reporting function.
 * We simply print to standard error.
 * @param[in] priority  the error severity (not used here).
 * @param[in] title  where in the application the error originated.
 * @param[in] message  the application error description.
 * @param[in] syserr  the system error description.
 */ 
static void
default_perror_messenger (int priority, const char * title,
                          const char * message, const char * syserr)
{
  if (title != NULL)
    fprintf(stderr, "%s: ", title);
  else
    fprintf(stderr, "From unknown: ");
  if (message != NULL)
    fprintf(stderr, "%s\n", message);
  else
    fprintf(stderr, "(unspecified message)\n");
  if (syserr != NULL)
    fprintf(stderr, "The system error message was:\n\t%s\n", syserr);
}

/**
 * Format the error message and pass it to the currently registered handler.
 * We format the error message text in a temporary buffer, invoke
 * whatever error handling function is currently registered as the
 * contents of our private application wide pointer, and abort if the
 * severity threshold is passed.
 * @param[in] priority  the severity for this error.
 * @param[in] title  where in the application the error originated.
 * @param[in] fmt  the format string for the remaining arguments.
 * @param[in] ap  the remaining arguments for generating the message.
 */
void
logerr_vdispatch (int priority, const char * title,
                  const char * fmt, va_list ap)
{
  char message[2048];

  vsnprintf (message, sizeof (message), fmt, ap);
  if (logerr_messenger == NULL)
    default_messenger (priority, title, message);
  else
    (*logerr_messenger) (priority, title, message);
  if (priority <= LOGERR_THRESHOLD_ABORT)
    abort();
}

/**
 * A wrapper for logerr_vdispatch, taking a variable number of arguments.
 * @param[in] priority  the severity for this error.
 * @param[in] title  where in the application the error originated.
 * @param[in] fmt  the format string for the remaining arguments.
 * @param[in] ...  the remaining arguments for generating the message.
 */
void
logerr_dispatch (int priority, const char * title, const char * fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  logerr_vdispatch (priority, title, fmt, ap);
  va_end (ap);
}

/**
 * Format a system error message, passing it to the registered handler.
 * We format the application error message text in a temporary buffer,
 * get the explanation of a system error number from the @c strerror
 * system function, and invoke whatever error system handling function
 * is currently registered as the contents of our private
 * application-wide pointer, aborting if the severity threshold is
 * passed.
 * @param[in] priority  the severity for this error.
 * @param[in] errnum the system error number.
 * @param[in] title  where in the application the error originated.
 * @param[in] fmt  the format string for the remaining arguments.
 * @param[in] ap  the remaining arguments for generating the message.
 */
void
logerr_vperror (int priority, int errnum, const char * title, const char * fmt, va_list ap)
{
  char message[2048];

  vsnprintf (message, sizeof (message), fmt, ap);
  if (logerr_perror_messenger == NULL)
    default_perror_messenger (priority, title, message, strerror(errnum));
  else
    (*logerr_perror_messenger) (priority, title, message, strerror(errnum));
  if (priority <= LOGERR_THRESHOLD_ABORT)
    abort();
}

/**
 * A wrapper for logerr_vperror, taking a variable number of arguments.
 * @param[in] priority  the severity for this error.
 * @param[in] errnum the system error number.
 * @param[in] title  where in the application the error originated.
 * @param[in] fmt  the format string for the remaining arguments.
 * @param[in] ...  the remaining arguments for generating the message.
 */
void
logerr_perror (int priority, int errnum, const char * title, const char * fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  logerr_vperror (priority, errnum, title, fmt, ap);
  va_end (ap);
}

/**
 * Set the severity threshold for aborting the application.
 * @param[in] threshold  the new threshold value.
 */
void
logerr_set_abort_threshold (int threshold)
{
  logerr_lock;
  logerr_abort_threshold = threshold;
  logerr_unlock;
}

/**
 * Register the normal error-reporting callback.
 * @param[in] messenger  the new error-reporting function.
 */
void
logerr_set_messenger (logerr_message_poster_t messenger)
{
  logerr_lock;
  logerr_messenger = messenger;
  logerr_unlock;
}

/**
 * Register the system error-reporting callback.
 * @param[in] poster  the new system error-reporting function.
 */
void
logerr_set_perror_messenger (logerr_perror_poster_t poster)
{
  logerr_lock;
  logerr_perror_messenger = poster;
  logerr_unlock;
}
