/** @file logerr.h
 * Flexible error reporting macros.
 * The collection of macros providing a friendly front end to a couple
 * of error reporting functions that an application can override with
 * its own callbacks if necessary (based partly on ideas lifted from
 * the error logging for GLib, the core library for GTK+ and GNOME).
 * Errors have a ranked series of severities, and there are no
 * messages reporting errors below a certain severity threshold.  For
 * reasons of efficiency, it would be nice if the compiler did not
 * generate code for messages that never appear, but simply having
 * them removed in the preprocessor (using ifdefs, for example) can
 * prevent the compiler finding syntax errors and other bugs that will
 * appear when the error reporting threshold is lowered for debugging.
 * So we go through a few minor contortions to encourage the
 * pre-processor to generate source that will be parsed by the
 * compiler, but which on recent versions of gcc and other such
 * optimizing compilers will not end up in the object code.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: logerr.h,v 1.1 2005/03/22 20:29:38 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#ifndef LOGERR_H
/** The standard guard against including the same header twice. */
#define LOGERR_H 1
#include <stdarg.h>
#include "project_utilities.h"

BEGIN_C_DECLARATIONS

/** Error severity codes (an ordered sequence). */
typedef enum
  {
    LOGERR_EMERG   = 0, /**< Emergency (the most severe error). */
    LOGERR_ALERT   = 1, /**< Alert. */
    LOGERR_CRIT    = 2, /**< Critical. */
    LOGERR_ERR     = 3, /**< Error. */
    LOGERR_WARNING = 4, /**< Warning. */
    LOGERR_NOTICE  = 5, /**< Notice. */
    LOGERR_INFO    = 6, /**< Information. */
    LOGERR_DEBUG   = 7  /**< Debug (the least severe error). */
  } logerr_priority;

#ifndef LOGERR_THRESHOLD_REPORT
/** The error severity threshold for reporting errors. */
#define LOGERR_THRESHOLD_REPORT LOGERR_DEBUG
#endif

#ifndef LOGERR_THRESHOLD_ABORT
/** The error threshold for aborting the application. */
#define LOGERR_THRESHOLD_ABORT LOGERR_ERR
#endif

/** A generic pointer to function reporting most errors. */
typedef void (*logerr_message_poster_t) (int priority, const char * title,
                                         const char * message);
/** A generic pointer to function reporting most system errors. */
typedef void (*logerr_perror_poster_t) (int priority, const char * title,
                                        const char * message, const char * syserr);

void logerr_vdispatch (int priority, const char * title,
                       const char * fmt, va_list ap);
void logerr_dispatch (int priority, const char * title,
                      const char * fmt, ...);

void logerr_vperror (int priority, int errnum, const char * title,
                     const char * fmt, va_list ap);

void logerr_perror (int priority, int errnum, const char * title,
                    const char * fmt, ...);

#if HAVE_C99_VARARGS
/** C99 standard varadic macro for general error reporting. */
#define logerr(priority, title, ...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (priority, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting the system error @a errnum. */
#define logerrno(priority, errnum, title, ...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_perror (priority, errnum, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting an emergency. */
#define logerr_emerg(title, ...) do { \
    if (LOGERR_EMERG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_EMERG, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting an alert. */
#define logerr_alert(title, ...) do { \
    if (LOGERR_ALERT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ALERT, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting a critical condition. */
#define logerr_crit(title, ...) do { \
    if (LOGERR_CRIT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_CRIT, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting an error. */
#define logerr_err(title, ...) do { \
    if (LOGERR_ERR <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ERR, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting a warning. */
#define logerr_warning(title, ...) do { \
    if (LOGERR_WARNING <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_WARNING, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting a notice. */
#define logerr_notice(title, ...) do { \
    if (LOGERR_NOTICE <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_NOTICE, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting information. */
#define logerr_info(title, ...) do { \
    if (LOGERR_INFO <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_INFO, title, __VA_ARGS__); \
  } while (0)
/** C99 standard varadic macro for reporting debugging information. */
#define logerr_debug(title, ...) do { \
    if (LOGERR_DEBUG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_DEBUG, title, __VA_ARGS__); \
  } while (0)
#elif HAVE_OLDGCC_VARARGS
/** Non-standard gcc varadic macro for general error reporting. */
#define logerr(priority, title, fmt...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (priority, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting the system error @a errnum. */
#define logerrno(priority, errnum, title, fmt...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_perror (priority, errnum, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting an emergency. */
#define logerr_emerg(title, fmt...) do { \
    if (LOGERR_EMERG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_EMERG, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting an alert. */
#define logerr_alert(title, fmt...) do { \
    if (LOGERR_ALERT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ALERT, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting a critical condition. */
#define logerr_crit(title, fmt...) do { \
    if (LOGERR_CRIT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_CRIT, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting an error. */
#define logerr_err(title, fmt...) do { \
    if (LOGERR_ERR <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ERR, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting a warning. */
#define logerr_warning(title, fmt...) do { \
    if (LOGERR_WARNING <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_WARNING, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting a notice. */
#define logerr_notice(title, fmt...) do { \
    if (LOGERR_NOTICE <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_NOTICE, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting information. */
#define logerr_info(title, fmt...) do { \
    if (LOGERR_INFO <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_INFO, title, fmt); \
  } while (0)
/** Non-standard gcc varadic macro for reporting debugging information. */
#define logerr_debug(title, fmt...) do { \
    if (LOGERR_DEBUG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_DEBUG, title, fmt); \
  } while (0)
#else /* we can't use varadic preprocessor macros */

/**
 * General prioritized error message if varadic macros don't work.
 * Wrap the logerr_vdispatch function in a variable-length argument
 * list handler.
 * @param[in] priority  the error severity code.
 * @param[in] title where in the application the error originated.
 * @param[in] fmt the format string for the message.
 * @param[in] ... any additional data needed for the message.
 */
static void
logerr (int priority, const char * title, const char * fmt, ...)
{
  va_list ap;

  if (priority <= LOGERR_THRESHOLD_REPORT)
    {
      va_start (ap, fmt);
      logerr_vdispatch (priority, title, fmt, ap);
      va_end (ap);
    }
}

/**
 * Prioritized error message for system errors if varadic macros don't work.
 * Wrap the logerr_vperror function in a variable-length argument
 * list handler.
 * @param[in] priority  the error severity code.
 * @param[in] errnum  the system error number.
 * @param[in] title where in the application the error originated.
 * @param[in] fmt the format string for the message.
 * @param[in] ... any additional data needed for the message.
 */
static void
logerrno (int priority, int errnum, const char * title, const char * fmt, ...)
{
  va_list ap;

  if (priority <= LOGERR_THRESHOLD_REPORT)
    {
      va_start (ap, fmt);
      logerr_vperror (priority, errnum, title, fmt, ap);
      va_end (ap);
    }
}

/**
 * Used to generate the entire family of named error handlers.
 * When varadic macros don't work, we generate error-handling functions
 * with the same easy-to-understand names.
 */
#define LOGERR_MAKE_EASY(NICENAME, UGLYCODE) static void \
logerr_##NICENAME (const char * title, const char * fmt, ...) \
{ \
  va_list ap; \
  if (UGLYCODE <= LOGERR_THRESHOLD_REPORT) \
    { \
      va_start (ap, fmt); \
      logerr_vdispatch (UGLYCODE, title, fmt, ap); \
      va_end (ap); \
    } \
}

LOGERR_MAKE_EASY (emerg, LOGERR_EMERG)
LOGERR_MAKE_EASY (alert, LOGERR_ALERT)
LOGERR_MAKE_EASY (crit, LOGERR_CRIT)
LOGERR_MAKE_EASY (err, LOGERR_ERR)
LOGERR_MAKE_EASY (warning, LOGERR_WARNING)
LOGERR_MAKE_EASY (notice, LOGERR_NOTICE)
LOGERR_MAKE_EASY (info, LOGERR_INFO)
LOGERR_MAKE_EASY (debug, LOGERR_DEBUG)

#endif /* !HAVE_*_VARARGS */

/** Make a prioritized assertion as a special case of an error message. */
#define logerr_assert(priority, title, condition) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) { \
        if (!(condition)) \
          logerr_dispatch(priority, title, \
                          "file %s: line %d: assertion (%s) failed", \
                          __FILE__, __LINE__, (#condition)); \
      } \
  } while (0);

void logerr_set_abort_threshold (int threshold);
void logerr_set_messenger (logerr_message_poster_t messenger);
void logerr_set_perror_messenger (logerr_perror_poster_t poster);

#if HAVE_C99_FUNC
/** Use the C99 standard function name by default. */
#define LOGERR_FUNC __func__
#elif HAVE_PRETTY_FUNCTION
/** Fall back on gcc's function name macro if it is available. */
#define LOGERR_FUNC __PRETTY_FUNCTION__
#else
/** Do not report function names if there's no obvious way to track them. */
#define LOGERR_FUNC NULL
#endif /* HAVE_*_FUNCTION */

END_C_DECLARATIONS

#endif /* !LOGERR_H */

