/** @file tclSadie.c
 * The main tclSadie program, joining a Tcl/Tk GUI to the SADIE library.
 * As an application based in C using Tcl/Tk for all the user
 * inteaction, a certain amount of boilerplate code is needed to
 * initialize the application and invoke the Tcl interpreter.  For
 * full details see Part VI (C Programming) in Welch, B.B., Jones,
 * K. & Hobbs, J. Practical Programming in Tcl/Tk (4th ed.). Prentice
 * Hall: Saddle River, New Jersey: 2003.  While tclSadie is currently
 * based on version 8.4 of Tcl/Tk, it includes some code to be
 * backwards-compatible with version 8.3, in particular some crude
 * substitutes for the Tcl 8.4 Tcl_FS functions.  The mechanism for
 * using the pure Tcl parts of tclSadie remains very primitive,
 * unchanged since version 7 of Tcl or earlier.  A single directory
 * holds all the script files, not structured into packages or in any
 * other way.  The initialization code adds the path to this script
 * library directory to the global variable @a auto_path (which has a
 * special significance to the Tcl interpreter as specifying places to
 * search for script files).
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: tclSadie.c,v 1.2 2005/03/25 04:05:10 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#include <tcl.h>
#include <tk.h>
#include <string.h>
#include <unistd.h>
#include <sadie.h>
#include "tclsadie.h"
#include "Sadie_Index.h"

/** @name Program name and version strings
 * Set the name version strings from autoconf and CVS information if possible.
 */
/*@{*/
#if HAVE_CONFIG_H
static const char package[] = PACKAGE;
static const char short_version[] = PACKAGE_STRING;
#else
static const char package[] = "tclsadie";
static const char short_version[] = "tclSadie";
#endif /* !HAVE_CONFIG_H */
static const char verbose_version[] = "$Id: tclSadie.c,v 1.2 2005/03/25 04:05:10 mmunro Exp $";
/*@}*/

/** The environment variable specifying the Tcl script library directory. */
static const char lib_env_var_name[] = "TCLSADIE_HOME";

/** The main Tcl script file name. */
static const char main_file_name[] = "tclSadie_main.tcl";

/** Name of a Tcl array whose elements include the library path variable. */
static const char sadievar_arrayname[] = SADIEVAR_ARRAYNAME;

/** Raw character library directory path from the command line arguments. */
static char *libdirpath = NULL;

/** @name Initialization function forward declarations.
 * These functions make the C commands available to the Tcl interpreter.
 */
/*@{*/
extern int Sadie_FileIO_Init (Tcl_Interp * interp);
extern int Sadie_General_Init (Tcl_Interp * interp);
extern int Sadie_Plot_Init (Tcl_Interp * interp);
extern int Sadie_NewFunctions_Init (Tcl_Interp * interp);
extern int Sadie_Contrast_Init (Tcl_Interp * interp);
extern int Sadie_Filter_Init (Tcl_Interp * interp);
extern int Sadie_Geometry_Init (Tcl_Interp * interp);
extern int Sadie_Multi_Init (Tcl_Interp * interp);
extern int Sadie_Classify_Init (Tcl_Interp * interp);
extern int Sadie_Tools_Init (Tcl_Interp * interp);
extern int Sadie_Image_Init (Tcl_Interp * interp);
extern int Sadie_Proto_Init (Tcl_Interp * interp);
/*@}*/

/** An application-wide global hash table for storing ROIs and other data. */
sad_doclistp_t global_doclist;

/** @name Tcl 8.3 backward-compatibility code. 
 * Provide substitutes for functions present in Tcl 8.4 but missing from 8.3.
 * Tcl 8.4 introduces a family of generic filesystem operations, the
 * Tcl_FS functions, and we use these if possible.  However they are
 * absent in Tcl 8.3, so to support it we here provide crude
 * substitutes for only those Tcl_FS functions that tclSadie uses.
 * See the man pages for the Tcl 8.4 versions for more details, but
 * much of the functionality is missing from the substitutes.
 */
/*@{*/
#if (TCL_MAJOR_VERSION < 8) || ((TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION <= 3))

/** The Tcl_FS substitutes need a global specifying the Tcl interpreter. */
static Tcl_Interp * global_interpreter = NULL;

/**
 * Convert the given Tcl_Obj to a valid Tcl path type.
 * In reality this is a placeholder that does nothing.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] pathPtr  the object to convert to a path.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
Tcl_FSConvertToPathType (Tcl_Interp * interp, Tcl_Obj * pathPtr)
{
  return
    ((interp == NULL) || (pathPtr == NULL)
     || (Tcl_GetString (pathPtr) == NULL))
    ? TCL_ERROR
    : TCL_OK;
}

/**
 * Check if we can read, write or test for existence of the specified file.
 * The mode is the file permission mask, whose bits determine the combination
 * of tests:
 * - @c R_OK test read access,
 * - @c W_OK test write access,
 * - @c X_OK test execute access and
 * - @c F_OK test checking for file existence.
 *
 * @param[in] pathPtr  the path to the file to be tested.
 * @param[in] mode  file permisson mask.
 * @return 0 on success (all requested permissions granted), -1 otherwise.
 */
static int
Tcl_FSAccess (Tcl_Obj * pathPtr, int mode)
{
  char *path = NULL;
  int status = -1;

  if (pathPtr != NULL)
    {
      path = Tcl_GetString (pathPtr);
      if (path != NULL)
        status = Tcl_Access (path, mode);
    }
  return status;
}

/**
 * Copy filesystem path segments into elements of a new Tcl list object.
 * The list of split path segments has an initial reference count of zero.
 * 
 * @param[in] pathPtr the path to be split.
 * @param[out] lenPtr if non-NULL, where to record the number of list elements.
 * @return the new Tcl list object holding the split path segments.
 */
static Tcl_Obj *
Tcl_FSSplitPath (Tcl_Obj * pathPtr, int * lenPtr)
{
  int argc, objc;
  char *path = NULL;
  CONST char **argv;
  Tcl_Obj **objv = NULL;
  Tcl_Obj *listobjPtr = NULL;
  int err = 0;

  if (pathPtr != NULL)
    {
      path = Tcl_GetString (pathPtr);
      if (path != NULL)
        {
          Tcl_SplitPath (path, &argc, &argv);
          if (argv != NULL)
            {
              objv = calloc (argc, sizeof (Tcl_Obj *));
              if (objv != NULL)
                {
                  for (objc = 0; (!err && (objc < argc)); objc++)
                    {
                      objv[objc] = Tcl_NewStringObj (argv[objc], -1);
                      err = (objv[objc] == NULL);
                    }
                  if (!err)
                    {
                      listobjPtr = Tcl_NewListObj (objc, objv);
                      err = (listobjPtr == NULL);
                    }
                  if (err)
                    {
                      while ((--objc) > 0)
                        Tcl_DecrRefCount (objv[objc]);
                    }
                  else if (lenPtr != NULL)
                    *lenPtr = objc;
                  free (objv);
                }
              Tcl_Free ((char *) argv);
            }              
        }
    }
  return listobjPtr;
}

/**
 * Join filsystem path segments, taken from a list, to form a new Tcl object.
 * Given a list whose elements represent filesystem path segments, join
 * the specified number of elements to form a single object representing
 * a path.
 *
 * @param[in] listObj the list of path segments to be joined.
 * @param[in] elements the number of elements to join (if < 0, everything).
 * @return the new Tcl list object holding the path.
 */
static Tcl_Obj *
Tcl_FSJoinPath (Tcl_Obj * listObj, int elements)
{
  int objc, argc, lolen;
  char **argv;
  Tcl_DString theresult;
  Tcl_Obj *elemobj = NULL;
  Tcl_Obj *pathobj = NULL;
  int err = TCL_OK;

  if ((listObj == NULL) || (global_interpreter == NULL)
      || (Tcl_ListObjLength (global_interpreter, listObj, &lolen) != TCL_OK))
    return NULL;
  argc = ((elements < 0) || (elements > lolen)) ? lolen : elements;
  if (argc > 0)
    {
      argv = calloc (argc, sizeof (char *));
      if (argv != NULL)
        {
          Tcl_DStringInit (&theresult);
          for (objc = 0;((err == TCL_OK) && (objc < argc)); objc++)
            {
              err = Tcl_ListObjIndex (global_interpreter, listObj, objc, &elemobj);
              if (err == TCL_OK)
                argv[objc] = Tcl_GetString (elemobj);
            }
          if (err == TCL_OK)
            pathobj = Tcl_NewStringObj (Tcl_JoinPath (argc, argv, &theresult), -1);
          Tcl_DStringFree (&theresult);
        }
      free (argv);
    }
  return pathobj;
}

/**
 * Append filesystem path segments to an existing path.
 * Given an existing path (which may be NULL) compose a new path by
 * appending segments taken from an array.
 *
 * @param[in] basePtr points to the existing path object.
 * @param[in] objc the number of path segment objects to append.
 * @param[in] objv the array filesystem path segment objects.
 * @return the new Tcl object (reference count 0) holding the joined path.
 */
Tcl_Obj *
Tcl_FSJoinToPath(Tcl_Obj * basePtr, int objc, Tcl_Obj * CONST objv[])
{
  int i, argc;
  char **argv;
  Tcl_DString theresult;
  Tcl_Obj *pathobj = NULL;
  Tcl_Obj **segobj = NULL;
  int err = TCL_OK;

  if ((objc <= 0) || (objv == NULL))
    pathobj = basePtr;
  else
    {
      argc = (basePtr == NULL) ? objc : objc + 1;
      argv = calloc (argc, sizeof (char *));
      if (argv != NULL)
        {
          i = 0;
          if (basePtr != NULL)
            argv[i++] = Tcl_GetString (basePtr);
          segobj = objv;
          while ((i < argc) && (segobj != NULL))
            argv[i++] = Tcl_GetString (*segobj++);
          Tcl_DStringInit (&theresult);
          pathobj = Tcl_NewStringObj (Tcl_JoinPath (i, argv, &theresult), -1);
          Tcl_DStringFree (&theresult);
          free (argv);
        }
    }
  return pathobj;
}

#endif /* (TCL_MAJOR_VERSION < 8) || ((TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION <= 3)) */
/*@}*/

/**
 * The exit handler function used to tear down the global hash table.
 * A wrapper to make the function that cleans up the global hash table
 * conform to the Tcl_ExitProc callback prototype so it can work as an
 * exit handler (registered by Tcl_CreateExitHandler).
 *
 * @param[in] clientData  really points to the application-global hash table.
 */
void
tclsadie_demolish_hashtable (ClientData clientData)
{
  dispose_doclist ((sad_doclistp_t) clientData);
}

/**
 * Display a summary of the main program command line options.
 * Write brief reminder of the options and their usage on standard output.
 *
 * @param[in] the name of the program (generally @c argv[0] from @c main).
 */
static void
usage (const char progname[])
{
  printf ("TclSadie, a Tcl/Tk front end to the SADIE image processing library.\n");
  printf ("Usage: %s [OPTION]\n", progname);
  printf ("\n");
  printf ("  -h           display this help and exit\n");
  printf ("  -v           display the short version number and exit\n");
  printf ("  -V           display the verbose version information and exit\n");
  printf ("  -d tclSadie_library_directory_path\n");
  printf ("\n");
#if HAVE_CONFIG_H
  /* Take the package maintainer email from the autoconf settings. */
  printf ("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
#endif /* HAVE_CONFIG_H */
}

/**
 * Try to set the Tcl script library directory from an environment variable.
 * Check if an environment variable defined by the constant string
 * @a lib_env_var_name holds the filesystem path of a location containing
 * the @c .tcl script files making up the pure Tcl parts of the
 * application, turning this into a new Tcl filesystem path object.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return a Tcl filesystem path object on success, NULL on failure.
 */
static Tcl_Obj *
setlibdir_env (Tcl_Interp *interp)
{
  Tcl_Obj *tclsadie_lib = NULL;
  Tcl_Obj *env = NULL;
  Tcl_Obj *valueobjPtr = NULL;

  env = Tcl_NewStringObj("env", -1);
  if (env != NULL)
    {
      Tcl_IncrRefCount (env);
      tclsadie_lib = Tcl_NewStringObj (lib_env_var_name, -1);
      if (tclsadie_lib != NULL)
        {
          Tcl_IncrRefCount (tclsadie_lib);
          valueobjPtr = Tcl_ObjGetVar2 (interp, env, tclsadie_lib,
                                        TCL_GLOBAL_ONLY);
          if (valueobjPtr)
            {
              Tcl_IncrRefCount (valueobjPtr);
              if ((Tcl_FSConvertToPathType (interp, valueobjPtr) != TCL_OK)
                  || (Tcl_FSAccess (valueobjPtr, F_OK)))
                {
                  Tcl_DecrRefCount (valueobjPtr);
                  valueobjPtr = NULL;
                }
            }
          Tcl_DecrRefCount (tclsadie_lib);
        }
      Tcl_DecrRefCount (env);
    }
  return valueobjPtr;
}

/**
 * Try to set the Tcl script library directory from the program's path.
 * Check if a filesystem path based on the parent directory of the
 * executable program is a location containing the @c .tcl script
 * files making up the pure Tcl parts of the application, turning this
 * into a new Tcl filesystem path object.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return a Tcl filesystem path object on success, NULL on failure.
 */
static Tcl_Obj *
setlibdir_default (Tcl_Interp *interp)
{
  Tcl_Obj *ourexepath = NULL;
  Tcl_Obj *pathlist = NULL;
  Tcl_Obj *dirnames[2] = { NULL, NULL };
  Tcl_Obj *libdirpath = NULL;
  int pathlen;
    
  ourexepath = Tcl_NewStringObj (Tcl_GetNameOfExecutable (), -1);
  if (ourexepath)
    {
      Tcl_IncrRefCount (ourexepath);
      if (Tcl_FSConvertToPathType (interp, ourexepath) == TCL_OK)
        {
          pathlist = Tcl_FSSplitPath (ourexepath, &pathlen);
          if (pathlist && (pathlen >= 2))
            {
              Tcl_IncrRefCount (pathlist);
              dirnames[0] = Tcl_NewStringObj ("share", -1);
              if (dirnames[0])
                {
                  Tcl_IncrRefCount (dirnames[0]);
                  dirnames[1] = Tcl_NewStringObj (package, -1);
                  if (dirnames[1])
                    {
                      Tcl_IncrRefCount (dirnames[1]);
                      if (Tcl_ListObjReplace (interp, pathlist,
                                          (pathlen - 2), 2, 2, dirnames)
                          == TCL_OK)
                        {
                          libdirpath = Tcl_FSJoinPath (pathlist, -1);
                          if (libdirpath)
                            {
                              Tcl_IncrRefCount (libdirpath);
                              if (Tcl_FSAccess (libdirpath, F_OK))
                                {
                                  Tcl_DecrRefCount (libdirpath);
                                  libdirpath = NULL;
                                }
                            }
                        }
                      Tcl_DecrRefCount (dirnames[1]);
                    }
                  Tcl_DecrRefCount (dirnames[0]);
                }
              Tcl_DecrRefCount (pathlist);
            }
        }
      Tcl_DecrRefCount (ourexepath);
    }
  return libdirpath;
}

/**
 * Try to set the Tcl script library directory from the source directory path.
 * Check if a filesystem path based on the parent directory of the
 * executable program in its pre-installation position in the source
 * directory subtree is a location containing the @c .tcl script files
 * making up the pure Tcl parts of the application, turning this into
 * a new Tcl filesystem path object.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @return a Tcl filesystem path object on success, NULL on failure.
 */
static Tcl_Obj *
setlibdir_build (Tcl_Interp *interp)
{
  Tcl_Obj *ourexepath = NULL;
  Tcl_Obj *pathlist = NULL;
  Tcl_Obj *dirname = NULL;
  Tcl_Obj *libdirpath = NULL;
  int pathlen;
    
  ourexepath = Tcl_NewStringObj (Tcl_GetNameOfExecutable (), -1);
  if (ourexepath)
    {
      Tcl_IncrRefCount (ourexepath);
      if (Tcl_FSConvertToPathType (interp, ourexepath) == TCL_OK)
        {
          pathlist = Tcl_FSSplitPath (ourexepath, &pathlen);
          if (pathlist && (pathlen >= 2))
            {
              Tcl_IncrRefCount (pathlist);
              dirname = Tcl_NewStringObj ("library", -1);
              if (dirname)
                {
                  Tcl_IncrRefCount (dirname);
                  if (Tcl_ListObjReplace (interp, pathlist,
                                          (pathlen - 2), 2, 1, &dirname)
                      == TCL_OK)
                    {
                      libdirpath = Tcl_FSJoinPath (pathlist, -1);
                      if (libdirpath)
                        {
                          Tcl_IncrRefCount (libdirpath);
                          if (Tcl_FSAccess (libdirpath, F_OK))
                            {
                              Tcl_DecrRefCount (libdirpath);
                              libdirpath = NULL;
                            }
                        }
                    }
                  Tcl_DecrRefCount (dirname);
                }
              Tcl_DecrRefCount (pathlist);
            }
        }
      Tcl_DecrRefCount (ourexepath);
    }
  return libdirpath;
}

/**
 * Inform the Tcl interpreter of the script library location.
 * The Tcl interpreter treats the global variable @c auto_path specially,
 * as a list of locations it should search for @c .tcl script files: we
 * append the script library directory to this, and also record it in the
 * the application's global array for holding SADIE settings.
 *
 * @param[in]  interp  the main Tcl interpreter.
 * @param[in]  pathObj  the path to the Tcl script library directory.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */ 
static int
set_tclsadie_libdir (Tcl_Interp *interp, Tcl_Obj *pathObj)
{
  Tcl_Obj *arrayobjPtr = NULL;
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *libdirobjPtr = NULL;
  Tcl_Obj *autopathname = NULL;
  Tcl_Obj *autopathlist = NULL;
  int err = TCL_OK;

  arrayobjPtr = Tcl_NewStringObj (sadievar_arrayname, -1);
  if (arrayobjPtr == NULL)
    err = TCL_ERROR;
  else
    {
      Tcl_IncrRefCount (arrayobjPtr);      
      indexobjPtr = Tcl_NewStringObj ("libdir", -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (indexobjPtr);      
          libdirobjPtr = Tcl_ObjSetVar2 (interp, arrayobjPtr, indexobjPtr,
                                         pathObj,
                                         TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
          if (libdirobjPtr == NULL)
            err = TCL_ERROR;
          else
            {
              Tcl_IncrRefCount (libdirobjPtr);      
              autopathname = Tcl_NewStringObj ("auto_path", -1);
              if (autopathname == NULL)
                err = TCL_ERROR;
              else
                {
                  Tcl_IncrRefCount (autopathname);
                  autopathlist = Tcl_ObjGetVar2 (interp, autopathname, NULL,
                                                 (TCL_GLOBAL_ONLY
                                                  | TCL_LEAVE_ERR_MSG));
                  if (autopathlist == NULL)
                    err = TCL_ERROR;
                  else
                    {
                      if (Tcl_IsShared (autopathlist))
                        autopathlist = Tcl_DuplicateObj (autopathlist);
                      err = Tcl_ListObjAppendElement (interp, autopathlist,
                                                      libdirobjPtr);
                    }
                  Tcl_DecrRefCount (autopathname);
                }
              Tcl_DecrRefCount (libdirobjPtr);      
            }
          Tcl_DecrRefCount (indexobjPtr);
        }
      Tcl_DecrRefCount (arrayobjPtr);      
    }
  return err;
}

/**
 * Search for the location of the Tcl script files (script library directory).
 * If a command line argument did not specify the script library
 * directory, we search first for a location specified by an
 * environment variable, then if this fails, attempt to find the
 * directory within the source build subtree (ignoring any installed
 * version), and finally search relative to the position of the
 * installed executable file.
 * 
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static int
setlibdir (Tcl_Interp *interp)   /* I  points to the command interpreter. */
{
  Tcl_Obj *basePtr = NULL;
  Tcl_Obj *fileobjPtr = NULL;
  Tcl_Obj *testobjPtr = NULL;
  int err = TCL_OK;

  if (libdirpath)
    basePtr = Tcl_NewStringObj (libdirpath, -1);
  else
    {
      basePtr = setlibdir_env (interp);
      if (basePtr == NULL)
        basePtr = setlibdir_build (interp);
      if (basePtr == NULL)
        basePtr = setlibdir_default (interp);
    }
  if (basePtr == NULL)
    err = TCL_ERROR;
  else
    {
      fileobjPtr = Tcl_NewStringObj (main_file_name, -1);
      if (fileobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (fileobjPtr);
          testobjPtr = Tcl_FSJoinToPath (basePtr, 1, &fileobjPtr);
          if (testobjPtr == NULL)
            err = TCL_ERROR;
          else
            {
              Tcl_IncrRefCount (testobjPtr);
              err = (Tcl_FSAccess (testobjPtr, R_OK) == 0)
                ? set_tclsadie_libdir (interp, basePtr)
                : TCL_ERROR;
              Tcl_DecrRefCount (testobjPtr);
            }
          Tcl_DecrRefCount (fileobjPtr);
        }
      Tcl_DecrRefCount (basePtr);
    }
  return err;
}

/**
 * Copy the short package name and version number into a global array.
 * Constant strings recording the program name and version description
 * are already available to the C parts of the application, but for
 * the Tcl scripts to use them we must transcribe them into the global
 * Tcl array that holds various application-wide settings.
 * 
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static int
set_tclsadie_version (Tcl_Interp *interp)
{
  Tcl_Obj *arrayobjPtr = NULL;
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *pnameobjPtr = NULL;
  int err = TCL_OK;

  arrayobjPtr = Tcl_NewStringObj (sadievar_arrayname, -1);
  if (arrayobjPtr == NULL)
    err = TCL_ERROR;
  else
    {
      Tcl_IncrRefCount (arrayobjPtr);      
      indexobjPtr = Tcl_NewStringObj ("version", -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (indexobjPtr); 
          pnameobjPtr = Tcl_NewStringObj (short_version, -1);
          if (pnameobjPtr == NULL)
            err = TCL_ERROR;
          else
            {
              Tcl_IncrRefCount (pnameobjPtr); 
              if (Tcl_ObjSetVar2 (interp, arrayobjPtr, indexobjPtr,
                                  pnameobjPtr, (TCL_GLOBAL_ONLY
                                                | TCL_LEAVE_ERR_MSG))
                  == NULL)
                err = TCL_ERROR;
              Tcl_DecrRefCount (pnameobjPtr); 
            }
          Tcl_DecrRefCount (indexobjPtr);
        }
      Tcl_DecrRefCount (arrayobjPtr);      
    }
  return err;
}

/**
 * The standard Tcl application initialization hook procedure.
 * Because tclSadie uses a custom C main program, rather than the
 * standard @a wish Tcl/Tk application, it provides its own
 * initialization code.  The Tcl library supports a generic hook
 * procedure, @a Tcl_AppInit, provided by the application; here it
 * invokes all the code needed to register the new Tcl commands (those
 * implemented as C functions) with the Tcl interpreter.The on-line
 * documents describing the Tcl library and Part VI of Welch
 * et. al. Practical Programming in Tcl/Tk describe this in greater
 * detail.
 * 
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
Tcl_AppInit (Tcl_Interp *interp)   /* I  points to the command interpreter. */
{
  if ((Tcl_Init (interp) != TCL_OK)
      || (Tk_Init (interp) != TCL_OK)
      || (tclsadie_private_data_init (interp) != TCL_OK)
      || (tclsadie_app_init () != TCL_OK))
    return TCL_ERROR;
#if (TCL_MAJOR_VERSION < 8) || ((TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION <= 3))
  global_interpreter = interp;
#endif
  /* Set up the hash table (modify the parameter to tune performance). */
  global_doclist = make_doclist (6);
  Tcl_CreateExitHandler (tclsadie_demolish_hashtable,
                         (ClientData) global_doclist);
  Application = APP_TCLSADIE;
  NAMES = ON;
  INIT ();
  /* Call the init procedures for newly defined commands. */
  if ((Sadie_FileIO_Init (interp) != TCL_OK)
      || (Sadie_General_Init (interp) != TCL_OK)
      || (Sadie_Plot_Init (interp) != TCL_OK)
      || (Sadie_NewFunctions_Init (interp) != TCL_OK)
      || (Sadie_Contrast_Init (interp) != TCL_OK)
      || (Sadie_Filter_Init (interp) != TCL_OK)
      || (Sadie_Geometry_Init (interp) != TCL_OK)
      || (Sadie_Multi_Init (interp) != TCL_OK)
      || (Sadie_Classify_Init (interp) != TCL_OK)
      || (Sadie_Tools_Init (interp) != TCL_OK)
      || (Sadie_Image_Init (interp) != TCL_OK)
      || (set_tclsadie_version (interp) != TCL_OK))
    return TCL_ERROR;
  /* Find the location of the library files */
  if (setlibdir (interp) == TCL_OK)
    {
      /* Initialize the sadie program */
      if (Tcl_Eval (interp, "tclSadie_main_init") != TCL_OK)
        return TCL_ERROR;
      Tk_MainLoop ();
    }
  else
    {
      fprintf (stderr,
               "Could not find the directory holding tclSadie's .tcl files;\n");
      fprintf (stderr,
               "  you can specify the path to this with the -d command line option,\n");
      fprintf (stderr,
               "  or with the %s environment variable.\n", lib_env_var_name);
      Tcl_Exit (1);
    }
  return TCL_OK;
}

/**
 * The tclSadie main program: parse command line arguments and start Tcl/Tk.
 * This version of @a main starts tclSadie without invoking the
 * standard Tcl applications such as @a wish.  The optarg utility
 * found in most C support libraries handles command line argument
 * processing: see the @a usage function for their descriptions.
 *
 * @param[in]  argc  the number of command-line arguments.
 * @param[in]  argv  the command line argument array.
 * @return zero on success, otherwise an error code.
 */
int
main (int argc, char *argv[])
{
  int optc;
  char *progname = NULL;
  extern char *optarg;
  extern int optind;
  extern int optopt;
  int err = 0;
  int h_flag = 0;
  int v_flag = 0;
  int V_flag = 0;

  progname = argv[0];
  while ((optc = getopt (argc, argv, ":hd:vV")) != -1)
    {
      switch (optc)
        {
        case 'h':
          h_flag = 1;
          break;
        case 'd':
          libdirpath = optarg;
          break;
        case 'v':
          v_flag = 1;
          break;
        case 'V':
          V_flag = 1;
          break;
        case ':':
          fprintf (stderr,
                   "%s: a path must follow the option -%c.\n",
                   progname, optopt);
          err = 1;
          break;
        case '?':
          fprintf (stderr,
                   "%s: could not recognize the option -%c.\n",
                   progname, optopt);
          err = 1;
        }
    }
  if (optind < argc)
    {
      fprintf (stderr, "%s: too many arguments.\n", progname);
      err = 1;
    }
  if (err)
    {
      fprintf (stderr, "Try `%s -h' for more information.\n", progname);
      exit (2);
    }
  if (h_flag)
    {
      usage (progname);
      exit (0);
    }
  if (v_flag || V_flag)
    {
      printf ("%s\n", short_version);
      if (V_flag)
        printf ("%s\n", verbose_version);
      printf ("Copyright (C) 1988--2004 University of Arizona Digital Image Analysis Lab.\n");
      exit (0);
    }
  /* Start the Tk main program, trimming the argument list. */
  printf ("Initializing tclSadie, please wait . . . . .\n");
  Tk_Main (1, argv, Tcl_AppInit);
  return 0;
}
