/** @file tclSadie_ROI.c
 * Manipulate region of interest (ROI) data structures associated with images.
 * The SADIE library does not provide a mechanism for associating regions
 * of interest with particular images (although some previous versions
 * used a modified image structure for this purpose).  Here we provide
 * not only an interface to the library's function for generating
 * region of interest data (ROISTATS), but also associate them with
 * the image from which they were derived, using a global hash table to
 * record the association for the lifetime of the image, or the the
 * duration of a particular tclSadie run (whichever is the shorter).
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: tclSadie_ROI.c,v 2.3 2005/03/22 21:53:21 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <sadie.h>
#include "Sadie_Index.h"
#include "tclSadie_ROI.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** Global hash table holding the ROIs for the lifetime of an image. */
extern sad_doclistp_t global_doclist;

/** Hash key for locating the list of ROIs associated with an image. */
static const char roiprop[] = ROI_PROPERTY_NAME;

/**
 * ROI destructor utility, for reclaiming storage from the ROI list.
 * The hash table houskeeping operations should invoke this
 * automatically when they remove a document from the table (delete
 * an image).
 * @param[in] data  is really a pointer to the head of the ROI list.
 */
static void
roi_destruct (void *data)
{
  ROIPtr rlist, victim;
  
  if (data != NULL)
    {
      rlist = (ROIPtr) data;
      while (rlist)
        {
          victim = rlist;
          free (victim->name);
          free (victim->zero);
          free (victim->gmin);
          free (victim->gmax);
          free (victim->mean);
          free (victim->dev);
          free (victim->cov);
          free (victim->cor);
          rlist = rlist->next;
          free (victim);
        }
    }
}

/** An empty ROI data structure (for initialization). */
static const ROI null_roi =
  { 0, 0, NULL, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };

/**
 * Create a data structure summarizing a rectangular ROI within an image.
 * Note that this is currently inefficient and inelegant: it operates on
 * a partial copy of the original image rather than simply examining the
 * original pixel values.  The @c sadie.h file defines a region of interest
 * data structure into which we place the various statistics, using the
 * ROISTATS operation from the SADIE library.  All the regions defined within
 * a particular image are strung together in a single list, and the global
 * hash table maintains the association between this and the image.
 *
 * @param[in] in    points to the image containing the region.
 * @param[in] name  a name to give the region.
 * @param[in] dkey  the ID of the image containing the region.
 * @param[in] xbgn  the region's initial horizontal coordinate.
 * @param[in] ybgn  the region's initial vertical coordinate.
 * @param[in] xend  the region's final horizontal coordinate.
 * @param[in] yend  the region's final vertical coordinate.
 * @param[in] matrix  compute covariances and correlations (1), or not (0).
 * @param[in] inc  the pixel subsample increment.
 * @param[in] nlev the number of histogram bins (or zero to suppress it).
 */
void
CREATE_ROI (IMAGE *in, const char *name, const char *dkey,
            uint32_t xbgn, uint32_t ybgn, uint32_t xend, uint32_t yend,
            int matrix, uint32_t inc, uint32_t nlev)
{
  char msg[SLEN];
  uint32_t i, j, k;
  uint64_t nval, n;
  PIXEL *gval = NULL;
  ROIPtr r = NULL;
  ROIPtr s = NULL;

  /* check input */
  if (!CHECKIMG (in))
    {
      MESSAGE('E', " Can't identify input image.");
      return;
    }
  if (!name)
    {
      MESSAGE('E', " Can't get a name for the region.");
      return;
    }
  if (!dkey)
    {
      MESSAGE('E', " Can't get a hash string for the image.");
      return;
    }
  if (NAMES)
    {
      MESSAGE('I', "");
      MESSAGE('I', "CREATE_ROI");
      MESSAGE('I', "");
      sprintf(msg, " Region Name:                %s", name);
      MESSAGE('I',msg);
      sprintf(msg," Region Start Point:         ( %lu , %lu )",
              (unsigned long) xbgn + 1, (unsigned long) ybgn + 1);
      MESSAGE('I',msg);
      sprintf(msg," Region End Point:           ( %lu , %lu )",
              (unsigned long) xend + 1, (unsigned long) yend + 1);
      MESSAGE('I',msg);
      sprintf(msg," Region Height:              %lu pixels",
              (unsigned long) (yend - ybgn) + 1);
      MESSAGE('I',msg);
      sprintf(msg," Region Width:               %lu pixels",
              (unsigned long) (xend - xbgn) + 1);
      MESSAGE('I',msg);
      MESSAGE('I'," ...............");
    }
  /* Determine the number of pixels in the region */
  nval = (uint64_t)(1 + xend - xbgn) * (uint64_t)(1 + yend - ybgn);
  /* Allocate space for the array of pixel values */
  gval = malloc (in->nbnd * nval * sizeof (PIXEL));
  if (gval)
    {
      r = malloc (sizeof (ROI));
      if (r == NULL)
        free (gval);
      else
        *r = null_roi;
    }
  if ((gval == NULL) || (r == NULL))
    {
      MESSAGE ('E', "Memory request failed.");
      return;
    }
  if (!((r->name = malloc ((strlen (name) + 1) * sizeof(char)))
        && (r->zero = malloc (in->nbnd * sizeof (uint64_t)))
        && (r->gmin = malloc (in->nbnd * sizeof (PIXEL)))
        && (r->gmax = malloc (in->nbnd * sizeof (PIXEL)))
        && (r->mean = malloc (in->nbnd * sizeof (double)))
        && (r->dev = malloc (in->nbnd * sizeof (double)))
        && (!matrix
            || (r->cov = malloc (in->nbnd * in->nbnd * sizeof(double))))
        && (!matrix
            || (r->cor = malloc (in->nbnd * in->nbnd * sizeof(double))))))
    {
      if (r->cor)
        free (r->cor);
      if (r->cov)
        free (r->cov);
      if (r->dev)
        free (r->dev);
      if (r->mean)
        free (r->mean);
      if (r->gmax)
        free (r->gmax);
      if (r->gmin)
        free (r->gmin);
      if (r->zero)
        free (r->zero);
      if (r->name)
        free (r->name);
      free (r);
      MESSAGE ('E', "Memory request failed.");
    }
  else
    {
      strcpy (r->name,name);
      r->nval = nval;
      for (n = i = 0; i < in->nbnd; i++)
        {
          for (j = ybgn; j <= yend; j++)
            {
              for (k = xbgn; k <= xend; k++, n++)
                gval[n] = in->data[i][j][k];
            }
        }
      ROISTATS (r->name, in->nbnd, nval, gval, inc, nlev, r->gmin, r->gmax,
                r->mean, r->dev, r->zero, r->cov, r->cor);
      r->next = NULL;
      s = sad_get_doc_property (global_doclist, dkey, roiprop);
      if (s == NULL)
        {
          if (sad_put_doc_property (global_doclist, dkey, roiprop,
                                    roi_destruct, (void *) r)
              != 0)
          MESSAGE ('E', " Cannot put ROI property.");
        }
      else
        {
          while (s->next != NULL)
            s = s->next;
          s->next = r;
        }
    }
  free (gval);
}





