/** @file tclSadie_app.c
 * Some tclSadie-specific application callbacks for the SADIE library.
 * The SADIE library allows an application to provide its own
 * callbacks for certain utility operations, such as displaying error
 * messages and updating the session log.  If we expect it to
 * manipulate parts of a user interface defined in Tcl/Tk (as in the
 * case of tclSadie), we need to provide callbacks that not only
 * conform to the requirements for SADIE's callbacks, but also are
 * properly Tcl/Tk aware.  For convenience, we also maintain an
 * application-wide data structure for details required by the
 * callbacks, such as the Tk canvas that represents the session log
 * and pre-parsed Tcl/Tk command templates for generating popup error
 * messages.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: tclSadie_app.c,v 1.1 2005/03/22 20:40:53 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "tclsadie.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/** Declare the plotting callback (defined elsewhere) so we can register it. */
extern int tclsadie_plotter (PIXEL * line, uint32_t nbnd, uint32_t npix,
                             const char *minxlbl, const char *maxxlbl,
                             int option, PIXEL gmin, PIXEL gmax, void *data);

/** Define slots in the command template for an error-reporting Tk popup. */
enum err_template_index
{
  ERR_TKDIALOG, /**< modal dialog @c tk_dialog Tcl/Tk command. */
  ERR_WINDOW, /**< the top-level window to use for dialog. */
  ERR_TITLE, /**< text to appear in the window manager's title bar. */
  ERR_TEXT, /**< message to appear in the top portion of the dialog box. */
  ERR_BITMAP, /**< a bitmap to display in the top portion of the dialog. */
  ERR_DEFAULT, /**< the index of the dialog's default button. */
  ERR_STRING, /**< text displayed to the right of the button. */
  max_err_index  /**< the command template length. */
};

/**
 * Free a dynamically allocated pre-parsed Tcl command.
 * Decrement the reference count of all the Tcl objects within
 * a pre-parsed command template before freeing the entire template.
 * @param[out]  template_handle  the template handle (ends up set to NULL).
 */
static void
dispose_template (tclsadie_cmd_skelp_t * template_handle)   /* O  free this. */
{
  int i, objc;
  Tcl_Obj ** objv;

  if ((template_handle != NULL) && (*template_handle != NULL))
    {
      objc = (*template_handle)->len;
      objv = (*template_handle)->cmd;
      if (objv != NULL)
        {
          for (i = 0; i < objc; i++)
            {
              if (objv[i])
                Tcl_DecrRefCount (objv[i]);
            }
          free (objv);
        }
      free (*template_handle);
      *template_handle = NULL;
    }
}

/**
 * Clean up the private data used by the SADIE library callbacks.
 * This conforms to the app_data_destructor_t prototype defined within the
 * SADIE library for something that disposes of application-specific data.
 * @param[out]  rawdata  really the tclSadie application-specific data.
 */
int
dispose_private_data (void * rawdata)
{
  tclsadie_app_datap_t thedata;

  if (rawdata == NULL)
    return 0;
  thedata = (tclsadie_app_datap_t) rawdata;
  dispose_template (&(thedata->log_refresh));
  dispose_template (&(thedata->err_popup));
  if (thedata->sadievar != NULL)
    Tcl_DecrRefCount (thedata->sadievar);
  if (thedata->sessionlog_canvas != NULL)
    Tcl_DecrRefCount (thedata->sessionlog_canvas);
  free (thedata);
  return 0;
}

/**
 * Register the SADIE application data cleanup with the Tcl interpreter.
 * In addition to registering Tcl-aware callbacks with the SADIE library,
 * we also register SADIE's generic application data cleanup operation with
 * the Tcl interpreter.  This wrapper conforms to the Tcl_ExitProc prototype
 * for a Tcl/Tk exit handler (registered by Tcl_CreateExitHandler).
 * @param[in]  clientData  required by the prototype, but not used.
 */
void
tclsadie_demolish_private_data (ClientData clientData)
{
  (void) sad_remove_app_data();
}

/**
 * Create a pre-parsed Tcl command template for refreshing the session log.
 * The default session log update command is very simple, so we can
 * generate the complete pre-parsed command without processing any more
 * arguments.
 * @param[out]  refresh_handle  a handle for the new command template.
 * @return @c TCL_ERROR on errors, otherwise @c TCL_OK.
 */
static int
make_refresh_template (tclsadie_cmd_skelp_t * refresh_handle)
{
  tclsadie_cmd_skelp_t refresh = NULL;
  int err = TCL_OK;

  if (refresh_handle == NULL)
    return TCL_ERROR;
  refresh = malloc (sizeof (struct Tclsadie_cmd_skel));
  if (refresh == NULL)
    err = TCL_ERROR;
  else
    {
      refresh->cmd = malloc (sizeof (Tcl_Obj *));
      if (refresh->cmd == NULL)
        err = TCL_ERROR;
      else
        {
          *(refresh->cmd) = Tcl_NewStringObj (DEFAULT_SESSION_LOG_REFRESH, -1);
          if (*(refresh->cmd) == NULL)
            {
              err = TCL_ERROR;
              free (refresh->cmd);
            }
          else
            {
              Tcl_IncrRefCount (*(refresh->cmd));
              refresh->len = 1;
            }
        }
      if (err != TCL_OK)
        free (refresh);
    }
  *refresh_handle = (err == TCL_OK) ? refresh : NULL;
  return err;
}

/**
 * Create a pre-parsed Tcl/Tk command template for error message pop-ups.
 * The error message command has to do something such as creating a Tk
 * modal dialog widget containing the error message text, so the
 * command template will not work without having a vacant slot filled
 * (the error text placeholder).
 * @param[out]  popup_handle  a handle for the new command template.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static int
make_popup_template (tclsadie_cmd_skelp_t * popup_handle)
{
  tclsadie_cmd_skelp_t popup = NULL;
  Tcl_Obj ** pcmd;
  int i, j, n_null;
  int err = TCL_OK;

  if (popup_handle == NULL)
    return TCL_ERROR;
  popup = malloc (sizeof (struct Tclsadie_cmd_skel));
  if (popup == NULL)
    err = TCL_ERROR;
  else
    {
      pcmd = calloc (max_err_index, sizeof (Tcl_Obj *));
      if (pcmd == NULL)
        err = TCL_ERROR;
      else
        {
          pcmd[ERR_TKDIALOG] = Tcl_NewStringObj ("tk_dialog", -1);
          pcmd[ERR_WINDOW] = Tcl_NewStringObj (DEFAULT_ERR_POPUP, -1);
          pcmd[ERR_TITLE] = Tcl_NewStringObj ("Warning", -1);
          pcmd[ERR_TEXT] = NULL;                /* Placeholder NULL. */
          pcmd[ERR_BITMAP] = Tcl_NewStringObj ("warning", -1);
          pcmd[ERR_DEFAULT] = Tcl_NewIntObj (0);
          pcmd[ERR_STRING] = Tcl_NewStringObj ("OK", -1);
          n_null = 0;
          for (i = 0; i < max_err_index; i++)
            {
              if (pcmd[i] == NULL)
                n_null++;
              else
                Tcl_IncrRefCount (pcmd[i]);
            }
          if (n_null == 1)
            {
              popup->cmd = pcmd;
              popup->len = max_err_index;
            }
          else
            {
              err = TCL_ERROR; /* Must match the number of placeholders. */
              j = max_err_index;
              while (j)
                {
                  if (pcmd[--j])
                    Tcl_DecrRefCount (pcmd[j]);
                }
              free (pcmd);
            }
        }
      if (err != TCL_OK)
        free (popup);
    }
  *popup_handle = (err == TCL_OK) ? popup : NULL;
  return err;
}

/**
 * Set up the tclSadie-specific private data structure for the SADIE library.
 * The tclSadie callbacks need access to information such as the
 * particular Tk canvas used for the SADIE session log, the name of
 * the top-level window to use for pop-up dialog boxes, and the
 * pre-parsed Tcl/Tk command templates, but cannot obtain it from
 * SADIE, so here we initialize the relevant global data structure for
 * their benefit.
 * @param[in]  interp  the main Tcl interpreter.
 * @return @c TCL_ERROR on errors, otherwise @c TCL_OK.
 */
int
tclsadie_private_data_init (Tcl_Interp * interp)
{
  tclsadie_app_datap_t data;
  int err = TCL_OK;
  
  data = malloc (sizeof (struct Tclsadie_app_data));
  if (data == NULL)
    return TCL_ERROR;
  data->main_interp = interp;
  data->sessionlog_canvas
    = Tcl_NewStringObj (DEFAULT_SESSION_LOG_CANVAS, -1);
  if (data->sessionlog_canvas == NULL)
    err = TCL_ERROR;
  else
    {
      Tcl_IncrRefCount (data->sessionlog_canvas);
      data->sadievar = Tcl_NewStringObj (SADIEVAR_ARRAYNAME, -1);
      if (data->sadievar == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (data->sadievar);
          err = make_refresh_template (&(data->log_refresh));
          if (err == TCL_OK)
            {
              err = make_popup_template (&(data->err_popup));
              if (err != TCL_OK)
                dispose_template (&(data->log_refresh));
            }
          if (err != TCL_OK)
            Tcl_DecrRefCount (data->sadievar);
        }
      if (err != TCL_OK)
        Tcl_DecrRefCount (data->sessionlog_canvas);
    }
  if (err != TCL_OK)
    free (data);
  else if ((sad_register_data_destructor (dispose_private_data) != 0)
           || (sad_add_app_data (data) != 0))
    {
      err = TCL_ERROR;
      dispose_private_data (data);
    }
  else
    Tcl_CreateExitHandler (tclsadie_demolish_private_data, NULL);
  return err;
}

/** String used to layout warning messages. */
static const char spacer[] = ":   ";

/** The index identifying the current message text in a Tcl array. */
static const char messix[] = DEFAULT_MESSAGE_INDEX;

/**
 * Update the session log with a new message.
 * We format the message text, adding a prefix that identifies the
 * message type, and place it in an element of the global Tcl array
 * dedicated to various SADIE settings.  Assuming the Tcl interpreter
 * is checking the array element for modifications (using the @c trace
 * mechanism), this will have the side-effect of adding the text to
 * the session log.
 * @param[in]  interp  the main Tcl interpreter.
 * @param[in]  sadievar  the Tcl array name.
 * @param[in]  option  the message type code.
 * @param[in]  msg  the error text.
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
static void
update_message (Tcl_Interp * interp, Tcl_Obj * sadievar,
                char option, char * msg)
{
  char messbuf[SLEN];
  size_t nfree = SLEN;
  char *c = NULL;
  Tcl_Obj * txt = NULL;
  Tcl_Obj * index = NULL;

  if ((!interp) || (!msg) || (!sadievar) || (nfree == 0))
    return;
  c = messbuf;
  *(c++) = option;
  if ((--nfree) > sizeof (spacer))
    {
      strncpy (c, spacer, sizeof (spacer));
      nfree -= (sizeof (spacer) - 1);
      c = messbuf + sizeof (spacer);
      strncpy (c, msg, nfree);
      txt = Tcl_NewStringObj (messbuf, -1);
      if (txt != NULL)
        {
          Tcl_IncrRefCount (txt);
          index = Tcl_NewStringObj (messix, -1);
          if (index != NULL)
            {
              Tcl_IncrRefCount (index);
              Tcl_ObjSetVar2 (interp, sadievar, index, txt, TCL_GLOBAL_ONLY);
              Tcl_DecrRefCount (index);
            }
          Tcl_DecrRefCount (txt);
        }
    }
}

/**
 * The callback allowing the SADIE library to generate log messages.
 * We write an informational or low-key warning message (as opposed to
 * an alert) to the session log without popping up a modal dialog box.
 * The message type code is the character supplied as the @a option
 * parameter of the SADIE library function MESSAGE, with the following
 * significance:
 * - @c E error,
 * - @c W warning,
 * - @c I informational message, and
 * - @c P picture (not used by tclSadie).
 *
 * @param[in]  option  the message type code.
 * @param[in]  msg  the error text.
 * @param[in]  data  points to the tclSadie application private data.
 */
void
tclsadie_warning_messenger (char option, char * msg, void * data)
{
  tclsadie_app_datap_t private_data;
  Tcl_Interp * interp;
  Tcl_Obj * svar;

  if (data == NULL)
    return;
  private_data = (tclsadie_app_datap_t) data;
  interp = private_data->main_interp;
  svar = private_data->sadievar;
  if ((interp != NULL) && (svar != NULL))
    update_message (interp, svar, option, msg);
}

/**
 * The callback allowing the SADIE library to generate error messages.
 * In this context an error is an alert that pops up a modal Tk dialog
 * widget, in addition to resetting the message text element of the
 * array holding Tcl settings.  The message type code is the character
 * supplied as the @a option parameter of the SADIE library function
 * MESSAGE, with the following significance:
 * - @c E error,
 * - @c W warning,
 * - @c I informational message, and
 * - @c P picture (not used by tclSadie).
 *
 * @param[in]  option  the message type code.
 * @param[in]  msg  the error text.
 * @param[in]  data  points to the tclSadie application private data.
 */
void
tclsadie_error_messenger (char option, char * msg, void * data)
{
  tclsadie_app_datap_t private_data;
  Tcl_Interp * interp;
  tclsadie_cmd_skelp_t popup;
  Tcl_Obj * svar;
  Tcl_Obj ** pcmd;
  int plen;

  if (data == NULL)
    return;
  private_data = (tclsadie_app_datap_t) data;
  interp = private_data->main_interp;
  popup = private_data->err_popup;
  svar = private_data->sadievar;
  if ((interp != NULL) && (popup != NULL) && (svar != NULL))
    {
      pcmd = popup->cmd;
      plen = popup->len;
      if ((pcmd != NULL) && (plen > 0))
        {
          pcmd[ERR_TEXT] = Tcl_NewStringObj (msg, -1);
          if (pcmd[ERR_TEXT] != NULL)
            {
              Tcl_IncrRefCount (pcmd[ERR_TEXT]);
              if (Tcl_EvalObjv (interp, plen, pcmd, TCL_EVAL_GLOBAL) == TCL_OK)
                update_message (interp, svar, option, msg);
              Tcl_DecrRefCount (pcmd[ERR_TEXT]);
              pcmd[ERR_TEXT] = NULL;
            }
        }
    }
}

/**
 * Register all the necessary callbacks with the SADIE library.
 * Note that this registers @e all the SADIE callbacks (including the
 * plotter and progress meter callbacks), not just the ones defined in
 * @c tclSadie_app.c.  The SADIE library provides the various
 * @c sad_register functions (see @c sadie.h for the interfaces).
 * @return @c TCL_OK on success, @c TCL_ERROR on errors.
 */
int
tclsadie_app_init (void)
{
  return ((sad_register_plotter_callback (tclsadie_plotter)
           || sad_register_creator_callback (ProgMeter_Create)
           || sad_register_updator_callback (ProgMeter_Update)
           || sad_register_destructor_callback (ProgMeter_Destroy)
           || sad_register_message_callback (SAD_ERROR,
                                             tclsadie_error_messenger)
           || sad_register_message_callback (SAD_WARNING,
                                             tclsadie_warning_messenger)
           || sad_register_message_callback (SAD_INFO,
                                             tclsadie_warning_messenger))
           || sad_register_message_callback (SAD_DEFAULT,
                                             tclsadie_warning_messenger))
          ? TCL_ERROR : TCL_OK;
}
