/** @file tclSadie_getset.c
 * Get tclSadie data from Tcl arrays and set them in the interpreter result.
 * The Tcl parts of tclSadie make extensive use of Tcl array elements
 * for data storage (even in some cases where Tcl lists would be more
 * appropriate); the functions with names beginning GetSadie recover
 * tclSadie data from Tcl array elements (specified by an array name
 * and index).  In the opposite direction, the Tcl interpreter needs
 * access to SADIE-specific objects, so the SetSadie functions return
 * these through the Tcl interpreter result.  The SADIE image IDs
 * manipulated by tclSadie are currently identical to the string
 * representations of their addresses, but anything that relies on
 * this implementation is deprecated, since future versions will use
 * arbitrary strings (the current implementation cannot handle
 * persistent or cross-platform image IDs).
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: tclSadie_getset.c,v 1.1 2005/03/22 20:43:50 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "tclsadie.h"

/**
 * Create a Tcl object representation from a SADIE image.
 * Note that the object has a zero reference count, and the particular
 * string representation may change in future.
 * @param[in] img  points to the SADIE image.
 * @return the new Tcl object representing the image ID.
 */
Tcl_Obj *
NewSadieImageObj (IMAGE *img)
{
  char strbuf[SLEN];

  if (img == NULL)
    return NULL;
  snprintf (strbuf, SLEN - 1, "%lx", (unsigned long) img);
  return Tcl_NewStringObj (strbuf, -1);
}

/**
 * Recover a pointer to a SADIE image from a Tcl object.
 * If possible, we deposit a pointer to a SADIE image in the specified
 * location, but otherwise set it to NULL.  Note that the string
 * representation we use here may change in future.
 * @param[in] objPtr  the Tcl object encoding the image ID.
 * @param[out] imagehandle  where to deposit the SADIE IMAGE pointer.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieImageFromObj (Tcl_Obj *objPtr, IMAGE **imagehandle)
{
  const char *hex_image = NULL;
  unsigned long raw_image;
  int len;

  if ((imagehandle == NULL) || (objPtr == NULL))
    return TCL_ERROR;
  *imagehandle = NULL;
  hex_image = Tcl_GetStringFromObj (objPtr, &len);
  if (len == 0)
    return TCL_ERROR;
  sscanf (hex_image, "%lx", &raw_image);
  *imagehandle = (IMAGE *) raw_image;
  return TCL_OK;
}

/**
 * Reset a SADIE image name, then return its ID as the Tcl interpreter result.
 * We create a new Tcl object representing the ID of a SADIE image,
 * set the name of this image to a string extracted from another Tcl
 * object, and make the image ID object the current interpreter
 * result.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] imagePtr  points to the SADIE image.
 * @param[in] nameobjPtr  points to the Tcl object providing the new name.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
SetSadieNamedImageResult (Tcl_Interp *interp, IMAGE *imagePtr,
                          Tcl_Obj *nameobjPtr)
{
  Tcl_Obj * objPtr;

  if ((interp == NULL) || (nameobjPtr == NULL))
    return TCL_ERROR;
  if (imagePtr != NULL)
    {
      strncpy (imagePtr->text, Tcl_GetString(nameobjPtr), TLEN - 1);
      objPtr = NewSadieImageObj (imagePtr);
      if (objPtr != NULL)
        Tcl_SetObjResult (interp, objPtr);
    }
  return TCL_OK;
}

/**
 * Recover a pointer to a SADIE image from a Tcl array element.
 * We find the Tcl value representing the SADIE image ID through a
 * pair of Tcl objects: an array name and an index (a straightforward
 * application of Tcl_ObjGetVar2).  The array name object must already
 * exist, but we convert a C string to the array index object.  If
 * possible, we deposit a pointer to this SADIE image in the specified
 * location, but otherwise set it to NULL.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] imagehandle  where to deposit the SADIE IMAGE pointer.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieImageFromObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                       const char *ixname, IMAGE **imagehandle)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  const char *hex_image = NULL;
  unsigned long raw_image;
  int len;
  int err = TCL_OK;

  if ((interp == NULL) || (imagehandle == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *imagehandle = NULL;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          hex_image = Tcl_GetStringFromObj (valueobjPtr, &len);
          if (len == 0)
            err = TCL_ERROR;
          else
            {
              sscanf (hex_image, "%lx", &raw_image);
              *imagehandle = (IMAGE *) raw_image;
            }
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Make a Tcl object representing a SADIE image ID, resetting the image name.
 * We create a new Tcl object (with a zero reference count) to
 * represent the ID of a SADIE image, resetting the image name to a
 * string taken from a Tcl array element.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[in,out] imagePtr  points to the SADIE image.
 * @return the new Tcl object representing the image ID.
 */
Tcl_Obj *
NewSadieImageObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                   const char *ixname, IMAGE *imagePtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *nameobjPtr = NULL;
  Tcl_Obj *newobjPtr = NULL;
  char strbuf[SLEN];
  int err = TCL_OK;

  if ((interp == NULL) || (imagePtr == NULL) || (arrayobjPtr == NULL)
      || (!CHECKIMG (imagePtr)))
    return NULL;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      nameobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                   TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (nameobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (nameobjPtr);
          strncpy (imagePtr->text, Tcl_GetString(nameobjPtr), TLEN - 1);
          snprintf (strbuf, SLEN - 1, "%lx", (unsigned long) imagePtr);
          newobjPtr = Tcl_NewStringObj (strbuf, -1);
          Tcl_DecrRefCount (nameobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return (err == TCL_OK) ? newobjPtr : NULL;
}

/**
 * Set the Tcl interpreter result to the ID of a renamed SADIE image.
 * We create a new Tcl object (with a zero reference count) to
 * represent the ID of a SADIE image, resetting the image name to a
 * string taken from a Tcl array element, and return the image ID
 * object through the Tcl interpreter result.  Note that the string
 * representation of the image ID  may change in future.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[in,out] imagePtr  points to the SADIE image.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
SetSadieImageObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                   const char *ixname, IMAGE *imagePtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *nameobjPtr = NULL;
  char strbuf[SLEN];
  int err = TCL_OK;

  if ((interp == NULL) || (imagePtr == NULL) || (arrayobjPtr == NULL)
      || (!CHECKIMG (imagePtr)))
    return TCL_ERROR;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      nameobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                   TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (nameobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (nameobjPtr);
          strncpy (imagePtr->text, Tcl_GetString(nameobjPtr), TLEN - 1);
          snprintf (strbuf, SLEN - 1, "%lx", (unsigned long) imagePtr);
          Tcl_SetObjResult (interp, Tcl_NewStringObj (strbuf, -1));
          Tcl_DecrRefCount (nameobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Set the Tcl interpreter result to a list representing a SADIE image pair.
 * We create new Tcl objects representing the IDs of two SADIE images,
 * renaming these to strings taken from two Tcl array elements (specified
 * by a Tcl object giving the array name, and a pair of index strings).
 * The list as a whole has a reference count of zero, but the image ID
 * objects end up with reference counts of one (they have their counts
 * incremented before inserting them in the list, so just after insertion
 * they are briefly two, and must be decremented).
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  a C array holding the two Tcl array index strings.
 * @param[in,out] comp_image  a C array holding two SADIE image pointers.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
SetSadieComponentObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                       const char **ixname, IMAGE **comp_image)
{
  Tcl_Obj *comp_obj[2] = { NULL, NULL };

  if ((!interp) || (!comp_image) || (!arrayobjPtr) || (!ixname)
      || (!ixname[0]) || (!ixname[1])
      || (!CHECKIMG (comp_image[0])) || (!CHECKIMG (comp_image[1])))
    return TCL_ERROR;
  comp_obj[0] = NewSadieImageObj2 (interp, arrayobjPtr, ixname[0],
                                   comp_image[0]);
  if (comp_obj[0] != NULL)
    Tcl_IncrRefCount (comp_obj[0]);
  else
    return TCL_ERROR;
  comp_obj[1] = NewSadieImageObj2 (interp, arrayobjPtr, ixname[1],
                                   comp_image[1]);
  if (comp_obj[1] != NULL)
    Tcl_IncrRefCount (comp_obj[1]);
  else
    {
      Tcl_DecrRefCount (comp_obj[0]);
      return TCL_ERROR;
    }
  Tcl_SetObjResult (interp, Tcl_NewListObj (2, comp_obj));
  Tcl_DecrRefCount (comp_obj[0]);
  Tcl_DecrRefCount (comp_obj[1]);
  return TCL_OK;
}

/**
 * Recover a PIXEL value from a Tcl array element.
 * A straightforward application of Tcl_ObjGetVar2 finds the PIXEL
 * value through a pair of Tcl objects: an array name (provided as an
 * existing Tcl object) and an index (newly created from the
 * specified C string). If possible, we deposit the value in the
 * specified location, but otherwise set it to zero.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] pixelPtr  where to deposit the PIXEL value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadiePixelFromObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                       const char *ixname, PIXEL *pixelPtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  double raw_pixel;
  int err = TCL_OK;

  if ((interp == NULL) || (pixelPtr == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *pixelPtr = 0;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          err = Tcl_GetDoubleFromObj (interp, valueobjPtr, &raw_pixel);
          if (err == TCL_OK)
            *pixelPtr = (PIXEL) raw_pixel;
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Recover a 32-bit unsigned integer value from a Tcl object.
 * We rely on the Tcl object's long value being large enough to encode
 * an unsigned 32-bit value, which we deposit in the specified
 * location if possible, but otherwise set it to zero.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] objPtr  the Tcl object holding the value.
 * @param[out] uintPtr  where to deposit the 32-bit unsigned integer value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieUintFromObj (Tcl_Interp *interp, Tcl_Obj *objPtr, uint32_t *uintPtr)
{
  long raw_value;
  int err = TCL_OK;

  if ((interp == NULL) || (uintPtr == NULL) || (objPtr == NULL))
    return TCL_ERROR;
  *uintPtr = 0;
  err = Tcl_GetLongFromObj (interp, objPtr, &raw_value);
  if (err == TCL_OK)
    *uintPtr = (uint32_t) raw_value;
  return err;
}

/**
 * Recover a 64-bit unsigned integer value from a Tcl object.
 * Newer versions of Tcl support 64-bit values through the Tcl_WideInt
 * type, but we fall back to using longs if this isn't available,
 * and in ether case set the value to zero on errors.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] objPtr  the Tcl object holding the value.
 * @param[out] uintPtr  where to deposit the 64-bit unsigned integer value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieUint64FromObj (Tcl_Interp *interp, /*  I  Tcl interpreter. */
                       Tcl_Obj *objPtr,    /*  I  the Tcl object. */
                       uint64_t *uintPtr)  /*  O  store the int here. */
{
#ifdef HAVE_TCL_WIDEINT
  Tcl_WideInt raw_value;
#else
  long raw_value;
#endif
  int err = TCL_OK;

  if ((interp == NULL) || (uintPtr == NULL) || (objPtr == NULL))
    return TCL_ERROR;
  *uintPtr = 0;
#ifdef HAVE_TCL_WIDEINT
  err = Tcl_GetWideIntFromObj (interp, objPtr, &raw_value);
#else
  err = Tcl_GetLongFromObj (interp, objPtr, &raw_value);
#endif
  if (err == TCL_OK)
    *uintPtr = (uint64_t) raw_value;
  return err;
}

/**
 * Recover a 32-bit unsigned integer value from a Tcl array element.
 * A straightforward application of Tcl_ObjGetVar2 finds a 32-bit
 * unsigned integer value through a pair of Tcl objects: an array name
 * (provided as an existing Tcl object) and an index (newly created
 * from the specified C string). If possible, we deposit the value in
 * the specified location, but otherwise set it to zero.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] uintPtr  where to deposit the 32-bit unsigned integer value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieUintFromObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                      const char *ixname, uint32_t *uintPtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  long raw_value;
  int err = TCL_OK;

  if ((interp == NULL) || (uintPtr == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *uintPtr = 0;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          err = Tcl_GetLongFromObj (interp, valueobjPtr, &raw_value);
          if (err == TCL_OK)
            *uintPtr = (uint32_t) raw_value;
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Recover an integer value from a Tcl array element.
 * We use Tcl_ObjGetVar2 to find the integer value through a pair of
 * Tcl objects: an array name (provided as an existing Tcl object) and
 * an index (newly created from a C string argument). If possible, we
 * deposit the value in the location specified by @a intPtr, but
 * otherwise set this to zero.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] intPtr  where to deposit the integer value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieIntFromObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                     const char *ixname, int *intPtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  int err = TCL_OK;

  if ((interp == NULL) || (intPtr == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *intPtr = 0;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          err = Tcl_GetIntFromObj (interp, valueobjPtr, intPtr);
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Recover a double value from a Tcl array element.
 * We use Tcl_ObjGetVar2 to find the double value through a pair of
 * Tcl objects: an array name (provided as an existing Tcl object) and
 * an index (newly created from a C string argument). If possible, we
 * deposit the value in the location specified by @a doublePtr, but
 * otherwise set this to zero.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] doublePtr  where to deposit the double value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieDoubleFromObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                        const char *ixname, double *doublePtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  int err = TCL_OK;

  if ((interp == NULL) || (doublePtr == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *doublePtr = 0;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          err = Tcl_GetDoubleFromObj (interp, valueobjPtr, doublePtr);
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Recover an Boolean value from a Tcl array element.
 * Tcl Boolean values are a little more versatile than just the
 * integer constants 0 and 1.  Objects whose string representation is
 * @c 0, @c false, @c no, or @c off (or the corresponding upper-case
 * or abbreviated versions) have a @a false Boolean value; similarly,
 * those whose string representation is @c 1, @c true, @c yes, or @c on
 * (or the corresponding variations) have a @a true Boolean value.  We
 * use Tcl_ObjGetVar2 to find an object holding such a Boolean value
 * through a pair of Tcl objects: an array name (provided as an
 * existing Tcl object) and an index (newly created from a C string
 * argument). If possible, we deposit the value in the location
 * specified by @a boolPtr, but otherwise set this to @a false.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] boolPtr  where to deposit the Boolean value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieBooleanFromObj2 (Tcl_Interp *interp, Tcl_Obj *arrayobjPtr,
                         const char *ixname, int *boolPtr)
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  int err = TCL_OK;

  if ((interp == NULL) || (boolPtr == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *boolPtr = 0;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          err = Tcl_GetBooleanFromObj (interp, valueobjPtr, boolPtr);
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}

/**
 * Obtain a hashable string from a Tcl object representing a SADIE image ID.
 * The string representation of the image ID object is already a reasonable
 * hash key.  Note however that we do not copy it here, and it should
 * generally be copied for subsequent use, lest the Tcl object be
 * nodified or freed.
 * @param[in] objPtr  the Tcl object represention of a SADIE image ID.
 * @param[out] dkeyhandle  where to deposit a pointer to the string.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
GetSadieDkeyFromObj (Tcl_Obj *objPtr, const char ** dkeyhandle)
{
  int len;

  if ((dkeyhandle == NULL) || (objPtr == NULL))
    return TCL_ERROR;
  *dkeyhandle = Tcl_GetStringFromObj (objPtr, &len);
  return (len == 0) ? TCL_ERROR : TCL_OK;
}

/**
 * Obtain a hashable string value from a Tcl array element.
 * The standard Tcl library function Tcl_ObjGetVar2 finds the string
 * value through a pair of Tcl objects: an array name (provided as an
 * existing Tcl object) and an index (newly created from the
 * specified C string), then GetSadieDkeyFromObj recovers the string
 * pointer, which we deposit in @a dkeyhandle, or set this NULL
 * if an error arises.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] arrayobjPtr  the name of an array variable (as a Tcl object).
 * @param[in] ixname  the array index (as a C string).
 * @param[out] dkeyhandle  where to deposit a pointer to the string value.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 2004 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This utility function obtains a hashable string from the Tcl object      */
/*   representaion of a SADIE image given by a Tcl array name and index.      */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int
GetSadieDkeyFromObj2 (Tcl_Interp *interp,   /*  I  Tcl interpreter. */
                      Tcl_Obj *arrayobjPtr, /*  I  the Tcl array object. */
                      const char *ixname,   /*  I  array index string. */
                      const char **dkeyhandle)  /*  O  string key handle. */
{
  Tcl_Obj *indexobjPtr = NULL;
  Tcl_Obj *valueobjPtr = NULL;
  int err = TCL_OK;

  if ((interp == NULL) || (dkeyhandle == NULL) || (arrayobjPtr == NULL))
    return TCL_ERROR;
  *dkeyhandle = NULL;
  if (ixname != NULL)
    {
      indexobjPtr = Tcl_NewStringObj (ixname, -1);
      if (indexobjPtr == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (indexobjPtr);
    }
  if (err == TCL_OK)
    {
      valueobjPtr = Tcl_ObjGetVar2 (interp, arrayobjPtr, indexobjPtr,
                                    TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      if (valueobjPtr == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (valueobjPtr);
          err = GetSadieDkeyFromObj (valueobjPtr, dkeyhandle);
          Tcl_DecrRefCount (valueobjPtr);
        }
      if (indexobjPtr != NULL)
        Tcl_DecrRefCount (indexobjPtr);
    }
  return err;
}


