/** @file tclSadie_plot.c
 * Provide a callback that lets the SADIE library plot bar graphs using Tk.
 * The SADIE library function PLOT can use a callback to invoke some
 * arbitrary external plotting operation.  Here we define one to meet the
 * requirements of tclSadie: the drawing operations use Tk, and the
 * resulting graphs appear directly within the tclSadie session log.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: tclSadie_plot.c,v 1.1 2005/03/22 20:45:52 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "tclsadie.h"

/** Constants for positioning parts of the plotted bar graphs. */
enum plot_pos
{
  TOP_MARGIN = 15, /**< vertical whitespace before the start of the plot. */
  RIGHT_MARGIN = 100, /**< horizontal whitespace before a band bar graph. */ 
  TOP_SPACING = 10, /**< additional space at the top of the @a y axis.*/
  LEFT_SPACING = 61,  /**< additional space to the left of the @a y axis.*/
  Y_HEIGHT = 100, /**< positive @a y axis height (maximum bar height). */
  AXIS_DEPTH = 1, /**< offset for the thickness of the axis line. */
  MINOR_DEPTH = 4,  /**< length of minor ticks on an axis. */ 
  MAJOR_DEPTH = 6,  /**< length of major ticks on an axis. */ 
  CAPTION_DEPTH = 8,  /**< offset for postioning the graph caption. */
  PLOT_HEIGHT = 140, /**< the total vertical distance taken by the plots. */
  TICK_SPACE = 16,  /**< the frequency of axis ticks (unit intervals). */
  MAJOR_SPACE = 2  /**< the frequency of major ticks (axis tick intervals). */
};

/** Array slots for axis labels. */
enum label_index
{
  LBL_XMIN, /**< minimum @a x axis slot. */
  LBL_XMAX, /**< maximum @a x axis slot. */
  LBL_YMIN, /**< minimum @a y axis slot. */
  LBL_YMAX, /**< maximum @a y axis slot. */
  max_label_index /**< the number of axis label slots. */
};

/** Slots for constants @sa initialize_plot_objs for the strings. */
enum plot_obj_index
{
  ANCHOR, AXISWIDTH, BLACK, BLUE, CREATE, E, FILL, GREEN, LINE, N, RED,
  TEXT, TEXTOPT, WIDTH, max_obj_index
};

/** Template slots for a pre-parsed Tcl/Tk axis line command. */
enum axis_line_template_index
{
  AXL_PATHNAME, /**< the Tcl/Tk canvas pathName. */
  AXL_CREATE, /**< the @c create widget command. */
  AXL_LINE, /**< the @c line item type. */
  AXL_X1, /**< the starting @a x coordinate of the line. */
  AXL_Y1, /**< the starting @a y coordinate of the line. */
  AXL_X2, /**< the ending @a x coordinate of the line. */
  AXL_Y2, /**< the ending @a y coordinate of the line. */
  AXL_WIDTH,  /**< the @c -width standard option. */
  AXL_OUTLINEWIDTH,  /**< the width value slot. */
  max_axl_index  /**< the number of axis line template slots. */
};

/** Template slots for a pre-parsed Tcl/Tk bar graph bar command. */
enum bar_template_index
{
  BAR_PATHNAME, /**< the Tcl/Tk canvas pathName. */
  BAR_CREATE, /**< the @c create widget command. */
  BAR_LINE, /**< the @c line item type. */
  BAR_X1, /**< the starting @a x coordinate of the line. */
  BAR_Y1, /**< the starting @a y coordinate of the line. */
  BAR_X2, /**< the ending @a x coordinate of the line. */
  BAR_Y2, /**< the ending @a y coordinate of the line. */
  BAR_FILL, /**< the @c -fill standard option. */
  BAR_COLOR, /**< the color value slot. */
  max_bar_index  /**< the number of bar graph bar template slots. */
};

/** Template slots for a pre-parsed Tcl/Tk text label command. */
enum text_template_index
{
  TXT_PATHNAME, /**< the Tcl/Tk canvas pathName. */
  TXT_CREATE, /**< the @c create widget command. */
  TXT_TEXT, /**< the @c text item type. */
  TXT_X, /**< the @a x coordinate of the text's positioning point. */
  TXT_Y, /**< the @a y coordinate of the text's positioning point. */
  TXT_ANCHOR, /** the @c -anchor option. */
  TXT_ANCHORPOS, /** how to position the text relative to the point. */
  TXT_FILL, /**< the @c -fill standard option. */
  TXT_COLOR, /**< the color value slot. */
  TXT_TEXTOPT, /**< @c -text standard option. */
  TXT_STRING, /**< the label text slot. */
  max_text_index  /**< the number of text label template slots. */
};

/** Slots for pre-set @a y coordinates. */
enum y_preset_index
{
  Y_TOP, /**< the top of the @a y axis. */
  Y_BASE, /**< starting @a y coordinate for drawing the graph bars. */
  Y_MINOR, /**< distal end of a minor axis tick. */
  Y_MAJOR, /**< distal end of a major axis tick. */
  Y_CAPTION, /**< vertical position at which to plot the caption. */
  max_y_preset_index  /**< the number of @a y coordinate pre-set slots. */
};

/** Slots for pre-set @a x coordinates. */
enum x_preset_index
{
  X_BASE, /**< just left of the @a x axis. */
  X_MAJOR,  /**< distal end of a major axis tick. */
  X_CAPTION,  /**< horizontal position at which to plot the caption. */
  X_RIGHT, /**< the right end of the @a x axis. */
  max_x_preset_index  /**< the number of @a x coordinate pre-set slots. */
};

/** An immutable coordinate range. */
typedef const struct Line_value_range * line_value_rangep_t;

/** We make enough use of coordinate ranges to merit defining a type. */
struct Line_value_range
{
  double min; /**< low end of the coordinate range. */
  double max; /**< high end of the coordinate range. */
};

/** Define a generic graph bar scaling operation. */
typedef double (*bar_sizer_t) (double val, line_value_rangep_t range);

/** A magic constant (no idea where this comes from). */
static const double magic_scale = 83.3;

/**
 * A variant linear bar scaling operation.
 * Scale the bar by the magic constant and truncate at the @a y axis ends.
 * @param[in] val  the value to scale.
 * @param[in] range  the range of relevant values.
 * @return the scaled bar length.
 */
static double
unit_bar_size (double val, line_value_rangep_t range)
{
  double size;

  size = (magic_scale * (val - range->min)) / (range->max - range->min);
  if (size < 0)
    size = 0;
  return (size > Y_HEIGHT) ? Y_HEIGHT : size;
}

/**
 * A linear bar scaling operation.
 * Scale the bar so the range limits fall at the ends of the @a y axis.
 * @param[in] val  the value to scale.
 * @param[in] range  the range of relevant values.
 * @return the scaled bar length.
 */
static double
norm_bar_size (double val, line_value_rangep_t range)
{
  return Y_HEIGHT * (val - range->min) / (range->max - range->min);
}

/**
 * A logarithmic bar scaling operation.
 * Scale the bar so the bars fit within the @a y axis.
 * @param[in] val  the value to scale.
 * @param[in] range  the range of relevant values.
 * @return the scaled bar length.
 */
static double
log_bar_size (double val, line_value_rangep_t range)
{
  return Y_HEIGHT * ((log10 (val + 1) - range->min)
                     / (range->max - range->min));
}

/**
 * Clean up a C array holding Tcl object pointers.
 * The objects themselves may persist if referenced elsewhere, but following
 * this operation it's safe to free the array.
 * @param[out] obj  the array of objects.
 * @param[in] max_index  the array length.
 */
static void
dispose_objs (Tcl_Obj * obj[], int max_index)
{
  int i;

  for (i = 0; (i < max_index); i++)
    {
      if (obj[i])
        Tcl_DecrRefCount (obj[i]);
    }
}

/**
 * Prepare a C array holding Tcl object pointers.
 * Assuming some of the Tcl objects are newly created, it is unsafe to use
 * them before their reference counts are incremented.
 * @param[in,out] obj  the array of objects.
 * @param[in] max_index  the array length.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
prepare_objs (Tcl_Obj * obj[], int max_index)
{
  int i;
  int err = TCL_OK;

  for (i = 0; ((err == TCL_OK) && (i < max_index)); i++)
    {
      if (obj[i] == NULL)
        err = TCL_ERROR;
      else
        Tcl_IncrRefCount (obj[i]);
    }
  if (err != TCL_OK)
    dispose_objs (obj, max_index);
  return err;
}

/**
 * Preset the Tcl objects for the axis limits.
 * @sa label_index for the array slot definitions.
 * @param[out] obj  the array of label objects.
 * @param[in] minxlbl  the label for the @a x axis minimum.
 * @param[in] maxxlbl  the label for the @a x axis maximum.
 * @param[in] range  the @a y axis range.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
initialize_label_objs (Tcl_Obj * lobj[],
                       const char *minxlbl, const char *maxxlbl,
                       line_value_rangep_t range)
{
  int i;
  char range_mintext[SLEN];
  char range_maxtext[SLEN];

  if ((!lobj) || (!minxlbl) || (!maxxlbl))
    return TCL_ERROR;
  for (i = 0; (i < max_label_index); i++)
    lobj[i] = NULL;
  snprintf (range_mintext, SLEN - 1, "%d", (int) range->min);
  snprintf (range_maxtext, SLEN - 1, "%d", (int) range->max);
  lobj[LBL_XMIN] = Tcl_NewStringObj (minxlbl, -1);
  lobj[LBL_XMAX] = Tcl_NewStringObj (maxxlbl, -1);
  lobj[LBL_YMIN] = Tcl_NewStringObj (range_mintext, -1);
  lobj[LBL_YMAX] = Tcl_NewStringObj (range_maxtext, -1);
  return prepare_objs (lobj, max_label_index);
}

/**
 * Pre-define various common building blocks for the Tcl/Tk commands.
 * The only reason we define these as C array elements is ease of management:
 * it's easy to re-use the operations that increment and decrement the
 * reference counts on arrays of Tcl objects.
 * @param[out] pobj  the array of Tcl objects.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
initialize_plot_objs (Tcl_Obj * pobj[])
{
  int i;

  if (pobj == NULL)
    return TCL_ERROR;
  for (i = 0; (i < max_obj_index); i++)
    pobj[i] = NULL;
  pobj[ANCHOR] = Tcl_NewStringObj ("-anchor", -1);
  pobj[AXISWIDTH] = Tcl_NewDoubleObj (1.5);
  pobj[BLACK] = Tcl_NewStringObj ("black", -1);
  pobj[BLUE] = Tcl_NewStringObj ("blue", -1);
  pobj[CREATE] = Tcl_NewStringObj ("create", -1);
  pobj[E] = Tcl_NewStringObj ("e", -1);
  pobj[FILL] = Tcl_NewStringObj ("-fill", -1);
  pobj[GREEN] = Tcl_NewStringObj ("green", -1);
  pobj[LINE] = Tcl_NewStringObj ("line", -1);
  pobj[N] = Tcl_NewStringObj ("n", -1);
  pobj[RED] = Tcl_NewStringObj ("red", -1);
  pobj[TEXT] = Tcl_NewStringObj ("text", -1);
  pobj[TEXTOPT] = Tcl_NewStringObj ("-text", -1);
  pobj[WIDTH] = Tcl_NewStringObj ("-width", -1);
  return prepare_objs (pobj, max_obj_index);
}

/**
 * Preset the Tcl objects for some vertical positions.
 * @sa y_preset_index for the array slot definitions.
 * @param[out] pos  the array of @a y coordinate objects.
 * @param[in] offset  the vertical offset of the session log insertion point.
 * @param[out] y_origin  where to record the @a y axis origin.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
initialize_y_presets (Tcl_Obj * pos[], uint32_t offset, uint32_t * y_origin)
{
  int i;

  if ((pos == NULL) || (y_origin == NULL))
    return TCL_ERROR;
  for (i = 0; (i < max_y_preset_index); i++)
    pos[i] = NULL;
  *y_origin = offset + TOP_SPACING + Y_HEIGHT;
  pos[Y_TOP] = Tcl_NewDoubleObj (offset + TOP_SPACING);
  pos[Y_BASE] = Tcl_NewDoubleObj (*y_origin + AXIS_DEPTH);
  pos[Y_MINOR] = Tcl_NewDoubleObj (*y_origin + MINOR_DEPTH);
  pos[Y_MAJOR] = Tcl_NewDoubleObj (*y_origin + MAJOR_DEPTH);
  pos[Y_CAPTION] = Tcl_NewDoubleObj (*y_origin + CAPTION_DEPTH);
  return prepare_objs (pos, max_y_preset_index);
}

/**
 * Preset the Tcl objects for some horizontal positions.
 * @sa x_preset_index for the array slot definitions.
 * @param[out] pos  the array of @a x coordinate objects.
 * @param[in] offset  the horizontal session log insertion point offset.
 * @param[in] x_len  the @a x axis length.
 * @param[out] x_origin  where to record the @a x axis origin.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
initialize_x_presets (Tcl_Obj * pos[], uint32_t offset, uint32_t xlen,
                      uint32_t * x_origin)
{
  int i;

  if ((pos == NULL) || (x_origin == NULL))
    return TCL_ERROR;
  for (i = 0; (i < max_x_preset_index); i++)
    pos[i] = NULL;
  *x_origin = offset + LEFT_SPACING;
  pos[X_BASE] = Tcl_NewDoubleObj (*x_origin - AXIS_DEPTH);
  pos[X_MAJOR] = Tcl_NewDoubleObj (*x_origin - MAJOR_DEPTH);
  pos[X_CAPTION] = Tcl_NewDoubleObj (*x_origin - CAPTION_DEPTH);
  pos[X_RIGHT] = Tcl_NewDoubleObj (*x_origin + xlen);
  return prepare_objs (pos, max_x_preset_index);
}

/**
 * Pre-define the template for the Tcl/Tk axis drawing command.
 * @sa axis_line_template_index for the template slot definitions.
 * @param[in] canvas  the Tk canvas for the session log.
 * @param[out] cmd  the axis drawing template.
 * @param[in] pobj  the array of pre-defined command components.
 */
static void
make_axis_line_template (Tcl_Obj * canvas, Tcl_Obj * cmd[], Tcl_Obj * pobj[])
{
  cmd[AXL_PATHNAME] = canvas;
  cmd[AXL_CREATE] = pobj[CREATE];
  cmd[AXL_LINE] = pobj[LINE];
  cmd[AXL_X1] = NULL;
  cmd[AXL_Y1] = NULL;
  cmd[AXL_X2] = NULL;
  cmd[AXL_Y2] = NULL;
  cmd[AXL_WIDTH] = pobj[WIDTH];
  cmd[AXL_OUTLINEWIDTH] = pobj[AXISWIDTH];
}

/**
 * Pre-define the template for the Tcl/Tk graph bar drawing command.
 * @sa bar_template_index for the template slot definitions.
 * @param[in] canvas  the Tk canvas for the session log.
 * @param[out] cmd  the bar drawing template.
 * @param[in] pobj  the array of pre-defined command components.
 */
static void
make_bar_template (Tcl_Obj * canvas, Tcl_Obj * cmd[], Tcl_Obj * pobj[])
{
  cmd[BAR_PATHNAME] = canvas;
  cmd[BAR_CREATE] = pobj[CREATE];
  cmd[BAR_LINE] = pobj[LINE];
  cmd[BAR_X1] = NULL;
  cmd[BAR_Y1] = NULL;
  cmd[BAR_X2] = NULL;
  cmd[BAR_Y2] = NULL;
  cmd[BAR_FILL] = pobj[FILL];
  cmd[BAR_COLOR] = NULL;
}

/**
 * Pre-define the template for the Tcl/Tk text label command.
 * @sa text_template_index for the template slot definitions.
 * @param[in] canvas  the Tk canvas for the session log.
 * @param[out] cmd  the text label drawing template.
 * @param[in] pobj  the array of pre-defined command components.
 */
static void
make_text_template (Tcl_Obj * canvas, Tcl_Obj * cmd[], Tcl_Obj * pobj[])
{
  cmd[TXT_PATHNAME] = canvas;
  cmd[TXT_CREATE] = pobj[CREATE];
  cmd[TXT_TEXT] = pobj[TEXT];
  cmd[TXT_X] = NULL;
  cmd[TXT_Y] = NULL;
  cmd[TXT_ANCHOR] = pobj[ANCHOR];
  cmd[TXT_ANCHORPOS] = NULL;
  cmd[TXT_FILL] = pobj[FILL];
  cmd[TXT_COLOR] = pobj[BLACK];
  cmd[TXT_TEXTOPT] = pobj[TEXTOPT];
  cmd[TXT_STRING] = NULL;
}

/**
 * Draw the @a y axis, using pre-defined Tcl objects.
 * We simply place the preset objects in order and invoke Tcl_EvalObjv.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] cmd  the axis drawing command template.
 * @param[in] x_pos  the pre-defined @a x axis positions.
 * @param[in] y_pos  the pre-defined @a y axis positions.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
draw_y_axis (Tcl_Interp * interp,
             Tcl_Obj * cmd[], Tcl_Obj * x_pos[], Tcl_Obj * y_pos[])
{
  int err = TCL_OK;

  if ((!interp) || (!cmd) || (!x_pos) || (!y_pos))
    return TCL_ERROR;
  cmd[AXL_X1] = x_pos[X_MAJOR];
  cmd[AXL_Y2] = cmd[AXL_Y1] = y_pos[Y_TOP];
  cmd[AXL_X2] = x_pos[X_BASE];
  err = Tcl_EvalObjv (interp, max_axl_index, cmd, TCL_EVAL_GLOBAL);
  if (err != TCL_OK)
    return err;
  cmd[AXL_X1] = cmd[AXL_X2];
  cmd[AXL_Y1] = cmd[AXL_Y2];
  cmd[AXL_Y2] = y_pos[Y_MAJOR];
  return Tcl_EvalObjv (interp, max_axl_index, cmd, TCL_EVAL_GLOBAL);
}

/**
 * Draw the @a x axis, using pre-defined Tcl objects.
 * Most of the coordinates needed to draw the @a x axis are pre-defined,
 * but we must generate new positions for each axis tick position.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] cmd  the axis drawing command template.
 * @param[in] x_pos  the pre-defined @a x axis positions.
 * @param[in] y_pos  the pre-defined @a y axis positions.
 * @param[in] x_origin  the @a x axis origin.
 * @param[in] axlen  the @a x axis length.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
draw_x_axis (Tcl_Interp * interp, Tcl_Obj * cmd[], Tcl_Obj * x_pos[],
             Tcl_Obj * y_pos[], uint32_t x_origin, uint32_t axlen)
{
  uint32_t tick;
  int err = TCL_OK;
  Tcl_Obj *x_tick = NULL;

  if ((!interp) || (!cmd) || (!x_pos) || (!y_pos))
    return TCL_ERROR;
  cmd[AXL_X1] = x_pos[X_MAJOR];
  cmd[AXL_Y2] = cmd[AXL_Y1] = y_pos[Y_BASE];
  cmd[AXL_X2] = x_pos[X_RIGHT];
  err = Tcl_EvalObjv (interp, max_axl_index, cmd, TCL_EVAL_GLOBAL);
  for (tick = TICK_SPACE;
       (err == TCL_OK) && (tick < axlen); tick += TICK_SPACE)
    {
      x_tick = Tcl_NewDoubleObj (x_origin + tick);
      if (x_tick == NULL)
        return TCL_ERROR;
      else
        Tcl_IncrRefCount (x_tick);
      cmd[AXL_X1] = cmd[AXL_X2] = x_tick;
      cmd[AXL_Y2] = (tick % (TICK_SPACE * MAJOR_SPACE))
        ? y_pos[Y_MINOR] : y_pos[Y_MAJOR];
      err = Tcl_EvalObjv (interp, max_axl_index, cmd, TCL_EVAL_GLOBAL);
      Tcl_DecrRefCount (x_tick);
    }
  return err;
}

/**
 * Label the @a y axis, using pre-defined Tcl objects.
 * We simply place the preset objects in order and invoke Tcl_EvalObjv.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] cmd  the axis labeling command template.
 * @param[in] pobj  the array of pre-defined command components.
 * @param[in] x_pos  the pre-defined @a x axis positions.
 * @param[in] y_pos  the pre-defined @a y axis positions.
 * @param[in] axlabel  the array of pre-defined axis labels.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
label_y_axis (Tcl_Interp * interp, Tcl_Obj * cmd[], Tcl_Obj * pobj[],
              Tcl_Obj * x_pos[], Tcl_Obj * y_pos[], Tcl_Obj * axlabel[])
{
  int err = TCL_OK;

  if ((!interp) || (!cmd) || (!pobj) || (!x_pos) || (!y_pos) || (!axlabel))
    return TCL_ERROR;
  cmd[TXT_X] = x_pos[X_CAPTION];
  cmd[TXT_Y] = y_pos[Y_TOP];
  cmd[TXT_ANCHORPOS] = pobj[E];
  cmd[TXT_STRING] = axlabel[LBL_YMAX];
  err = Tcl_EvalObjv (interp, max_text_index, cmd, TCL_EVAL_GLOBAL);
  if (err != TCL_OK)
    return err;
  cmd[TXT_Y] = y_pos[Y_BASE];
  cmd[TXT_STRING] = axlabel[LBL_YMIN];
  return Tcl_EvalObjv (interp, max_text_index, cmd, TCL_EVAL_GLOBAL);
}

/**
 * Label the @a x axis, using pre-defined Tcl objects.
 * We simply place the preset objects in order and invoke Tcl_EvalObjv.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] cmd  the axis labeling command template.
 * @param[in] pobj  the array of pre-defined command components.
 * @param[in] x_pos  the pre-defined @a x axis positions.
 * @param[in] y_pos  the pre-defined @a y axis positions.
 * @param[in] axlabel  the array of pre-defined axis labels.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
label_x_axis (Tcl_Interp * interp, Tcl_Obj * cmd[], Tcl_Obj * pobj[],
              Tcl_Obj * x_pos[], Tcl_Obj * y_pos[], Tcl_Obj * axlabel[])
{
  int err = TCL_OK;

  if ((!interp) || (!cmd) || (!pobj) || (!x_pos) || (!y_pos) || (!axlabel))
    return TCL_ERROR;
  cmd[TXT_X] = x_pos[X_BASE];
  cmd[TXT_Y] = y_pos[Y_CAPTION];
  cmd[TXT_ANCHORPOS] = pobj[N];
  cmd[TXT_STRING] = axlabel[LBL_XMIN];
  err = Tcl_EvalObjv (interp, max_text_index, cmd, TCL_EVAL_GLOBAL);
  if (err != TCL_OK)
    return err;
  cmd[TXT_X] = x_pos[X_RIGHT];
  cmd[TXT_STRING] = axlabel[LBL_XMAX];
  return Tcl_EvalObjv (interp, max_text_index, cmd, TCL_EVAL_GLOBAL);
}

/**
 * Draw all the bars in a bar graph.
 * Given a series of pixel values, a set of pre-defined vertical positions,
 * and a starting position, we use the specified bar scaling function
 * to find the size of each bar, and plug the coordinates into the Tcl/Tk
 * command template to draw them.
 * @param[in] interp  the Tcl interpreter.
 * @param[in,out] cmd  the axis labeling command template.
 * @param[in] y_pos  the pre-defined @a y axis positions.
 * @param[in] x_origin  the horizontal origin position.
 * @param[in] y_origin  the vertical origin position.
 * @param[in] npix  the number of bars to plot.
 * @param[in] bline  the base of the series of values to plot.
 * @param[in] find_bar_size  the bar scaling function.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
draw_bars (Tcl_Interp * interp, Tcl_Obj * cmd[],
           Tcl_Obj * y_pos[], line_value_rangep_t range, double x_origin,
           double y_origin, uint32_t npix, PIXEL * bline,
           bar_sizer_t find_bar_size)
{
  int err = TCL_OK;
  uint32_t x;
  Tcl_Obj *bar_x = NULL;
  Tcl_Obj *bar_len = NULL;
  PIXEL *val = NULL;

  if ((!interp) || (!cmd) || (!y_pos) || (!range) || (!bline)
      || (!find_bar_size))
    return TCL_ERROR;
  for (x = 0, val = bline; ((err == TCL_OK) && (x < npix)); x++, val++)
    {
      bar_x = Tcl_NewDoubleObj (x_origin + x);
      if (bar_x != NULL)
        {
          Tcl_IncrRefCount (bar_x);
          bar_len = Tcl_NewDoubleObj (y_origin
                                      - (*find_bar_size) (*val, range));
          if (bar_len == NULL)
            {
              Tcl_DecrRefCount (bar_x);
            }
          else
            {
              Tcl_IncrRefCount (bar_len);
            }
        }
      if (bar_len == NULL)
        err = TCL_ERROR;
      else
        {
          cmd[BAR_X1] = cmd[BAR_X2] = bar_x;
          cmd[BAR_Y1] = y_pos[Y_BASE];
          cmd[BAR_Y2] = bar_len;
          err = Tcl_EvalObjv (interp, max_bar_index, cmd, TCL_EVAL_GLOBAL);
          Tcl_DecrRefCount (bar_len);
          Tcl_DecrRefCount (bar_x);
        }
    }
  return err;
}

/**
 * Refresh the session log display after plotting the bar graphs.
 * There should already be a pre-parsed Tcl command for refreshing the
 * session log (obtained through tclSadie's private application data
 * structure), so we simply invoke this, then update the session log
 * insertion point, advancing it by the total height of the bar graph
 * plots.
 * @param[in] interp  the Tcl interpreter.
 * @param[in] refresh  the Tcl command to refresh the session log.
 * @param[in] sadievar  the global Tcl array for SADIE-specific settings.
 * @param[in,out] y_offset  the location holding the current vertical offset.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
static int
update_log_view (Tcl_Interp * interp, tclsadie_cmd_skelp_t refresh,
                 Tcl_Obj * sadievar, uint32_t * y_offset)
{
  Tcl_Obj *index = NULL;
  Tcl_Obj *offset = NULL;
  int err = TCL_OK;

  if ((!interp) || (!refresh) || (!refresh->cmd) || (!sadievar))
    return TCL_ERROR;
  err = Tcl_EvalObjv (interp, refresh->len, refresh->cmd, TCL_EVAL_GLOBAL);
  if (err != TCL_OK)
    return err;
  (*y_offset) += PLOT_HEIGHT;
  offset = Tcl_NewLongObj ((long) *y_offset);
  if (offset == NULL)
    err = TCL_ERROR;
  else
    {
      Tcl_IncrRefCount (offset);
      index = Tcl_NewStringObj (DEFAULT_SESSION_LOG_POS_INDEX, -1);
      if (index == NULL)
        err = TCL_ERROR;
      else
        {
          Tcl_IncrRefCount (index);
          if (Tcl_ObjSetVar2
              (interp, sadievar, index, offset, TCL_GLOBAL_ONLY) == NULL)
            err = TCL_ERROR;
          Tcl_DecrRefCount (index);
        }
      Tcl_DecrRefCount (offset);
    }
  return err;
}

/**
 * A callback allowing the SADIE library to plot bar graphs using Tk.
 * Because SADIE knows nothing of the Tcl/Tk specifics of tclSadie's
 * session log (notably the Tk canvas it uses and the current vertical
 * position within this), we first recover the relevant details from
 * the private application-specific data.  Then for each band in the
 * image we plot a bar graph, arranging them horizontally within
 * the session log (for large hyperspectral images this may be unwieldy).
 * The first three bands of a multi-band image get plotted in red,
 * green and blue respectively: everything else in black.  The
 * bar lengths are scaled according to the @a option code, one of the
 * values defined in @c sadie.h :
 * - @c UNIT  values truncated at the unit interval (broken?),
 * - @c NORM  linear scaling, and
 * - @c LOG   logarithmic scaling.
 *
 * @param[in] line  a band-sequential array of pixel values to plot.
 * @param[in] nbnd  the number of bands (number of bar graphs).
 * @param[in] npix  the number of bars to draw for each graph.
 * @param[in] minxlbl  the label for the @a x axis minimum.
 * @param[in] maxxlbl  the label for the @a x axis maximum.
 * @param[in] option  the scaling option code.
 * @param[in] gmin  the low end of the range of pixel values.
 * @param[in] gmax  the high end of the range of pixel values.
 * @param[in] data  the tclSadie application private data structure.
 * @return @c TCL_OK on success, @c TCL_ERROR on failure.
 */
int
tclsadie_plotter (PIXEL * line, uint32_t nbnd, uint32_t npix,
                  const char *minxlbl, const char *maxxlbl,
                  int option, PIXEL gmin, PIXEL gmax, void *data)
{
  tclsadie_app_datap_t tclsadie_private = NULL;
  Tcl_Interp *interp = NULL;
  Tcl_Obj *canvas = NULL;
  Tcl_Obj *sadievar = NULL;
  Tcl_Obj *y_pos[max_y_preset_index];
  Tcl_Obj *x_pos[max_x_preset_index];
  Tcl_Obj *axlabel[max_label_index];
  Tcl_Obj *labelcmd[max_text_index];
  Tcl_Obj *pobj[max_obj_index];
  Tcl_Obj *axlcmd[max_axl_index];
  Tcl_Obj *barcmd[max_bar_index];
  tclsadie_cmd_skelp_t refresh;
  uint32_t x_offset, y_offset, x_origin, y_origin, band;
  struct Line_value_range full_range;
  bar_sizer_t find_bar_size;
  int err = TCL_OK;

  if ((!line) || (!minxlbl) || (!maxxlbl) || (!data))
    return TCL_ERROR;
  tclsadie_private = data;
  interp = tclsadie_private->main_interp;
  canvas = tclsadie_private->sessionlog_canvas;
  sadievar = tclsadie_private->sadievar;
  refresh = tclsadie_private->log_refresh;
  if ((!interp) || (!canvas) || (!sadievar) || (!refresh))
    return TCL_ERROR;
  switch (option)
    {
    case UNIT:
      find_bar_size = unit_bar_size;
      break;
    case NORM:
      find_bar_size = norm_bar_size;
      break;
    case LOG:
      find_bar_size = log_bar_size;
      break;
    default:
      return TCL_ERROR;
    }
  err = GetSadieUintFromObj2 (interp, sadievar, DEFAULT_SESSION_LOG_POS_INDEX,
                              &y_offset);
  if (err != TCL_OK)
    {
      Tcl_BackgroundError (interp);
      return err;
    }
  full_range.min = gmin;
  full_range.max = gmax;
  y_offset += TOP_MARGIN;
  err = initialize_plot_objs (pobj);
  if (err == TCL_OK)
    {
      err = initialize_label_objs (axlabel, minxlbl, maxxlbl, &full_range);
      if (err == TCL_OK)
        {
          err = initialize_y_presets (y_pos, y_offset, &y_origin);
          if (err != TCL_OK)
            dispose_objs (axlabel, max_label_index);
        }
      if (err != TCL_OK)
        dispose_objs (pobj, max_obj_index);
    }
  if (err != TCL_OK)
    {
      Tcl_BackgroundError (interp);
      return err;
    }
  make_axis_line_template (canvas, axlcmd, pobj);
  make_text_template (canvas, labelcmd, pobj);
  make_bar_template (canvas, barcmd, pobj);
  for (band = 0; ((err == TCL_OK) && (band < nbnd)); band++)
    {
      x_offset = (band * (npix + RIGHT_MARGIN));
      if ((nbnd == 1) || (band > 2))
        barcmd[BAR_COLOR] = pobj[BLACK];
      else if (band == 0)
        barcmd[BAR_COLOR] = pobj[RED];
      else if (band == 1)
        barcmd[BAR_COLOR] = pobj[GREEN];
      else
        barcmd[BAR_COLOR] = pobj[BLUE];
      if ((initialize_x_presets (x_pos, x_offset, npix, &x_origin) != TCL_OK)
          || (draw_y_axis (interp, axlcmd, x_pos, y_pos) != TCL_OK)
          || (draw_x_axis (interp, axlcmd, x_pos, y_pos, x_origin, npix)
              != TCL_OK)
          || (label_y_axis (interp, labelcmd, pobj, x_pos, y_pos, axlabel)
              != TCL_OK)
          || (label_x_axis (interp, labelcmd, pobj, x_pos, y_pos, axlabel)
              != TCL_OK)
          || (draw_bars (interp, barcmd, y_pos, &full_range, x_origin,
                         y_origin, npix, (line + (band * npix)),
                         find_bar_size) != TCL_OK))
        {
          err = TCL_ERROR;
          Tcl_BackgroundError (interp);
        }
      dispose_objs (x_pos, max_x_preset_index);
    }
  dispose_objs (y_pos, max_y_preset_index);
  dispose_objs (axlabel, max_label_index);
  dispose_objs (pobj, max_obj_index);
  if (err == TCL_OK)
    {
      err = update_log_view (interp, refresh, sadievar, &y_offset);
      if (err != TCL_OK)
        Tcl_BackgroundError (interp);
    }
  return err;
}
