/** @file tclsadie.h
 * Application-specific data and definitions for tclSadie.
 * Visible interfaces for the utility functions that recover data from
 * Tcl arrays, implement progress bars, and maintain tclSadie's private
 * data within the SADIE library.
 *
 * @author University of Arizona Digital Image Analysis Lab
 * @date 1999, 2004
 * @version $Id: tclsadie.h,v 1.5 2005/03/22 21:21:16 mmunro Exp $
 */

/* This file is part of tclSadie.

   tclSadie is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   tclSadie is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tclSadie; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.  */

#ifndef TCLSADIE_H
/** The standard guard against including the same header twice. */
#define TCLSADIE_H 1
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "project_utilities.h"

BEGIN_C_DECLARATIONS

/** @name Strings used to reference Tcl data.
 * Default or constant strings serving as Tcl names or array indexes.
 * We follow a convention that defines the strings in one place
 * (here), then copies them into private character arrays within the
 * files that need to reference them.
 */
/*@{*/
/** The name of the global Tcl array for SADIE settings. */
#define SADIEVAR_ARRAYNAME "SadieVar"
/** The usual index for finding a SADIE image ID within a Tcl array. */
#define DEFAULT_IN_IMAGE_ADDR_INDEX "inimg,addr"
/** The usual index for finding an output image name within a Tcl array. */
#define DEFAULT_OUT_IMAGE_NAME_INDEX "outname"
/** A trace on an array element with this index updates the session log. */
#define DEFAULT_MESSAGE_INDEX "message"
/** The canvas name for pop-up modal dialogs. */
#define DEFAULT_ERR_POPUP ".dlg"
/** The canvas name for the session log window. */
#define DEFAULT_SESSION_LOG_CANVAS ".sessionlog.canvas"
/** The index of a Tcl array element tracking the session log insert point. */
#define DEFAULT_SESSION_LOG_POS_INDEX "sessionlog,pos"
/** The name of a Tcl command for updating the session log. */
#define DEFAULT_SESSION_LOG_REFRESH "update_sessionlog_view"
/** The name to use for a Tcl array in error messages, etc. */
#define DEFAULT_TCL_ARRAY_TITLE "arrayname"
/*@}*/

/** The pointer to the private data, recovered from @c void* in callbacks. */
typedef struct Tclsadie_app_data * tclsadie_app_datap_t;
/** A pointer to a pre-parsed Tcl object command. */ 
typedef struct Tclsadie_cmd_skel * tclsadie_cmd_skelp_t;

/** A pre-parsed Tcl command, suitable for passing to Tcl_EvalObjv. */
struct Tclsadie_cmd_skel
{
  int len;  /**< the number of command components (words). */
  Tcl_Obj ** cmd; /**< points to the array of components. */
};

/** Application private data for tclSadie (for SADIE library callbacks). */
struct Tclsadie_app_data
{
  Tcl_Interp * main_interp; /**< The main Tcl interpreter. */
  Tcl_Obj * sessionlog_canvas; /**< the Tk canvas for the session log. */
  Tcl_Obj * sadievar;  /**< the global Tcl array holding SADIE settings. */
  tclsadie_cmd_skelp_t log_refresh; /**< the session log refresh command. */
  tclsadie_cmd_skelp_t err_popup; /**< the error message pop-up command. */
};

extern int tclsadie_private_data_init (Tcl_Interp * interp);

extern int tclsadie_app_init (void);

extern int ProgMeter_Create (const char * title, void * data);

extern int ProgMeter_Update (int gaugenum, double value, void * data);

extern int ProgMeter_Destroy (int gaugenum, void * data);

extern Tcl_Obj * NewSadieImageObj (IMAGE * img);

extern Tcl_Obj * NewSadieImageObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                    const char * ixname, IMAGE * imagePtr);
extern int SetSadieNamedImageResult (Tcl_Interp * interp, IMAGE * imagePtr,
                                     Tcl_Obj * nameobjPtr);
extern int GetSadieImageFromObj (Tcl_Obj * objPtr, IMAGE ** imagehandle);

extern int GetSadieDkeyFromObj (Tcl_Obj * objPtr, const char ** dkeyhandle);

extern int GetSadieImageFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                   const char * ixname, IMAGE ** imagehandle);
extern int SetSadieImageObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                              const char * ixname, IMAGE * imagePtr);
extern int SetSadieComponentObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                  const char ** ixname, IMAGE ** comp_image);
extern int GetSadiePixelFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                  const char * ixname, PIXEL * pixelPtr);
extern int GetSadieUintFromObj (Tcl_Interp * interp, Tcl_Obj * objPtr,
                                uint32_t * uintPtr);
extern int GetSadieUint64FromObj (Tcl_Interp * interp, Tcl_Obj * objPtr,
                                  uint64_t * uintPtr);
extern int GetSadieUintFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                 const char * ixname, uint32_t * uintPtr);
extern int GetSadieIntFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                const char * ixname, int * intPtr);
extern int GetSadieDoubleFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                   const char * ixname, double * doublePtr);
extern int GetSadieBooleanFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                    const char * ixname, int * boolPtr);
extern int GetSadieDkeyFromObj (Tcl_Obj * objPtr, const char ** dkeyhandle);

extern int GetSadieDkeyFromObj2 (Tcl_Interp * interp, Tcl_Obj * arrayobjPtr,
                                 const char * ixname,
                                 const char ** dkeyhandle);
END_C_DECLARATIONS

#endif /* !TCLSADIE_H */
